/*******************************************************/
 /*
 * artech2ncid.c - This file is part of artech2ncid.
 *
 * Copyright (c) 2021-2024
 * by Bruno Grasland  <b.sletteland@orange.fr>
 *
 * Some parts of the code have been inspired from NCID by
 * John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * artech2ncid is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * artech2ncid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with artech2ncid.  If not, see <http://www.gnu.org/licenses/>.
 */

/********************************************************/
#include  "artech2ncid.h"
/* globals */
int trace ;
int debug, test ;
int sd                    = -1;
int HostnameFlag, outcall = 1;
int ncidport              = NCIDPORT;
int verbose               = 1;
int disable_incoming      = 0;

volatile int TERM_flag;

hid_device *handle;
struct hid_device_info *devs, *cur_dev;

const char * pidfile;
const char * ncidhost = LOCALHOST;
const char * lineid = LINEID;
const char * logfile  = LOGFILE;
const char * tracefile  = TRACEFILE;
char *progname, *readfile, *writefile;
char hname[HOST_NAME_MAX + 1];
pid_t pid;
FILE *logptr;
FILE *traceptr;

int conferr, OSXlaunchd;

#define MAX_STR 255
#define FALSE 0
#define TRUE  1

/* "pseudo voltage" values (can be changed in config file to match the local phone system )  */
int VAL_OFF_LINE = 32;
int MIN_ON_HOOK = 46;
int MAX_ON_HOOK = 50;
int MIN_OFF_HOOK = 30;
int MAX_OFF_HOOK = 45;
int ON_HOOK_LIMIT = 1300;  //100 for Europe
int OFF_HOOK_LIMIT = 100;
int OFF_LINE_LIMIT = 10000;  // high value > 10000 (not really used for now )
int OPEN_LOOP_DISCONNECT_LIMIT = 10 ;
int RING_PERIOD = 30000 ;         //  15000 for Europe
int RING_GAP = 5000000 ;          //  2000000  for Europe (uSec)
int LOOP_START = 1 ;             // available in EU (1)  only sometimes in US (0 or 1) see https://en.wikipedia.org/wiki/Loop_start
int OPEN_LOOP_DISCONNECT = 10 ;  // see https://en.wikipedia.org/wiki/Loop_start   detection only (not used )
int NON_HEX55_LIMIT = 35 ; // 55 for  Europe   (triggers  ON_LINE via speech detection)    (Only used if LOOP_START == 0 )
int HOOK_FLASH_LIMIT = 500; // 24 for Europe
const char * HOOK_FLASH_ZONE  ="US";  // can be "US" or "EU"
int DELAY = 5 ;
int MAX_TRY = 20 ;
short int volt_int ;
int voltage_abs ;
int voltage ;
int old_voltage;
int polarity= -1 ;
int polarity_change= FALSE ;
int old_polarity= -1 ;

/*
 Parameters ending in LIMIT  are expressed in "number of read-loop iteration"
 100 =~ 0.03 sec  ....  3000 =~ 1 seconde  (depending how fast your machine is ... )

 ON_HOOK_LIMIT should be greater than HOOK_FLASH_LIMIT  ,
 otherwise flash-hook will be detected as hang-up

 * flash hook duration  ( 0.1 to 0.12 sec in Europe ) ( 0.3 to  1 sec in US )

 set ON_HOOK_LIMIT to approximately  100 for Europe , set to approximately  1300 for US ) (this may vary depending on phone company )

 Flash hook processing is different in North America and Europe  , set HOOK_FLASH_ZONEૌ  acordingly ( "US" (default)  or "EU" )
 */


/* call state */
#define OFF 0
#define CONNECT -1
#define WAIT -2
#define HOLD -1
#define ON  1

 int call1_state= OFF ;
 int call2_state= OFF ;

struct setword setword[] = {
   /* char *word;   int type; char **buf; int *value;   int min; int max */
   {"HostnameFlag",     WORDNUM,  0,          &HostnameFlag,     0,       1},
   {"ncidhost",         WORDSTR,  &ncidhost,  0,                 0,       0},
   {"ncidport",         WORDNUM,  0,          &ncidport,         0,       0},
   {"disable-incoming", WORDNUM,  0,          &disable_incoming, 0,       0xFF},
   {"pidfile",          WORDSTR,  &pidfile,   0,                 0,       0},
   {"lineid",           WORDSTR,  &lineid ,   0,                 0,       0},
   {"verbose",          WORDNUM,  0,          &verbose,          1,       MAXLEVEL - 1},
   {"mem_trace",        WORDNUM,  0,          &trace,            0,       2},
   {"tracefile",        WORDSTR,  &tracefile, 0,                 0,       0},
   {"VAL_OFF_LINE",     WORDNUM,  0,       &VAL_OFF_LINE,     0x00,    0xFF},
   {"MIN_ON_HOOK",      WORDNUM,  0,       &MIN_ON_HOOK,      0x00,    0xFF},
   {"MAX_ON_HOOK",      WORDNUM,  0,       &MAX_ON_HOOK,      0x00,    0xFF},
   {"MIN_OFF_HOOK",     WORDNUM,  0,       &MIN_OFF_HOOK,     0x00,    0xFF},
   {"MAX_OFF_HOOK",     WORDNUM,  0,       &MAX_OFF_HOOK,     0x00,    0xFF},
   {"ON_HOOK_LIMIT",    WORDNUM,  0,       &ON_HOOK_LIMIT,    0,     100000},
   {"OFF_HOOK_LIMIT",   WORDNUM,  0,       &OFF_HOOK_LIMIT,   0,     100000},
   {"OFF_LINE_LIMIT",   WORDNUM,  0,       &OFF_LINE_LIMIT,   0,     10000000},
   {"OPEN_LOOP_DISCONNECT_LIMIT",   WORDNUM,  0,       &OPEN_LOOP_DISCONNECT_LIMIT,   0,     1000},
   {"RING_PERIOD",   WORDNUM,  0,          &RING_PERIOD,      0,     1000000},
   {"RING_GAP",         WORDNUM,  0,       &RING_GAP      ,   0,    10000000},
   {"LOOP_START",         WORDNUM,  0,     &LOOP_START    ,   0,           1},
   {"NON_HEX55_LIMIT",    WORDNUM,  0,     &NON_HEX55_LIMIT,  5,        1000},
   {"HOOK_FLASH_ZONE",    WORDSTR,  &HOOK_FLASH_ZONE,  0,     0,           0},
   {"HOOK_FLASH_LIMIT",   WORDNUM,  0,     &HOOK_FLASH_LIMIT,   0,    100000},
   {"DELAY",              WORDNUM,  0,     &DELAY,              0,   INT_MAX},
   {"MAX_TRY",            WORDNUM,  0,     &MAX_TRY,           -1,   INT_MAX},
   {0,                  0,        0,          0,                0,         0}
};

/* Needed for config.c in the library */
struct sendclient sendclient[1];


/* modem or another gateway and per-call info */
struct cid
{
    int status;
    int calltype;   /* CID HUP OUT etc */
    int cidsent;    /* used to give gencid on 2nd ring */
    char ciddate[CIDSIZE];
    char cidtime[CIDSIZE];
    char cidnmbr[CIDSIZE];
    char cidname[CIDSIZE];
    char cidmesg[CIDSIZE];
    char cidline[CIDSIZE];
    char * infoBuffer;
};
struct cid incomingCall_cid = {0, CID, 0, "", "", "", "", NOMESG, ONELINE};


char outgoing_call_date_time[MAX_STR];
char outgoing_call_start_time[MAX_STR];
char outgoing_call_pickup_time[MAX_STR];
char outgoing_call_end_time[MAX_STR];
char outgoing_call2_date_time[MAX_STR];
char outgoing_call2_start_time[MAX_STR];
char outgoing_call2_pickup_time[MAX_STR];
char outgoing_call2_end_time[MAX_STR];

char incoming_call_date_time[MAX_STR];
char incoming_call_start_time[MAX_STR];
char incoming_call_pickup_time[MAX_STR];
char incoming_call_end_time[MAX_STR];
char incoming_call2_date_time[MAX_STR];
char incoming_call2_start_time[MAX_STR];
char incoming_call2_pickup_time[MAX_STR];
char incoming_call2_end_time[MAX_STR];
/*dialled_number is the number being dialled it evolves (one digit added ) every time the user pushes a key on the telephone key pad */
/*called number is the final state of "dialled_number" : all the digits have beeen keyed and the call is actually getting through ...  */
char dialled_number[MAX_STR];
char dialled_number2[MAX_STR];
char called_number[MAX_STR];
char called_number_HF[MAX_STR];
char called_number2[MAX_STR];
char calling_number[MAX_STR];
char calling_name[MAX_STR];
char calling_number2[MAX_STR];
char calling_name2[MAX_STR];
char calling_callType[MAX_STR];
int cpu_id;
char cpu_version[MAX_STR];


int getOptions(int argc, char *argv[]);
void  outgoingCall(const char * number );
void  outgoingCallInfo(const char  * ptype ,const char * number );
void  incomingCall(void);
void  incomingCallInfo(const char * ptytpe ,  const char * number , const char * name );
void  outgoingCall2(const char * number );
void  outgoingCallInfo2(const char  * ptype ,const char * number );
void  incomingCallInfo2(const char * ptytpe ,  const char * number , const char * name );
void  try_to_reconnect( void) ;
int readLine(int sd, char * buf);
void errorExit(int error, const char * msg, const char * arg);
void setup_signal_handlers(void);
void USRhandler(int sig);
void TERMexit(int sig);
void socketConnect(int fatal);
int xdmfCID( unsigned char * buf ,int length);
void hex2uchar(const char * buf, unsigned char *ucbuf);
void cleanup( void ) ;
unsigned char checksum(unsigned char *ptr, size_t size);
long long timeval_diff(struct timeval *difference, struct timeval *end_time, struct timeval *start_time) ;
void  process_flash_hook_EU(const char * number2) ;
void  process_flash_hook_US(const char * number2) ;
struct timeval WID_start_time ;

int main(int argc, char* argv[])
{
   int argind, i, errnum = 0;
   char msgbuf[BUFSIZ];


    int res;
    unsigned char write_buf[64];// 1 extra byte for the report ID
    unsigned char read_buf[84];// 1 extra byte for the report ID
    unsigned char old_buf_4 = 0x00 ;
    unsigned char write_index =0x00 ;
    #define MAX_LINE_STATE_COUNT  10
    #define MAX_END_CALL 30
    # define CONNECT1 1
    # define CONNECT2 2
    #define UNKNOWN  -1
    #define OFF_LINE 0
    #define ON_HOOK  1
    #define OFF_HOOK 2
    #define BEEP_BEEP 3  // can be "remote call getting through(after dialing)" or "remote ringing" or "busy tone" or "remote has hung up" ..that kind of tone ...
    #define ON_LINE  4
    #define RING     5
    #define RING_STOPPED    6

     #define ZERO 0
     #define NON_ZERO 1

    int OFF_LINE_counter=0;
    int ON_HOOK_counter=0;
    int OFF_HOOK_counter=0;
    int ring_detect=0 ;

    wchar_t wstr[MAX_STR];
    int line_state= OFF_LINE ;
    int hook_flash = 0 ;
    int hook_flash_counter = 0 ;
    // hook_flash can be 0 or 1  (pre-hook flash)  or 2 (hook-flash + dial-tone x55)
    int x55_counter=0 ;
    int non_hex55_counter =0 ;
    int old_non_hex55_counter =0 ;
    int CID_byte_count=0 ;
    int CID_byte_length=100 ;
    int CID_offset=0 ;
    #define MAX_CID 3500
    unsigned char CID_hex[MAX_CID];
    char byte_as_a_string[MAX_STR];
    char CID_as_a_string[15000];
    int  started_reading_CID= FALSE ;
    struct timeval start_dial_digit ,interval ;
    struct timeval now  ,last_ring_time , hook_flash_time ,last_poke_time ,last_polarity_change_time;
    long long int uSecs = 0;


    /* zeroes out cpu_version string */
    cpu_id=0;
    bzero(cpu_version,MAX_STR);

    /* global containing name of program */
    progname = strrchr(argv[0], (int) '/');
    progname = progname ? progname + 1 : argv[0];

    /* Get the default config file */
    config = CONFIG;

    /* process options from the command line */
    argind = getOptions(argc, argv);

    /* should not be any arguments */
    if (argc - argind != 0)
    {
        fprintf(stderr, NOOPT, progname, argv[argind]);
        fprintf(stderr, USAGE, progname);
        exit(0);
    }

    else
    {
       /* create or open existing logfile */
        logptr = fopen(logfile, "a");
        errnum = errno;
    }
    logMsg(LEVEL1, "Started: %s\nGateway: %s %s\n",strdate(WITHSEP),
            progname, VERSION);

    /* log command line and any options on separate lines */
    sprintf(msgbuf, "Command line: %s", argv[0]);
    for (i = 1; i < argc; i++)
    {
        if (*argv[i] == '-')
            strcat(strcat(msgbuf, "\n              "), argv[i]);
        else strcat(strcat(msgbuf, " "), argv[i]);
    }
    strcat(msgbuf, "\n");
    logMsg(LEVEL1, msgbuf);

    if (logptr)
    {
       /* logfile opened */
       logMsg(LEVEL1, "logfile: %s\n", logfile);
    }
    else
    {
       /* logfile open failed */
       logMsg(LEVEL1, "%s: %s\n", logfile, strerror(errnum));
    }

    /*
     * read config file, if present, exit on any errors
     * do not override any options set on the command line
     */
    if (doConf()) errorExit(-104, 0, 0);

    logMsg(LEVEL2, "VAL_OFF_LINE:                 %02d\n", VAL_OFF_LINE );
    logMsg(LEVEL2, "MIN_ON_HOOK:                  %02d\n", MIN_ON_HOOK );
    logMsg(LEVEL2, "MAX_ON_HOOK:                  %02d\n", MAX_ON_HOOK );
    logMsg(LEVEL2, "MIN_OFF_HOOK:                 %02d\n", MIN_OFF_HOOK );
    logMsg(LEVEL2, "MAX_OFF_HOOK:                 %02d\n", MAX_OFF_HOOK );
    logMsg(LEVEL2, "ON_HOOK_LIMIT:                %d\n",     ON_HOOK_LIMIT );
    logMsg(LEVEL2, "OFF_HOOK_LIMIT:               %d\n",    OFF_HOOK_LIMIT );
    logMsg(LEVEL2, "OFF_LINE_LIMIT:               %d\n",    OFF_LINE_LIMIT );
    logMsg(LEVEL2, "OPEN_LOOP_DISCONNECT_LIMIT:   %d\n",    OPEN_LOOP_DISCONNECT_LIMIT );
    logMsg(LEVEL2, "RING_PERIOD:                  %d\n",    RING_PERIOD );
    logMsg(LEVEL2, "RING_GAP:                     %d\n",    RING_GAP );
    logMsg(LEVEL2, "LOOP_START:                   %d\n",  LOOP_START );
    logMsg(LEVEL2, "NON_HEX55_LIMIT:              %d\n",  NON_HEX55_LIMIT );
    logMsg(LEVEL2, "HOOK_FLASH_ZONE:              %s\n",  HOOK_FLASH_ZONE  );
    logMsg(LEVEL2, "HOOK_FLASH_LIMIT:             %d\n",  HOOK_FLASH_LIMIT  );
    logMsg(LEVEL2, "DELAY:                        %d\n",  DELAY  );
    logMsg(LEVEL2, "MAX_TRY:                      %d\n",  MAX_TRY  );

    logMsg(LEVEL1, "Verbose level:  %d\n", verbose);

    /* global containing hostname */
    if (gethostname(hname, HOST_NAME_MAX) == -1)
    {
        logMsg(LEVEL1, "gethostname(): %s\n", strerror(errno));
        if (HostnameFlag)
        {
            HostnameFlag = 0;
            logMsg(LEVEL1, "Hostname flag reset to 0\n");
        }
    }
    else
    {
        logMsg(LEVEL1, "Hostname flag:  %d\n", HostnameFlag);
        if (HostnameFlag)
        {
        logMsg(LEVEL1, "Hostname: %s\n", hname);
        strcat(hname, "/");
        }
    }

    if (test) logMsg(LEVEL1, "Test Mode\n");

    if (debug || OSXlaunchd)
    {
        if (debug)
	{
	  logMsg(LEVEL1, "Debug Mode\n");
	}
        else
	{
	  logMsg(LEVEL1, "OSX Launchd Mode\n");
	}
    }

    if (!test && doPID(&pid, pidfile))
    {
       sprintf(msgbuf,"%s already exists", pidfile);
       errorExit(-110, "Fatal", msgbuf);
    }
    switch (trace)
    {
    case 0 : logMsg(LEVEL1, "trace = 0 ( no trace )\n") ;
              break ;
    case 1 : logMsg(LEVEL1, "trace = 1 ( medium  trace )\n") ;
              break ;
    case 2 : logMsg(LEVEL1, "trace = 2 ( heavy trace )\n") ;
              break ;
    }                                                     ;

    logMsg(LEVEL1, "disable_incoming = %#04x\n", disable_incoming ) ;

    if ( (disable_incoming & NO_CID)  != 0 )
    {
      logMsg(LEVEL1, "Disabling reporting of CID calls to ncidd server to avoid duplication with a modem or another gateway)\n");
    }
    if ((disable_incoming & NO_RID)  != 0 )
    {
      logMsg(LEVEL1, "Disabling reporting of RID calls to ncidd server to avoid duplication with a modem or another gateway)\n");
    }
    if ((disable_incoming & NO_WID) != 0 )
    {
      logMsg(LEVEL1, "Disabling reporting of WID calls to ncidd server to avoid duplication with a modem or another gateway)\n");
    }
    if ((disable_incoming & NO_MWI)  != 0 )
    {
      logMsg(LEVEL1, "Disabling reporting of MWI calls to ncidd server to avoid duplication with a modem or another gateway)\n");
    }
    if (disable_incoming != 0x0F  && disable_incoming != 0x00  )
    {
      logMsg(LEVEL1, "Enabling reporting of all other incoming calls to ncidd server \n");
    }
    if (disable_incoming == 0x00  )
    {
      logMsg(LEVEL1, "Enabling reporting of all  incoming calls to ncidd server \n");
    }

    voltage=VAL_OFF_LINE ;
    old_voltage=VAL_OFF_LINE ;

    if (trace >= 1 )
    {
       /* create  tracefile */
        traceptr = fopen(tracefile, "a");
        errnum = errno;
       if (traceptr)
       {
         /* tracefile opened */
         logMsg(LEVEL1, "Opened tracefile: %s\n", tracefile);
       }
       else
       {
         /* tracefile open failed */
	 logMsg(LEVEL1, "%s: %s\n", tracefile, strerror(errnum));
         errorExit(-105, progname, "tracefile open failed");
       }
    }

    // initialize the read buffer (just in case )
    memset(read_buf,0x00,sizeof(read_buf));
    memset(read_buf,0x00,sizeof(read_buf));
    memset(read_buf,0x00,sizeof(read_buf));

    // initialize various string and buffers
    memset(CID_hex,0x00,sizeof(CID));
    strncpy(byte_as_a_string,"",sizeof(byte_as_a_string));
    strncpy(CID_as_a_string,"",sizeof(CID_as_a_string));
    //
    // Set up the command buffer.
    memset(write_buf,0x00,sizeof(write_buf));
    write_buf[0] = 0x01;
    write_buf[1] = 0x81;
/////////////////////////////////////////////////////////////////////////////////////////
// enumerate HID  device in the unlikely cas that there are several Artech AD 102 devices

   /* Initialize the HIDAPI library */
    res = hid_init();
    if (res !=0)
    {
       errorExit(-105, progname, "Failed to initialize HIDAPI library\n");
    }
        logMsg(LEVEL2, "Initialized HIDAPI library\n");

#if defined(__APPLE__)
    /*
     * To work properly needs to be called before hid_open/hid_open_path
     * after hid_init.  Best is to call it right after hid_init.
     */
    hid_darwin_set_open_exclusive(0);
#endif

    /* find Artech AD102  Devices */
    if (!(devs = hid_enumerate(0xc5cb, 0x0002)))
    {
        errorExit(-104, progname, "No  Artech AD102 Devices Found");
    }
    logMsg(LEVEL2, "enumerating Artech AD102  Device(s)\n");

    cur_dev = devs;
    for (i = 0; cur_dev; ++i) {
        logMsg(LEVEL2, "  Artech AD102 Device Found\n");
        logMsg(LEVEL2, "  path:          %s\n", cur_dev->path);
        logMsg(LEVEL2, "  Release:       %hx\n", cur_dev->release_number);
        logMsg(LEVEL2, "  Interface:     %d\n",  cur_dev->interface_number);

#ifdef HID_API_VERSION_MAJOR
        /* library contains the HID version */
        logMsg(LEVEL2, "  HID Version:   %s\n", hid_version_str());
#else
        logMsg(LEVEL2, "  HID Version:   < 0.11, unable to obtain it\n");
#endif

        logMsg(LEVEL2, "\n");
        cur_dev = cur_dev->next;
    }

    logMsg(LEVEL2, "%d Artech AD102  Device(s) found.\n",i );
    hid_free_enumeration(devs);

    // Artech AD102  Devices have no serial number  , so it is very difficult to distinguish between devices
    // in case there are more than one devices plugged in , we exit with an error and ask the user to unplug
    // the additional device .

    if ( i >=  2 )
        {
          errorExit(-105, progname, "Only One Artech AD102 device is supported.  Unplug the additionnal devices.");
        }

////////////////////////////////////////////////////////////////////////////////////////

    // Open the device using the VID, PID,
    // and try to read  Manufacturer String , the Product String and the Serial number.
    handle = hid_open(0xc5cb, 0x0002, NULL);
    if (!handle) {
       errorExit(-105, progname, "HID, unable to open Artech AD102 device.");
    }

    // Read the Manufacturer String
    wstr[0] = 0x0000;
    res = hid_get_manufacturer_string(handle, wstr, MAX_STR);
    if (res < 0)
        logMsg(LEVEL1,"HID ,Unable to read manufacturer string \n");
    logMsg(LEVEL1,"Manufacturer String: %ls\n", wstr);

    // Read the Product String
    wstr[0] = 0x0000;
    res = hid_get_product_string(handle, wstr, MAX_STR);
    if (res < 0)
        logMsg(LEVEL1,"HID,Unable to read product string\n");
    logMsg(LEVEL1,"Product String: %ls\n", wstr);

    // Read the Serial Number String
    wstr[0] = 0x0000;
    res = hid_get_serial_number_string(handle, wstr, MAX_STR);
    if (res < 0)
        logMsg(LEVEL1,"Unable to read serial number string\n");
    logMsg(LEVEL1,"Serial Number String:  \"%ls\"\n", wstr);

    // Set the hid_read() function to be non-blocking.
    hid_set_nonblocking(handle, 1);

    //initialize the Artech AD102 Device


   // As for Write buffer  don't really know the significance of what is being written ,
   // this just mimics  what was seen in the Windows Artech demo program

   //  0x07
    memset(write_buf,0x00,sizeof(write_buf));
    write_buf[0] = 0x02;
    write_buf[1] = 0x07;
    write_buf[2] = write_index ;
    write_buf[3] = 0x00 ;
    write_buf[4] = 0x00 ;
    write_buf[5] = 0x00 ;
    write_buf[6] = 0x30 ;
    write_buf[7] = 0x01 ;
    write_buf[8] = 0x03 ;
    write_index++ ;
    res = hid_write(handle, write_buf, sizeof(write_buf));
    if (res < 0) {
       logMsg(LEVEL1,"HID Error: %ls\n", hid_error(handle));
       errorExit(-105, progname, "HID, unable to write buffer.");
    }

    // 0x09
    write_buf[0] = 0x02;
    write_buf[1] = 0x09;
    write_buf[2] = write_index ;
    write_buf[3] = 0x00 ;
    write_buf[4] = 0x00 ;
    write_buf[5] = 0x00 ;
    write_buf[6] = 0x30 ;
    write_buf[7] = 0x03 ;
    write_buf[8] = 0x11 ;
    write_buf[9] = 0x00 ;
    write_buf[10] = 0x10 ;
    write_index++ ;
    res = hid_write(handle, write_buf, sizeof(write_buf));
    if (res < 0) {
       logMsg(LEVEL1,"HID Error: %ls\n", hid_error(handle));
       errorExit(-105, progname, "HID, unable to write buffer.");
    }

    write_buf[0] = 0x02;
    write_buf[1] = 0x09;
    write_buf[2] = write_index ;
    write_buf[3] = 0x00 ;
    write_buf[4] = 0x00 ;
    write_buf[5] = 0x00 ;
    write_buf[6] = 0x30 ;
    write_buf[7] = 0x03 ;
    write_buf[8] = 0x11 ;
    write_buf[9] = 0x10 ;
    write_buf[10] = 0x10 ;
    write_index++ ;
    res = hid_write(handle, write_buf, sizeof(write_buf));
    if (res < 0) {
       logMsg(LEVEL1,"HID Error: %ls\n", hid_error(handle));
       errorExit(-105, progname, "HID, unable to write buffer.");
    }

    write_buf[0] = 0x02;
    write_buf[1] = 0x09;
    write_buf[2] = write_index ;
    write_buf[3] = 0x00 ;
    write_buf[4] = 0x00 ;
    write_buf[5] = 0x00 ;
    write_buf[6] = 0x30 ;
    write_buf[7] = 0x03 ;
    write_buf[8] = 0x11 ;
    write_buf[9] = 0x20 ;
    write_buf[10] = 0x10 ;
    write_index++ ;
    res = hid_write(handle, write_buf, sizeof(write_buf));
    if (res < 0) {
       logMsg(LEVEL1,"HID Error: %ls\n", hid_error(handle));
       errorExit(-105, progname, "HID, unable to write buffer.");
    }

    write_buf[0] = 0x02;
    write_buf[1] = 0x09;
    write_buf[2] = write_index ;
    write_buf[3] = 0x00 ;
    write_buf[4] = 0x00 ;
    write_buf[5] = 0x00 ;
    write_buf[6] = 0x30 ;
    write_buf[7] = 0x03 ;
    write_buf[8] = 0x11 ;
    write_buf[9] = 0x30 ;
    write_buf[10] = 0x10 ;
    write_index++ ;
    res = hid_write(handle, write_buf, sizeof(write_buf));
    if (res < 0) {
       logMsg(LEVEL1,"HID Error: %ls\n", hid_error(handle));
       errorExit(-105, progname, "HID, unable to write buffer.");
    }

    write_buf[0] = 0x02;
    write_buf[1] = 0x09;
    write_buf[2] = write_index ;
    write_buf[3] = 0x00 ;
    write_buf[4] = 0x00 ;
    write_buf[5] = 0x00 ;
    write_buf[6] = 0x30 ;
    write_buf[7] = 0x03 ;
    write_buf[8] = 0x11 ;
    write_buf[9] = 0x40 ;
    write_buf[10] = 0x10 ;
    write_index++ ;
    res = hid_write(handle, write_buf, sizeof(write_buf));
    if (res < 0) {
       logMsg(LEVEL1,"HID Error: %ls\n", hid_error(handle));
       errorExit(-105, progname, "HID, unable to write buffer.");
    }

    write_buf[0] = 0x02;
    write_buf[1] = 0x09;
    write_buf[2] = write_index ;
    write_buf[3] = 0x00 ;
    write_buf[4] = 0x00 ;
    write_buf[5] = 0x00 ;
    write_buf[6] = 0x30 ;
    write_buf[7] = 0x03 ;
    write_buf[8] = 0x11 ;
    write_buf[9] = 0x50 ;
    write_buf[10] = 0x10 ;
    write_index++ ;
    res = hid_write(handle, write_buf, sizeof(write_buf));
    if (res < 0) {
       logMsg(LEVEL1,"HID Error: %ls\n", hid_error(handle));
       errorExit(-105, progname, "HID, unable to write buffer.");
    }

    write_buf[0] = 0x02;
    write_buf[1] = 0x09;
    write_buf[2] = write_index ;
    write_buf[3] = 0x00 ;
    write_buf[4] = 0x00 ;
    write_buf[5] = 0x00 ;
    write_buf[6] = 0x00 ;
    write_buf[7] = 0x30 ;
    write_buf[8] = 0x03 ;
    write_buf[9] = 0x11 ;
    write_buf[10] = 0x60 ;
    write_buf[11] = 0x10 ;
    write_index++ ;
    res = hid_write(handle, write_buf, sizeof(write_buf));
    if (res < 0) {
       logMsg(LEVEL1,"HID Error: %ls\n", hid_error(handle));
       errorExit(-105, progname, "HID, unable to write buffer.");
    }

    write_buf[0] = 0x02;
    write_buf[1] = 0x09;
    write_buf[2] = write_index ;
    write_buf[3] = 0x00 ;
    write_buf[4] = 0x00 ;
    write_buf[5] = 0x00 ;
    write_buf[6] = 0x00 ;
    write_buf[7] = 0x30 ;
    write_buf[8] = 0x03 ;
    write_buf[9] = 0x11 ;
    write_buf[10] = 0x70 ;
    write_buf[11] = 0x10 ;
    write_index++ ;
    res = hid_write(handle, write_buf, sizeof(write_buf));
    if (res < 0) {
       logMsg(LEVEL1,"HID Error: %ls\n", hid_error(handle));
       errorExit(-105, progname, "HID, unable to write buffer.");
    }

    // set the "poke" request
    memset(write_buf,0x00,sizeof(write_buf));
    write_buf[0] = 0x02;
    write_buf[1] = 0x07;
    write_buf[2] = write_index ;
    write_buf[3] = 0x00 ;
    write_buf[4] = 0x00 ;
    write_buf[5] = 0x00 ;
    write_buf[6] = 0x30 ;
    write_buf[7] = 0x01 ;
    write_buf[8] = 0x03 ;
    // sends a "poke" every 30  sec , and receives reports forever

       setup_signal_handlers();
       logMsg(LEVEL1,"NCID server at %s:%d\n", ncidhost, ncidport);
       if (!test)
       {
         //open connection to ncidd server
         socketConnect(FATAL);
       }
       strncpy (outgoing_call_start_time,"",sizeof(outgoing_call_start_time));
       strncpy (outgoing_call_pickup_time,"",sizeof(outgoing_call_pickup_time));
       strncpy (outgoing_call_end_time,"",sizeof(outgoing_call_end_time));
       strncpy (incoming_call_start_time,"",sizeof(incoming_call_start_time));
       strncpy (incoming_call_pickup_time,"",sizeof(incoming_call_pickup_time));
       strncpy (incoming_call_end_time,"",sizeof(incoming_call_end_time));
       strncpy (called_number,"",sizeof(called_number));
       strncpy (dialled_number,"",sizeof(dialled_number));
       strncpy (calling_number,"",sizeof(calling_number));
       strncpy (calling_number,"",sizeof(calling_number));
       strncpy (outgoing_call2_start_time,"",sizeof(outgoing_call_start_time));
       strncpy (outgoing_call2_pickup_time,"",sizeof(outgoing_call_pickup_time));
       strncpy (outgoing_call2_end_time,"",sizeof(outgoing_call_end_time));
       strncpy (incoming_call2_start_time,"",sizeof(incoming_call_start_time));
       strncpy (incoming_call2_pickup_time,"",sizeof(incoming_call_pickup_time));
       strncpy (incoming_call2_end_time,"",sizeof(incoming_call_end_time));
       strncpy (called_number2,"",sizeof(called_number));
       strncpy (called_number_HF,"",sizeof(called_number_HF));
       strncpy (dialled_number2,"",sizeof(dialled_number2));
       strncpy (calling_number2,"",sizeof(calling_number));
       strncpy (calling_callType,"",sizeof(calling_callType));
       line_state = OFF_LINE;
       read_buf[0] =0x03 ;
       res = hid_write(handle, write_buf, sizeof(write_buf));
       if (res < 0) {
          errorExit(-105, progname, "HID, unable to write buffer.");
        }
        else
        {
           logMsg(LEVEL9,"%s hid_write() OK ;  (poking) \n" , strdate(LOGFILETIME));
            write_index++ ;
            write_buf[2] = write_index ;
            gettimeofday(&last_poke_time, NULL);
	}
       logMsg(LEVEL3,"%s line_state=  OFF_LINE \n",strdate(LOGFILETIME) );
       gettimeofday(&last_polarity_change_time, NULL);
      for ( ; ; ) {

        // Read requested state. hid_read() has been set to be
        // non-blocking by the call to hid_set_nonblocking() above.
        // This loop demonstrates the non-blocking nature of hid_read().
        res = 0;
        while (res == 0) {
          res = hid_read(handle,read_buf, sizeof(read_buf) +1 );
          if (res == 0)
          {
            usleep (1000);
          }
          if (res < 0)
            {
	     logMsg(LEVEL1,"HID Error: %ls\n",hid_error(handle));
             errorExit(-105, progname, "HID, unable to read.");
            }
          if (res < 64  && res > 0 )
            {
             logMsg(LEVEL1,"hid_read() res=%d ( res < 64 ) \n",res );
             if (  trace >= 1 )
             {
               fprintf(traceptr,"hid_read() res=%d:  ( res < 64 ) \n",res );
               fprintf(traceptr,"\n");
             }
            }
         }

         // trace of memory content of artech 102 device (for debug/reverse engineering  ).
         // As "0x04" lines are apparently "repeats"  (no changes ,but read_buf[4] (voltage?) ) ,
         // We do print them only if trace >= 1 (heavy ).
         // By default , written to /var/log/artech2ncid.trace  .

         //  these 6 lines allow further debug (traces "0x04" lines )
         if (trace == 2)
         {
         //  fprintf(traceptr,"%s    ", strdate(LOGFILETIME) );
         //  fprintf(traceptr,"%02hhx ", read_buf[4]);
         //  fprintf(traceptr,"\n");
           fprintf(traceptr,"%s    ", strdate(LOGFILETIME) );
            for (i = 0; i < res ; i++)
            {
              fprintf(traceptr,"%02hhx ", read_buf[i]);
            }
            fprintf(traceptr,"\n");
         }

         if (trace >= 1 && read_buf[1] != 0x04 )
         {
            fprintf(traceptr,"%s    ", strdate(LOGFILETIME) );
            for (i = 0; i < res ; i++)
            {
              fprintf(traceptr,"%02hhx ", read_buf[i]);
            }
            fprintf(traceptr,"\n");
         }

         if ( read_buf[1] == 0x18 )  //cpu_id
         {
	          logMsg(LEVEL8,"%s  cpu_id %02hhx %02hhx %02hhx \n" ,strdate(LOGFILETIME) ,read_buf[11],read_buf[12],read_buf[13]);
	       if ( cpu_id == 0 && read_buf[11] != 0xFF && read_buf[12] != 0xFF && read_buf[13] != 0xFF && read_buf[14] != 0xFF && read_buf[15] != 0xFF && read_buf[16] != 0xFF )
	       {
	          cpu_id = (int) read_buf[11] + (int) read_buf[12] * 256 + (int) read_buf[13] * 256 * 256 ;
	          logMsg(LEVEL8,"%s  cpu_id %02hhx %02hhx %02hhx \n" ,strdate(LOGFILETIME) ,read_buf[11],read_buf[12],read_buf[13]);
	          logMsg(LEVEL1,"%s  Artech Device cpu_id: %d \n" ,strdate(LOGFILETIME), cpu_id);
               }
	       continue ;

        }
        if ( read_buf[1] == 0x0b )  // cpu_version
        {

	       if ( strcmp(cpu_version ,"") == 0 && read_buf[10] != 0xFF && read_buf[11] != 0xFF && read_buf[12] != 0xFF   )
	       {
	          sprintf(cpu_version,"%d.%d.%d",read_buf[10],read_buf[11],read_buf[12]);
	          logMsg(LEVEL1,"%s  Artech Device cpu_version: %s \n" ,strdate(LOGFILETIME), cpu_version);
	       }
	       continue ;
        }

	 if (read_buf[4] != old_buf_4 )
	 {

	   /* Line voltage seems to be encoded in read_buf[4] (and maybe read_buf[5] ) */
	   /* value on Hook is arround -48 V  or +48V (+/- 2V )  which is consistent with POT specifications */

	   old_voltage = voltage ;
	   volt_int=  (((short)read_buf[4]) << 8 | read_buf[5]) ;
	   voltage  = (int) (volt_int)  /1000 ;
	   voltage_abs = abs(voltage) ;
	   old_polarity = polarity ;
	   // we don't want to crash with divide by zero .....
	   if ( voltage_abs != 0 )
	   {
	     polarity = voltage / voltage_abs  ;
	   }
	   else
	   {
	     polarity = 1 ;
	   }
	   if ( voltage != 0 )
	   {
	     voltage = ( polarity * 64 ) -  voltage ;
	   }
	   voltage_abs = abs(voltage) ;
           if ( voltage != old_voltage )
	   {
             logMsg(LEVEL7,"%s voltage =%d V  polarity = %d \n",strdate(LOGFILETIME) , voltage , polarity  );
	   }
	   if ( old_polarity !=  polarity )
	   {
	     polarity_change=TRUE;
             logMsg(LEVEL5,"%s Polarity change !!    \n", strdate(LOGFILETIME) , polarity );
             OFF_LINE_counter=0;
             ON_HOOK_counter=0;
             OFF_HOOK_counter=0;
              gettimeofday(&now, NULL);
              uSecs=timeval_diff(&interval,&now , &last_polarity_change_time );
              if ( uSecs < RING_PERIOD ) // ( ring period < 15000 usec  (default )
	      {
	        ring_detect = ring_detect + 1 ;
                logMsg(LEVEL9,"%s On polarity change !!  ring_detect = %d  ring period  <  %d uSec   \n", strdate(LOGFILETIME) , ring_detect , RING_PERIOD);
	      }
	      else
	      {
	        ring_detect = 0 ;
	      }
              gettimeofday(&last_polarity_change_time, NULL);
	   }
	   else
	   {
	     polarity_change=FALSE;
	     if ( ring_detect != 0)
	     {
	       gettimeofday(&now, NULL);
               uSecs=timeval_diff(&interval,&now , &last_polarity_change_time );
	       if ( uSecs > 3 * RING_PERIOD )
	       {
	         ring_detect = 0 ;
	       }
	     }

	   }

	   old_buf_4 = read_buf[4] ;
         }
         if (voltage_abs >=  MIN_ON_HOOK && voltage_abs  <=  MAX_ON_HOOK && line_state != ON_HOOK   )
         {
	      OPEN_LOOP_DISCONNECT = 0;
              if (line_state == RING  )
                 {
                   ON_HOOK_counter++ ;
                   gettimeofday(&now, NULL);
                   uSecs=timeval_diff(&interval,&now ,&last_ring_time);
                   logMsg(LEVEL9,"%s line state is RING , we test for  RING_STOPPED uSecs=%lld  ring_detect=%d  \n", strdate(LOGFILETIME), uSecs,  ring_detect);
                   if ( uSecs > 8000000 ) //  8  sec without ring  (ringing stopped) Default = 8000000
                   {
                      OFF_HOOK_counter=0;
                      ON_HOOK_counter=0;
		      OFF_LINE_counter=0;
                      line_state = RING_STOPPED;
                      logMsg(LEVEL3,"%s line state changed  to RING_STOPPED  \n", strdate(LOGFILETIME));
                      if (  trace >= 1 )
                      {
                       fprintf(traceptr,"%s    ", strdate(LOGFILETIME) );
                       fprintf(traceptr,"%02hhx ", read_buf[4]);
                       fprintf(traceptr," line state changed to RING_STOPPED  ");
                       fprintf(traceptr,"\n");
                     }
                   }
                 }
                 else
                 {
		   if (line_state != ON_HOOK  )
                   {
                   if (ON_HOOK_counter >= ON_HOOK_LIMIT )
                   {
                     if (  trace >= 1 )
                     {
                       fprintf(traceptr,"%s    ", strdate(LOGFILETIME) );
                       fprintf(traceptr,"%02hhx ", read_buf[4]);
                       fprintf(traceptr," line state changed to ON_HOOK ");
                       fprintf(traceptr,"\n");
                     }
                     line_state = ON_HOOK ;
		     hook_flash = 0 ;
                     strcpy (dialled_number2, "");
                    // logMsg(LEVEL5,"%s  (resetting  hook_flash to 0 ) since line state changed  to ON_HOOK ; called_number_HF =  \n" , strdate(LOGFILETIME) , called_number_HF);
                     logMsg(LEVEL3,"%s line state changed to ON_HOOK   \n" , strdate(LOGFILETIME));
                     if (  0  != strncmp(incoming_call_start_time ,"",sizeof(incoming_call_start_time)))
                     // "" means no incoming call
                        {
                           if (  0  != strncmp(incoming_call_pickup_time ,"-",sizeof(incoming_call_pickup_time)))
                           // "-" means incoming call  ringing , but not yet picked up
                             {
                               incomingCallInfo( BYE ,calling_number , calling_name);
                               /*call was ringing , then was picked up */
                             }
                           else
                             {
                               incomingCallInfo( CANCEL ,calling_number , calling_name);
                               /*call was ringing but  was not  picked up */
                             }
                           /* reset everything */
                           strncpy(incoming_call_start_time,"",sizeof(incoming_call_start_time));
                           strncpy(incoming_call_pickup_time,"",sizeof(incoming_call_pickup_time));
                           strncpy(incoming_call_date_time,"",sizeof(incoming_call_date_time));
                           strncpy(calling_number,"",sizeof(calling_number));
                           strncpy(calling_name,"",sizeof(calling_name));
                        }
                     if (  0  != strncmp( called_number,"",sizeof(called_number)))
                     /*called_number == ""  means no outgoing call got through
		     (if even an outgoing  call was attempted ?) */
                     {
                             if (  0  != strncmp(outgoing_call_pickup_time ,"-",sizeof(outgoing_call_pickup_time)))
                             /* outgoing_call_pickup_time == "-" means dialled call did not  connect through
			        (not picked up by callee , or busy ,etc ) */
                             {
                                outgoingCallInfo( BYE  , called_number);
                             }
                             else
                             {
			        /* not sure if we will ever get here but it does not hurt */
                                outgoingCallInfo( CANCEL , called_number );
                             }
                     }
		     else
		     {
		        if ( 0  != strncmp( dialled_number,"",sizeof(dialled_number)))
			/* a number was dialled but we hung up before the call got through */
			/* or it was not answered ...*/
			{
                                outgoingCall( dialled_number );
                                outgoingCallInfo( CANCEL , dialled_number );
			}
		     }

                     if (  0  != strncmp(incoming_call2_start_time ,"",sizeof(incoming_call2_start_time)))
                     // "" means no 2nd incoming call  (WID)
                        {
                           if (  0  != strncmp(incoming_call2_pickup_time ,"-",sizeof(incoming_call2_pickup_time)))
                           // "-" means incoming 2nd call (WID) waiting , but not yet picked up
                             {
                               incomingCallInfo2( BYE ,calling_number2 , calling_name2);
                               /*2nd call (WID) was accepted  */
                             }
                           else
                             {
                               incomingCallInfo( CANCEL ,calling_number , calling_name);
                               /*2nd call (WID)  was not accepted or timed out  */
                             }
                           /* reset everything */
                           strncpy(incoming_call2_start_time,"",sizeof(incoming_call2_start_time));
                           strncpy(incoming_call2_pickup_time,"",sizeof(incoming_call2_pickup_time));
                           strncpy(incoming_call2_date_time,"",sizeof(incoming_call2_date_time));
                           strncpy(calling_number2,"",sizeof(calling_number2));
                           strncpy(calling_name2,"",sizeof(calling_name2));
                        }
                     if (  0  != strncmp(outgoing_call2_start_time ,"",sizeof(outgoing_call_start_time)))
                     /* "" means no 2nd outgoing call  */
                     {
                             if (  0  != strncmp(outgoing_call2_pickup_time ,"-",sizeof(outgoing_call2_pickup_time)))
                             /* "-" means dialled call did not  connect through (not picked up by callee , or busy ,etc ) */
                             {
                                outgoingCallInfo2( BYE  , called_number2);
                             }
                             else
                             {
                                outgoingCallInfo2( CANCEL , called_number2 );
                             }
                     }


                     strncpy (incoming_call_start_time, "",sizeof(incoming_call_pickup_time));
                     strncpy(incoming_call_pickup_time,"",sizeof(incoming_call_pickup_time));
		     strncpy(incoming_call_date_time,"",sizeof(incoming_call_date_time));

                     strncpy (incoming_call_start_time, "",sizeof(incoming_call2_pickup_time));
                     strncpy(incoming_call2_pickup_time,"",sizeof(incoming_call2_pickup_time));
		     strncpy(incoming_call2_date_time,"",sizeof(incoming_call2_date_time));

		     strncpy (outgoing_call_start_time, "",sizeof(incoming_call_pickup_time));
                     strncpy(outgoing_call_date_time,"",sizeof(outgoing_call_date_time));
		     strncpy(outgoing_call_pickup_time,"",sizeof(outgoing_call_pickup_time));

		     strncpy (outgoing_call2_start_time, "",sizeof(incoming_call2_pickup_time));
		     strncpy(outgoing_call2_pickup_time,"",sizeof(outgoing_call2_pickup_time));
                     strncpy(outgoing_call2_date_time,"",sizeof(outgoing_call2_date_time));

		     strcpy (dialled_number, "");
                     strcpy (called_number, "");
                     strcpy (calling_number, "");
                     strcpy (calling_name, "");
                     strcpy (dialled_number2, "");
                     strcpy (called_number2, "");
                     strcpy (calling_number2, "");
                     strcpy (calling_name2, "");
                     OFF_HOOK_counter = 0 ;
		     OFF_LINE_counter=0;
                   }
		   else
		   {
		     ON_HOOK_counter++ ;
                     if (  trace >= 1 )
                     {
                        fprintf(traceptr,"%s    ", strdate(LOGFILETIME) );
                        fprintf(traceptr,"%02hhx ", read_buf[4]);
                        fprintf(traceptr,"ON_HOOK ?? counter=%d",ON_HOOK_counter);
                        fprintf(traceptr,"\n");
                      }
		   }
		  }
		  else  // line_state = ON_HOOK
		  {
                   if (ON_HOOK_counter >= ON_HOOK_LIMIT )
                   {
                     OFF_HOOK_counter = 0 ;
		     OFF_LINE_counter=0;
		   }

		  }
                }
         }

	 if (  voltage_abs >= MIN_ON_HOOK  &&  voltage_abs  <= MAX_ON_HOOK   && line_state == ON_LINE  )
	 {
	      OPEN_LOOP_DISCONNECT = 0;
              logMsg(LEVEL9,"%s  hook flash (?) (voltage == on_hook level)  voltage = %2d  hook_flash_counter=%d  \n" , strdate(LOGFILETIME),voltage  , hook_flash_counter );

	 }
	 else
	 {
	        if (hook_flash == 1 ||  hook_flash == 2 )
		{
		  gettimeofday(&now, NULL);
                  uSecs=timeval_diff(&interval,&now ,&hook_flash_time);
		  if ( uSecs > 1000000  && hook_flash != 0 ) // 0.5  Sec max between end of hook flash and second polarity change .
                   {
                     hook_flash = 0 ;
                     strcpy (called_number_HF, dialled_number2);
                     logMsg(LEVEL5,"%s 1 seconds since hook flash (and no polarity change yet ) --> resetting hook flash ;  hook_flash  = 0 \n" , strdate(LOGFILETIME)  );
		     strcpy (dialled_number2, "");
                     strcpy (called_number_HF, "");
                   }
		}
	        hook_flash_counter= 0 ;
	 }


	 if (strlen(dialled_number2) >  1  && hook_flash == 4 && line_state == ON_LINE )
	 {
                   gettimeofday(&now, NULL);
                   uSecs=timeval_diff(&interval,&now ,&start_dial_digit);
                   if ( uSecs > 3000000  && hook_flash != 0 ) //  more than 3  sec since last dialled digit
		   {
	             strcpy (called_number_HF, dialled_number2);
                     logMsg(LEVEL5,"%s call1_state=%d  call2_state=%d   \n", strdate(LOGFILETIME) , call1_state , call2_state );
                     logMsg(LEVEL5,"%s 3 seconds since last dialled digit (calling 2nd  number) : processing  hook flash ; called_number_HF = %s  , read_buf[6] = %02hhx  ; read_buf[7] = %02hhx  ; read_buf[8] = %02hhx \n" , strdate(LOGFILETIME) , called_number_HF,read_buf[6],read_buf[7] ,read_buf[8] );
		     if ( 0 == strcmp( HOOK_FLASH_ZONE , "EU" ) )
		     {
	               process_flash_hook_EU(called_number_HF);
		     }
                     else // "US" is default
		     {
		       process_flash_hook_US(called_number_HF);
		     }
	             hook_flash = 0 ;
                     logMsg(LEVEL5,"%s resetting hook flash ;  hook_flash  = 0  ;   \n", strdate(LOGFILETIME) );
                     strcpy (dialled_number2, "");
		  }
         }

	 if ( strlen(dialled_number2) == 1 &&  call1_state != OFF && call2_state != OFF  && hook_flash == 4 && line_state == ON_LINE && 0 == strcmp( HOOK_FLASH_ZONE , "EU" ) )
	 {
	    strcpy (called_number_HF, dialled_number2);
            logMsg(LEVEL5,"%s call1_state=%d  call2_state=%d   \n", strdate(LOGFILETIME) , call1_state , call2_state );
            logMsg(LEVEL5,"%s Only  one digit dialled (command) + 0.5 sec (2 calls in progress)  -->processing  hook flash  ;  called_number_HF = %s  , read_buf[6] = %02hhx  ; read_buf[7] = %02hhx  ; read_buf[8] = %02hhx \n" , strdate(LOGFILETIME) , called_number_HF,read_buf[6],read_buf[7] ,read_buf[8] );
	    process_flash_hook_EU(called_number_HF);
	    hook_flash = 0 ;
            logMsg(LEVEL5,"%s resetting hook flash ;  hook_flash  = 0  ;   \n", strdate(LOGFILETIME) );
            strcpy (dialled_number2, "");
	 }

         if (   hook_flash == 3 )  // reset stall hook flash
	 {
            gettimeofday(&now, NULL);
            uSecs=timeval_diff(&interval,&now ,&hook_flash_time);
	    if ( uSecs > 16000000   ) // 16  Sec max between start of hook flash and DTMF event (if not we reset hook_flash . )
            {
              hook_flash = 0 ;
              logMsg(LEVEL5,"%s 16 seconds since hook flash and no DTMF (digit)  yet --> resetting hook flash ;  hook_flash  = 0 \n" , strdate(LOGFILETIME)  );
	      strcpy (dialled_number2, "");
              strcpy (called_number_HF, "");
            }
	 }

         if (  read_buf[6] == 0x32  &&  read_buf[7] == 0x01 && ( hook_flash == 3 ) && line_state == ON_LINE )   // DTMF　 event
	 {
	       hook_flash = 4 ;
               logMsg(LEVEL5,"%s !!!!!!  hook flash =%d !!!!! (pre hook_flash + two polarity changes + DTMF event(s) \n" , strdate(LOGFILETIME), hook_flash );
	 }


         if (   polarity_change == 1    && (  hook_flash  == 1 ||  hook_flash  == 2 )  && line_state == ON_LINE  )
	 {
	         hook_flash ++ ;
                 strcpy (dialled_number2, "");  // just to be safe ....
		 if ( 0 != strcmp( HOOK_FLASH_ZONE , "EU" ) && hook_flash == 3 &&  call1_state != OFF && call2_state != OFF )  //HOOK_FLASH_ZONE  is  NOT  "EU" (default to "US" )
		 {
                   logMsg(LEVEL5,"%s  HOOK_FLASH_ZONE == 'US' , No digits dialled , (2 calls are in progress)  --> processing  hook flash ;\n" , strdate(LOGFILETIME) );
		   process_flash_hook_US(called_number_HF);
	           hook_flash = 0 ;
                   logMsg(LEVEL5,"%s resetting hook flash ;  hook_flash  = 0  ;   \n", strdate(LOGFILETIME) );
                   strcpy (dialled_number2, "");
                 }
		 else
		 {
                   logMsg(LEVEL5,"%s !!!!!!  hook flash =%d (pre hook_flash + %d  polarity change(s) !!!!! \n" , strdate(LOGFILETIME), hook_flash , hook_flash -1 );
	         }
	 }



	 if ( hook_flash_counter >= HOOK_FLASH_LIMIT )
	 {
           if (hook_flash == 0)
	   {
             hook_flash  = 1 ;// pre hook flash
	     gettimeofday(&hook_flash_time, NULL);
             logMsg(LEVEL5,"%s  hook flash =%d (pre-hook flash)!!!!! \n" , strdate(LOGFILETIME), hook_flash );
           }
	 }
	 else
	 {
	   hook_flash_counter ++ ;
	 }







	 if (  voltage_abs  >= MIN_OFF_HOOK  &&  voltage_abs  <= MAX_OFF_HOOK   && line_state != ON_LINE )
           {
	    OPEN_LOOP_DISCONNECT = 0 ;
            if (line_state == OFF_HOOK  )
	    {
	      ON_HOOK_counter = 0 ;
              OFF_LINE_counter=0;
	      logMsg(LEVEL9,"%s  x55_counter %d  \n" , strdate(LOGFILETIME) , x55_counter);
              if ( x55_counter >= 1600  )
	      {
	         line_state = BEEP_BEEP ;
                 logMsg(LEVEL3,"%s  line state changed to BEEP_BEEP \n" , strdate(LOGFILETIME));
	      }
	    }
	    else
	    {
	      if ( line_state == BEEP_BEEP )
	      {
	      ON_HOOK_counter = 0 ;
	      OFF_HOOK_counter = 0 ;
              OFF_LINE_counter=0;
              }
	      else /* line_state not  OFF_HOOK nor BEEP_BEEP */
	      {
                 if (OFF_HOOK_counter >= OFF_HOOK_LIMIT )
                 {
		   if (line_state == RING )
		    {
                      gettimeofday(&now, NULL);
                      uSecs=timeval_diff(&interval,&now ,&last_ring_time);
                      if ( uSecs > RING_GAP ) //  2  sec without ring  (phone was picked up ) (= 2000000 usec default)
                      {
		        line_state = ON_LINE;
			call1_state = ON;
                        logMsg(LEVEL3,"%s  line state changed to ON_LINE (after RING) \n" , strdate(LOGFILETIME));
                        logMsg(LEVEL5,"%s call1_state=%d  call2_state=%d   \n", strdate(LOGFILETIME) , call1_state , call2_state );
                        if ( 0 == strncmp(incoming_call_pickup_time,"-",sizeof(incoming_call_pickup_time)))
                        {
                           logMsg(LEVEL4, "%s line state changing to ON_LINE  , incoming_call_pickup_time = 0 (it means it's  ringing and we are picking  up...) \n",strdate(LOGFILETIME));
                           strncpy(incoming_call_pickup_time,strdate(WITHSEP),sizeof(incoming_call_pickup_time));
	                   call1_state= ON ;
                           logMsg(LEVEL4, "%s line state changing to ON_LINE  ,we set  incoming_call_pickup_time = %s \n",incoming_call_pickup_time,strdate(LOGFILETIME));
                        }
                        ON_HOOK_counter = 0 ;
                        OFF_LINE_counter=0;
		        ring_detect=0;
		      }
		    }
		    else
		    {
                      line_state = OFF_HOOK ;
                      logMsg(LEVEL3,"%s  line state changed to OFF_HOOK \n" , strdate(LOGFILETIME));
                      ON_HOOK_counter = 0 ;
                      OFF_LINE_counter=0;
		      /* We are now OFF_HOOK , we will very probably start dialling soon
		         So we set  outgoing_call_date_time and outgoing_call_start_time now */
		      outgoing_call_date_time[0] = 0;
                      strncpy(outgoing_call_date_time,strdate(NOYEAR),sizeof(outgoing_call_date_time));
                      logMsg(LEVEL4, "outgoing_call_date_time: \n", outgoing_call_date_time);
                      strncpy(outgoing_call_start_time,strdate(WITHSEP),sizeof(outgoing_call_start_time));
                      strncpy(outgoing_call_pickup_time,"-",sizeof(outgoing_call_start_time));

		      ring_detect=0;
                      if (  trace >= 1 )
                      {
                        fprintf(traceptr,"%s    ", strdate(LOGFILETIME) );
                        fprintf(traceptr,"%02hhx ", read_buf[4]);
                        fprintf(traceptr," line state changed to OFF_HOOK  ") ;
                        fprintf(traceptr,"\n");
                      }
		    }
                 }
	         else
	         {
                    OFF_HOOK_counter++ ;
                    if (  trace >= 1 )
                    {
                       fprintf(traceptr,"%s    ", strdate(LOGFILETIME) );
                       fprintf(traceptr,"%02hhx ", read_buf[4]);
                       fprintf(traceptr,"OFF_HOOK ?? counter=%d",OFF_HOOK_counter);
                       fprintf(traceptr,"\n");
                    }
	         }
	       }
	     }
            }
         // polarity change triggers the state change to ON_LINE (LOOP_START) , however lopp start is not always available ; In that case we do speech detection (NON_HEX55 )
	 // see https://en.wikipedia.org/wiki/Loop_start
	 if ( ( LOOP_START == 1  &&  polarity_change==TRUE   &&  line_state == OFF_HOOK && 0 != strcmp(dialled_number,"") )  || ( LOOP_START == 0  &&  non_hex55_counter  > NON_HEX55_LIMIT   &&  line_state == OFF_HOOK && 0 != strcmp(dialled_number,"") ) )   // remote party picks  up the call ....
            {
              if (  line_state == OFF_HOOK  &&  0 != strcmp(dialled_number,"")  )  //  we have been  dialling , i.e. calling OUT and are just now having  connection event (going from OFF_HOOK to ON_LINE)
              {
                 logMsg(LEVEL2, "%s dialled number = %s \n",strdate(LOGFILETIME) , dialled_number);
                 strncpy(called_number ,dialled_number ,sizeof(called_number));
                 strncpy(dialled_number ,"",sizeof(dialled_number));
                 call1_state= ON ;
                 outgoingCall(called_number );   //send called number to ncidd
	      }
              if ( line_state == OFF_HOOK &&  0 != strcmp ( called_number , "" ) )
              {
		call1_state= ON ;
                strncpy(outgoing_call_pickup_time,strdate(WITHSEP),sizeof(outgoing_call_pickup_time));
              }
              if (  trace >= 1 )
              {
                fprintf(traceptr,"%s    ", strdate(LOGFILETIME) );
                fprintf(traceptr,"%02hhx ", read_buf[4]);
                fprintf(traceptr," line state changed to ON_LINE  ") ;
                fprintf(traceptr,"\n");
              }
              line_state = ON_LINE ;
              logMsg(LEVEL3,"%s  line state changed to ON_LINE \n" , strdate(LOGFILETIME));
              ON_HOOK_counter = 0 ;
              OFF_HOOK_counter = 0 ;
              OFF_LINE_counter=0;
            }

	 if (  voltage_abs >= MIN_OFF_HOOK  &&  voltage_abs <= MAX_OFF_HOOK  &&  line_state == RING && OFF_HOOK_counter > OFF_HOOK_LIMIT )   // we have just picked a ringing phone . ...
   	 {
	        OPEN_LOOP_DISCONNECT = 0 ;
                gettimeofday(&now, NULL);
                uSecs=timeval_diff(&interval,&now ,&last_ring_time);
                if ( uSecs > 2000000 ) //  2  sec without ring  (phone was picked up )
                {
                  OFF_HOOK_counter=0;
	          line_state = ON_LINE;
                  if ( 0 == strncmp(incoming_call_pickup_time,"-",sizeof(incoming_call_pickup_time)))
                  {
                    logMsg(LEVEL3, "%s line state changing to ON_LINE (after RING and pickup ) , incoming_call_pickup_time = 0 (it means it's  ringing and we are picking  up...) \n",strdate(LOGFILETIME));
                    strncpy(incoming_call_pickup_time,strdate(WITHSEP),sizeof(incoming_call_pickup_time));
	            call1_state= ON ;
                    logMsg(LEVEL3, "%s line state changing to ON_LINE  ,we set  incoming_call_pickup_time = %s \n",incoming_call_pickup_time,strdate(LOGFILETIME));
                  }
                  if (  trace >= 1 )
                  {
                    fprintf(traceptr,"%s    ", strdate(LOGFILETIME) );
                    fprintf(traceptr,"%02hhx ", read_buf[4]);
                    fprintf(traceptr," line state changed to ON_LINE (after RING and pickup )   ") ;
                    fprintf(traceptr,"\n");
                  }
		  hook_flash = 0 ;
                  OFF_HOOK_counter=0;
                  ON_HOOK_counter=0;
		  ring_detect=0;
		  call1_state=OFF;
		  call2_state=OFF;
		}
          }
          if ( voltage_abs == VAL_OFF_LINE )
          {
          // counters are for stability of line state ....
            if (line_state != OFF_LINE )
            {
	      if (OFF_LINE_counter > OPEN_LOOP_DISCONNECT_LIMIT)
	      {
	         OPEN_LOOP_DISCONNECT= 1 ;
                 logMsg(LEVEL3,"%s Got an OPEN_LOOP_DISCONNECT (briefly no voltage) \n",strdate(LOGFILETIME));
              }
              if (OFF_LINE_counter > OFF_LINE_LIMIT)
              {
                if (  trace >= 1 )
                {
                  fprintf(traceptr,"%s    ", strdate(LOGFILETIME) );
                  fprintf(traceptr,"%02hhx ", read_buf[4]);
                  fprintf(traceptr," line state changed to OFF_LINE (disconnected) (no voltage)  ") ;
                  fprintf(traceptr,"\n");
                }
                line_state = OFF_LINE;
		hook_flash = 0 ;
                logMsg(LEVEL3,"%s line_state changed to OFF_LINE (disconnected) (no voltage) \n",strdate(LOGFILETIME));
                OFF_HOOK_counter=0;
                ON_HOOK_counter=0;
		ring_detect=0;
		call1_state=OFF;
		call2_state=OFF;
              }
	      else
	      {
                OFF_LINE_counter ++ ;
	      }
            }
          }
          else
          {
	    if ( ring_detect > 40   )    // our Phone is Ringing ........
            {
	       if ( line_state == ON_HOOK )
	       {
                 line_state = RING ;
                 logMsg(LEVEL3,"%s line state changed  to   RING \n",strdate(LOGFILETIME));
                 if (  trace >= 1 )
                 {
                    fprintf(traceptr,"%s    ", strdate(LOGFILETIME) );
                    fprintf(traceptr,"%02hhx ", read_buf[4]);
                    fprintf(traceptr," line state changed to RING  ") ;
                    fprintf(traceptr,"\n");
                  }
               }
	       else
	       {
	         if ( line_state == RING )
                 {
                   gettimeofday(&last_ring_time, NULL);
                   logMsg(LEVEL9,"%s RING continues ...\n", strdate(LOGFILETIME) );
	         }
	       }
	    }
        }
        if (  read_buf[1] == 0x07 && read_buf[6] == 0x32  &&  read_buf[7] == 0x01 )   // DTMF　 event
        {

        switch (line_state)
        {
        case OFF_LINE :
          // do nothing
          break;
        case BEEP_BEEP:
          // do nothing
          break;
        case RING:
          // do nothing
          break;
        case OFF_HOOK :
          //parsing dialling procedure
          if ( read_buf[6] == 0x32  && read_buf[1] == 0x07)   // DTMF  detected
          {
            logMsg(LEVEL5,"%s Parse dialling  ; read_buf[7]=%02hhx read_buf[8]=%02hhx  \n" ,strdate(LOGFILETIME ) ,read_buf[7] , read_buf[8] ) ;
                    switch (read_buf[8])
                    {
                      case 0x0A :
                      strncat(dialled_number , "0",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x01 :
                      strncat(dialled_number , "1",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x02 :
                      strncat(dialled_number , "2",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x03 :
                      strncat(dialled_number , "3",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x04 :
                      strncat(dialled_number , "4",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x05 :
                      strncat(dialled_number , "5",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x06 :
                      strncat(dialled_number , "6",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x07 :
                      strncat(dialled_number , "7",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x08 :
                      strncat(dialled_number , "8",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x09 :
                      strncat(dialled_number , "9",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x0B :
                      strncat(dialled_number , "*",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x0C :
                      strncat(dialled_number , "#",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      default:
                      // should never get there
                      logMsg(LEVEL5,"%s UNKNOWN DIGIT read_buf[8]=　%02hhx",strdate(LOGFILETIME),read_buf[8]);
                      break;
                    }
                    x55_counter=0 ;
                    logMsg(LEVEL5,"%s dialled_number = %s \n",strdate(LOGFILETIME),dialled_number);
           }
        break;
        case ON_HOOK:
            logMsg(LEVEL5,"%s Parse dialling (but ON_HOOK) !!!! ; read_buf[7]=%02hhx read_buf[8]=%02hhx  \n" ,strdate(LOGFILETIME ) ,read_buf[7] , read_buf[8] ) ;
           break;
        case ON_LINE:
            logMsg(LEVEL5,"%s Parse dialling (but ON_LINE) !!!! ; read_buf[7]=%02hhx read_buf[8]=%02hhx  \n" ,strdate(LOGFILETIME ) ,read_buf[7] , read_buf[8] ) ;
          //parsing dialling procedure
          if ( read_buf[6] == 0x32  && read_buf[1] == 0x07 && hook_flash > 0 )   // DTMF  detected
          {
            logMsg(LEVEL5,"%s Parse dialling  ; read_buf[7]=%02hhx read_buf[8]=%02hhx  \n" ,strdate(LOGFILETIME ) ,read_buf[7] , read_buf[8] ) ;
                    switch (read_buf[8])
                    {
                      case 0x0A :
                      strncat(dialled_number2 , "0",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x01 :
                      strncat(dialled_number2 , "1",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x02 :
                      strncat(dialled_number2 , "2",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x03 :
                      strncat(dialled_number2 , "3",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x04 :
                      strncat(dialled_number2 , "4",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x05 :
                      strncat(dialled_number2 , "5",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x06 :
                      strncat(dialled_number2 , "6",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x07 :
                      strncat(dialled_number2 , "7",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x08 :
                      strncat(dialled_number2 , "8",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x09 :
                      strncat(dialled_number2 , "9",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x0B :
                      strncat(dialled_number2 , "*",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      case 0x0C :
                      strncat(dialled_number2 , "#",2);
                      // reset start time
                      gettimeofday(&start_dial_digit, NULL);
                      break;
                      default:
                      // should never get there
                      logMsg(LEVEL5,"%s UNKNOWN DIGIT read_buf[8]=　%02hhx",strdate(LOGFILETIME),read_buf[8]);
                      break;
                    }
                    logMsg(LEVEL5,"%s dialled_number2 = %s \n",strdate(LOGFILETIME),dialled_number2);

	   }
           break;
         }
        }
         /*------------------------------------------------------------------*/
         /*------------------------------------------------------------------*/
          if (read_buf[1] == 0x07 )
          {
            if (   read_buf[6] == 0x31 && read_buf[7] == 0x01 )
            {
               //  just logging x55  (hex55) ************************
	       if (   read_buf[8] == 0x55 )
	       {
	         non_hex55_counter = 0 ;
               }
	       else
	       {
	         non_hex55_counter++ ;
	       }
	       if ( non_hex55_counter != old_non_hex55_counter &&  line_state == OFF_HOOK  )
	       {
                 logMsg(LEVEL4,"%s non_hex55_counter = %d )\n" ,strdate(WITHSEP), non_hex55_counter);
	       }
	       old_non_hex55_counter = non_hex55_counter ;
               //*****************************************************

              // catching CID  : we wait for a long sequence of x55 in read_buf[8]
              //if ( x55_counter >=9  && line_state != OFF_HOOK && line_state != OFF_LINE)
              if ( x55_counter >=9  )
              // we already got the long x55 sequence
              // we now try to catch CID
              {
                logMsg(LEVEL9,"%s x55_counter >=  9 , ready to catch CID \n" ,strdate(WITHSEP));
                   logMsg(LEVEL9," CID_byte_count=%d  CID_byte_length=%d \n" , CID_byte_count , CID_byte_length);
               logMsg(LEVEL9," started_reading_CID=%d  read_buf[8] =%02hhx  read_buf[6] =%02hhx  read_buf[7] =%02hhx\n" , started_reading_CID,read_buf[8]  , read_buf[6]  , read_buf[7]);
                if (   read_buf[8] != 0x55   ||  started_reading_CID==TRUE || CID_offset > 0   )
                {
                   if (CID_byte_count < CID_byte_length && CID_byte_count < 100  )
                   /* we catch a max of 100 bytes , we will sort it out later */
                   {
                     logMsg(LEVEL9," read_buf[1]=%02hhx read_buf[8]=%02hhx   ; \n" , read_buf[1] , read_buf[8] ) ;
                     CID_hex[CID_byte_count]=read_buf[8];
                     if ( CID_byte_count == 0 )
                     {
                       switch (CID_hex[0])
                       {
                         case 0x80:
                           started_reading_CID=TRUE ;
                           /* good ! start of MDMF message "call setup" */
                         break;
                         case 0x82:
                           started_reading_CID=TRUE ;
                           /* good ! start of MDMF message "waiting indicator" */
                         break;
                         case 0x86:
                           started_reading_CID=TRUE ;
                           /* good ! start of MDMF message "advice of charge" */
                           logMsg(LEVEL2," start of MDMF message 'advice of charge' (not implemented ) \n" );
                           x55_counter=0 ;
                           started_reading_CID=FALSE ;
                           CID_byte_count=0;
                           CID_byte_length=100;
                           strncpy(CID_as_a_string,"",sizeof(CID_as_a_string));
                           strncpy(byte_as_a_string,"",sizeof(byte_as_a_string));
                           memset(CID_hex,0x00,sizeof(CID));
                         break;
                         case 0x89:
                           /* good ! start of MDMF message "Short Message Service" */
                           logMsg(LEVEL2," start of MDMF message 'Short Message Service' (not implemented ) \n" );
                           x55_counter=0 ;
                           started_reading_CID=FALSE ;
                           CID_byte_count=0;
                           CID_byte_length=100;
                           strncpy(CID_as_a_string,"",sizeof(CID_as_a_string));
                           strncpy(byte_as_a_string,"",sizeof(byte_as_a_string));
                           memset(CID_hex,0x00,sizeof(CID));
                         break;
                         case 0x04:
                           started_reading_CID=TRUE ;
                           /* good ! start of SDMF message signaling */
                         break;
                         default:
			   /*sometimes , there is some garbage after the x55 sequence ( Todd's line)  , in that case  we seek , up to
			    * 5 bytes further to find a good CID　start marker */
			   if ( CID_offset < 5 )
			   {
                             CID_byte_count=0;
                             CID_byte_length=100;
			     CID_offset ++ ;
                             started_reading_CID=FALSE ;
                             logMsg(LEVEL9," (not the start of SDNS or MDMS) (garbage) ( --> we seek one byte further ( max 5 ) ) ,started_reading_CID=%d  CID_offset=%d read_buf[8] =%02hhx  read_buf[6] =%02hhx  read_buf[7] =%02hhx\n" , started_reading_CID , CID_offset ,read_buf[8]  , read_buf[6]  , read_buf[7]);
                           }
                           else
                           {
                             /* CID_ofset is now exhausted ,  (not good) , we reset X55 counter  */
                             x55_counter=0 ;
                             started_reading_CID=FALSE ;
                             CID_byte_count=0;
                             CID_byte_length=100;
			     CID_offset=0;
                             strncpy(CID_as_a_string,"",sizeof(CID_as_a_string));
                             strncpy(byte_as_a_string,"",sizeof(byte_as_a_string));
                             memset(CID_hex,0x00,sizeof(CID));
                             logMsg(LEVEL9," resetting ,x55_counter (not the start of SDNS or MDMS) (and CID_offset is exhausted ) ,started_reading_CID=%d  read_buf[8] =%02hhx  read_buf[6] =%02hhx  read_buf[7] =%02hhx\n" , started_reading_CID,read_buf[8]  , read_buf[6]  , read_buf[7]);
			   }
                           break;
                       }
                     }
                     if ( CID_byte_count == 1 )
                     {
                       CID_byte_length = CID_hex[CID_byte_count] + 3 ;
                       logMsg(LEVEL9,"  length of  CID (?) (%d)\n",CID_byte_length ) ;
                     }
                     if ( started_reading_CID == TRUE )  /*checking it is still at TRUE ....just in case the X55 counter got resetted in the switch 20 lines above */
                     {
                       CID_byte_count++;
                       sprintf(byte_as_a_string,"%02hhx",read_buf[8]);
                       strncat(CID_as_a_string,byte_as_a_string,2);
                       logMsg(LEVEL9,"  add a  byte  to CID (%d)\n",CID_byte_count - 1) ;
                     }
                   }
                   if (CID_byte_count >=  CID_byte_length  || CID_byte_count >= 100 )
                   {
                     logMsg(LEVEL9,"CID_byte_count >= CID_byte_length (%d) \n", CID_byte_count );
                     logMsg(LEVEL9," CID_as_a_string =  %s \n" , CID_as_a_string) ;
                     logMsg(LEVEL9," CID_hex[0] (header): %02hhx  ,CID_hex[1] (length):  %02hhx ,CID_hex[2]: %02hhx \n" ,  CID_hex[0] ,  CID_hex[1] ,  CID_hex[2]) ;
                     logMsg(LEVEL9," We start  processing up CID  \n");
                     if ( CID_hex[0] == 0x80 )
                     {
                       //MDMF
                       CID_byte_count =  CID_hex[1]  + 3 ;
                       logMsg(LEVEL5,"%s CID_byte_count=%02hhx \n CID_byte_count=%d \n",strdate(LOGFILETIME) ,CID_byte_count, CID_byte_count );
                       if(0 == xdmfCID(CID_hex ,CID_byte_count) )
                       {
                          logMsg(LEVEL2,"%s We got a good  CID  (MDMF)\n",strdate(LOGFILETIME));
                         incomingCall();
                       }
                     }
                     if ( CID_hex[0] == 0x82 )  // MDMF signaling
                     {
                       CID_byte_count =  CID_hex[1]  + 3 ;
                       logMsg(LEVEL9,"$%s CID_byte_count=%02hhx \n CID_byte_count=%d \n",strdate(LOGFILETIME),CID_byte_count, CID_byte_count );
                       if(0 == xdmfCID(CID_hex ,CID_byte_count) )
                       {
                          logMsg(LEVEL2,"%s We got a good  MWI or else   \n",strdate(LOGFILETIME));
                          incomingCall();
                       }
                     }

                     if ( CID_hex[0] == 0x04 ) // SDMF
                     {
                       CID_byte_count =  CID_hex[1]  + 1 ;
                       if(0 == xdmfCID(CID_hex ,CID_byte_count) )
                       {
                          logMsg(LEVEL2,"%s We got a good  CID  (SDMF)\n",strdate(LOGFILETIME));
                         incomingCall();
                       }
                     }
                     x55_counter=0 ;
                     started_reading_CID=FALSE ;
		     CID_offset=0;
		     /* adding this block (5 next lines) should solve the "garbage in NAME" problem (CID_hex was not re-zeroed after a CID decoding ) */
                     CID_byte_count=0;
                     CID_byte_length=100;
                     strncpy(CID_as_a_string,"",sizeof(CID_as_a_string));
                     strncpy(byte_as_a_string,"",sizeof(byte_as_a_string));
                     memset(CID_hex,0x00,sizeof(CID));
                     logMsg(LEVEL5,"$%s CID_hex is re-zeroed \n",strdate(LOGFILETIME) );
                   }
                 }
                 else
                 {
                   if ( read_buf[8]== 0x55   )
                   {
		    if ( CID_offset >=1 )
                      {
			CID_byte_count=0;
                        CID_byte_length=100;
			CID_offset ++ ;
                        started_reading_CID=FALSE ;
                        logMsg(LEVEL9,"(spurious x55) (not the start of SDNS or MDMS)  (garbage) ( --> we seek one byte further ( max 5 ) ) ,started_reading_CID=%d  CID_offset=%d read_buf[8] =%02hhx  read_buf[6] =%02hhx  read_buf[7] =%02hhx\n" , started_reading_CID , CID_offset ,read_buf[8]  , read_buf[6]  , read_buf[7]);
		        if ( CID_offset > 5 )
			{
                             x55_counter=0 ;
                             started_reading_CID=FALSE ;
                             CID_byte_count=0;
                             CID_byte_length=100;
			     CID_offset=0;
                             strncpy(CID_as_a_string,"",sizeof(CID_as_a_string));
                             strncpy(byte_as_a_string,"",sizeof(byte_as_a_string));
                             memset(CID_hex,0x00,sizeof(CID));
                             logMsg(LEVEL9,"%s (spurious x55) resetting ,x55_counter (not the start of SDNS or MDMS) (and CID_offset is exhausted ) ,started_reading_CID=%d  read_buf[8] =%02hhx  read_buf[6] =%02hhx  read_buf[7] =%02hhx\n" ,strdate(LOGFILETIME), started_reading_CID,read_buf[8]  , read_buf[6]  , read_buf[7]);
			}
                    }
		  else
		    {
		      x55_counter++ ;
                      CID_byte_count=0;
		      CID_offset=0;
                      logMsg(LEVEL9,"%s  x55_counter %d  line_state=%d  \n",strdate(LOGFILETIME) ,x55_counter,  line_state );
		    }
                   }
                 }
           }
           else
           {
             if ( read_buf[8]== 0x55   )
             {
               x55_counter++ ;
               CID_byte_count=0;
               logMsg(LEVEL9,"%s x55_counter %d  line_state=%d  \n",strdate(LOGFILETIME), x55_counter , line_state );
             }
             else
             {
             //reset x55_counter
               x55_counter=0 ;
               started_reading_CID=FALSE ;
               CID_byte_count=0;
               CID_byte_length=100;
	       CID_offset=0;
               strncpy(CID_as_a_string,"",sizeof(CID_as_a_string));
               strncpy(byte_as_a_string,"",sizeof(byte_as_a_string));
               memset(CID_hex,0x00,sizeof(CID));
               logMsg(LEVEL9,"%s resetting ,x55_counter  ,started_reading_CID=%d  read_buf[8] =%02hhx  read_buf[6] =%02hhx  read_buf[7] =%02hhx\n",strdate(LOGFILETIME) , started_reading_CID,read_buf[8]  , read_buf[6]  , read_buf[7]);
             }
          }
       }
       }
	           if (call2_state == WAIT )
		   {
                     gettimeofday(&now, NULL);
                     uSecs=timeval_diff(&interval,&now , &WID_start_time );
                     if ( uSecs > 18000000 ) //  18  sec (reset call2)
                     {
                       logMsg(LEVEL4, "Time out on WID call %s \n", calling_number2);
                       incomingCallInfo2( CANCEL ,calling_number2 , calling_name2);
		       call2_state=OFF;
		       strcpy(incoming_call2_start_time,"");
		       strcpy(incoming_call2_pickup_time,"");
		       strcpy(incoming_call2_end_time,"");
		       strcpy(calling_number2,"");
		       strcpy(calling_name2,"");
                       logMsg(LEVEL4, "Time out on WID call %s \n", calling_number2);
                       logMsg(LEVEL5,"%s call1_state=%d  call2_state=%d   \n", strdate(LOGFILETIME) , call1_state , call2_state );
	             }
                  }

	 gettimeofday(&now, NULL);
	 uSecs=timeval_diff(&interval,&now , &last_poke_time ) ;
	 if ( uSecs > 14920*1000 ) //  14.92  sec ( poke every 14.92  sec )
	 {
	      res = hid_write(handle, write_buf, sizeof(write_buf));
              if (res < 0) {
	         logMsg(LEVEL6,"HID Error: %ls\n",hid_error(handle));
                 errorExit(-105, progname, "HID, unable to write buffer.");
              }
              else
              {
                logMsg(LEVEL9,"%s hid_write() OK ;  (poking) \n" , strdate(LOGFILETIME));
                write_index++ ;
                write_buf[2] = write_index ;
		gettimeofday(&last_poke_time, NULL);
              }
	 }
     } 	 //end infinite loop


// clean up function


    return 0;
}




int getOptions(int argc, char *argv[])
{
    int c, num;
    int option_index = 0;
    char *ptr;
    static struct option long_options[] = {
        {"configfile", 1, 0, 'C'},
        {"debug", 0, 0, 'D'},
        {"disable-incoming", 1, 0, 'd'},
        {"help", 0, 0, 'h'},
        {"hostname-flag", 0, 0, 'H'},
        {"logfile", 1, 0, 'L'},
        {"mem_trace", 1, 0, 'm'},
        {"ncid", 1, 0, 'n'},
        {"pidfile", 1, 0, 'P'},
        {"test", 0, 0, 't'},
        {"tracefile", 1, 0, 'T'},
        {"verbose", 1, 0, 'v'},
        {"version", 0, 0, 'V'},
	{"osx-launchd", 0, 0, '0'},
        {0, 0, 0, 0}
    };

    while ((c = getopt_long (argc, argv, "C:d:DhH:L:m:n:P:v:tT:V",
        long_options, &option_index)) != -1)
    {
        switch (c)
        {
	    case '0':
                ++OSXlaunchd;
                break;
            case 'C':
                if (!(config = strdup(optarg)))
                    errorExit(-1, progname, 0);
                break;
            case 'd':
                 disable_incoming =  atoi(optarg);
                /* valid range: 0x00-0x0F */
                if (strlen(optarg) != 0 || strlen(optarg) > 4 || disable_incoming < 0x00 || disable_incoming > 0x0F )
                {
                    disable_incoming = 0;
                    errorExit(-107, "Invalid number", optarg);
                }
                if ((num = findWord("disable-incoming")) >= 0) setword[num].type = 0;
                break;
                 disable_incoming=0;
                 break;
            case 'D':
                 debug = 1;
                 verbose = 3;
                 break;
            case 'h': /* help message */
                fprintf(stderr, DESC, progname);
                fprintf(stderr, USAGE, progname);
                exit(0);
            case 'H': /* add hostname to IDENT */
                HostnameFlag = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(HostnameFlag == 0 && *optarg == '0') && HostnameFlag != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("HostnameFlag")) >= 0) setword[num].type = 0;
                break;
            case 'L':
                if (!(logfile = strdup(optarg))) errorExit(-1, progname, 0);
                break;
            case 'm':
                trace =  atoi(optarg);
                /* valid range: 0-2 */
                if (strlen(optarg) != 1 )
                {
                    trace = 0;
                    errorExit(-107, "Invalid number", optarg);
                }
                if ((num = findWord("mem_trace")) >= 0) setword[num].type = 0;
                break;
                trace=0;
                break;
            case 'n': /* [host][:port] must contain host or port or both */
                if ((ptr = index(optarg, (int) ':')))
                {
                    if ((ncidport = atoi(ptr + 1)) == 0)
                        errorExit(-101, "Invalid port number", optarg);
                    if ((num = findWord("ncidport")) >= 0)
                        setword[num].type = 0;
                }
                if (optarg != ptr)
                {
                    if (ptr) *ptr = '\0';
                    if (!(ncidhost = strdup(optarg))) errorExit(-1, progname, 0);
                    if ((num = findWord("ncidhost")) >= 0)
                        setword[num].type = 0;
                }
                break;
            case 'P':
                if (!(pidfile = strdup(optarg))) errorExit(-1, progname, 0);
                break;
            case 't':
                test = 1;
                debug = 1;
                verbose = 3;
                break;
            case 'T':
                if (!(tracefile = strdup(optarg))) errorExit(-1, progname, 0);
                break;
           case 'v':
                verbose = atoi(optarg);
                /* valid range: 1-9 */
                if (strlen(optarg) != 1 || (verbose == 0))
                {
                    verbose = 1;
                    errorExit(-107, "Invalid number", optarg);
                }
                if ((num = findWord("verbose")) >= 0) setword[num].type = 0;
                break;
            case 'V': /* version */
                fprintf(stderr, SHOWVER, progname, VERSION, API);
                exit(0);
                break;
            case '?': /* bad option */
                fprintf(stderr, USAGE, progname);
                exit(-100);
        }
    }
    return optind;
}


void try_to_reconnect()
{
        int attempt = 0 ;
        logMsg(LEVEL2, "could not send packets  to ncidd  (Connection lost)\n");
        close(sd);
        sd = 0;
	if ( !test)
	while  (( attempt <  MAX_TRY || MAX_TRY == -1 ) && sd==0 )  /* -1 means infinity ... */
	{
        sleep ( DELAY ) ;
	    attempt ++ ;
	    logMsg(LEVEL2, "Attempting to reconnect to ncidd server .. (try %d )\n", attempt);
	    socketConnect(NOTFATAL);
    }
	if ( sd == 0 )
	{
          errorExit(-105, progname, "Could not reconnect to ncidd server .. Giving up ...");
	}
}


/*
 * connect to the NCID server
 *     exit on connect error if fatal = 1
 *     continue on connect error if fatal = 0
 */
void socketConnect(int fatal)
{
    int size;
    char msgbuf[BUFSIZ];
    struct sockaddr_in pin;
    struct hostent *hp;

    /* find out about the desired host machine */
    if ((hp = gethostbyname(ncidhost)) == 0) {
        logMsg(LEVEL1, "Could not connect to ncidd at %s:%d\n",
           ncidhost, ncidport);
        errorExit(-1, "gethostbyname", hstrerror(h_errno));
    }

    /* fill in the socket structure with host information */
    memset(&pin, 0, sizeof(pin));
    pin.sin_family = AF_INET;
    pin.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
    pin.sin_port = htons(ncidport);

    /* grab an Internet domain socket */
    if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
    errorExit(-1, "socket", 0);

    /* connect to PORT on HOST */
    if (connect(sd, (struct sockaddr *) &pin, sizeof(pin)) == -1)
    {
        if (fatal) errorExit(-1, "NCID server", 0);
        logMsg(LEVEL2, "Warning: could not connect to ncidd at %s:%d\n",
                ncidhost, ncidport);
        /* if connect fails, close socket */
        close(sd);
        sd = 0;

    }
    else
    {
        /* connected */
        logMsg(LEVEL1, "Connected to ncidd at %s:%d\n", ncidhost, ncidport);

      /* send ident to server */
      sprintf(msgbuf, "%sgateway %s%s %s%s\n", IDENT, HostnameFlag ? hname : "",
              progname, VERSION, outcall ? GWOUT : "");
      send(sd, msgbuf, strlen(msgbuf), 0);
      logMsg(LEVEL1, "Sent: %sgateway %s%s %s%s\n", IDENT, HostnameFlag ? hname : "",
              progname, VERSION, outcall ? GWOUT : "");

      /* make sure call log not sent */
      sprintf(msgbuf, "%sno_log\n", COMMAND);
      send(sd, msgbuf, strlen(msgbuf), 0);
      logMsg(LEVEL1, "Sent: %sno_log\n", COMMAND);

      /* log server startup messages */
      while ((size = readLine(sd, msgbuf)))
      {
        if (size < 0)
        {
           if (fatal && errno != EINTR) errorExit(-1, progname, 0);
           continue;
        }
        else if (size == 0) break;

        /* log 2x0 and 25x line messages */
        if (msgbuf[0] == '2' && (msgbuf[1] == '5' || msgbuf[2] == '0'))
            logMsg(LEVEL1, msgbuf);
        /* log Options */
        else if (!strncmp(msgbuf, "OPT: ", strlen("OPT: "))) logMsg(LEVEL1, msgbuf);
        /* log 300 line message and break - End of connection startup */
        else if (msgbuf[0] == '3' && msgbuf[2] == '0')
        {
          logMsg(LEVEL1, msgbuf);
          break;
        }
        /* log lines */
        else logMsg(LEVEL5, msgbuf);
      }
    if (fcntl(sd, F_SETFL, O_NONBLOCK) < 0) errorExit(-1, "socket", 0);
   }
}


void  process_flash_hook_US(const char * number_H)
{
         logMsg(LEVEL5, "called_number = %s \n",called_number);
         logMsg(LEVEL5, "calling_number = %s \n",calling_number);
         logMsg(LEVEL5, "called_number2 = %s \n",called_number2);
         logMsg(LEVEL5, "calling_number2 = %s \n",calling_number2);
  if (  0  == strcmp( number_H,"") )
  {
     logMsg(LEVEL5, "US Hook flash with no DTMF digits  ...\n");
     if ( 0 == strncmp(incoming_call2_pickup_time,"-",sizeof(incoming_call2_pickup_time)))
     {
         strncpy(incoming_call2_pickup_time,strdate(WITHSEP),sizeof(incoming_call2_pickup_time));
         logMsg(LEVEL4, "%s (Accepting  call2 ) ...  We set  incoming_call2_pickup_time = %s \n",incoming_call2_pickup_time,strdate(LOGFILETIME));
	 call2_state=ON ;
	 call1_state=HOLD;
     }
     else
     {
       logMsg(LEVEL5,  "(Switching between  calls) ...\n");
       if ( call2_state==ON && call2_state==ON) //conference
       {
         call1_state=ON;
	 call2_state=HOLD;
       }
       else if (  call2_state==HOLD  && call1_state==ON)
       {
	 call1_state=HOLD;
         call2_state=ON;
       }
       else if (  call1_state==HOLD  && call2_state==ON)
       {
         call1_state=ON;
	 call2_state=HOLD;
       }
       else
       {
         logMsg(LEVEL5,  "(Something wrong with call(s)_states ; should never get there ) ...\n");
       }
     }
  }


  else
  {
    logMsg(LEVEL5,  "US Hook flash followed by DTMF digits ...(A second call has been dialled : %s) ...\n",number_H) ;
    strcpy(called_number2,number_H);
    outgoingCall2(called_number2) ;
    call2_state=ON ;
    call1_state=HOLD;
  }

}







void  process_flash_hook_EU(const char * number_H)
{
	 logMsg(LEVEL5, "called_number = %s \n",called_number);
	 logMsg(LEVEL5, "calling_number = %s \n",calling_number);
	 logMsg(LEVEL5, "called_number2 = %s \n",called_number2);
	 logMsg(LEVEL5, "calling_number2 = %s \n",calling_number2);
  if (  0  == strcmp( number_H,"") )
  {
        logMsg(LEVEL5, "EU Hook flash with nothing ... we do nothing ...\n");
  }
  else if (  0  == strcmp( number_H,"0") )
  {
     logMsg(LEVEL5, "EU Hook flash with 0 (Terminating waiting call) ...call1=%d call2=%d  \n",call1_state,call2_state);
       if ( call1_state==ON && call2_state==ON) //conference
       {
         logMsg(LEVEL5, "call1_state==ON && call2_state==ON (conf) ...\n");
         call2_state=OFF;
	 call1_state=ON;
	 if ( strcmp(calling_number ,"") !=0 )
	 {
	   logMsg(LEVEL5, "BYE incoming call1 ...\n");
           incomingCallInfo( BYE ,calling_number , calling_name);
	 }
	 else
	 {
	   logMsg(LEVEL5, "BYE outgoing  call1 ...\n");
	  outgoingCallInfo( BYE , called_number ) ;
	 }
	// call2 should become call1 ?
	strcpy(called_number,called_number2);
	strcpy(calling_number,calling_number2);
	   strcpy(called_number2,"");
	   strcpy(calling_number2,"");
	   strcpy(calling_name2,"");
	   call2_state=OFF;
           call1_state=ON;
       }
       else if (   call2_state==WAIT   && call1_state==ON)
       {
     if ( 0 == strncmp(incoming_call2_pickup_time,"-",sizeof(incoming_call2_pickup_time)))
     {
         logMsg(LEVEL4, " (Rejecting  call2 ) ...\n",calling_number2);
         incomingCallInfo2( CANCEL ,calling_number2 , calling_name2);
         strcpy(incoming_call2_start_time,"");
         strcpy(incoming_call2_pickup_time,"");
         strcpy(incoming_call2_end_time,"");
	 strcpy(calling_number2,"");
	 strcpy(calling_name2,"");
	 call1_state=ON ;
	 call2_state=OFF;
     }
       }
       else if (   call2_state==HOLD  && call1_state==ON)
       {
         logMsg(LEVEL5, "call1_state==ON && call2_state==HOLD ...\n");
	 call2_state=OFF;
         call1_state=ON;
	 if ( strcmp(calling_number2 ,"") !=0 )
	 {
	   logMsg(LEVEL5, "BYE incoming  call2 ...\n");
           incomingCallInfo2( BYE ,calling_number2 , calling_name2);
	   strcpy(calling_number2,"");
	 }
	 else
	 {
	   logMsg(LEVEL5, "BYE outgoing  call2 ...\n");
	   outgoingCallInfo2( BYE , called_number2 ) ;
	   strcpy(called_number2,"");
	 }
       }
       else if (  call1_state==HOLD  && call2_state==ON)
       {
         logMsg(LEVEL5, "call1_state==HOLD && call2_state==ON ...\n");
         call2_state=ON;
	 call1_state=OFF;
	  if ( strcmp(calling_number  ,"") !=0 )
         {
	   logMsg(LEVEL5, "BYE incoming  call1 ...\n");
           incomingCallInfo( BYE ,calling_number , calling_name);
         }
         else
         {
	   logMsg(LEVEL5, "BYE outgoing  call1 ...\n");
          outgoingCallInfo( BYE , called_number ) ;
         }

	 // call2 should become call1 ?
	 strcpy(called_number,called_number2);
	 strcpy(calling_number,calling_number2);
	 strcpy(incoming_call_start_time ,incoming_call2_start_time);
	 strcpy(incoming_call2_start_time,"") ;
	 strcpy(incoming_call_pickup_time ,incoming_call2_pickup_time);
	 strcpy(incoming_call2_pickup_time,"") ;
	 strcpy(incoming_call_end_time ,incoming_call2_end_time);
	 strcpy(incoming_call2_end_time,"") ;
	 strcpy(outgoing_call_start_time ,outgoing_call2_start_time);
	 strcpy(outgoing_call2_start_time,"") ;
	 strcpy(outgoing_call_pickup_time ,outgoing_call2_pickup_time);
	 strcpy(outgoing_call2_pickup_time,"") ;
	 strcpy(outgoing_call_end_time ,outgoing_call2_end_time);
	 strcpy(called_number2,"");
	 strcpy(calling_number2,"");
	 strcpy(calling_name2,"");
	 call2_state=OFF;
       }
       else
       {
         logMsg(LEVEL5,  "(Something wrong with call(s)_states ; should never get there ) ...\n");
       }
   }
   else if (  0  == strcmp( number_H,"1") )
   {
     logMsg(LEVEL5, "EU Hook flash with 1 (Terminating Active call) ...\n");
       if ( call2_state==ON && call2_state==ON) //conference
       {
         call1_state=ON;
	 call2_state=OFF;
	 if ( strcmp(calling_number2 ,"") !=0 )
	 {
	   logMsg(LEVEL5, "BYE incoming  call2 ...\n");
           incomingCallInfo2( BYE ,calling_number2 , calling_name2);
	   strcpy(calling_number2,"");
	 }
	 else
	 {
	   logMsg(LEVEL5, "BYE outgoing  call2 ...\n");
	   outgoingCallInfo2( BYE , called_number2 ) ;
	   strcpy(called_number2,"");
	 }
       }
       else if (  call2_state==HOLD  && call1_state==ON)
       {
         call2_state=OFF;
	 call1_state=ON;
	 if ( strcmp(calling_number ,"") !=0 )
	 {
	   logMsg(LEVEL5, "BYE incoming call1 ...\n");
           incomingCallInfo( BYE ,calling_number , calling_name);
	 }
	 else
	 {
	   logMsg(LEVEL5, "BYE outgoing  call1 ...\n");
	  outgoingCallInfo( BYE , called_number ) ;
	 }
	 // call2 should become call1 ?
	 strcpy(called_number,called_number2);
	 strcpy(calling_number,calling_number2);
	 strcpy(incoming_call_start_time ,incoming_call2_start_time);
	 strcpy(incoming_call2_start_time,"") ;
	 strcpy(incoming_call_pickup_time ,incoming_call2_pickup_time);
	 strcpy(incoming_call2_pickup_time,"") ;
	 strcpy(incoming_call_end_time ,incoming_call2_end_time);
	 strcpy(incoming_call2_end_time,"") ;
	 strcpy(outgoing_call_start_time ,outgoing_call2_start_time);
	 strcpy(outgoing_call2_start_time,"") ;
	 strcpy(outgoing_call_pickup_time ,outgoing_call2_pickup_time);
	 strcpy(outgoing_call2_pickup_time,"") ;
	 strcpy(outgoing_call_end_time ,outgoing_call2_end_time);
	 strcpy(outgoing_call2_end_time,"") ;
	 strcpy(called_number2,"");
	 strcpy(calling_number2,"");
	 strcpy(calling_name2,"");
	 call2_state=OFF;
         call1_state=ON;
       }
       else if (  call1_state==HOLD  && call2_state==ON)
       {
         call1_state=ON;
	 call2_state=OFF;
	 if ( strcmp(calling_number2 ,"") !=0 )
	 {
	   logMsg(LEVEL5, "BYE incoming  call2 ...\n");
           incomingCallInfo2( BYE ,calling_number2 , calling_name2);
	   strcpy(calling_number2,"");
	 }
	 else
	 {
	   logMsg(LEVEL5, "BYE outgoing  call2 ...\n");
	   outgoingCallInfo2( BYE , called_number2 ) ;
	   strcpy(called_number2,"");
	 }
       }
       else
       {
         logMsg(LEVEL5,  "(Something wrong with call(s)_states ; should never get there ) ...\n");
       }
   }
   else if (  0  == strcmp( number_H,"2") )
   {
     logMsg(LEVEL5, "EU Hook flash with 2 \n");
     if ( 0 == strncmp(incoming_call2_pickup_time,"-",sizeof(incoming_call2_pickup_time)))
     {
         strncpy(incoming_call2_pickup_time,strdate(WITHSEP),sizeof(incoming_call2_pickup_time));
         logMsg(LEVEL4, "%s (Accepting  call2 ) ...  We set  incoming_call2_pickup_time = %s \n",incoming_call2_pickup_time,strdate(LOGFILETIME));
	 call2_state=ON ;
	 call1_state=HOLD;
     }
     else
     {
       logMsg(LEVEL5,  "(Switching between  calls) ...\n");
       if ( call2_state==ON && call2_state==ON) //conference
       {
         call1_state=ON;
	 call2_state=HOLD;
       }
       else if (  call2_state==HOLD  && call1_state==ON)
       {
	 call1_state=HOLD;
         call2_state=ON;
       }
       else if (  call1_state==HOLD  && call2_state==ON)
       {
         call1_state=ON;
	 call2_state=HOLD;
       }
       else
       {
         logMsg(LEVEL5,  "(Something wrong with call(s)_states ; should never get there ) ...\n");
       }
     }
   }
   else if (  0  == strcmp( number_H,"3") )
   {
     logMsg(LEVEL5, "EU Hook flash with 3 (Conference  call) ...\n");
      call2_state=ON;
      call1_state=ON;
   }
   else
   {
      logMsg(LEVEL5,  "EU Hook flash ..(A second call has been dialled : %s) ...\n",number_H) ;
      strcpy(called_number2,number_H);
      outgoingCall2(called_number2) ;
      call2_state=ON ;
      call1_state=HOLD;
   }




}

void  outgoingCall(const char * number )
{
    char cidmsg[BUFSIZ];
    ssize_t retval=-1;

            /* Outgoing Call */
       //      name = NONAME;
       //      type = CALLOUT;
        sprintf(cidmsg, CALL,outgoing_call_date_time ,CALLOUT,lineid, number, NONAME);

        if (sd) retval =  write(sd, cidmsg, strlen(cidmsg));
        logMsg(LEVEL2, cidmsg);
	if( retval == -1 ) try_to_reconnect() ;
}

void  outgoingCall2(const char * number )
{
    char cidmsg[BUFSIZ];
    ssize_t retval=-1;

            /* Outgoing Call2 */
      //      name = NONAME;
      //      type = CALLOUT;
        outgoing_call2_date_time[0] = 0;
        strncpy(outgoing_call2_date_time,strdate(NOYEAR),sizeof(outgoing_call2_date_time));
            logMsg(LEVEL4, "outgoing_call2_date_time: \n", outgoing_call2_date_time);
        strncpy(outgoing_call2_start_time,strdate(WITHSEP),sizeof(outgoing_call2_start_time));
	// This is the second call (outgoing ) , there is no way to know if it has been picked up
	// S0  we set pickup_time to start_time
     // strncpy(outgoing_call2_pickup_time,"-",sizeof(outgoing_call2_start_time));
        strncpy(outgoing_call2_pickup_time,strdate(WITHSEP),sizeof(outgoing_call2_start_time));

        sprintf(cidmsg, CALL,outgoing_call2_date_time ,CALLOUT,lineid, number, NONAME);

        if (sd) retval =  write(sd, cidmsg, strlen(cidmsg));
        logMsg(LEVEL2, cidmsg);
	if( retval == -1 ) try_to_reconnect() ;
}

void  incomingCall(void )
//This covers both call1  and call2  (MWI as well)
{
    char cidmsg[BUFSIZ];
    char date_time [MAX_STR] ;
    ssize_t retval=-1;
    char calltypeString [10] = {0};
    int send=1 ;

       /* incoming Call */
      //  call Type = "CID" /"WID" / "RID" / "MWI" ;
         switch (incomingCall_cid.calltype)
         {
           case CID :
           strcpy(calltypeString,"CID");
           strcpy(calling_callType,"CID");
           if ((disable_incoming & NO_CID)  != 0)
           {
             logMsg(LEVEL2, "No CID  line sent to ncidd server as --disable-incoming option has value  %#4X (to avoid duplicates with a modem or another gateway)\n",disable_incoming );
	     send=0;
           }
           break;
           case RID :
           strcpy(calltypeString,"RID");
           strcpy(calling_callType,"RID");
           if ((disable_incoming & NO_RID)  != 0)
           {
             logMsg(LEVEL2, "No RID  line sent to ncidd server as --disable-incoming option has value  %#4X (to avoid duplicates with a modem or another gateway)\n",disable_incoming );
	     send=0;
           }
           break;
           case WID :
           strcpy(calltypeString,"WID");
	   call2_state = WAIT;
           if ((disable_incoming & NO_WID)  != 0)
           {
             logMsg(LEVEL2, "No WID  line sent to ncidd server as --disable-incoming option has value  %#4X (to avoid duplicates with a modem or another gateway)\n",disable_incoming );
	     send=0;
           }
           logMsg(LEVEL5,"%s call1_state=%d  call2_state=%d   \n", strdate(LOGFILETIME) , call1_state , call2_state );
           break;
           case MWI :
           strcpy(calltypeString,"MWI");
           if ((disable_incoming & NO_MWI)  != 0)
           {
             logMsg(LEVEL2, "No MWI  line sent to ncidd server as --disable-incoming option has value  %#4X (to avoid duplicates with a modem or another gateway)\n",disable_incoming );
	     send=0;
           }
           break;
         }
        date_time[0] = 0 ;
        //first copy  only the  first 4 characters of incomingCall_cid.ciddate (remove the year)
        strncat(date_time,incomingCall_cid.ciddate ,4);
        //then add  the time (4 characters )
        strncat(date_time ,incomingCall_cid.cidtime ,4);
            logMsg(LEVEL4, "date_time:%s \n", date_time);
            logMsg(LEVEL4, "incomingCall_cid.ciddate:%s \n", incomingCall_cid.ciddate);
            logMsg(LEVEL4, "incomingCall_cid.cidtime:%s \n", incomingCall_cid.cidtime);
            logMsg(LEVEL4, "incomingCall_cid.calltype:%s \n", calltypeString);
        if ( incomingCall_cid.calltype == CID  || incomingCall_cid.calltype == RID )
        {
         strncpy( incoming_call_date_time ,date_time ,sizeof(incoming_call_date_time));
         logMsg(LEVEL4, "Resulting  incoming_call_date_time:%s \n", incoming_call_date_time);
         strncpy(incoming_call_start_time,strdate(WITHSEP),sizeof(incoming_call_start_time));
         strncpy(incoming_call_pickup_time,"-",sizeof(incoming_call_pickup_time));
         strncpy(calling_number,incomingCall_cid.cidnmbr,sizeof(calling_number));
         strncpy(calling_name,incomingCall_cid.cidname,sizeof(calling_name));
        }
	if ( incomingCall_cid.calltype == WID )
        {
         strncpy( incoming_call2_date_time ,date_time ,sizeof(incoming_call2_date_time));
         logMsg(LEVEL4, "Resulting  incoming_call_date_time:%s \n", incoming_call2_date_time);
         strncpy(incoming_call2_start_time,strdate(WITHSEP),sizeof(incoming_call2_start_time));
         strncpy(incoming_call2_pickup_time,"-",sizeof(incoming_call2_pickup_time));
         strncpy(calling_number2,incomingCall_cid.cidnmbr,sizeof(calling_number2));
         strncpy(calling_name2,incomingCall_cid.cidname,sizeof(calling_name2));
	 call2_state=WAIT;
	 gettimeofday(&WID_start_time,NULL);
        }
        sprintf(cidmsg, CALL,date_time ,calltypeString ,incomingCall_cid.cidline , incomingCall_cid.cidnmbr,incomingCall_cid.cidname);
        if (send == 1)
        {
          if (sd) retval =  write(sd, cidmsg, strlen(cidmsg));
          logMsg(LEVEL2, cidmsg);
	  if( retval == -1 ) try_to_reconnect() ;
        }
}






void  outgoingCallInfo(const char * ptype, const char * number )
{
    char cidmsg[BUFSIZ];
    ssize_t retval=-1;


            /* Outgoing Call */
        //        name = NONAME;
        //        type = CALLOUT;
        strncpy(outgoing_call_end_time,strdate(WITHSEP),sizeof(outgoing_call_end_time));
        logMsg(LEVEL4, "outgoing_call_start_time = %s \n", outgoing_call_start_time);
        logMsg(LEVEL4, "outgoing_call_pickup_time = %s \n", outgoing_call_pickup_time);
        logMsg(LEVEL4, "outgoing_call_end_time = %s \n", outgoing_call_end_time);
        sprintf(cidmsg, CALLINFO,ptype ,outgoing_call_date_time , outgoing_call_start_time ,outgoing_call_pickup_time ,outgoing_call_end_time,CALLOUT,lineid, number, NONAME);

        strncpy(outgoing_call_start_time ,"",sizeof(outgoing_call_start_time));
        strncpy(outgoing_call_pickup_time ,"",sizeof(outgoing_call_pickup_time));
        if (sd) retval =  write(sd, cidmsg, strlen(cidmsg));
        logMsg(LEVEL2, cidmsg);
	if( retval == -1 ) try_to_reconnect() ;
}
void  outgoingCallInfo2(const char * ptype, const char * number2 )
{
    char cidmsg[BUFSIZ];
    ssize_t retval=-1;


            /* Outgoing Call */
        //        name = NONAME;
        //        type = CALLOUT;
        strncpy(outgoing_call2_end_time,strdate(WITHSEP),sizeof(outgoing_call2_end_time));
        logMsg(LEVEL4, "outgoing_call2_start_time = %s \n", outgoing_call2_start_time);
        logMsg(LEVEL4, "outgoing_call2_pickup_time = %s \n", outgoing_call2_pickup_time);
        logMsg(LEVEL4, "outgoing_call2_end_time = %s \n", outgoing_call2_end_time);
        sprintf(cidmsg, CALLINFO,ptype ,outgoing_call2_date_time , outgoing_call2_start_time ,outgoing_call2_pickup_time ,outgoing_call2_end_time,CALLOUT,lineid, number2, NONAME);

        strncpy(outgoing_call2_start_time ,"",sizeof(outgoing_call2_start_time));
        strncpy(outgoing_call2_pickup_time ,"",sizeof(outgoing_call2_pickup_time));
        if (sd) retval =  write(sd, cidmsg, strlen(cidmsg));
        logMsg(LEVEL2, cidmsg);
	if( retval == -1 ) try_to_reconnect() ;
}


void  incomingCallInfo(const  char * ptype , const char * number ,const  char * name )
{
    char cidmsg[BUFSIZ];
    ssize_t retval=-1;


            /* Incoming Call */
      //      type = CALLIN;

        strncpy(incoming_call_end_time,strdate(WITHSEP),sizeof(incoming_call_end_time));
        logMsg(LEVEL4, "incoming_call_start_time = %s \n", incoming_call_start_time);
        logMsg(LEVEL4, "incoming_call_pickup_time = %s \n", incoming_call_pickup_time);
        logMsg(LEVEL4, "incoming_call_end_time = %s \n", incoming_call_end_time);
        sprintf(cidmsg, CALLINFO,ptype ,incoming_call_date_time ,incoming_call_start_time ,incoming_call_pickup_time ,incoming_call_end_time,CALLIN,incomingCall_cid.cidline, number,name);
        strncpy(incoming_call_start_time ,"",sizeof(incoming_call_start_time));
        strncpy(incoming_call_pickup_time ,"",sizeof(incoming_call_pickup_time));

        if (sd) retval =  write(sd, cidmsg, strlen(cidmsg));
        logMsg(LEVEL2, cidmsg);
	if( retval == -1 ) try_to_reconnect() ;
}

void  incomingCallInfo2(const  char * ptype , const char * number2 ,const  char * name2 )
{
    char cidmsg[BUFSIZ];
    ssize_t retval=-1;


            /* Incoming Call2 */
      //      type = CALLIN;

        strncpy(incoming_call2_end_time,strdate(WITHSEP),sizeof(incoming_call2_end_time));
        logMsg(LEVEL4, "incoming_call2_start_time = %s \n", incoming_call2_start_time);
        logMsg(LEVEL4, "incoming_call2_pickup_time = %s \n", incoming_call2_pickup_time);
        logMsg(LEVEL4, "incoming_call2_end_time = %s \n", incoming_call2_end_time);
        sprintf(cidmsg, CALLINFO,ptype ,incoming_call2_date_time ,incoming_call2_start_time ,incoming_call2_pickup_time ,incoming_call2_end_time,CALLIN,incomingCall_cid.cidline, number2,name2);
        strncpy(incoming_call2_start_time ,"",sizeof(incoming_call2_start_time));
        strncpy(incoming_call2_pickup_time ,"",sizeof(incoming_call2_pickup_time));

        if (sd) retval =  write(sd, cidmsg, strlen(cidmsg));
        logMsg(LEVEL2, cidmsg);
	if( retval == -1 ) try_to_reconnect() ;
}

/* Log error, call cleanup() and exit */
void errorExit(int error, const char * msg, const char * arg)
{
    if (error == -1 && arg == 0)
    {
        /* system error */
        error = errno;
        logMsg(LEVEL1, "%s: %s\n", msg, strerror(errno));
    }
    else if (msg != 0)
    {
        /* internal program error */
        logMsg(LEVEL1, "%s: %s\n", msg, arg);
    }

    logMsg(LEVEL1, "Terminated:  %s\n", strdate(WITHSEP));
    cleanup();

    exit(error);
}


int readLine(int sd, char * buf)
{
    int size;
    char *eptr;

    /* read data from socket without removing the data */
    size = recv(sd, buf, BUFSIZ, MSG_PEEK);
    if (size > 0)
    {
        /* find first line in buf */
        if ((eptr = strchr(buf, (int) '\n'))) size = (eptr - buf) + 1;

        memset(buf, '\0', BUFSIZ);
        /* read a line if there or all data in buf */
        size = recv(sd, buf, size, 0);
    }

    return size;
}

void cleanup( void )
{
    /* remove pid file, if it was created and we are in parent (reading) process */
        logMsg(LEVEL5, "In cleanup \n");
    if (pid  )
    {
        unlink(pidfile);
        logMsg(LEVEL1, "Removed pidfile: %s  pid= %d\n", pidfile ,pid );
    }

    /* close open files and sockets */
    if (traceptr)
    {
       logMsg(LEVEL7,"closing tracefile \n");
       fclose(traceptr); /* trace file */
    }
    if (sd >= 0)
    {
       logMsg(LEVEL7,"closing connection to ncidd \n");
       close(sd); /* connection to ncidd */
    }
    if (handle)
    {
      logMsg(LEVEL7,"closing hid device \n");
      hid_close(handle); /* hid device */
    }
    /* Free static HIDAPI objects. */
    logMsg(LEVEL7,"Free static HIDAPI objects.\n");
    hid_exit();

    if (logptr)
    {
      logMsg(LEVEL7,"closing log file.\n");
      fclose(logptr); /* logfile) */
    }
}






/*
 * set up signal handlers
 */

void setup_signal_handlers(void)
{
   /* SIGHUP, SIGUSR1 and SIGUSR2 not used */
   static struct sigaction user_sigaction;
   user_sigaction.sa_handler = USRhandler;
   sigfillset( &user_sigaction.sa_mask);
   user_sigaction.sa_flags = 0;
   sigaction(SIGHUP, &user_sigaction, NULL);
   sigaction(SIGUSR1, &user_sigaction, NULL);
   sigaction(SIGUSR2, &user_sigaction, NULL);

   /* SIGINT, etc, will stop artech2ncid */
   TERM_flag = 0;
   static struct sigaction leave_sigaction;
   leave_sigaction.sa_handler = TERMexit;
   sigfillset( &leave_sigaction.sa_mask);
   leave_sigaction.sa_flags = 0;
   sigaction(SIGINT, &leave_sigaction, NULL);  /* Interrupt from keyboard */
   sigaction(SIGQUIT, &leave_sigaction, NULL); /* Quit from keyboard */
   sigaction(SIGABRT, &leave_sigaction, NULL); /* Abort signal from abort(3) */
   sigaction(SIGTERM, &leave_sigaction, NULL); /* Termination signal */
   sigaction(SIGCHLD, &leave_sigaction, NULL);    /* child process stopped */


   /* ignore a bunch of them */
   static struct sigaction ignore_sigaction;
   ignore_sigaction.sa_handler = SIG_IGN;
   sigemptyset( &ignore_sigaction.sa_mask);
   ignore_sigaction.sa_flags = 0;
   sigaction(SIGTSTP, &ignore_sigaction, NULL);
   sigaction(SIGPIPE, &ignore_sigaction, NULL);    /* Broken pipe: write to pipe with no readers */
   sigaction(SIGTTIN, &ignore_sigaction, NULL);
   sigaction(SIGTTOU, &ignore_sigaction, NULL);
   sigaction(SIGIO, &ignore_sigaction, NULL);    /* io is waiting */
   sigaction(SIGWINCH, &ignore_sigaction, NULL); /* Window resize signal */

   /* set these back to the default */
   static struct sigaction default_sigaction;
   default_sigaction.sa_handler = SIG_DFL;
   sigfillset( &default_sigaction.sa_mask);
   default_sigaction.sa_flags = 0;
   sigaction(SIGSEGV, &default_sigaction, NULL);
   sigaction(SIGXCPU, &default_sigaction, NULL);
   sigaction(SIGXFSZ, &default_sigaction, NULL);
}



/*
 * USRhandler
 */

void    USRhandler(int sig)
{
  switch (sig)
  {
    case SIGHUP:
    case SIGUSR1:
    case SIGUSR2:
      logMsg(LEVEL6, "Received Signal %d: %s\nIgnored: %s\n",
             sig, strsignal(sig), strdate(WITHSEP));
      break;
    default:
      /* well that's odd */
      logMsg(LEVEL6, "Received unexpected Signal %d: %s\nIgnored: %s\n",
             sig, strsignal(sig), strdate(WITHSEP));
      break;
  }
}



void TERMexit(int sig)
{

   /* got signal SIGTERM via TERMhandler, clean up and leave */
   logMsg(LEVEL1, "Received Signal %d: %s\nTerminated: %s\n",
           sig, strsignal(sig), strdate(WITHSEP));

   cleanup();
   exit(0);
}



int  xdmfCID( unsigned char *ucbuf ,int length)
{
    unsigned char  *ptr, chksum, reclen, nmss;
    char *tptr;
    size_t msglen;
    int nameLength , i , ChecksumReached ;

      logMsg(LEVEL5, "Detected MDMF or SDMF message\n");

      if (verbose >= HEXLEVEL) hexdump(ucbuf, length + 10  );

      ptr = ucbuf;
      incomingCall_cid.calltype = CID;
      strcpy(incomingCall_cid.cidline, lineid);
      incomingCall_cid.cidsent = 0;
      ChecksumReached=0;
      chksum = checksum(ptr, (msglen = length ));
      logMsg(LEVEL5, "Checksum %s: 0x%.2x\n", chksum ? "bad" : "good", chksum);
      if (chksum != 0 )
      {
        return (-1);
      }
      // is it a WID or a CID ?
      // if calling_number != "" or called_number != ""  it's a WID (we are already on line )
      logMsg(LEVEL5, "Checking calling_number (%s) and called_number (%s)  (Is one of them not empty ? --> This is a WID) \n", calling_number , called_number);
      if (( 0 !=  strcmp(calling_number ,"" )) || ( 0 != strcmp (called_number, "" ) ) )
      {
        incomingCall_cid.calltype = WID;
      }
      else
      {
        incomingCall_cid.calltype = CID;
      }
     // need to properly zeroe out all strings beforehand (otherwise , strncat() could leave some garbage ) (string missing terminating zero )
      bzero(incomingCall_cid.ciddate,CIDSIZE);
      bzero(incomingCall_cid.cidtime,CIDSIZE);
      bzero(incomingCall_cid.cidname,CIDSIZE);
      bzero(incomingCall_cid.cidnmbr,CIDSIZE);

      if (*ptr == SDMF_MESG_TYPE)
      {
        logMsg(LEVEL5, "Detected SDMF message\n");

        /* SDMF date and time are at fixed byte positions */
        reclen = 8;

        /* start at second byte */
        ptr += 2;

        strncat(incomingCall_cid.ciddate, (char *) ptr, 4);
        incomingCall_cid.status |= CIDDATE;

        strncat(incomingCall_cid.cidtime, (char *) ptr + 4, 4);
        incomingCall_cid.status |= CIDTIME;

        logMsg(LEVEL5, "Date & Time: %s %s\n", incomingCall_cid.ciddate, incomingCall_cid.cidtime);

        /* need to get year */
        tptr = strdate(ONLYYEAR);        /* returns: YYYY */
        strcat(incomingCall_cid.ciddate, tptr);       /* MMDDYYYY */

        ptr += reclen;

        /*
         * SDMF has NMBR only, no NAME
         * starts at a fixed position
         * length varies
         */

        strcpy(incomingCall_cid.cidname, NONAME);
        incomingCall_cid.status |= CIDNAME;

        reclen = msglen - reclen - 3;

        strncat(incomingCall_cid.cidnmbr, (char *) ptr, reclen);
        incomingCall_cid.status |= CIDNMBR;

        if (reclen == 1)
        {
          logMsg(LEVEL5, "Why No Number: %c\n", *ptr);
          switch (*ptr)
          {
             case LETTER_P:
               strcpy(incomingCall_cid.cidnmbr, WORDFOR_P);
               break;
             case LETTER_O:
               strcpy(incomingCall_cid.cidnmbr, WORDFOR_O);
               break;
          }
        }


        logMsg(LEVEL5, "Message Checksum: 0x%.2x\n", *(ucbuf + msglen - 1));
      }
      else
      {
        logMsg(LEVEL5, "Detected MDMF message\n");

        /* Check packet ID  for call type*/
        if (*ptr == MDMF_MSG_WAITING) incomingCall_cid.calltype = MWI;

        /* start at second parameter */
        ptr += 2;

        while (*ptr != 0 && ChecksumReached == 0 )
        {
          switch (*ptr)
          {
            case MDMF_VISUAL_IND:
              logMsg(LEVEL5, "Parameter 0x%.2x: US Visual Indicator\n", *ptr);
              ++ptr;
              reclen = *ptr++;
              if (*ptr) strcpy(incomingCall_cid.cidname, "Message(s) Waiting");
              else strcpy(incomingCall_cid.cidname, "No Messages Waiting");
              strcpy(incomingCall_cid.cidnmbr, "Message");
              incomingCall_cid.status = CIDNAME | CIDNMBR;
              ptr += reclen;
              incomingCall_cid.calltype = MWI;
              break;
            case CALL_TYPE_UK:
              ++ptr;
              reclen = *ptr++;
              logMsg(LEVEL4, "Parameter 0x%.2x: UK Call Type 0x%.2x; ",
                     *(ptr - 2), *ptr);
              switch (*ptr)
              {
                case VOICE_UK:
                  //incomingCall_cid.calltype = CID;
                  //or
                  //incomingCall_cid.calltype = WID;
                  //this has been determined before
                  logMsg(LEVEL5, "CallType 0x%.2x\n", incomingCall_cid.calltype);
                  break;
                case RINGBACK_UK:
                  incomingCall_cid.calltype = RID;
                  logMsg(LEVEL5, "CallType 0x%.2x\n", incomingCall_cid.calltype);
                  break;
                case MSG_WAITING_UK:
                  incomingCall_cid.calltype = MWI;
                  ++ptr;
                  nmss = *ptr++;
                  reclen = *ptr++;
                  switch (*ptr)
                  {
                    case 0:
                      strcpy(incomingCall_cid.cidname, "No Messages Waiting");
                      break;
                    case 1:
                      strcpy(incomingCall_cid.cidname, "1 Message Waiting");
                      break;
                    default:
                      sprintf(incomingCall_cid.cidname, "%u Messages Waiting", *ptr);
                      break;
                  }
                  logMsg(LEVEL5, "CallType 0x%.2x; NMSS 0x%.2x\n",
                         incomingCall_cid.calltype, nmss);
                  strcpy(incomingCall_cid.cidnmbr, "Voicemail");
                  incomingCall_cid.status = CIDNAME | CIDNMBR;
                  break;
              }
              ptr += reclen;
              break;
            case MDMF_DATETIME:
              logMsg(LEVEL5, "Parameter 0x%.2x: Date & Time\n", *ptr);
              ++ptr;
              reclen = *ptr++;
              strncat(incomingCall_cid.ciddate, (char *) ptr, 4);
              incomingCall_cid.status |= CIDDATE;

              strncat(incomingCall_cid.cidtime, (char *) ptr + 4, 4);
              incomingCall_cid.status |= CIDTIME;

              /* need to get year */
              tptr = strdate(ONLYYEAR);        /* returns: YYYY */
              strcat(incomingCall_cid.ciddate, tptr);       /* MMDDYYYY */

              ptr += reclen;
              break;
            case MDMF_CALLER_NMBR:
              logMsg(LEVEL5, "Parameter 0x%.2x: Calling Number\n", *ptr);
              ++ptr;
              reclen = *ptr++;
              if (reclen)
              {
                strncat(incomingCall_cid.cidnmbr, (char *) ptr, reclen);
                logMsg(LEVEL5, "incomingCall_cid.cidnmbr %s   reclen= %d\n",incomingCall_cid.cidnmbr , reclen );
                incomingCall_cid.status |= CIDNMBR;
              }
              ptr += reclen;
              break;
            case MDMF_WHY_NO_NMBR:
              logMsg(LEVEL5, "Parameter 0x%.2x: Why No Number\n", *ptr);
              ++ptr;
              reclen = *ptr++;
              if (*ptr == LETTER_P) strcpy(incomingCall_cid.cidnmbr, WORDFOR_P);
              else if (*ptr == LETTER_O) strcpy(incomingCall_cid.cidnmbr, WORDFOR_O);
              else strcpy(incomingCall_cid.cidnmbr, NONMBR);
              incomingCall_cid.status |= CIDNMBR;
              ++ptr;
              break;
            case MDMF_CALLER_NAME:
              logMsg(LEVEL5, "Parameter 0x%.2x: Calling Name\n", *ptr);
              ++ptr;
              reclen = *ptr++;
              if (reclen)
              {
                strncat(incomingCall_cid.cidname, (char *) ptr, reclen);
                logMsg(LEVEL5, "incomingCall_cid.cidname %s   reclen= %d\n",incomingCall_cid.cidname , reclen );
                incomingCall_cid.status |= CIDNAME;
              }
              ptr += reclen;
	      // check that incomingCall_cid.cidname contains only 7 bit char( pure ASCII )  AND is printable
              nameLength = strlen(incomingCall_cid.cidname) ;
              for ( i = 0 ; i < nameLength ; i ++ )
              {
                 if ( 0 == isascii(incomingCall_cid.cidname[i])  && 0 == isprint(incomingCall_cid.cidname[i] )  )
                 {
                    logMsg(LEVEL5, "Non ascii or non printable  char found in Name: %s \n", incomingCall_cid.cidname);
                    // we replace by NULL ( snips  the string )
                    incomingCall_cid.cidname[i] = 0 ;
                    break ;
                 }
              }
              break;
            case MDMF_WHY_NO_NAME:
              logMsg(LEVEL5, "Parameter 0x%.2x: Why No Name\n", *ptr);
              ++ptr;
              reclen = *ptr++;
              if (*ptr == LETTER_P) strcpy(incomingCall_cid.cidname, WORDFOR_P);
              else if (*ptr == LETTER_O) strcpy(incomingCall_cid.cidname, WORDFOR_O);
              else strcpy(incomingCall_cid.cidname, NONAME);
              incomingCall_cid.status |= CIDNAME;
              ++ptr;
              break;
            default:
              ++ptr;
              if (ptr == ucbuf + msglen)
              {
                logMsg(LEVEL5, "Parameter 0x%.2x: Message Checksum\n", *(ptr - 1));
		ChecksumReached=1;
              }
              else
              {
                logMsg(LEVEL5, "Parameter 0x%.2x: Unknown\n", *(ptr - 1));
                /* skip parameter data */
                reclen = *ptr++;
                ptr += reclen;
              }
              break;
          }
        }

        /* make sure there is a date, time, number and name */
        if (incomingCall_cid.cidnmbr[0] == '\0')
        {
          strcpy(incomingCall_cid.cidnmbr, NONMBR);
          incomingCall_cid.status |= CIDNMBR;
        }
        if (incomingCall_cid.cidname[0] == '\0')
        {
          strcpy(incomingCall_cid.cidname, NONAME);
          incomingCall_cid.status |= CIDNAME;
        }
        logMsg(LEVEL5, "incomingCall_cid.ciddate %s incomingCall_cid.cidtime: %s \n",incomingCall_cid.ciddate ,incomingCall_cid.cidtime );
        if (incomingCall_cid.ciddate[0] == '\0' || incomingCall_cid.cidtime[0] == '\0')
        {
          logMsg(LEVEL5, "filling up date \n");
          tptr = strdate(NOSEP);
          strncat(incomingCall_cid.ciddate, tptr, 8);
          strncat(incomingCall_cid.cidtime, tptr + 9, 4);
          incomingCall_cid.status |= (CIDDATE | CIDTIME);
        }

      }
      logMsg(LEVEL2, "Calling Number: %s\n", incomingCall_cid.cidnmbr);
      logMsg(LEVEL2, "Call Type: %d\n", incomingCall_cid.calltype);

      return(0);
}

/*
 * Convert hex string to unsigned characters
 */
void hex2uchar(const char * buf, unsigned char * ucbuf)
{
    int buf_len = strlen(buf)/2, i;
    char tmpchar[5];

    bzero(tmpchar, sizeof(tmpchar));
    for (i = 0; i < buf_len; i++)
    {
        strncpy(tmpchar, buf + 2 * i, 2);
        ucbuf[i] = (unsigned char) strtol(tmpchar, NULL, 16);
    }
}

/*
 * 2's complement 8-bit checksum
 * Modified from: https://stackoverflow.com/questions/3463976/c-file-checksum
 */
unsigned char checksum (unsigned char *ptr, size_t size) {
  unsigned int sum = 0;
  while (size-- != 0)
  {
    sum += *ptr++;
  }
  sum = (sum ^ 0xFF) + 1;
  return (unsigned char)sum;
}




long long
timeval_diff(struct timeval *difference,
             struct timeval *end_time,
             struct timeval *start_time
            )
{
  struct timeval temp_diff;

  if(difference==NULL)
  {
    difference=&temp_diff;
  }

  difference->tv_sec =end_time->tv_sec -start_time->tv_sec ;
  difference->tv_usec=end_time->tv_usec-start_time->tv_usec;

  /* Using while instead of if below makes the code slightly more robust. */

  while(difference->tv_usec<0)
  {
    difference->tv_usec+=1000000;
    difference->tv_sec -=1;
  }

  return 1000000LL*difference->tv_sec+
                   difference->tv_usec;

} /* timeval_diff() */

/*******************************************************/
 /*
 * artech2ncid.c - This file is part of artech2ncid.
 *
 * Copyright (c) 2021
 * by Bruno Grasland  <b.sletteland@orange.fr>
 *
 * Some parts of the code have been inspired from Ncid by
 * John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * artech2ncid is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * artech2ncid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. */
