/*
 * cideasy2ncid.c - This file is part of cideasy2ncid.
 *
 * Copyright (c) 2022-2024
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * cideasy2ncid is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * cideasy2ncid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with cideasy2ncid.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "cideasy2ncid.h"

uid_t getuid(void);
int writeCIDEasy(hid_device * _handle, const char * reply);
int readCIDEasy(hid_device * _handle, unsigned char * _read_str, int _read_str_len, int _timeout);
int readLine(int sd, char * buf);
void callIndex(void);
void socketConnect(int fatal);
void decodeCIDEasy(unsigned char * data);
void sendCALL(void);
void sendCALLINFO(char * htype);
void doRingTimer(void);
void restartRingTimer(void);
void clearCall(void);
void setup_signal_handlers(void);
void TERMhandler(int sig);
void SEGVhandler(int sig);
void TERMexit(void);

void cleanup(void);
void errorExit(int error, const char * msg, const char * arg);
int getOptions(int argc, char *argv[]);

int debug, test;
int sd = -1;
int cepos;                  /* current position in calls array */;
char lineid[GWSIZE];        /* current lineid */
char offhookLineid[GWSIZE]; /* determines which port went off hook */
char onhookLineid[GWSIZE];  /* determines which port went on hook */
char ringLineid[GWSIZE];    /* determines which port ringing stopped */
int ringcnt;                /* one ringie-dingie ... */
time_t ringtimer;           /* to detect end of ringing */
int HostnameFlag             = 1;
int outcall                  = 0;
int verbose                  = 1;
int ncidport                 = NCIDPORT;
const char * ncidhost        = LOCALHOST;
const char * gatewayid       = GW;
const char * logfile         = LOGFILE;
const char * pidfile;
char *progname;
char hname[HOST_NAME_MAX + 1];
pid_t pid;
FILE *logptr;

int conferr, OSXlaunchd;

volatile int TERM_flag;

struct setword setword[] =
{
   /* *word         type      **buf        *value             min     max */
   {"HostnameFlag", WORDNUM,  0,           &HostnameFlag,     OFF,    ON},
   {"gatewayid",    WORDSTR,  &gatewayid,  0,                 0,      0},
   {"ncidhost",     WORDSTR,  &ncidhost,   0,                 0,      0},
   {"ncidport",     WORDNUM,  0,           &ncidport,         0,      0},
   {"outcall",      WORDNUM,  0,           &outcall,          0,      1},
   {"pidfile",      WORDSTR,  &pidfile,    0,                 0,      0},
   {"verbose",      WORDNUM,  0,           &verbose,          1,      MAXLEVEL},
};

struct call_t
{
    int  valid;         /* valid == 1 when CID data is received */
    hid_device * handle;
    /* start of call variables */
    char ctype[GWSIZE];
    char  date[GWSIZE];
    char  lineid[GWSIZE];
    char  nmbr[GWSIZE];
    char  name[GWSIZE];
    /* end of call variables */
    char htype[GWSIZE];
    char scall[GWSIZE];
    char pcall[GWSIZE];
    char ecall[GWSIZE];
} calls[CALLS];

/* hid device */
hid_device * handle;

/* Needed for config.c in the library */
struct sendclient sendclient[1];

int main(int argc, char **argv)
{
    int argind, i, ret = 0, errnum = 0;
    uid_t uid;
    char msgbuf[BUFSIZ];
    char *ptr, serverline[BUFSIZ];

    /* global containing name of program */
    progname = strrchr(argv[0], (int) '/');
    progname = progname ? progname + 1 : argv[0];

    /* Get the default config file */
    config = CONFIG;

    /* process options from the command line */
    argind = getOptions(argc, argv);

    /* should not be any arguments */
    if (argc - argind != 0)
    {
        fprintf(stderr, NOOPT, progname, argv[argind]);
        fprintf(stderr, USAGE, progname);
        exit(0);
    }

    /* open or create logfile */
    logptr = fopen(logfile, "a");
    errnum = errno;

    logMsg(LEVEL1, "Started: %s\nGateway: %s %s\n%s\n",
           strdate(WITHSEP), progname, VERSION, API);

    /* log command line and any options on separate lines */
    sprintf(msgbuf, "Command line: %s", argv[0]);
    for (i = 1; i < argc; i++)
    {
        if (*argv[i] == '-')
            strcat(strcat(msgbuf, "\n              "), argv[i]);
        else strcat(strcat(msgbuf, " "), argv[i]);
    }
    strcat(msgbuf, "\n");
    logMsg(LEVEL1, msgbuf);

    /* check status of logfile */
    if (logptr)
    {
        /* logfile opened */
        logMsg(LEVEL1, "logfile: %s\n", logfile);
    }
    else
    {
        /* logfile open failed */
        logMsg(LEVEL1, "%s: %s\n", logfile, strerror(errnum));
    }

    /*
     * read config file, if present, exit on any errors
     * do not override any options set on the command line
     */
    if (doConf()) errorExit(-104, 0, 0);

    logMsg(LEVEL1, "Verbose level: %d\n", verbose);

    logMsg(LEVEL1, "gatewayid: %s\n", gatewayid);

    if (gethostname(hname, HOST_NAME_MAX) == -1)
    {
        logMsg(LEVELE, "gethostname(): %s\n", strerror(errno));
        if (HostnameFlag)
        {
            HostnameFlag = 0;
            logMsg(LEVEL1, "HostnameFlag reset to 0\n");
        }
    }
    else
    {
        logMsg(LEVEL1, "hostname flag: %d\nHostname: %s\n", HostnameFlag, hname);
        strcat(hname, "/");
    }

    /* set up signal handlers */
    setup_signal_handlers();

    if (test) logMsg(LEVEL1, "Test Mode\n");
    if (debug || OSXlaunchd)
    {
        if (debug) logMsg(LEVEL1, "Debug Mode\n");
        else logMsg(LEVEL1, "OSX Launchd Mode\n");
    }
    else
    {
        /* fork and exit parent */
        if(fork() != 0) return 0;

        /* close stdin and make fd 0 unavailable */
        close(0);
        if (open("/dev/null",  O_WRONLY | O_SYNC) < 0)
        {
            errorExit(-1, "/dev/null", 0);
        }

        /* become session leader */
        setsid();
    }

    /* must not be in test mode to create a PID file and connect to NCID */
    if (!test)
    {
        logMsg(LEVEL1,"NCID server at %s:%d\n", ncidhost, ncidport);

        /*create a PID file */
        if (doPID(&pid, pidfile))
        {
            sprintf(msgbuf,"%s already exists", pidfile);
            errorExit(-110, "Fatal", msgbuf);
        }

        socketConnect(FATAL);
    }

    /* CID Easy */
    int result;
    unsigned char data[DATASIZE];
    struct hid_device_info *devs, *cur_dev;

    /* Initialize the HIDAPI library */
    result = hid_init();
    if (result !=0)
    {
        logMsg(LEVEL1, "Failed to initialize HIDAPI library\n");
        errorExit(-104, 0, 0);
    }

#if defined(__APPLE__)
    /*
     * To work properly needs to be called before hid_open/hid_open_path
     * after hid_init.  Best is to call it right after hid_init.
     */
    hid_darwin_set_open_exclusive(0);
#endif

    /* find all CID Easy Devices */
    if (!(devs = hid_enumerate(VID, PID)))
        errorExit(-104, progname, "No CID Easy Devices Found");
    logMsg(LEVEL2, "enumerating CIDEasy Device(s)\n");
    cur_dev = devs;

    for (i = 0; cur_dev; ++i) {
        logMsg(LEVEL2, "Device Found\n");
        logMsg(LEVEL2, "  path:              %s\n", cur_dev->path);
        logMsg(LEVEL2, "  Vendor ID:         %04hx\n", cur_dev->vendor_id);
        logMsg(LEVEL2, "  Product ID:        %04hx\n", cur_dev->product_id);
        logMsg(LEVEL2, "  Manufacturer:      %ls\n", cur_dev->manufacturer_string);
        logMsg(LEVEL2, "  Product:           %ls\n", cur_dev->product_string);
        logMsg(LEVEL2, "  BCD Serial Number: %ls\n", cur_dev->serial_number);
        logMsg(LEVEL2, "  Release:           %hx\n", cur_dev->release_number);
        logMsg(LEVEL2, "  Interface:         %d\n",  cur_dev->interface_number);

#ifdef HID_API_VERSION_MAJOR
        /* library contains the HID version */
        logMsg(LEVEL2, "  HID Version:       %s\n", hid_version_str());
#else
        logMsg(LEVEL2, "  HID Version:       < 0.11, unable to obtain it\n");
#endif

        /* Open the CID Easy device using its VID, PID, and S/N.
         */
         handle = hid_open(VID, PID, cur_dev->serial_number);

        /*
         * Open the CID Easy device using its VID and PID.
         * handle = hid_open(VID, PID, NULL);
         */

        if (handle == NULL)
        {
            logMsg(LEVEL1, "Unable to open CID Easy device using:\n");
            logMsg(LEVEL1, "    manufacture ID: %X\n", VID);
            logMsg(LEVEL1, "  & product ID:     %.4X\n", PID);
            logMsg(LEVEL1, "  & Serial Number:  %ls\n", cur_dev->serial_number);
            uid = getuid();
            if (uid != 0)
            {

              printf("Warning: UID = %d,  must be 0 (root) to open device\n",
                      uid);
            }
            errorExit(-104, 0, 0);
        }
        else
        {
          logMsg(LEVEL1, "Opened CID Easy device using:\n");
          logMsg(LEVEL1, "    manufacture ID: %X\n", VID);
          logMsg(LEVEL1, "  & product ID:     %X\n", PID);
          logMsg(LEVEL1, "  & Serial Number:  %ls\n", cur_dev->serial_number);
          logMsg(LEVEL1, "  CID Easy Device Handle: %d\n", handle);
        }

        cur_dev = cur_dev->next;
    }
    if (i) logMsg(LEVEL1, "Found %d CID Easy %s\n", i,
                  i == 1 ? "Device" : "Devices");
    if (i > 1) logMsg(LEVEL1,
            "WARNING: Only one device is supported.  Using last one found.\n");

    hid_free_enumeration(devs);

    logMsg(LEVEL1,"Wait for data: %d.%d seconds\n", DATAWAIT/1000, DATAWAIT%1000);
    logMsg(LEVEL1,"Wait for ring: %d seconds\n", RINGWAIT);
    logMsg(LEVEL1,"End of startup: %s\n", strdate(WITHSEP));

    /* read data from CIDEasy device */
    while (1)
    {
        if (TERM_flag)
        {
           /* a TERM signal asked us to exit */
           TERMexit();
        }

        if ((ret = read(sd, serverline, BUFSIZ - 1) == -1))
        {
            /* if no data available, connection dropped */
            if (ret != EAGAIN || ret != EWOULDBLOCK) socketConnect(NOTFATAL);
        }
        serverline[ret] = '\0';
        if ((ptr = strchr(serverline, '\r'))) *ptr = '\0';
        if ((ptr = strchr(serverline, '\n'))) *ptr = '\0';
        logMsg(LEVEL3, "Received: %s\n", serverline);

        ret = readCIDEasy(handle, data, sizeof(data) - 1, DATAWAIT);

        switch (ret)
        {
            case 0: /* timeout */
                if (ringtimer) doRingTimer();
                break;
            case -1: /* bad data */
                writeCIDEasy(handle, NCK);
                logMsg(LEVEL1, "%s\n%s\n", data, NCK);
                break;
            default: /* good data */
                if (!strncmp((char *) data, ENQ, 3) || !strncmp((char *) data, ETB, 3))
                {
                    writeCIDEasy(handle, ACK);
                }
                else
                {
                    writeCIDEasy(handle, DCK);
                    decodeCIDEasy(data);
                }
        }
    }

    /* Close the device */
    hid_close(handle);

    /* Finalize the HIDAPI library */
    result = hid_exit();

    return 0;
}

int getOptions(int argc, char *argv[])
{
    int c, num;
    int option_index = 0;
    char *ptr;
    static struct option long_options[] = {
        {"config", 1, 0, 'C'},
        {"debug", 0, 0, 'D'},
        {"help", 0, 0, 'h'},
        {"hostname-flag", 1, 0, 'H'},
        {"gatewayid", 1, 0, 'g'},
        {"logfile", 1, 0, 'L'},
        {"ncidhost", 1, 0, 'n'},
        {"pidfile", 1, 0, 'P'},
        {"test", 0, 0, 't'},
        {"usage", 0, 0, 'u'},
        {"verbose", 1, 0, 'v'},
        {"version", 0, 0, 'V'},
        {"osx-launchd", 0, 0, '0'},
        {0, 0, 0, 0}
    };

    while ((c = getopt_long (argc, argv, "C:Dg:hH:L:n:P:tuv:V",
        long_options, &option_index)) != -1)
    {
        switch (c)
        {
            case '0':
                ++OSXlaunchd;
                break;
            case 'C':
                if (!(config = strdup(optarg)))
                    errorExit(-1, progname, 0);
                break;
            case 'D':
                debug = 1;
                verbose = 3;
                break;
            case 'g':
                if (!(gatewayid = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("gatewayid")) >= 0) setword[num].type = 0;
                break;
            case 'h': /* help message */
                fprintf(stderr, DESC, progname);
                fprintf(stderr, USAGE, progname);
                exit(0);
            case 'H': /* add hostname to IDENT */
                HostnameFlag = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(HostnameFlag == 0 && *optarg == '0') && HostnameFlag != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("HostnameFlag")) >= 0) setword[num].type = 0;
                break;
            case 'L':
                if (!(logfile = strdup(optarg))) errorExit(-1, progname, 0);
                break;
            case 'n': /* [host][:port] must contain host or port or both */
                if ((ptr = index(optarg, (int) ':')))
                {
                    if ((ncidport = atoi(ptr + 1)) == 0)
                        errorExit(-101, "Invalid port number", optarg);
                    if ((num = findWord("ncidport")) >= 0)
                        setword[num].type = 0;
                }
                if (optarg != ptr)
                {
                    if (ptr) *ptr = '\0';
                    if (!(ncidhost = strdup(optarg))) errorExit(-1, progname, 0);
                    if ((num = findWord("ncidhost")) >= 0)
                        setword[num].type = 0;
                }
                break;
            case 'P':
                if (!(pidfile = strdup(optarg))) errorExit(-1, progname, 0);
                break;
            case 't':
                test = 1;
                debug = 1;
                verbose = 3;
                break;
            case 'u': /* usage message */
                fprintf(stderr, USAGE, progname);
                exit(0);
            case 'v':
                verbose = atoi(optarg);
                /* valid range: 1-9 */
                if (strlen(optarg) != 1 || (verbose == 0))
                {
                    verbose = 1;
                    errorExit(-107, "Invalid number", optarg);
                }
                if ((num = findWord("verbose")) >= 0) setword[num].type = 0;
                break;
            case 'V': /* version */
                fprintf(stderr, SHOWVER, progname, VERSION, API);
                exit(0);
            case '?': /* bad option */
                fprintf(stderr, USAGE, progname);
                exit(-100);
        }
    }
    return optind;
}


/*
 * Decode CIDEasy Message
 */
void decodeCIDEasy(unsigned char * data)
{
    size_t msglen, gwmsglen = 0;
    unsigned char *ptr, reclen;
    char *line = NULL;
    char calltype[] = {0, 0, 0, 0, 0};
    char xdmftype[] = {0, 0, 0, 0, 0, 0, 0, 0};

    strcpy(calltype, "CID");
    msglen = strlen((char *) data);
    logMsg(LEVEL3, "Total Message Length: %d\n", msglen);
    logMsg(LEVEL3, "Message: %s\n", data);
    ptr = data;
    while (*ptr != 0)
    {
        if (msglen < 10)
        {
            logMsg(LEVEL1, "Received unknown message: ");
            while (*ptr) if (isprint (*ptr)) printf("%d", *ptr++);
            printf("\n");
            return;
        }

        /* STA/?:RING, STA/?:OFF HOOK, STA/?ON HOOK, etc */
        if (!strncmp((const char *) ptr, "STA/", 4))
        {
            sprintf(lineid,"%s%c%c", gatewayid, '-', *(ptr + 4));
            callIndex();
            line =  (char *) (ptr + 6);
            logMsg(LEVEL3, "Received on port %c: %s\n", *(ptr + 4), line);
            if (!strcmp(line, "RING"))
            {
                ringcnt++;
                if (ringcnt == 1) strcpy(calls[cepos].scall, strdate(WITHSEP));
                strcpy(ringLineid, lineid);
                restartRingTimer();
            }
            if (!strcmp(line, "OFF HOOK"))
            {
                ringtimer = 0;
                strcpy(offhookLineid, lineid);
                strcpy(calls[cepos].pcall, strdate(WITHSEP));
            }
            if (!strcmp(line, "ON HOOK"))
            {
                strcpy(onhookLineid, lineid);
                sendCALLINFO(BYE); /* call was answered */
            }
            calls[cepos].valid = 0;
            break;
        }

        /* Caller ID */
        if (ptr == data)
        {
            logMsg(LEVEL3, "Received on port %c: %s\n", *ptr, "Caller ID");
            sprintf(lineid,"%s%c%c", gatewayid, '-', *ptr);
            callIndex();
            strcpy(calls[cepos].lineid, lineid);
            /* RING, not received, start of call is when CID is received */
            if (!ringcnt) strcpy(calls[cepos].scall, strdate(WITHSEP));
            strcpy(calls[cepos].pcall, NOPICKUP);
            logMsg(LEVEL4, "lineid: %s\n", lineid);
            ++ptr;
        }

        if (ptr == data + 1)
        {
            logMsg(LEVEL4, "  Parameter 0x%.2x: xmdf\n", *ptr);
            if (*ptr == 0x80) strcpy(xdmftype, "MDMF");
            else if (*ptr == 0x4) strcpy(xdmftype, "SDMF");
            else strcpy(xdmftype, "Unknown");
            logMsg(LEVEL4, "  XDMF Type: %s\n", xdmftype);
            strcpy(calls[cepos].ctype, CALLIN);
            logMsg(LEVEL4, "  Call: %s\n", calls[cepos].ctype);
            calls[cepos].valid = 1;
            ++ptr;
        }

        if (ptr == data + 2)
        {
            logMsg(LEVEL4, "  Parameter 0x%.2x: message length\n", *ptr);
            gwmsglen = *ptr;
            logMsg(LEVEL4, "  CID Message Length: %d\n", gwmsglen);
            ++ptr;
        }

        if (*ptr == 1)
        {
            logMsg(LEVEL4, "  Parameter 0x%.2x: Date & Time\n", *ptr);
            ++ptr;
            reclen = *ptr++;
            strncpy((char *) calls[cepos].date, (char *) ptr, 8);

            ptr += reclen;
            logMsg(LEVEL4, "  Date: %s\n", calls[cepos].date);
        }

        if (*ptr == 2)
        {
            logMsg(LEVEL4, "  Parameter 0x%.2x: Calling Number\n", *ptr);
            ++ptr;
            reclen = *ptr++;
            if (reclen != 1) strncpy((char *) calls[cepos].nmbr, (char *) ptr, reclen);
            else
            {
                if (*ptr == LETTER_P) strcpy(calls[cepos].nmbr, WORDFOR_P);
                else if (*ptr == LETTER_O) strcpy(calls[cepos].nmbr, WORDFOR_O);
                else strcpy(calls[cepos].nmbr, NONMBR);
            }
            ptr += reclen;
            logMsg(LEVEL4, "  Number: %s\n", calls[cepos].nmbr);
        }

        if (*ptr == 7)
        {
            logMsg(LEVEL4, "  Parameter 0x%.2x: Calling Name\n", *ptr);
            ++ptr;
            reclen = *ptr++;
            if (reclen)
            {
                strncpy((char *) calls[cepos].name, (char *) (char *) ptr, reclen);
                logMsg(LEVEL4, "  Name: %s\n", calls[cepos].name);
            }
            ptr += reclen;
        }
        if (*ptr++)
            logMsg(LEVEL4, "  parameter 0x%.2x: Unknown\n", *ptr);
    }

    if (gwmsglen)
    {
        if (gwmsglen == msglen -3)
            logMsg(LEVEL4, "  Message length OK\n");
        else
            logMsg(LEVEL4, "  Message length %d differs from gwmsglen %d + 2\n",
                    msglen, gwmsglen);
    }
    if (calls[cepos].valid) sendCALL();
}


/*
 * Write a reply to an CIDEasy device
 * based on the provided reply strings: "ACK" "DCK" "NAK"
 * Returns -1 on failure, number of bytes written on success
 */
int writeCIDEasy(hid_device * _handle, const char * reply)
{
    unsigned char replybuf[REPLYSIZE];
    int status;

    /* Get the length of the command string */
    const int replylen = strlen(reply);

    /* Zero out buffer to pad with null values */
    memset(replybuf, 0, REPLYSIZE);

    if (replylen > REPLYSIZE - 1)
    {
        logMsg(LEVEL1, "Error: reply string is larger than %i\n", REPLYSIZE - 1);
        return -1;
    }
    replybuf[0] = 0x0;
    memcpy( &replybuf[1], reply, replylen);

    /* write the command to the device */
    if ((status = hid_write(_handle, replybuf, REPLYSIZE)) != -1)
        logMsg(LEVEL4, "wrote:    %s\n", reply);
    else
    {
        logMsg(LEVEL1, "Error writing to device: %ls\n", hid_error(_handle));
        cleanup();
        exit(-1);
    }

    return status;
}


/*
 * Read a response from an CIDEasy device with a specified timeout
 */
int readCIDEasy(hid_device * _handle, unsigned char * str, int strlen, int timeout)
{
    int readlen;
    char *ptr;

    /* Attempt to read data from the device */
    readlen = hid_read_timeout(_handle, str, strlen, timeout);
    if (readlen > 0)
    {
        /* zero "\r\n" in string to terminate it at '\r' */
        if ((ptr = strchr((const char *) str, '\r')))
            *ptr = '\0', *(++ptr) = '\0';

        logMsg(LEVEL5, "Data received: %s\n", strdate(WITHSEP));
        logMsg(LEVEL5, "Length of data received: %d\n", readlen);
        hexdump(str, (int) strlen);
        logMsg(LEVEL4, "Received: %s\n", str);
    }

    /* returns 0 on success, a negative number on error, or bytes read */
    return readlen;
}

/*
 * connect to the NCID server
 *     exit on connect error if fatal = 1
 *     continue on connect error if fatal = 0
 */
void socketConnect(int fatal)
{
    int size;
    char msgbuf[BUFSIZ];
    struct sockaddr_in pin;
    struct hostent *hp;

    /* find out about the desired host machine */
    if ((hp = gethostbyname(ncidhost)) == 0) {
        logMsg(LEVEL1, "Could not connect to ncidd at %s:%d\n",
           ncidhost, ncidport);
        errorExit(-1, "gethostbyname", hstrerror(h_errno));
    }

    /* fill in the socket structure with host information */
    memset(&pin, 0, sizeof(pin));
    pin.sin_family = AF_INET;
    pin.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
    pin.sin_port = htons(ncidport);

    /* grab an Internet domain socket */
    if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
    errorExit(-1, "socket", 0);

    /* connect to PORT on HOST */
    if (connect(sd, (struct sockaddr *) &pin, sizeof(pin)) == -1)
    {
        if (fatal) errorExit(-1, "NCID server", 0);
        logMsg(LEVEL2, "Warning: could not connect to ncidd at %s:%d\n",
                ncidhost, ncidport);
        /* if connect fails, close socket */
        close(sd);
        sd = 0;

    }
    else
    {
        /* connected */
        logMsg(LEVEL1, "Connected to ncidd at %s:%d\n", ncidhost, ncidport);

      /* send ident to server */
      sprintf(msgbuf, "%sgateway %s%s %s%s\n", IDENT, HostnameFlag ? hname : "",
              progname, VERSION, outcall ? GWOUT : "");
      send(sd, msgbuf, strlen(msgbuf), 0);
      logMsg(LEVEL1, "Sent: %sgateway %s%s %s%s\n", IDENT, HostnameFlag ? hname : "",
              progname, VERSION, outcall ? GWOUT : "");

      /* make sure call log not sent */
      sprintf(msgbuf, "%sno_log\n", COMMAND);
      send(sd, msgbuf, strlen(msgbuf), 0);
      logMsg(LEVEL1, "Sent: %sno_log\n", COMMAND);

      /* log server startup messages */
      while ((size = readLine(sd, msgbuf)))
      {
        if (size < 0)
        {
           if (fatal && errno != EINTR) errorExit(-1, progname, 0);
           continue;
        }
        else if (size == 0) break;

        /* log 2x0 and 25x line messages */
        if (msgbuf[0] == '2' && (msgbuf[1] == '5' || msgbuf[2] == '0'))
            logMsg(LEVEL1, msgbuf);
        /* log Options */
        else if (!strncmp(msgbuf, "OPT: ", strlen("OPT: "))) logMsg(LEVEL1, msgbuf);
        /* log 300 line message and break - End of connection startup */
        else if (msgbuf[0] == '3' && msgbuf[2] == '0')
        {
          logMsg(LEVEL1, msgbuf);
          break;
        }
        /* log lines */
        else logMsg(LEVEL4, msgbuf);
      }
    if (fcntl(sd, F_SETFL, O_NONBLOCK) < 0) errorExit(-1, "socket", 0);
   }
}

void sendCALL()
{
    char msgbuf[BUFSIZ], * ptr;

    sprintf(msgbuf, CALL, calls[cepos].date,
            calls[cepos].ctype,
            calls[cepos].lineid,
            calls[cepos].nmbr,
            calls[cepos].name,
            CRLF);
    if (!test) send(sd, msgbuf, strlen(msgbuf), 0);

    /* replace CRLF with NL and '\0' */
    ptr = strstr(msgbuf, CRLF);
    *ptr = NL;
    *(++ptr) = '\0';

    logMsg(LEVEL3, "Sent: %s\n", msgbuf);
}

void sendCALLINFO(char * htype)
{
    char msgbuf[BUFSIZ], * ptr;

    strcpy(calls[cepos].ecall, strdate(WITHSEP));

    /* end of call hangup type: CANCEL or BYE */
    strcpy(calls[cepos].htype, htype);

    sprintf(msgbuf, CALLINFO,
            calls[cepos].htype,
            calls[cepos].date,
            calls[cepos].scall,
            calls[cepos].pcall,
            calls[cepos].ecall,
            calls[cepos].ctype,
            calls[cepos].lineid,
            calls[cepos].nmbr,
            calls[cepos].name,
            CRLF);
    if (!test) send(sd, msgbuf, strlen(msgbuf), 0);

    /* replace CRLF with NL and '\0' */
    ptr = strstr(msgbuf, CRLF);
    *ptr = NL;
    *(++ptr) = '\0';

    logMsg(LEVEL3, "Ring Count: %d\n", ringcnt);
    logMsg(LEVEL3, "Sent: %s\n", msgbuf);

    ringcnt = ringtimer = 0;
    clearCall();
}

/*
 * determine position of call in calls array
 *   returns position in calls array for current call or new call
 *   returns -1 if no room in calls array
 */
void callIndex(void)
{
    int pos;
    for (pos = 1; pos < CALLS; pos++)
    {
        /* find current call position in calls array */
        if (!strcmp(calls[pos].lineid, lineid))
        {
            logMsg(LEVEL4, "current call array position: calls[%d]\n", pos);
            break;
        }
    }
    if (pos == CALLS)
    {
        /* find a new call position in calls array */
        for (pos = 1; pos < CALLS; pos++)
        {
            if (calls[pos].lineid[0] == '\0')
            {
                logMsg(LEVEL4, "new call array position: calls[%d]\n", pos);
                break;
            }
        }
    }
    if (pos == CALLS)
    {
        /* no room in calls array */
        logMsg(LEVEL1, "call array filled: %d current calls\n", pos);
        pos = -1;
    }

    cepos = pos;
}

/*
 * zero out call in the calls array
 */
void clearCall(void)
{
        calls[cepos].valid = 0;
        calls[cepos].handle = 0;
        memset(calls[cepos].ctype,  0, GWSIZE);
        memset(calls[cepos].date,   0, GWSIZE);
        memset(calls[cepos].lineid, 0, GWSIZE);
        memset(calls[cepos].nmbr,   0, GWSIZE);
        memset(calls[cepos].name,   0, GWSIZE);
        memset(calls[cepos].htype,  0, GWSIZE);
        memset(calls[cepos].scall,  0, GWSIZE);
        memset(calls[cepos].pcall,  0, GWSIZE);
        memset(calls[cepos].ecall,  0, GWSIZE);
}

/*
 * Reads one line from a socket, even if there is more than
 * one line.  If no lines, reads all data from socket.
 *
 * Must be called multiple times until zero is returned.
 * Returns -1, 0, or number of bytes read.
 *
 * Should properly handle the case where data does not end in '\n',
 * but this should only happen when receiving large log file so no
 * need to reconstruct a line before returning the data.
 */

int readLine(int sd, char * buf)
{
    int size;
    char *eptr;

    /* read data from socket without removing the data */
    size = recv(sd, buf, BUFSIZ, MSG_PEEK);
    if (size > 0)
    {
        /* find first line in buf */
        if ((eptr = strchr(buf, (int) '\n'))) size = (eptr - buf) + 1;

        memset(buf, '\0', BUFSIZ);
        /* read a line if there or all data in buf */
        size = recv(sd, buf, size, 0);
    }

    return size;
}

/*
 * the modem hasn't said RING in a while (RINGWAIT seconds)
 */
void doRingTimer(void)
{
   if (ringtimer < time(0))
   {
       /* turn the timer off */
       ringtimer = 0;
       logMsg(LEVEL3, "ring: %d timeout at: %s\n", ringcnt, strdate(ONLYTIME));
       /* call was unanswered */
       sendCALLINFO(CANCEL);
   }
}

/*
 * increment the RING count and reset the timer
 */
void restartRingTimer(void)
{
      /* call starts on receiving ring */
      /* timer to announce that ringing has stopped */
      ringtimer = time(0) + RINGWAIT;
}

/*
 * exit on a SIGINT, SIGQUIT, SIGABRT, SIGTERM signal
 */

void TERMhandler(int sig)
{
   /* set a flag so we can close everything neatly before leaving */
   TERM_flag = sig;
}

void TERMexit(void)
{
   cleanup();

   /* got signal SIGTERM via TERMhandler, clean up and leave */
   logMsg(LEVEL1, "Received Signal %d: %s\nTerminated: %s\n",
           TERM_flag, strsignal(TERM_flag), strdate(WITHSEP));

   /* close open files and sockets */
   exit(0);
}

/*
 * exit on a SIGSEGV signal
 * do minimal clean up after a memory exception
 */

void SEGVhandler(int sig)
{
    /* remove pid file, if it was created */
    if (pid) {
        unlink(pidfile);
    }

    /* and exit */
    _exit(sig);
}

/*
 * set up signal handlers
 */

void setup_signal_handlers(void)
{
/* SIGINT, etc, will stop 3c cideasy2ncid */
   TERM_flag = 0;
   static struct sigaction leave_sigaction;
   leave_sigaction.sa_handler = TERMhandler;
   sigfillset( &leave_sigaction.sa_mask);
   leave_sigaction.sa_flags = 0;
   sigaction(SIGINT, &leave_sigaction, NULL);  /* Interrupt from keyboard */
   sigaction(SIGQUIT, &leave_sigaction, NULL); /* Quit from keyboard */
   sigaction(SIGABRT, &leave_sigaction, NULL); /* Abort signal from abort(3) */
   sigaction(SIGTERM, &leave_sigaction, NULL); /* Termination signal */

   /* SIGSEGV, will try to remove the lock file */
   static struct sigaction segv_sigaction;
   segv_sigaction.sa_handler = SEGVhandler;
   sigfillset( &segv_sigaction.sa_mask);
   segv_sigaction.sa_flags = 0;
   sigaction(SIGSEGV, &segv_sigaction, NULL);    /* Invalid memory reference */
}

/*
 * Remove PID files, close open files and free HID memory
 * Must be called by errorExit or followed by exit(-1)
 */
void cleanup(void)
{
    /* remove pid file, if it was created */
    if (pid)
    {
        unlink(pidfile);
        logMsg(LEVEL1, "Removed pidfile: %s\n", pidfile);
    }

    /* close open files and sockets */
    if (logptr) fclose(logptr); /* logfile) */
    if (sd >= 0) close(sd); /* connection to ncidd */
#ifdef HIDAPI_H__
    /* hidapi.h header file was included */
    if (handle) hid_close(handle); /* hid device */

    /* Free static HIDAPI objects. */
    hid_exit();
#endif
}

/* Log error, call cleanup() and exit() */
void errorExit(int error, const char * msg, const char * arg)
{
    if (error == -1 && arg == 0)
    {
        /*
         * system error
         * print msg, arg should be zero
         */
        error = errno;
        logMsg(LEVEL1, "%s: %s\n", msg, strerror(errno));
    }
    else if (msg != 0)
    {
        /* internal program error */
        logMsg(LEVEL1, "%s: %s\n", msg, arg);
    }

    /* do not print terminated message, or cleanup, if option error */
    if (error != -100 && error != -101 && error != -106 && error != -107 &&
        error != -108 && error != -113)
    {
        cleanup();
        logMsg(LEVEL1, "Terminated:  %s\n", strdate(WITHSEP));
    }

    exit(error);
}
