/*
 * ncid2ncid.c - This file is part of ncid2ncid.
 *
 * Copyright (c) 2005-2024
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncid2ncid is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncid2ncid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncid2ncid.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ncid2ncid.h"

/* globals */
int debug, warn, HostnameFlag;
int verbose   = 1;
char *name;
char hname[HOST_NAME_MAX + 1];
const char * pidfile;
const char * logfile  = LOGFILE;
pid_t pid;
FILE *logptr;

int conferr, OSXlaunchd;

volatile int USR2_flag, TERM_flag;

int getOptions(int argc, char *argv[]);
int socketConnect(int pos);
int readLine(int sd, char *buf);

void cleanup(void);
void errorExit(int error, const char * msg, const char * arg);
void showConnected(void);
void setup_signal_handlers(void);
void USRhandler(int sig);
void TERMhandler(int sig);
void SEGVhandler(int sig);
void TERMexit(void);
void  handle_incoming_message(int pos, char * um, int length);
void  handle_accumulated_message(int pos, const char * buf);

/*
 * NONSERVERS must match the number of configuration lines below
 *
 * tohost and a number of fromhost? are added when gateway starts up
 * the size of SERVERS determines the number of fromhost?
 *    tohost = setword[2]
 *    toport = setword[3]
 *    fromhost1 = setword[4]
 *    fromport1 = setword[5]
 *    etc
 */

#define NONSERVERS  4
#define SERVERS     10           /*number of 'TO' and 'FROM' servers */
struct setword setword[NONSERVERS + SERVERS * 2] =
{
   /* *word         type      **buf    *value          min     max */
   {"HostnameFlag", WORDNUM,  0,        &HostnameFlag, OFF,    ON},
   {"pidfile",      WORDSTR,  &pidfile, 0,             0,      0},
   {"verbose",      WORDNUM,  0,        &verbose,      1,      MAXLEVEL},
   {"warn",         WORDNUM,  0,        &warn,         OFF,    ON}
};

/*
 * a number of fromhost? are added when gateway starts up
 * the size of SERVERS determines the number of fromhost?
 *    tohost = ns[0]
 *    fromhost1 = ns[1]
 *    fromhost2 = ns[2]
 *    etc
 */
struct server
{
    const char * name;
    const char * host;
    int port;
    int tries;
    time_t next_try;
};
struct server ns[SERVERS] =
{
   /* *name   *host  port try now */
    {"tohost", TOHOST, PORT, 0, 0}
};

/* Needed for config.c in the library */
struct sendclient sendclient[1];


/* file descriptor list for poll() */
struct pollfd polld[SERVERS];

/* All line types of interest to a server, new types must be added */
/* sorted alphabetically here for readability */

char *lineType[] =
{
    "BLK: ",
    "CID: ",
    "CIDINFO: ",
    "END: ",
    "HUP: ",
    "MSG: ",
    "MWI: ",
    "NOT: ",
    "RLY: ",
    "OUT: ",
    "PID: ",
    "PUT: ",
    "RID: ",
    "WID: ",
    ""
};

int main(int argc, char *argv[])
{
    int    argind, events, pos, i, errnum;
    char   msgbuf[BUFSIZ], rcvbuf[BUFSIZ], tmp[100];

    /* global containing name of program */
    name = strrchr(argv[0], (int) '/');
    name = name ? name + 1 : argv[0];

    /* Get the default config file */
    config = CONFIG;

    /* initialize ns[] */
    for(i = 1; i < SERVERS; ++i)
    {
        ns[i].port = PORT;
        sprintf(tmp, "fromhost%d", i);
        if (!(ns[i].name = strdup(tmp))) errorExit(-1, name, 0);
    }

    /* initialize safesend */
    initQueuedMsg(SERVERS);

    /* add tohost, fromhost1, etc to the setword[] options */
    for(pos = 0, i = NONSERVERS; pos < SERVERS; ++pos)
    {
        setword[i].word = ns[pos].name;
        setword[i].type = WORDSTR;
        setword[i++].buf = &ns[pos].host;
        if (!pos) sprintf(tmp, "toport");
        else sprintf(tmp, "fromport%d", pos);
        setword[i].type = WORDNUM;
        if (!(setword[i].word = strdup(tmp))) errorExit(-1, name, 0);
        setword[i++].value = &ns[pos].port;
    }

    /* process options from the command line */
    argind = getOptions(argc, argv);

    /* should not be any arguments */
    if (argc - argind != 0)
    {
        fprintf(stderr, NOOPT, name, argv[argind]);
        fprintf(stderr, USAGE, name);
        exit(0);
    }

    /* open or create logfile */
    logptr = fopen(logfile, "a");
    errnum = errno;

    logMsg(LEVEL1, "Started: %s\nGateway: %s %s\n%s\n",
           strdate(WITHSEP), name, VERSION, API);

    /* log command line and any options on separate lines */
    sprintf(msgbuf, "Command line: %s", argv[0]);
    for (i = 1; i < argc; i++)
    {
        if (*argv[i] == '-')
            strcat(strcat(msgbuf, "\n              "), argv[i]);
        else strcat(strcat(msgbuf, " "), argv[i]);
    }
    strcat(msgbuf, "\n");
    logMsg(LEVEL1, msgbuf);

    /* check status of logfile */
    if (logptr)
    {
        /* logfile opened */
        logMsg(LEVEL1, "logfile: %s\n", logfile);
    }
    else
    {
        /* logfile open failed */
        logMsg(LEVEL1, "%s: %s\n", logfile, strerror(errnum));
    }

    /*
     * read config file, if present, exit on any errors
     * do not override any options set on the command line
     */
    if (doConf()) errorExit(-104, 0, 0);

    logMsg(LEVEL1, "Total servers supported: %d\n", SERVERS);

    logMsg(LEVEL1, "Verbose level: %d\n", verbose);

    /* global containing hostname */
    if (gethostname(hname, HOST_NAME_MAX) == -1)
    {
        logMsg(LEVELE, "gethostname(): %s\n", strerror(errno));
        if (HostnameFlag)
        {
            HostnameFlag = 0;
            logMsg(LEVEL1, "HostnameFlag reset to 0\n");
        }
    }
    else
    {
        logMsg(LEVEL1, "hostname flag: %d\nHostname: %s\n", HostnameFlag, hname);
        strcat(hname, "/");
    }

    logMsg(LEVEL1,
            "Send server disconnect and reconnect messages to clients? %s\n",
            warn ? "YES" : "NO");

    /* first 'FROM' host is required */
    if (!ns[1].host) errorExit(-108, NOFROM, REQOPT);

    /* set up signal handlers */
    setup_signal_handlers();

    if (debug || OSXlaunchd)
    {
        if (debug) logMsg(LEVEL1, "Debug Mode\n");
        else logMsg(LEVEL1, "OSX Launchd Mode\n");
    }
    else
    {
        /* fork and exit parent */
        if(fork() != 0) return 0;

        /* close stdin and make fd 0 unavailable */
        close(0);
        if (open("/dev/null",  O_WRONLY | O_SYNC) < 0)
        {
            errorExit(-1, "/dev/null", 0);
        }

        /* become session leader */
        setsid();
    }

    /*create a PID file */
    if (doPID(&pid, pidfile))
    {
        sprintf(msgbuf,"%s already exists", pidfile);
        errorExit(-110, "Fatal", msgbuf);
    }

    /*
     * connect to the NCID servers configured
     * add the socket to poll()
     */
    int confservers = 0;
    int connservers = 0;
    for (i = 0; i < SERVERS; ++i)
    {
        if (ns[i].host)
        {
            polld[i].fd = socketConnect(i);
            confservers++;

            QueuedMsg_begin_connect(i);

            if ( polld[i].fd > 0 ) {
               connservers++;
               polld[i].revents = 0;
               polld[i].events = (POLLIN | POLLPRI);
            }
            else {
               /* try again in 30 seconds */
               ns[i].tries = 1;
               ns[i].next_try = time(0) + 30;
               logMsg(LEVEL1, "Next try after %d seconds.\n", 30);
            }
        }
        else
        {
            /* -1 tells poll() to ignore it */
            polld[i].fd = -1;
        }
    }

    logMsg(LEVEL1, "Successfully connected to %d of %d configured servers.\n", connservers, confservers);

    logMsg(LEVEL1,"End of startup: %s\n",
            strdate(WITHSEP));

    while (1)
    {
        int poll_sleep = POLL_TIMEOUT;

        if (TERM_flag || USR2_flag) poll_sleep = 0;
        switch (events = poll(polld, SERVERS, poll_sleep))
        {
            case -1:    /* error */
                if (errno != EINTR) /* No error for signals */
                    errorExit(-1, "poll", 0);
                /* fall through to treat interrupts the same as a time out */

            case 0:     /* timeout */
                if (TERM_flag)
                {
                   /* a TERM signal asked us to exit */
                   TERMexit();
                }
                if (USR2_flag)
                {
                   /* got signal SIGUSR2, list the connected clients */
                   USR2_flag = 0;
                   showConnected();
                }

                /* try to reconnect servers that disconnected */
                for (pos = 0; pos < SERVERS; pos++)
                {
                    if (polld[pos].fd < 0 && ns[pos].host)
                    {
                        /*
                         * try to connect when the server is disconnected
                         * polld[] and ns[] are in the same order
                         */
                        polld[pos].fd = socketConnect(pos);
                        if (polld[pos].fd > 0)
                        {
                            QueuedMsg_begin_connect(pos);

                            polld[pos].events = (POLLIN | POLLPRI);
                            if (warn && pos>0)
                            {
                                /* tell the tohost about this */
                                sprintf(rcvbuf, "%sMSG: %s%s %s:%d reconnected\n",
                                  PREFIX, hname, ns[pos].name, ns[pos].host, ns[pos].port);

                                safeSend(0, rcvbuf);
                            }
                            ns[pos].next_try = time(0) + 30;
                            ns[pos].tries = 0;
                        }
                        else {
                            /* try again in tries*30 seconds */
                            int nexttry = (ns[pos].tries<10?ns[pos].tries*30:300);
                            logMsg(LEVEL1, "Next try after %d seconds.\n", nexttry);
                            ns[pos].next_try = time(0) + nexttry;
                            ns[pos].tries++;
                        }
                    }
                    else
                    {
                       /* poke servers to see if they're still there */
                       safeSend(pos, "\n");

                    }
                    /* if a signal arrived, break out of this loop */
                    if (TERM_flag || USR2_flag) break;
                }
                break;
            default:    /* 1 or more polld entries have events */
                if (TERM_flag)
                {
                   /* a TERM signal asked ncidd to exit */
                   TERMexit();
                }
                for (pos = 0; events && pos < SERVERS; ++pos)
                {
                    if (!polld[pos].revents) continue; /* no events on this one */
                    --events;

                    if (polld[pos].revents & POLLHUP)
                    {
                        /* Disconnected (output only) */
                        logMsg(LEVEL1, "Client %d pos %d Disconnected: %s %s:%d %s\n",
                          polld[pos].fd, pos, ns[pos].name, ns[pos].host, ns[pos].port, strdate(WITHSEP));
                        disconnectClient(pos);
                    }
                    if (polld[pos].revents & POLLERR)
                    {
                        /* Poll Error (output only) */
                        logMsg(LEVEL1, "Poll Error, closed client %d pos %d: %s %s:%d %s\n",
                          polld[pos].fd, pos, ns[pos].name, ns[pos].host, ns[pos].port, strdate(WITHSEP));
                        disconnectClient(pos);
                    }
                    if (polld[pos].revents & POLLNVAL)
                    {
                        /* Invalid Request (output only) */
                        logMsg(LEVEL1,
                                "Removed client %d pos %d, invalid request: %s %s:%d %s\n",
                          polld[pos].fd, pos, ns[pos].name, ns[pos].host, ns[pos].port, strdate(WITHSEP));
                        disconnectClient(pos);
                    }
                    if (polld[pos].revents & POLLOUT)
                    {
                       /* a previously blocked socket has become writable */
                       send_blocked(pos);
                    }
                    if (polld[pos].revents & (POLLIN | POLLPRI))
                    {
                       int size;
                       char buf[BUFSIZ];

                        /* There is data to read */
                        logMsg(LEVEL4, "Reading client %d pos %d: %s %s:%d %s\n",
                          polld[pos].fd, pos, ns[pos].name, ns[pos].host, ns[pos].port, strdate(WITHSEP));

                        size = read(polld[pos].fd, buf, BUFSIZ-1);
                        if (size < 0)
                        {
                           logMsg(LEVEL1,
                              "Removed client %d pos %d, read failed: %s %s:%d %s\n",
                              polld[pos].fd, pos, ns[pos].name, ns[pos].host, ns[pos].port, strdate(WITHSEP));
                           disconnectClient(pos);
                        }
                        /* read will return 0 for a disconnect */
                        else if (size == 0)
                        {
                          /* TCP/IP Client End Connection */
                          disconnectClient(pos);
                          logMsg(LEVEL1, "%s %s:%d disconnected\n",
                             ns[pos].name, ns[pos].host, ns[pos].port);
                          if (warn && pos > 0)
                          {
                              /* tell tohost about lost fromhosts */
                              sprintf(msgbuf, "%sMSG: %s%s %s:%d disconnected\n",
                                    PREFIX, hname, ns[pos].name, ns[pos].host, ns[pos].port);
                              safeSend(0, msgbuf);
                          }
                        }
                        else
                        {
                          /*
                           * Client sent message to server
                           */

                          /* Terminate String */
                          buf[size] = '\0';

                          /*
                           * there is no guarantee that the message is complete, it's
                           * a stream, so call handle_incoming_message() to
                           * accumulate data until a newline character is found
                           */
                          handle_incoming_message(pos, buf, size);
                        }
                    }
                    polld[pos].revents = 0;
                }
                break;
        }
    }

    /* should not get here */
    cleanup();
    exit(0);
}


int getOptions(int argc, char *argv[])
{
    int c, num;
    int option_index = 0;
    char *ptr;
    static struct option long_options[] = {
        {"config", 1, 0, 'C'},
        {"debug", 0, 0, 'D'},
        {"fromhost", 1, 0, 'f'},
        {"help", 0, 0, 'h'},
        {"hostname-flag", 1, 0, 'H'},
        {"logfile", 1, 0, 'L'},
        {"pidfile", 1, 0, 'P'},
        {"tohost", 1, 0, 't'},
        {"verbose", 1, 0, 'v'},
        {"version", 0, 0, 'V'},
        {"warn", 1, 0, 'W'},
        {"osx-launchd", 0, 0, '0'},
        {0, 0, 0, 0}
    };

    while ((c = getopt_long (argc, argv, "f:hs:t:v:C:DH:L:P:VW:",
        long_options, &option_index)) != -1)
    {
        switch (c)
        {
            case '0':
                ++OSXlaunchd;
                break;
            case 'f': /* [host][:port] must contain host or port or both */
                if ((ptr = index(optarg, (int) ':')))
                {
                    if ((ns[1].port = atoi(ptr + 1)) == 0)
                        errorExit(-101, "Invalid port number", optarg);
                    if ((num = findWord("fromport1")) >= 0)
                        setword[num].type = 0;
                }
                if (optarg != ptr)
                {
                    if (ptr) *ptr = '\0';
                    if (!(ns[1].host = strdup(optarg))) errorExit(-1, name, 0);
                    if ((num = findWord("fromhost1")) >= 0)
                        setword[num].type = 0;
                }
                break;
            case 'h': /* help message */
                fprintf(stderr, DESC, name);
                fprintf(stderr, USAGE, name);
                exit(0);
            case 't': /* [host][:port] must contain host or port or both */
                if ((ptr = index(optarg, (int) ':')))
                {
                    if ((ns[0].port = atoi(ptr + 1)) == 0)
                        errorExit(-101, "Invalid port number", optarg);
                    if ((num = findWord("toport")) >= 0)
                        setword[num].type = 0;
                }
                if (optarg != ptr)
                {
                    if (ptr) *ptr = '\0';
                    if (!(ns[0].host = strdup(optarg))) errorExit(-1, name, 0);
                    if ((num = findWord("tohost")) >= 0)
                        setword[num].type = 0;
                }
                break;
            case 'v':
                verbose = atoi(optarg);
                /* valid range: 1-9 */
                if (strlen(optarg) != 1 || (verbose == 0))
                {
                    verbose = 1;
                    errorExit(-107, "Invalid number", optarg);
                }
                if ((num = findWord("verbose")) >= 0) setword[num].type = 0;
                break;
            case 'C':
                if (!(config = strdup(optarg)))
                    errorExit(-1, name, 0);
                break;
            case 'D':
                debug = 1;
                break;
            case 'H': /* add hostname to IDENT */
                HostnameFlag = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(HostnameFlag == 0 && *optarg == '0') && HostnameFlag != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("HostnameFlag")) >= 0) setword[num].type = 0;
                break;
            case 'L':
                if (!(logfile = strdup(optarg))) errorExit(-1, name, 0);
                break;
            case 'P':
                if (!(pidfile = strdup(optarg))) errorExit(-1, name, 0);
                break;
            case 'V': /* version */
                fprintf(stderr, SHOWVER, name, VERSION, API);
                exit(0);
            case 'W': /* warn users */
                warn = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(warn == 0 && *optarg == '0') && warn != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("warn")) >= 0) setword[num].type = 0;
                break;
            case '?': /* bad option */
                fprintf(stderr, USAGE, name);
                exit(-100);
        }
    }
    return optind;
}

/*
 * connect to the NCID server
 *
 *     returns socket created or -1 if not connected
 */
int socketConnect(int pos)
{
    int sd, size;
    char msgbuf[BUFSIZ], tmpbuf[BUFSIZ+120];
    struct sockaddr_in pin;
    struct hostent *hp;

    /* find out about the desired host machine */
    if ((hp = gethostbyname(ns[pos].host)) == 0) {
        logMsg(LEVEL1, "Connecting to pos %d: %s %s:%d %s\n",
           pos, ns[pos].name, ns[pos].host, ns[pos].port);
        errorExit(-1, "gethostbyname", hstrerror(h_errno));
    }

    /* fill in the socket structure with host information */
    memset(&pin, 0, sizeof(pin));
    pin.sin_family = AF_INET;
    pin.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
    pin.sin_port = htons(ns[pos].port);

    /* grab an Internet domain socket */
    if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1) errorExit(-1, "socket", 0);

    logMsg(LEVEL1, "Connecting to %d pos %d: %s %s:%d %s\n",
           sd, pos, ns[pos].name, ns[pos].host, ns[pos].port, strdate(WITHSEP));

    /* connect to ns[pos].port on ns[pos].host */
    if (connect(sd, (struct sockaddr *) &pin, sizeof(pin)) == -1)
    {
        /* connect failed */
        logMsg(LEVEL1,"Connect failed %d pos %d: %s  %s\n", sd, pos,
                    ns[pos].name, strerror(errno));
        /* if connect fails, close socket */
        close(sd);
        sd = -1;
    }
    else
    {
        /* connected */
        logMsg(LEVEL1, "Connected %d pos %d: %s %s:%d %s\n",
                sd, pos, ns[pos].name, ns[pos].host, ns[pos].port, strdate(WITHSEP));

      /* send ident to server */
      sprintf(msgbuf, "%sgateway %s%s %s\n", IDENT, HostnameFlag ? hname : "", name, VERSION);
      send(sd, msgbuf, strlen(msgbuf), 0);
      logMsg(LEVEL1, "%s%s=>%s: %sgateway %s%s %s\n", PRETTYINDENT, "", ns[pos].name,
              IDENT, HostnameFlag ? hname : "", name, VERSION);

      /* make sure call log not sent */
      sprintf(msgbuf, "%sno_log\n", COMMAND);
      send(sd, msgbuf, strlen(msgbuf), 0);
      logMsg(LEVEL1, "%s%s=>%s: %sno_log\n", PRETTYINDENT, "", ns[pos].name, COMMAND);

      /* log server startup messages */
      while ((size = readLine(sd, msgbuf)))
      {
        if (size < 0)
        {
           continue;
        }
        else if (size == 0) break;

        /* log 2x0 and 25x line messages */
        if (msgbuf[0] == '2' && (msgbuf[1] == '5' || msgbuf[2] == '0'))
        {
           sprintf(tmpbuf, "%s%s<=%s: %s", PRETTYINDENT, "", ns[pos].name, msgbuf);
           logMsg(LEVEL1, tmpbuf);
        }
        /* log Options */
        else if (!strncmp(msgbuf, "OPT: ", strlen("OPT: ")))
        {
           sprintf(tmpbuf, "%s%s<=%s: %s", PRETTYINDENT, "", ns[pos].name, msgbuf);
           logMsg(LEVEL1, tmpbuf);
        }
        /* log 300 line message and break - End of connection startup */
        else if (msgbuf[0] == '3' && msgbuf[2] == '0')
        {
           sprintf(tmpbuf, "%s%s<=%s: %s", PRETTYINDENT, "", ns[pos].name, msgbuf);
           logMsg(LEVEL1, tmpbuf);
           break;
        }
        /* lOGxxx: lines */
        else
        {
           sprintf(tmpbuf, "%s%s<=%s: %s", PRETTYINDENT, "", ns[pos].name, msgbuf);
           logMsg(LEVEL4, tmpbuf);
        }
      }
      logMsg(LEVEL1,"\n");

       /* set the client socket to non-blocking */
       if (fcntl(sd, F_SETFL, O_NONBLOCK) < 0)
       {
         logMsg(LEVEL2, "NONBLOCK Error: %s, sd: %d\n",
           strerror(errno), sd);
         close(sd);
         sd = -1;
       }
    }

    return sd;

}

/*
 * Reads one line from a socket, even if there is more than
 * one line.  If no lines, reads all data from socket.
 *
 * Must be called multiple times until zero is returned.
 * Returns -1, 0, or number of bytes read.
 *
 * Should properly handle the case where data does not end in '\n',
 * but this should only happen when receiving large log file so no
 * need to reconstruct a line before returning the data.
 */

int readLine(int sd, char *buf)
{
    int size;
    char *eptr;

    /* read data from socket without removing the data */
    size = recv(sd, buf, BUFSIZ, MSG_PEEK);
    if (size > 0)
    {
        /* find first line in buf */
        if ((eptr = strchr(buf, (int) '\n'))) size = (eptr - buf) + 1;

        memset(buf, '\0', BUFSIZ);
        /* read a line if there or all data in buf */
        size = recv(sd, buf, size, 0);
    }

    return size;
}

void cleanup(void)
{
    /* remove pid file, if it was created */
    if (pid)
    {
        unlink(pidfile);
        logMsg(LEVEL1, "Removed pidfile: %s\n", pidfile);
    }
}

/*
 * show connected clients
 */
void showConnected(void)
{
    int pos = 0;
    int cnt = 0;

    logMsg(LEVEL1, "Received Signal USR2: show connected at %s\n", strdate(WITHSEP));

    if (polld[0].fd > 0)
    {
        /* the tohost */
        logMsg(LEVEL1, "tohost %03d pos %03d %9s %s:%d is connected\n",
            polld[0].fd, pos, ns[0].name, ns[0].host, ns[0].port);
    }
    else
    {
        logMsg(LEVEL1, "tohost --- pos %03d %9s %s:%d is not connected\n",
            pos, ns[0].name, ns[0].host, ns[0].port);
    }


    for (pos = 1; pos < SERVERS; ++pos)
    {
        if (polld[pos].fd > 0)
        {
            /* empty polld slots are not clients */
            logMsg(LEVEL1, "Client %03d pos %03d %9s %s:%d is connected\n",
                polld[pos].fd, pos, ns[pos].name, ns[pos].host, ns[pos].port);
            cnt++;
        }
        else if (ns[pos].host)
        {
            logMsg(LEVEL1, "Client --- pos %03d %9s %s:%d is not connected after %d tries\n",
                pos, ns[pos].name, ns[pos].host, ns[pos].port, ns[pos].tries);
        }

    }
    if (cnt == 0)
    {
        logMsg(LEVEL1, "There are no clients connected at this time\n" );
    }
}

/*
 * USRhandler
 */

void    USRhandler(int sig)
{
  switch (sig)
  {
    case SIGHUP:
    case SIGUSR1:
    case SIGALRM:
      logMsg(LEVEL6, "Received Signal %d: %s\nIgnored: %s\n",
             sig, strsignal(sig), strdate(WITHSEP));
      break;
    case SIGUSR2:
       /* SIGUSR2 lists the connected clients */
      USR2_flag = 1;
      break;
    default:
      /* well that's odd */
      logMsg(LEVEL1, "Received unexpected Signal %d: %s\nIgnored: %s\n",
             sig, strsignal(sig), strdate(WITHSEP));
      break;
  }
}

/*
 * exit on a SIGINT, SIGQUIT, SIGABRT, SIGTERM signal
 */

void TERMhandler(int sig)
{
   /* set a flag so we can close everything neatly before leaving */
   TERM_flag = sig;
}

void TERMexit(void)
{
   cleanup();

   /* got signal SIGTERM via TERMhandler, clean up and leave */
   logMsg(LEVEL1, "Received Signal %d: %s\nTerminated: %s\n",
           TERM_flag, strsignal(TERM_flag), strdate(WITHSEP));

   /* close open files and sockets */
   exit(0);
}

/*
 * exit on a SIGSEGV signal
 * do minimal clean up after a memory exception
 */

void SEGVhandler(int sig)
{
    /* remove pid file, if it was created */
    if (pid) {
        unlink(pidfile);
    }

    /* and exit */
    _exit(sig);
}

/*
 * set up signal handlers
 */

void setup_signal_handlers(void)
{
   /* SIGHUP and SIGUSR1 not used */
   /* SIGUSR2 lists the connected clients */
   USR2_flag = 0;
   static struct sigaction user_sigaction;
   user_sigaction.sa_handler = USRhandler;
   sigfillset( &user_sigaction.sa_mask);
   user_sigaction.sa_flags = 0;
   sigaction(SIGHUP, &user_sigaction, NULL);
   sigaction(SIGUSR1, &user_sigaction, NULL);
   sigaction(SIGUSR2, &user_sigaction, NULL);
   sigaction(SIGALRM, &user_sigaction, NULL);

   /* SIGINT, etc, will stop ncid2ncid */
   TERM_flag = 0;
   static struct sigaction leave_sigaction;
   leave_sigaction.sa_handler = TERMhandler;
   sigfillset( &leave_sigaction.sa_mask);
   leave_sigaction.sa_flags = 0;
   sigaction(SIGINT, &leave_sigaction, NULL);  /* Interrupt from keyboard */
   sigaction(SIGQUIT, &leave_sigaction, NULL); /* Quit from keyboard */
   sigaction(SIGABRT, &leave_sigaction, NULL); /* Abort signal from abort(3) */
   sigaction(SIGTERM, &leave_sigaction, NULL); /* Termination signal */

   /* SIGSEGV, will try to remove the lock file */
   static struct sigaction segv_sigaction;
   segv_sigaction.sa_handler = SEGVhandler;
   sigfillset( &segv_sigaction.sa_mask);
   segv_sigaction.sa_flags = 0;
   sigaction(SIGSEGV, &segv_sigaction, NULL);    /* Invalid memory reference */

   /* ignore a bunch of them */
   static struct sigaction ignore_sigaction;
   ignore_sigaction.sa_handler = SIG_IGN;
   sigemptyset( &ignore_sigaction.sa_mask);
   ignore_sigaction.sa_flags = 0;
   sigaction(SIGTSTP, &ignore_sigaction, NULL);
   sigaction(SIGCHLD, &ignore_sigaction, NULL);    /* child process stopped */
   sigaction(SIGPIPE, &ignore_sigaction, NULL);    /* Broken pipe: write to pipe with no readers */
   sigaction(SIGTTIN, &ignore_sigaction, NULL);
   sigaction(SIGTTOU, &ignore_sigaction, NULL);
   sigaction(SIGIO, &ignore_sigaction, NULL);    /* io is waiting */
   sigaction(SIGWINCH, &ignore_sigaction, NULL); /* Window resize signal */

   /* set these back to the default */
   static struct sigaction default_sigaction;
   default_sigaction.sa_handler = SIG_DFL;
   sigfillset( &default_sigaction.sa_mask);
   default_sigaction.sa_flags = 0;
   /* sigaction(SIGSEGV, &default_sigaction, NULL);*/
   sigaction(SIGXCPU, &default_sigaction, NULL);
   sigaction(SIGXFSZ, &default_sigaction, NULL);
}

/* Log error, call cleanup() and exit */
void errorExit(int error, const char * msg, const char * arg)
{
    if (error == -1 && arg == 0)
    {
        /*
         * system error
         * print msg, arg should be zero
         */
        error = errno;
        logMsg(LEVEL1, "%s: %s\n", msg, strerror(errno));
    }
    else if (msg != 0)
    {
        /* internal program error */
        logMsg(LEVEL1, "%s: %s\n", msg, arg);
    }

    /* do not print terminated message, or cleanup, if option error */
    if (error != -100 && error != -101 && error != -106 && error != -107 &&
        error != -108 && error != -113)
    {
        cleanup();
        logMsg(LEVEL1, "Terminated:  %s\n", strdate(WITHSEP));
    }

    exit(error);
}


/*
 * Process a line from a client.
 */
void handle_accumulated_message(int pos, const char * rcvbuf)
{
  int ret = 0;
  char fwdbuf[BUFSIZ];

  /* Make sure there is data in the message line */
  if ((strlen(rcvbuf)) != 0)
  {
    /* handle an incoming line */
    if (pos == 0)
    {
      /* discard lines from the tohost */
      logMsg(LEVEL2, "\nLine received but ignored from client %d pos %d: %s %s:%d %s\n%s%s\n",
         polld[0].fd, 0, ns[0].name, ns[0].host, ns[0].port, strdate(WITHSEP),
         PRETTYINDENT, rcvbuf);
    }
    else
    {
      int i;
      /* data is from a 'FROM' server, keep lines with MSG: HUP: etc */
      for (i = 0; *lineType[i]; i++)
      {
        /* search for specific line types */
        if (!(ret = strncmp(lineType[i], rcvbuf,
            strlen(lineType[i])))) break;
      }
      if (!ret)
      {
        /* found line type to forward, add the +prefix and end with a newline */
        sprintf(fwdbuf, "%s%s\n", PREFIX, rcvbuf);
        safeSend(0, fwdbuf);

        /* line sent to 'TO' host */
        logMsg(LEVEL2, "\nLine received and forwarded to client %d pos %d: %s %s:%d %s\n%s%s",
                polld[0].fd, 0, ns[0].name, ns[0].host, ns[0].port, strdate(WITHSEP),
                PRETTYINDENT, fwdbuf);
      }
      else
      {
        logMsg(LEVEL2, "\nIgnored line from client %d pos %d, unrecognized line type: %s %s:%d %s\n%s%s\n",
           polld[pos].fd, pos, ns[0].name, ns[0].host, ns[0].port, strdate(WITHSEP),
           PRETTYINDENT, rcvbuf);
      }
    }
  }
  else
  {
     /* Found empty line */
     logMsg(LEVEL6, "Client %d sent empty line %s\n", polld[pos].fd, strdate(WITHSEP));
  }
} /* handle_accumulated_message */


/*
 * called from safesend disconnectClient to finish up anything
 * that is specific to ncid2ncid
 */
void finishDisconnect(int pos)
{
   /* ncid2ncid doesn't do anything extra */
}
