/*
 * sip2ncid.c - This file is part of sip2ncid.
 *
 * Copyright (c) 2005-2022
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * sip2ncid is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * sip2ncid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with sip2ncid.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "sip2ncid.h"

/* globals */
int debug, listdevs, msgsent, nofilter, test, sd, warn, rmdups, OSXlaunchd;
int HostnameFlag, outcall = 1;
int ncidport   = NCIDPORT;
int sipport    = SIPPORT;
int verbose    = 1;
int LineidSize = 4;
int ReadfilePacketDate = 1;
int linktype = -1;
int link_layer_header_size = 0;

volatile int TERM_flag;

const char * pidfile;
const char * ncidhost = LOCALHOST;
const char * logfile  = LOGFILE;
const char * device;
const char * siphost;
const char * packetstrdate(int format, struct timeval ts);
char *progname, *readfile, *writefile;
char hname[HOST_NAME_MAX + 1];
pid_t pid;
FILE *logptr;
pcap_t *descr;
pcap_dumper_t *dumpfile;

struct calls                    /* calls in progress */
{
    char * callid;
    char * callstart;
    char * calldate;
    char * line;
    int    outcall;
    int    invite;
    char * tonumber;
    char * fromname;
    char * fromnumber;
    time_t calltime;
} calls[MAXCALL];

char * linenum[MAXLINE];         /* telephone lines */

struct setword setword[] = {
   /* char *word;   int type; char **buf; int *value;   int min; int max */
   {"HostnameFlag",  WORDNUM,  0,          &HostnameFlag,  0,       1},
   {"interface",     WORDSTR,  &device,    0,              0,       0},
   {"LineidSize",    WORDNUM,  0,          &LineidSize,    1,       9},
   {"ncidhost",      WORDSTR,  &ncidhost,  0,              0,       0},
   {"ncidport",      WORDNUM,  0,          &ncidport,      0,       0},
   {"outcall",       WORDNUM,  0,          &outcall,       0,       1},
   {"pidfile",       WORDSTR,  &pidfile,   0,              0,       0},
   {"siphost",       WORDSTR,  &siphost,   0,              0,       0},
   {"sipport",       WORDNUM,  0,          &sipport,       0,       0},
   {"verbose",       WORDNUM,  0,          &verbose,       1,       MAXLEVEL - 1},
   {"warn",          WORDNUM,  0,          &warn,          OFF,     ON},
   {"rmdups",        WORDNUM,  0,          &rmdups,        OFF,     ON},
   {0,               0,        0,          0,              0,       0}
};

/* Needed for config.c in the library */
struct sendclient sendclient[1];

int getOptions(int argc, char *argv[]);
int pcapListDevs(void);
int parseLine(const char * sipbuf, const char * plabel, const char * llabel, char ** name, char ** number);
int getCallID(const char * sipbuf, char * callid, int size, const char * label);
int findCallID(const char * callid);
int addCallID(const char * callid, const char * line, struct timeval pktdate, int outcall, const char * fromname, const char * fromnumber, const char * tonumber);
void rmCallID(int pos);
int socketRead(void);
int callType(const char * sipbuf);
int readLine(int sd, char * buf);
void keepLineNumber(const char * sipnumber);
void cleanup(int error);
void doPCAP(void);
void errorExit(int error, const char * msg, const char * arg);
void setup_signal_handlers(void);
void USRhandler(int sig);
void TERMhandler(int sig);
void SEGVhandler(int sig);
void TERMexit(void);
void socketConnect(int fatal);
void processPackets(u_char *args, const struct pcap_pkthdr* pkthdr, const u_char* packet);

void printIndented(int level, const char * sipbuf, int indent);
void logRequest(int level, const char * request, const char * sipbuf);

void  handleInvite(const char * sipbuf, struct timeval pktdate);
void  handleRegister(const char * sipbuf, struct timeval pktdate);
void  handleCallEnd(const char * sipbuf, struct timeval pktdate, const char * ptype);
void  handleNotify(const char * sipbuf, struct timeval pktdate);
void  handleRefer(const char * sipbuf, struct timeval pktdate);
void  handleSubscribe(const char * sipbuf, struct timeval pktdate);
void  handleAck(const char * sipbuf, struct timeval pktdate);
void  handleInfo(const char * sipbuf, struct timeval pktdate);
void  handleResponse(const char * sipbuf, struct timeval pktdate);
void  doCallEnd(int pos, struct timeval pktdate, const char * ptype);

const char * lineLabel(const char * number);
const char * getToken(const char * inptr, char * wdptr);
const char * getLine(const char * inptr, char * lineptr);

int main(int argc, char *argv[])
{
    int argind, i, errnum = 0;
    char msgbuf[BUFSIZ];

    /* global containing name of program */
    progname = strrchr(argv[0], (int) '/');
    progname = progname ? progname + 1 : argv[0];

    /* Get the default config file */
    config = CONFIG;

    /* process options from the command line */
    argind = getOptions(argc, argv);

    /* should not be any arguments */
    if (argc - argind != 0)
    {
        fprintf(stderr, NOOPT, progname, argv[argind]);
        fprintf(stderr, USAGE, progname);
        exit(0);
    }

    if (listdevs)
    {
        pcapListDevs();
        exit(0);
    }

    if (test || readfile)
    {
        /* test mode and read dump file includes debug mode */
        debug = 1;
    }
    else
    {
        /* not in test or read dump mode, create or open existing logfile */
        logptr = fopen(logfile, "a");
        errnum = errno;
    }

    logMsg(LEVEL1, "Started: %s\nGateway: %s %s\n%s\n",strdate(WITHSEP),
            progname, VERSION, API);

    /* log command line and any options on separate lines */
    sprintf(msgbuf, "Command line: %s", argv[0]);
    for (i = 1; i < argc; i++)
    {
        if (*argv[i] == '-')
            strcat(strcat(msgbuf, "\n              "), argv[i]);
        else strcat(strcat(msgbuf, " "), argv[i]);
    }
    strcat(msgbuf, "\n");
    logMsg(LEVEL1, msgbuf);

    if (!test && !readfile)
    {
        /* not in test mode or dump mode, check status of logfile */
        if (logptr)
        {
            /* logfile opened */
            logMsg(LEVEL1, "logfile: %s\n", logfile);
        }
        else
        {
            /* logfile open failed */
            logMsg(LEVEL1, "%s: %s\n", logfile, strerror(errnum));
        }
    }

    /*
     * read config file, if present, exit on any errors
     * do not override any options set on the command line
     */
    if (doConf()) errorExit(-104, 0, 0);

    logMsg(LEVEL1, "Verbose level:  %d\n", verbose);

    logMsg(LEVEL1, "Size of displayed lineid: %d\n", LineidSize);

    /* global containing hostname */
    if (gethostname(hname, HOST_NAME_MAX) == -1)
    {
        logMsg(LEVEL1, "gethostname(): %s\n", strerror(errno));
        if (HostnameFlag)
        {
            HostnameFlag = 0;
            logMsg(LEVEL1, "Hostname flag reset to 0\n");
        }
    }
    else
    {
        logMsg(LEVEL1, "Hostname flag:  %d\n", HostnameFlag);
        if (HostnameFlag)
        {
        logMsg(LEVEL1, "Hostname: %s\n", hname);
        strcat(hname, "/");
        }
    }

    if (readfile)
    {
        logMsg(LEVEL1, "Reading from dumpfile: %s\n\n", readfile);
        logMsg(LEVEL1, "Use current or packet date/time?     : %s\n",
                ReadfilePacketDate ? "PACKET":  "CURRENT");
    }
    else if (writefile)
    {
        logMsg(LEVEL1, "Writing to dumpfile:   %s\n", writefile);
    }

    logMsg(LEVEL1,
        "Warn clients:   'No SIP packets' or 'SIP packets returned' message? %s\n",
            warn ? "YES" : "NO");

    if (rmdups) logMsg(LEVEL1, "The 'rmdups' option is obsolete and should be removed.\n\n");

    /* set up signal handlers */
    setup_signal_handlers();

    if (debug || test || readfile || OSXlaunchd)
    {
        if (debug) logMsg(LEVEL1, "Debug Mode");
        if (test)
        {
            logMsg(LEVEL1, " & Test Mode");
        }
        if (readfile)
        {
            logMsg(LEVEL1, " & Read Dump Mode");
        }
        if (debug) logMsg(LEVEL1, "\n");
        else logMsg(LEVEL1, "OSX Launchd Mode\n");
    }
    else
    {
        /* fork and exit parent */
        if(fork() != 0) return 0;

        /* close stdin and make fd 0 unavailable */
        close(0);
        if (open("/dev/null",  O_WRONLY | O_SYNC) < 0)
        {
            errorExit(-1, "/dev/null", 0);
        }

        /* become session leader */
        setsid();
    }

    /* must not be in test mode to create a PID file and connect to NCID */
    if (!test)
    {
        logMsg(LEVEL1,"NCID server at %s:%d\n", ncidhost, ncidport);

        if (doPID(&pid, pidfile))
        {
            sprintf(msgbuf,"%s already exists", pidfile);
            errorExit(-110, "Fatal", msgbuf);
        }

        socketConnect(FATAL);
    }

    doPCAP();

    /* should only get here after reading a dump file */
    cleanup(0);
    exit(0);
}


int getOptions(int argc, char *argv[])
{
    int c, num;
    int option_index = 0;
    char *ptr;
    static struct option long_options[] = {
        {"config", 1, 0, 'C'},
        {"debug", 0, 0, 'D'},
        {"help", 0, 0, 'h'},
        {"hostname-flag", 1, 0, 'H'},
        {"interface", 1, 0, 'i'},
        {"listdevs", 0, 0, 'l'},
        {"lineid-size", 1, 0, 'S'},
        {"logfile", 1, 0, 'L'},
        {"ncid", 1, 0, 'n'},
        {"pidfile", 1, 0, 'P'},
        {"readfile", 1, 0, 'r'},
        {"sip", 1, 0, 's'},
        {"testudp", 0, 0, 't'},
        {"testsip", 0, 0, 't'},
        {"testall", 0, 0, 'T'},
        {"verbose", 1, 0, 'v'},
        {"version", 0, 0, 'V'},
        {"writefile", 1, 0, 'w'},
        {"warn", 1, 0, 'W'},
        {"osx-launchd", 0, 0, '0'},
        {0, 0, 0, 0}
    };

    while ((c = getopt_long (argc, argv, "C:DhH:i:lL:n:P:r:s:S:tTv:Vw:W:",
        long_options, &option_index)) != -1)
    {
        switch (c)
        {
            case '0':
                ++OSXlaunchd;
                break;
            case 'C':
                if (!(config = strdup(optarg)))
                    errorExit(-1, progname, 0);
                break;
            case 'D':
                debug = 1;
                break;
            case 'h': /* help message */
                fprintf(stderr, DESC, progname);
                fprintf(stderr, USAGE, progname);
                exit(0);
            case 'H': /* add hostname to IDENT */
                HostnameFlag = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(HostnameFlag == 0 && *optarg == '0') && HostnameFlag != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("HostnameFlag")) >= 0) setword[num].type = 0;
                break;
            case 'i':
                if (!(device = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("interface")) >= 0) setword[num].type = 0;
                break;
            case 'l':
                listdevs = 1;
                break;
            case 'L':
                if (!(logfile = strdup(optarg))) errorExit(-1, progname, 0);
                break;
            case 'n': /* [host][:port] must contain host or port or both */
                if ((ptr = index(optarg, (int) ':')))
                {
                    if ((ncidport = atoi(ptr + 1)) == 0)
                        errorExit(-101, "Invalid port number", optarg);
                    if ((num = findWord("ncidport")) >= 0)
                        setword[num].type = 0;
                }
                if (optarg != ptr)
                {
                    if (ptr) *ptr = '\0';
                    if (!(ncidhost = strdup(optarg))) errorExit(-1, progname, 0);
                    if ((num = findWord("ncidhost")) >= 0)
                        setword[num].type = 0;
                }
                break;
            case 'P':
                if (!(pidfile = strdup(optarg))) errorExit(-1, progname, 0);
                break;
            case 'r':
                if (!(readfile = strdup(optarg))) errorExit(-1, progname, 0);
                test = 1;
                verbose = 3;
                break;
            case 's':
                if ((ptr = index(optarg, (int) ':')))
                {
                    if ((sipport = atoi(ptr + 1)) == 0)
                        errorExit(-101, "Invalid port number", optarg);
                    if ((num = findWord("sipport")) >= 0)
                        setword[num].type = 0;
                }
                if (optarg != ptr)
                {
                    if (ptr) *ptr = '\0';
                    if (!(siphost = strdup(optarg))) errorExit(-1, progname, 0);
                    if ((num = findWord("siphost")) >= 0)
                        setword[num].type = 0;
                }
                break;
            case 'S':
                LineidSize = atoi(optarg);
                /* valid range: 1-9 */
                if (strlen(optarg) != 1 || (LineidSize == 0))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("LineidSize")) >= 0) setword[num].type = 0;
                break;
            case 't':
                test = 1;
                verbose = 3;
                break;
            case 'T':
                test = 1;
                nofilter = 1;
                verbose = 3;
                break;
            case 'v':
                verbose = atoi(optarg);
                /* valid range: 1-9 */
                if (strlen(optarg) != 1 || (verbose == 0))
                {
                    verbose = 1;
                    errorExit(-107, "Invalid number", optarg);
                }
                if ((num = findWord("verbose")) >= 0) setword[num].type = 0;
                break;
            case 'V': /* version */
                fprintf(stderr, SHOWVER, progname, VERSION, API);
                exit(0);
            case 'w':
                if (!(writefile = strdup(optarg))) errorExit(-1, progname, 0);
                debug = 1;
                verbose = 3;
                break;
            case 'W': /* warn users */
                warn = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(warn == 0 && *optarg == '0') && warn != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("warn")) >= 0) setword[num].type = 0;
                break;
            case '?': /* bad option */
                fprintf(stderr, USAGE, progname);
                exit(-100);
        }
    }
    return optind;
}

/*
 * connect to the NCID server
 *     exit on connect error if fatal = 1
 *     continue on connect error if fatal = 0
 */
void socketConnect(int fatal)
{
    int size;
    char msgbuf[BUFSIZ];
    struct sockaddr_in pin;
    struct hostent *hp;

    /* find out about the desired host machine */
    if ((hp = gethostbyname(ncidhost)) == 0) {
        logMsg(LEVEL1, "Could not connect to ncidd at %s:%d\n",
           ncidhost, ncidport);
        errorExit(-1, "gethostbyname", hstrerror(h_errno));
    }

    /* fill in the socket structure with host information */
    memset(&pin, 0, sizeof(pin));
    pin.sin_family = AF_INET;
    pin.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
    pin.sin_port = htons(ncidport);

    /* grab an Internet domain socket */
    if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
    errorExit(-1, "socket", 0);

    /* connect to PORT on HOST */
    if (connect(sd, (struct sockaddr *) &pin, sizeof(pin)) == -1)
    {
        if (fatal) errorExit(-1, "NCID server", 0);
        logMsg(LEVEL2, "Warning: could not connect to ncidd at %s:%d\n",
                ncidhost, ncidport);
        /* if connect fails, close socket */
        close(sd);
        sd = 0;

    }
    else
    {
        /* connected */
        logMsg(LEVEL1, "Connected to ncidd at %s:%d\n", ncidhost, ncidport);

      /* send ident to server */
      sprintf(msgbuf, "%sgateway %s%s %s%s\n", IDENT, HostnameFlag ? hname : "",
              progname, VERSION, outcall ? GWOUT : "");
      send(sd, msgbuf, strlen(msgbuf), 0);
      logMsg(LEVEL1, "Sent: %sgateway %s%s %s%s\n", IDENT, HostnameFlag ? hname : "",
              progname, VERSION, outcall ? GWOUT : "");

      /* make sure call log not sent */
      sprintf(msgbuf, "%sno_log\n", COMMAND);
      send(sd, msgbuf, strlen(msgbuf), 0);
      logMsg(LEVEL1, "Sent: %sno_log\n", COMMAND);

      /* log server startup messages */
      while ((size = readLine(sd, msgbuf)))
      {
        if (size < 0)
        {
           if (fatal && errno != EINTR) errorExit(-1, progname, 0);
           continue;
        }
        else if (size == 0) break;

        /* log 2x0 and 25x line messages */
        if (msgbuf[0] == '2' && (msgbuf[1] == '5' || msgbuf[2] == '0'))
            logMsg(LEVEL1, msgbuf);
        /* log Options */
        else if (!strncmp(msgbuf, "OPT: ", strlen("OPT: "))) logMsg(LEVEL1, msgbuf);
        /* log 300 line message and break - End of connection startup */
        else if (msgbuf[0] == '3' && msgbuf[2] == '0')
        {
          logMsg(LEVEL1, msgbuf);
          break;
        }
        /* log lines */
        else logMsg(LEVEL4, msgbuf);
      }
    if (fcntl(sd, F_SETFL, O_NONBLOCK) < 0) errorExit(-1, "socket", 0);
   }
}

/*
 * Reads one line from a socket, even if there is more than
 * one line.  If no lines, reads all data from socket.
 *
 * Must be called multiple times until zero is returned.
 * Returns -1, 0, or number of bytes read.
 *
 * Should properly handle the case where data does not end in '\n',
 * but this should only happen when receiving large log file so no
 * need to reconstruct a line before returning the data.
 */

int readLine(int sd, char * buf)
{
    int size;
    char *eptr;

    /* read data from socket without removing the data */
    size = recv(sd, buf, BUFSIZ, MSG_PEEK);
    if (size > 0)
    {
        /* find first line in buf */
        if ((eptr = strchr(buf, (int) '\n'))) size = (eptr - buf) + 1;

        memset(buf, '\0', BUFSIZ);
        /* read a line if there or all data in buf */
        size = recv(sd, buf, size, 0);
    }

    return size;
}

int socketRead(void)
{
    int num;
    char msgbuf[BUFSIZ];

    if ((num = read(sd, msgbuf, BUFSIZ-1)) > 0)
    {
        msgbuf[num] = '\0';
        logMsg(LEVEL9, msgbuf);
    }
    return(num);
}

/*
 * List all network Devices
 */
int pcapListDevs(void)
{
    char errbuf[PCAP_ERRBUF_SIZE];
    char *desc, *ptr;
    pcap_if_t *alldevsp;

    if (pcap_findalldevs(&alldevsp, errbuf) < 0)
    {
        errorExit(-1, "pcap_findalldevs()", errbuf);
    }
    while (alldevsp)
    {
        if ((desc = alldevsp->description) == 0)
        {
            if (alldevsp->flags == PCAP_IF_LOOPBACK) desc = LOOPBACK;
            else if (!strncmp(alldevsp->name, "virbr", 5)) desc = VIRBR;
            else desc = NODESC;
        }

        /* If using WinPcap, path elements use '\' instead of '/' */
        while ((ptr = index(alldevsp->name, (int) '\\'))) *ptr = '/';

        fprintf(stdout, "%s : %s\n", alldevsp->name, desc);
        alldevsp = alldevsp->next;
    }
    pcap_freealldevs(alldevsp);
    return(0);
}

/* callback function that is passed to pcap_loop(..) and called each time
 * a packet is received
 */
void processPackets(u_char *args,
                    const struct pcap_pkthdr* pkthdr,
                    const u_char* packet)
{
    static unsigned int pktnum = 0;        /* packet number */
    static unsigned int charcnt;           /* character count */

    /* declare pointers to packet headers */
    const struct ip *ip;                   /* IP Header */
    const struct udphdr *udp;              /* UDP Header */
    const struct tcphdr *tcp;              /* TCP Header */
    const char   *pdata;                   /* Packet Data */
    struct timeval pktdate;

    int size_ip, size_udp, size_tcp, cnt, retval;
    int size_pdata = 0;
    int data_offset = 0;        /* the payload is in here somewhere */

    (void) retval;

    char sipbuf[BUFSIZ], warnmsg[BUFSIZ];

    char proto[4];

    if (!readfile) {alarm(PKTWAIT);} /* reset SIP packet timeout alarm */
    if (msgsent & 0x1)
    {
        /* log only one SIP packets returned message */
        logMsg(LEVEL1, "SIP packets returned: port %d %s\n",
                sipport, strdate(WITHSEP));

        if (warn && sd)
        {
            /*
            * send clients SIP OK message if warn option
            * set and if connected to the NCID server
            */
            sprintf(warnmsg, "MSG: SIP packets returned: port %d %s\n",
                sipport, strdate(WITHSEP));
            retval =  write(sd, warnmsg, strlen(warnmsg));
        }
    }
    msgsent = 0;    /* reset message log flag */

    /*
     * if socket is open:
     *  try to read data from server
     *  close socket if connection gone
     */
    if (sd)
    {
        /*
         * socketRead:
         *  returns  0 when socket no longer connected
         *  returns -1 when socket has no data to read
         *  returns a number when data read from socket
         */
        if ((cnt = socketRead()) == 0)
        {
            close(sd);
            sd = 0;
        }
        else if (cnt == -1)
        {
            if (charcnt)
            {
                logMsg(LEVEL3, "NCID server sent %d characters\n", charcnt);
                charcnt = 0;
            }
        }
        else charcnt += (unsigned) cnt;
    }

    /* write packet to dumpfile */
    if (writefile) pcap_dump(args, pkthdr, packet);

    /* if no open socket and not in test mode, try to connect again */
    if (!sd && !test) socketConnect(NOTFATAL);

    logMsg(LEVEL3, "\n%s\n\n", EQUALS);

    logMsg(LEVEL2, "Packet number: %u\n", ++pktnum);

    logMsg(LEVEL2, "  Packet date: %s\n", packetstrdate(WITHSEP, pkthdr->ts));

/*    if (debug && verbose >= HEXLEVEL) {
        hexdump(packet, pkthdr->caplen);
        logMsg(HEXLEVEL, "caplen %d len %d\n", pkthdr->caplen, pkthdr->len);
    } */

    /* packet may have BSD loopback encapsulation */
    if (linktype == DLT_NULL)
    {
        /* the loopback encapsulation is 4 bytes in host byte order
         * containing a value of
         * 2 for IPv4 packets
         * 24, 28, or 30 for IPv6 packets
         */
        int enc;
        memcpy(&enc, packet, sizeof(enc));
        logMsg(LEVEL4, "Loopback encapsulation type (IPv4 is 2) %d\n", enc);
        /* IPv4 can be enforced if not reading from a file */
        if (!readfile && enc != 2) return;
        data_offset += 4;
    }
    else {   /* DLT_EN10MB an ethernet packet */
        /* the packet starts with an ethernet header,
         * 12 bytes of mac address and a 2 byte type */
        struct ether_header * hd = (struct ether_header *)packet;
        data_offset += 14;        /* sizeof(struct ether_header) */

        /* a packet header may have a VLAN tag https://en.m.wikipedia.org/wiki/IEEE_802.1Q */
        switch (ntohs(hd->ether_type)) {
            case ETHERTYPE_IP:
                logMsg(LEVEL3, "Ethernet Type: IP\n");
                break;
            case ETHERTYPE_ARP:
                logMsg(LEVEL3, "Ethernet Type: ARP\n");
                return;
            case ETHERTYPE_VLAN:
                logMsg(LEVEL3, "Ethernet Type: VLAN tag 0x8100\n");
                data_offset += 4;
                /* check the type again */
                hd = (struct ether_header *)(packet+4);
                if (ntohs(hd->ether_type) == ETHERTYPE_ARP) {
                    logMsg(LEVEL3, "Ethernet Type is ARP\n");
                    return;
                }
                if (ntohs(hd->ether_type) != ETHERTYPE_IP) {
                    logMsg(LEVEL3, "Ethernet Type %x is not IP\n", ntohs(hd->ether_type));
                    return;
                }
                break;
            default:
                if (ntohs(hd->ether_type) < 0x0600) {
                    /* it's a length in 802.3 */
                    logMsg(LEVEL3, "Ethernet Type %x is a length\n", ntohs(hd->ether_type));
                    /* hexdump(packet, pkthdr->caplen);
                       logMsg(HEXLEVEL, "caplen %d len %d\n", pkthdr->caplen, pkthdr->len); */
                    return;
                }
                else {
                    logMsg(LEVEL3, "Ethernet Type %x is not IP\n", ntohs(hd->ether_type));
                    return;
                }
                break;
        }
    }

    /* the IP layer has a header of its own */
    ip = (struct ip *)(packet + data_offset);
    /* the length of the header is 4 bits of the header */
    size_ip = (ip->ip_hl)*4;
    data_offset += size_ip;
    if (size_ip < 20)
    {
        logMsg(LEVEL1, "   * Invalid IP header length: %u bytes\n", size_ip);
        return;
    }

    /* log source and destination IP addresses */
    logMsg(LEVEL4, "         From: %s\n           To: %s\n",
            inet_ntoa(ip->ip_src), inet_ntoa(ip->ip_dst));

    /* determine the IP protocol */
    switch(ip->ip_p) {
        case IPPROTO_TCP:
            logMsg(LEVEL3, "     Protocol: TCP\n");
            if (nofilter) return;
            strcpy(proto, PROTOTCP);
            break;
        case IPPROTO_UDP:
            logMsg(LEVEL3, "     Protocol: UDP\n");
            if (nofilter) return;
            strcpy(proto, PROTOUDP);
            break;
        case IPPROTO_ICMP:
            logMsg(LEVEL3, "     Protocol: ICMP\n");
            return;
        case IPPROTO_IP:
            logMsg(LEVEL3, "     Protocol: IP\n");
            return;
        default:
            logMsg(LEVEL3, "     Protocol: unknown %0x\n", ip->ip_p);
            return;
    }

    if (ip->ip_p == IPPROTO_UDP)
    {
       /* the UDP layer has a header of its own with two port numbers */
       udp = (struct udphdr *)(packet + data_offset);

       size_udp = sizeof (struct udphdr);
       data_offset += size_udp;

       /* log source and destination ports */
       logMsg(LEVEL4, "     Src port: %d\n     Dst port: %d\n",
               ntohs(udp->uh_sport), ntohs(udp->uh_dport));

       /* point pdata to the UDP payload */
       pdata = (char *)(packet + data_offset);

       /* compute UDP payload size */
       size_pdata = ntohs(ip->ip_len) - (size_ip + size_udp);
    }

    else if (ip->ip_p == IPPROTO_TCP)
    {
       /* the TCP layer has a header of its own with two port numbers */
       tcp = (struct tcphdr *)(packet + data_offset);

       size_tcp = tcp->th_off*4;
       data_offset += size_tcp;

       /* log source and destination ports */
       logMsg(LEVEL4, "     Src port: %d\n     Dst port: %d\n",
               ntohs(tcp->th_sport), ntohs(tcp->th_dport));

       /* point pdata to the TCP payload */
       pdata = (char *)(packet + data_offset);

       /* compute TCP payload size */
       size_pdata = ntohs(ip->ip_len) - (size_ip + size_tcp);
    }

    /* if packet has data */
    if (size_pdata > 0)
    {
        logMsg(LEVEL4, "     %s data: %d bytes:\n", proto, size_pdata);

        if (size_pdata < BUFSIZ)
            *(strncpy(sipbuf, pdata, size_pdata) + size_pdata) = '\0';
        else
        {
            *(strncpy(sipbuf, pdata, BUFSIZ - 1) + BUFSIZ - 1) = '\0';
            logMsg(LEVEL1, "Warning: SIP Packet truncated: %u > %u\n",
                    size_pdata, BUFSIZ - 1);
        }
        printIndented(LEVEL3, sipbuf, 15);
        logMsg(LEVEL3, "\n");
        if (verbose >= HEXLEVEL)
        {
            hexdump(pdata, size_pdata);
            logMsg(HEXLEVEL, "\n");
        }


        /* get the time from the packet or use the current time */
        if (ReadfilePacketDate)
            pktdate = pkthdr->ts;
        else
            (void) gettimeofday(&pktdate, 0);

        /*
         * Must be a SIP/2 packet
         */

        /* look for specific packets like INVITE, REGISTER, BYE, and CANCEL */
        /* look for SIP/2.0 nnn explanation responses */
        /* discard rubbish from clipped TCP messages */
        /* like http headers, these are new-line delimited */

        if (strncmp(sipbuf, "INVITE ", 7) == 0)
        {
           handleInvite(sipbuf, pktdate);
        }
        else if (strncmp(sipbuf, "REGISTER ", 9) == 0)
        {
           handleRegister(sipbuf, pktdate);
        }
        else if (strncmp(sipbuf, "BYE ", 4) == 0)
        {
           handleCallEnd(sipbuf, pktdate, BYE);
        }
        else if (strncmp(sipbuf, "CANCEL ", 7) == 0)
        {
           handleCallEnd(sipbuf, pktdate, CANCEL);
        }
        else if (strncmp(sipbuf, "NOTIFY ", 7) == 0)
        {
           handleNotify(sipbuf, pktdate);
        }
        else if (strncmp(sipbuf, "REFER ", 6) == 0)
        {
           handleRefer(sipbuf, pktdate);
        }
        else if (strncmp(sipbuf, "SUBSCRIBE ", 9) == 0)
        {
           handleSubscribe(sipbuf, pktdate);
        }
        else if (strncmp(sipbuf, "ACK ", 4) == 0)
        {
           handleAck(sipbuf, pktdate);
        }
        else if (strncmp(sipbuf, "INFO ", 5) == 0)
        {
           handleInfo(sipbuf, pktdate);
        }
        /* SIP/2.0 200 OK */
        else if (strncmp(sipbuf, "SIP/2.0 ", 8) == 0)
        {
           handleResponse(sipbuf, pktdate);
        }

    }

    return;
}


/*
 * INVITE Packets:
 *   INVITE sip:15553331212@192.168.111.21:10000 SIP/2.0
 *
 * An INVITE initiates a session dialog to set up a phone call.
 * Either end can send this, so the direction is deduced by
 * matching the contact on the INVITE line to an earlier REGISTER
 * request.
 */
void  handleInvite(const char * sipbuf, struct timeval pktdate)
{
    int outcall;
    char *sipnumber, *sipname, *type;
    char cidmsg[BUFSIZ], callid[CIDSIZ], tonumber[CIDSIZ],
         fromnumber[CIDSIZ], fromname[CIDSIZ];
    const char * line;
    int retval;

    (void) retval;

    /* log the request line */
    logRequest(LEVEL2, "Request", sipbuf);

    /* Get the unique Call-ID */
    getCallID(sipbuf, callid, sizeof(callid), INVITE);

    /* if Call-ID found, call in progress */
    if (findCallID(callid) != -1) return;

    /*
     * Get the destination from the INVITE request line:
     * INVITE sip:abc39f90-df4e-e7fc-2180-dbd4b476ccdc@10.11.12.5:51197 SIP/2.0
     *
     * if it matches the end that sends REGISTER then the call is incoming
     */
    outcall = callType(sipbuf);
    /* logMsg(LEVEL2, "outcall %d\n", outcall); */

    /*
     * Get called number from To line:
     *
     * To: <sip:CALLED_NMBR@IP_Address>;tag=TAG_NMBR
     */
    if (parseLine(sipbuf, INVITE, TO, 0, &sipnumber) == 0)
    {
        strcpy(tonumber, sipnumber);
    }
    else
    {
        tonumber[0]= '\0';
    }

    /*
     * Get Caller ID information from From line
     *
     * From: [["]NAME["]] <sip:CALLING_NMBR@IP_ADDR>;tag=TAG_NMBR
     */
    if (parseLine(sipbuf, INVITE, FROM, &sipname, &sipnumber) == 0)
    {
        strcpy(fromname, sipname);
        strcpy(fromnumber, sipnumber);

        /* set call info according to the call direction. */
        if (outcall)
        {
            /* Outgoing Call */
            line = lineLabel(fromnumber);
            sipnumber = tonumber;
            sipname = NONAME;
            type = CALLOUT;
        }
        else
        {
            /* Incoming Call */
            line = lineLabel(tonumber);
            sipnumber = fromnumber;
            sipname = fromname;
            type = CALLIN;
        }

        /* add Call-ID to calls in-progress table, if not filled */
        (void) addCallID(callid, line, pktdate, outcall, fromname, fromnumber, tonumber);

        sprintf(cidmsg, CALL, packetstrdate(NOYEAR, pktdate),
                    type, line, sipnumber, sipname);

        if (sd) retval =  write(sd, cidmsg, strlen(cidmsg));
        logMsg(LEVEL2, cidmsg);
    }
}


/*
 * REGISTER Packets:
 *   REGISTER sip:h.voncp.com:10000 SIP/2.0
 *
 * REGISTER is sent to a SIP registrar so the SIP user-agent
 * (the phone) can be found by others.
 *
 * sip2ncid stores the Contact info from these for later
 * use in deciding which direction the INVITE requests
 * are being sent
 *
 */
void  handleRegister(const char * sipbuf, struct timeval pktdate)
{
    char * sipnumber;

    /* log the request line */
    logRequest(LEVEL2, "Request", sipbuf);

    /*
     * Get the telephone line number from CONTACT line
     *
     * Contact: ["]NAME["]] <sip:NMBR@IP_ADDR:PORT>;expires=TIME
     *
     * Normally the contact line is the same as the "From:" line
     * but on a PBX the "From:" line is the extension and the
     * contact line is the number called.  Extensions need to
     * be handled properly.
     */
    if (parseLine(sipbuf, REGISTER, CONTACT, 0, &sipnumber) == 0)
    {
       keepLineNumber(sipnumber);
    }
}


/*
 * BYE Packets:
 *   BYE sip:15553331212@192.168.111.21:10000 SIP/2.0
 *
 *   Hangup After Answer -- end an already-established session
 *
 * CANCEL Packets:
 *   CANCEL sip:15553331212@192.168.111.21:10000 SIP/2.0
 *
 *   Hangup Before Answer -- cancel an INVITE in progress but unanswered
 *
 */
void  handleCallEnd(const char * sipbuf, struct timeval pktdate, const char * ptype)
{
    char callid[CIDSIZ];
    int pos;

    /* log the request line */
    logRequest(LEVEL2, "Request", sipbuf);

    /* Get the unique Call-ID */
    getCallID(sipbuf, callid, sizeof(callid), ptype);

    /* If Call-ID not found, return */
    if ((pos = findCallID(callid)) == -1) return;

    /*
     * Called and calling numbers came from a "To:" or "From:" line
     * in the original INVITE message for this Call-ID
     * To: [["]NAME["]] <sip:CALLING_NMBR@IP_ADDR>;tag=TAG_NMBR
     */
    doCallEnd(pos, pktdate, ptype);

}

/*
 * NOTIFY Packets:
 *   NOTIFY sip:abc39f90-df4e-e7fc-2180-dbd4b476ccdc@10.100.200.250:65135;transport=tcp SIP/2.0
 *   Via: SIP/2.0/TCP 10.100.200.250:5060;branch=z9hG4bK93419e5dcbbe8d
 *   From: <sip:57132@10.100.200.250>;tag=1765021653
 *   To: <sip:57132@10.176.221.3>
 *   Call-ID: 4abe3580-b191fd60-8a390c-2980300a@10.100.200.250
 *   CSeq: 101 NOTIFY
 *   Contact: <sip:57132@10.100.200.250:5060;transport=tcp>
 *   Messages-Waiting: yes
 *
 *   Mostly used by the server to communicate statistics
 *   which sip2ncid ignores.
 *   We're looking for a message waiting MWI indicator that
 *   can also be sent this way.
 */
void  handleNotify(const char * sipbuf, struct timeval pktdate)
{
    char *sipnumber, *sipname, *type;
    char cidmsg[BUFSIZ], tonumber[CIDSIZ];
    const char * line;
    int retval;

    (void) retval;

    /* log the request line */
    logRequest(LEVEL2, "Request", sipbuf);

   /*
    * Check for a Messages-Waiting packet
    */
   if ((!strstr(sipbuf, MESSAGES_ON)) && (!strstr(sipbuf, MESSAGES_OFF))) return;

   /*
    * Get called number from To line:
    *
    * To: <sip:CALLED_NMBR@IP_Address>
    */
   if (parseLine(sipbuf, NOTIFY, TO, 0, &sipnumber) == 0)
   {
       strcpy(tonumber, sipnumber);
       line = lineLabel(tonumber);
       sipnumber = MWINUMBER;
       sipname = strstr(sipbuf, MESSAGES_ON) ? MWINAME_ON : MWINAME_OFF;
       type = CALLMWI;

       sprintf(cidmsg, CALL, packetstrdate(NOYEAR, pktdate), type, line,
                   sipnumber, sipname);

       if (sd) retval =  write(sd, cidmsg, strlen(cidmsg));
       logMsg(LEVEL2, cidmsg);
   }
}


/*
 * REFER Packets:
 *   REFER sip:15553331212@192.168.111.21:10000 SIP/2.0
 *
 * Used by a client to subscribe for notification of
changes in call flow, for instance. Typically seen when a phone
requests to transfer a call to another party.
 */
void  handleRefer(const char * sipbuf, struct timeval pktdate)
{

    /* log the request line */
    logRequest(LEVEL2, "Request", sipbuf);

}


/*
 * SUBSCRIBE Packets:
 *   SUBSCRIBE sip:15553331212@192.168.111.21:10000 SIP/2.0
 *
 * Used by a SIP Phone to ask for statistics such as MWI information.
 *
 */
void  handleSubscribe(const char * sipbuf, struct timeval pktdate)
{

    /* log the request line */
    logRequest(LEVEL2, "Request", sipbuf);

}


/*
 * ACK Packets:
 *   ACK sip:15553331212@192.168.111.21:10000 SIP/2.0
 *
 * Used to acknowledge the end of the INVITE dialog.
 *
 */
void  handleAck(const char * sipbuf, struct timeval pktdate)
{
    char callid[CIDSIZ];
    int pos;

    /* log the request line */
    logRequest(LEVEL2, "Request", sipbuf);

    /* Get the unique Call-ID */
    getCallID(sipbuf, callid, sizeof(callid), ACK);

    /* If Call-ID not found, return */
    if ((pos = findCallID(callid)) == -1) return;

    /* No longer doing INVITE dialog */
    calls[pos].invite = 0;

}


/*
 * INFO Packets:
 *   INFO sip:15553331212@192.168.111.21:10000 SIP/2.0
 *
 * Provides additional signaling, possibly not for a call in progress.
 * Might contain an MWI (message waiting indicator)?
 *
 */
void  handleInfo(const char * sipbuf, struct timeval pktdate)
{

    /* log the request line */
    logRequest(LEVEL2, "Request", sipbuf);

}


/*
 *   SIP/2.0 100 Trying
 *   SIP/2.0 180 Ringing
 *   SIP/2.0 200 OK
 *   SIP/2.0 nnn Trying
 *   SIP/2.0 nnn Ringing
 *   SIP/2.0 183 Session Progress
 *   SIP/2.0 401 Unauthorized
 *   SIP/2.0 407 Proxy Authentication Required
 *   SIP/2.0 486 Busy Here
 *   SIP/2.0 487 Request Cancelled
 *   SIP/2.0 487 Request Terminated/terminated
 *
 * SIP Status Messages
 *  100-199 a temporary status like 180 Ringing
 *  200-299 a final success like 200 OK
 *  300-399 a final failure with forwarding information
 *  400-499 a final error from the SIP Proxy, such as 407 which says a password is needed
 *  500-599 the PBX Server has had an internal error
 *  600-699 a final error that is global and can't be handled by another SIP Proxy
 *
 * sip2ncid logs these, but ignores all but 487, using 487 to
 * make sure the call information has been deleted
 */
void  handleResponse(const char * sipbuf, struct timeval pktdate)
{
    char callid[CIDSIZ];
    int pos;

    const char * inptr;    /* words from the packet header */
    char word[BUFSIZ];     /* words from the packet header */
    char respcode[BUFSIZ]; /* words from the packet header */

    /* log the response line */
    logRequest(LEVEL2, "Status Response", sipbuf);

    /* e.g. SIP/2.0 100 Trying
     * skip over the SIP/2.0 part to get the code and text
     */
    inptr = getToken(sipbuf, word);
    if (inptr != NULL) {
       inptr = getToken(inptr, respcode);
       if (inptr != NULL) {
          /* got a reason code */
          /* rest of line is text
          inptr = getLine(inptr, word);
          logMsg(LEVEL2, " %s\n", word); */

          /* Get the unique Call-ID */
          getCallID(sipbuf, callid, sizeof(callid), respcode);

          /* If Call-ID not found, return */
          if ((pos = findCallID(callid)) == -1) return;

          /*
           * handle a 487 Request Terminated or Cancelled
           */
          if (strcmp(respcode, REQTERM) == 0)
          {
              /* The call wasn't cleared with the expected CANCEL or BYE */
              if (calls[pos].invite)
                 doCallEnd(pos, pktdate, CANCEL);
              else
                 doCallEnd(pos, pktdate, BYE);
              return;
          }
          /*
           * Handle a 486 Busy Here
           */
          if (strcmp(respcode, REQBUSY) == 0 )
          {
              /* The Busy response came without a CANCEL */
              doCallEnd(pos, pktdate, BUSY);
              return;
          }
       }
    }
}


/*
 * Send a CALLINFO message to the server, using the information
 * that has been stored in the calls table and then
 * remove the entry from the calls table
 *
 */
void  doCallEnd(int pos, struct timeval pktdate, const char * ptype)
{
    char *sipnumber, *sipname, *type;
    char cidmsg[BUFSIZ];
    int retval;

    (void) retval;

    /*
     * Called and calling numbers came from a "To:" or "From:" line
     * in the original INVITE message for this Call-ID
     * To: [["]NAME["]] <sip:CALLING_NMBR@IP_ADDR>;tag=TAG_NMBR
     */

    if (calls[pos].outcall)
    {
        /* Outgoing Call */
        sipnumber = calls[pos].tonumber;
        sipname = NONAME;
        type = CALLOUT;
    }
    else
    {
        /* Incoming Call */
        sipnumber = calls[pos].fromnumber;
        sipname = calls[pos].fromname;
        type = CALLIN;
    }

    sprintf(cidmsg, CALLINFO, ptype, calls[pos].calldate,
                calls[pos].callstart, packetstrdate(WITHSEP, pktdate),
                type, calls[pos].line, sipnumber, sipname);


    if (sd) retval =  write(sd, cidmsg, strlen(cidmsg));
    logMsg(LEVEL2, cidmsg);

    /* Remove Call-ID */
    rmCallID(pos);

}

/*
 * Determine the line label, usually last 4 digits
 * of the receiving call telephone number, but it
 * might be masked in which case it is anon [ymous]
 */
const char * lineLabel(const char * number)
{
    const char * label;

    if (isdigit(number[0]) && isdigit(number[1]))
        label = number + strlen(number) - LineidSize;
    else label = "ANON";

    return label;
}


/*
 * remember the Contact information from the REGISTER requests
 *
 */
void keepLineNumber(const char * sipnumber)
{
   int cnt, pos, empty;

   /* add phone number if not seen before */
   for (cnt = pos = empty = 0; cnt < MAXLINE; ++cnt)
   {
       /* finding the empty spot means they've all been examined */
       if (!linenum[cnt])
       {
           /* get 1st empty position in table */
           if (!empty) pos = cnt;
           empty = 1;
           continue;
       }
       if (!strcmp(linenum[cnt], sipnumber)) return;
   }


   if (empty)
   {
       linenum[pos] = strdup(sipnumber);
   }
   else
   {
       logMsg(LEVEL1,
           "Number of telephone lines exceeds: %u\n", MAXLINE);
   }
}


/*
 * Determine if call is incoming or outgoing
 * by comparing the INVITE destination
 * to the Contact that is advertised in the REGISTER requests
 *  Returns 0 if incoming call
 *  Returns 1 if outgoing call
 */

int callType(const char * sipbuf)
{
    int cnt, outcall = 1;
    char *sptr, *eptr, *elptr, *contact;
    char linebuf[BUFSIZ];

    /* make copy of input buffer */
    strcpy(linebuf, sipbuf);

    /* Terminate end of line */
    if ((elptr = index(linebuf, (int) NL)))
    {
        *elptr = '\0';

        /* Look for contact number */
        if ((sptr = strstr(linebuf, SIPCOLON)))
        {
            if ((eptr = index(sptr, (int) SIPAT)))
            {
                *eptr = '\0';
                contact = sptr + strlen(SIPCOLON);
            }
            else
            {
                logMsg(LEVEL1, "INVITE packet lacks a contact field %s\n", linebuf);
                contact = BADNUMBER;
            }
        }
        /* No contact number found on INVITE line */
        else contact = NONUMBER;
    }
    else contact = NONUMBER;


    /*
     * Determine if the call is incoming or outgoing.
     * If the INVITE contact is the same as the one being advertised
     * in REGISTER requests (stored in linenum), it is an incoming call.
     */
    for (cnt = 0; cnt < MAXLINE; ++cnt)
    {
        if (linenum[cnt])
        {
            logMsg(LEVEL4,
                "linenum[%d]=%s  contact=%s\n", cnt, linenum[cnt], contact);
        }
        else
        {
           /* empty table entry */
           break;
        }

        if (linenum[cnt] && !strcmp(linenum[cnt], contact))
        {
            logMsg(LEVEL4,
                "Found incoming INVITE contact in linenum[%d]=%s\n", cnt, linenum[cnt]);
            outcall = 0;
            break;
        }
    }

    /*
     * At least one PBX does not send out any REGISTER
     * packets so the linenum array is not populated
     * with telephone numbers receiving calls.
     */
    if (cnt >= MAXLINE)
    {
        /*
         * The linenum array has no entries or is filled.
         * Look for "Call-ID: call-" at the beginning of
         * the buffer.  If there, it is an outgoing call.
         *
         * This may not be always true and another check
         * or a different check may be needed.
         */
        outcall = 0;
        if (strstr(sipbuf, OUTCALL) != NULL) outcall = 1;
        logMsg(LEVEL4, "Checked for \"%s\" outcall= %d\n",
            OUTCALL, outcall);
    }
    else if (outcall) {
        logMsg(LEVEL4, "INVITE contact was not REGISTERed %s, outcall=1\n", contact);
    }
    return outcall;
}


/*
 * Parse Line for number and name, if requested.
 * Returns pointers into the static linebuf[].
 */

int parseLine(const char * sipbuf, const char * plabel, const char * llabel,
               char ** name, char ** number)
{
    int ret = 0;
    char *sptr, *eptr, *elptr;
    static char linebuf[BUFSIZ];

    /* make copy of input buffer */
    strcpy(linebuf, sipbuf);

    logMsg(LEVEL3, "Examining '%s' packet for '%s'\n", plabel, llabel);

    if ((sptr = strstr(linebuf, llabel)))
    {

        logMsg(LEVEL3, "Found '%s' ", llabel);

        /* Terminate end of line */
        if ((elptr = index(sptr, (int) NL)))
        {
            *elptr = '\0';

            if (name)
            {
                /* Look for NAME in quotes */
                if ((*name = index(sptr, (int) QUOTE)))
                {
                    if ((eptr = index(++*name, (int) QUOTE)))
                    {
                        *eptr = '\0';
                        sptr = ++eptr;
                    }
                    else
                    {
                        logMsg(LEVEL1,
                                "%s packet: Missing end of Name quote\n",
                                plabel);
                        *name = "BADNAME";
                    }
                }
                else
                {
                    /* NAME may not be in quotes */
                    *name = sptr + strlen(llabel) + 1;
                    if (**name != '<')
                    {
                        for (eptr=*name; !isspace((int) *eptr); ++eptr);
                        *eptr = '\0';
                        sptr = ++eptr;
                    }
                    else *name = NONAME;
                }
            }

            /* Look for calling number */
            if ((sptr = strstr(sptr, SIPNUM)))
            {
                if ((eptr = index(sptr, (int) SIPAT)))
                {
                    *eptr = '\0';
                    *number = sptr + strlen(SIPNUM);
                    /* skip a + or * at the start of the number */
                    if (**number == '+'|| **number == '*') (*number)++;
                }
                else
                {
                    logMsg(LEVEL1, "%s packet: %s Number Bad\n",
                            plabel, llabel);
                    *number = BADNUMBER;
                }
            }
            /* No calling number in packet */
            else *number = NONUMBER;

        }
        else
        {
            /* should not happen */
            logMsg(LEVEL1, "%s packet: %s line bad\n", plabel, llabel);
            ret = 1;
        }
    }
    else
    {
        /* should not happen */
        logMsg(LEVEL1, "%s packet: %s line not found\n", plabel, llabel);
        ret = 1;
    }
    if (name) logMsg(LEVEL3, " name '%s'", *name);
    logMsg(LEVEL3, " number '%s'\n", *number);

    return ret;
}


/*
 * indent contents of each line of a string
 * where a line is terminated by a newline
 */

void printIndented(int level, const char * sipbuf, int indent)
{

    char oneLine[BUFSIZ];
    char * oneptr = oneLine;
    const char * sipptr = sipbuf;

    while (1) {

       *oneptr = *sipptr;

       if (*sipptr == '\n' || *sipptr == '\r' || *sipptr == '\0')
       {
          /* break the line and log it if it's not empty */
          *oneptr = '\0';
          if (oneptr != oneLine)
             logMsg(level, "%*s%s\n", indent, " ", oneLine);

          if (*sipptr == '\0') break;

          /* start over */
          oneptr = oneLine;
          /* skip past the \n or \r */
          sipptr++;
       }
       else
       {
          oneptr++;
          sipptr++;
       }

    }
}


/*
 * print the first line of the request
 * where a line is terminated by a newline
 */
void    logRequest(int level, const char * request, const char * sipbuf)
{
    char oneLine[BUFSIZ];

    /*
    LEVEL1 - all the time
    LEVEL2 - warning, CALL, CALLINFO, and packet number/date
    LEVEL3 - packet parse progress & results
    LEVEL4 - everything
    LEVEL5 - reserved for hex dumps
    */
    if (verbose >= level)
    {
        /* get the first line from the SIP buffer */
        getLine(sipbuf, oneLine);

        if (oneLine[0] != '\0')
                 logMsg(level, "%s: %s\n", request, oneLine);
    }
}


/*
 * Returns the packet date and time as a string in the format:
 *      WITHSEP:     MM/DD/YYYY HH:MM:SS
 *      NOSEP:       MMDDYYYY HHMM
 *      ONLYTIME:    HH:MM:SS
 *      NOYEAR:      MMDDHHMM
 *      ONLYYEAR:    YYYY
 */
const char *packetstrdate(int format, struct timeval ts)
{
    static char buf[BUFSIZ];
    struct tm *tm;

    tm = localtime((time_t *) &(ts.tv_sec));
    switch (format)
    {
      case WITHSEP:
        sprintf(buf, "%.2d/%.2d/%.4d %.2d:%.2d:%.2d", tm->tm_mon + 1,
            tm->tm_mday, tm->tm_year + 1900, tm->tm_hour, tm->tm_min,
            tm->tm_sec);
        break;
      case NOSEP:
        sprintf(buf, "%.2d%.2d%.4d %.2d%.2d", tm->tm_mon + 1, tm->tm_mday,
            tm->tm_year + 1900, tm->tm_hour, tm->tm_min);
        break;
      case ONLYTIME:
        sprintf(buf, "%.2d:%.2d:%.2d",  tm->tm_hour, tm->tm_min, tm->tm_sec);
        break;
      case NOYEAR:
        sprintf(buf, "%.2d%.2d%.2d%.2d", tm->tm_mon + 1,
                tm->tm_mday, tm->tm_hour, tm->tm_min);
        break;
      case ONLYYEAR: sprintf(buf, "%.4d", tm->tm_year + 1900);
        break;
      default: /* error */
        logMsg(LEVEL2, "Unknown format: %d\n");
        break;
    }
    return buf;
}


/*
 * Get the unique Call-ID and return it as callid
 * return 0 if Call-ID not found
 * return 1 Call-ID found
 */
int getCallID(const char * sipbuf, char * callid, int size, const char * label)
{
    int len, ret = 0;
    char *sptr, *eptr;

    if ((sptr = strstr(sipbuf, CALLID)))
    {
        /* using sizeof() will skip end space */
        sptr += sizeof(CALLID);
        if ((eptr = index(sptr, (int) NL)))
        {
            len = eptr - sptr;
            if (len < size)
                *(strncpy(callid, sptr, len) + len) = '\0';
            else
            {
                *(strncpy(callid, sptr, size) + size - 1) = '\0';
                logMsg(LEVEL1, "Warning: %s packet - %s truncated: %u > %u\n",
                        label, CALLID, len, size - 1);
            }
            ret = 1;
        }
    }
    if (!ret)
    {
        *callid = '\0';
        logMsg(LEVEL1, "Warning: cannot get %s in %s\n", CALLID, label);
    }

    return ret;
}

/*
 * find the CallID in calls array
 * if found: return position
 * if not found: return -1
 */
int findCallID(const char * callid)
{
    int pos;

    for (pos = 0; pos < MAXCALL; ++pos)
    {
        if (calls[pos].callid && !strcmp(calls[pos].callid, callid))
        {
            /* found call in the calls in-progress table */
            logMsg(LEVEL3, "Found calls[%d]=%s\n", pos, calls[pos].callid);
            /* update the calltime to avoid reclaiming the entry */
            calls[pos].calltime = time(0);
            break;
        }
    }
    return (pos == MAXCALL ? -1 : pos);
}


/*
 * add the CallID, CallStart and other call info from the INVITE to calls array
 * if added: return position
 * if not added: return -1
 */
int addCallID(const char * callid, const char * line, struct timeval pktdate, int outcall,
        const char * fromname, const char * fromnumber, const char * tonumber)
{
    int pos, oldest = 0;
    time_t oldest_time = time(0);

    for (pos = 0; pos < MAXCALL; ++pos)
    {
        if (!calls[pos].callid)
        {
            /* add call to in-progress table */
            calls[pos].callid = strdup(callid);
            calls[pos].outcall = outcall;
            calls[pos].invite = 1;

            calls[pos].fromname = strdup(fromname);
            calls[pos].fromnumber = strdup(fromnumber);
            calls[pos].tonumber = strdup(tonumber);

            calls[pos].callstart = strdup(packetstrdate(WITHSEP, pktdate));
            calls[pos].calldate  = strdup(packetstrdate(NOYEAR, pktdate));
            calls[pos].calltime = time(0);

            calls[pos].line = strdup(line);
            logMsg(LEVEL3, "Added calls[%d].callid=%s\n", pos, callid);
            break;
        }
        else {
           /* note the oldest entry */
           if (calls[pos].calltime < oldest_time)
           {
              oldest_time = calls[pos].calltime;
              oldest = pos;
           }
        }
    }
    if (pos == MAXCALL)
    {
        /* The in-progress table is full. */
        logMsg(LEVEL1, "%d simultaneous calls exceeded\n", pos);
        /* reclaim the oldest calls entry to get some space
         * and issue CALLINFO for it using the current packet's date
         */
        pos = oldest;
        doCallEnd(pos, pktdate, UNKNOWN);

        /* add call to in-progress table */
        calls[pos].callid = strdup(callid);
        calls[pos].outcall = outcall;
        calls[pos].invite = 1;

        calls[pos].fromname = strdup(fromname);
        calls[pos].fromnumber = strdup(fromnumber);
        calls[pos].tonumber = strdup(tonumber);

        calls[pos].callstart = strdup(packetstrdate(WITHSEP, pktdate));
        calls[pos].calldate  = strdup(packetstrdate(NOYEAR, pktdate));
        calls[pos].calltime = time(0);

        calls[pos].line = strdup(line);
        logMsg(LEVEL3, "Added calls[%d].callid=%s\n", pos, callid);
    }
    return pos;
}

/*
 * Remove CallID, Call Start and line label from calls array
 * at a known position
 */
void rmCallID(int pos)
{
   logMsg(LEVEL3, "Removed calls[%d]=%s\n", pos, calls[pos].callid);
   free(calls[pos].callid);
   free(calls[pos].callstart);
   free(calls[pos].calldate);
   free(calls[pos].line);
   free(calls[pos].fromname);
   free(calls[pos].fromnumber);
   free(calls[pos].tonumber);
   calls[pos].callid = 0;
}

void doPCAP(void)
{
    int pcapret, retval;
    char errbuf[PCAP_ERRBUF_SIZE], filter_exp[BUFSIZ];
    char warnmsg[BUFSIZ+10];
    struct bpf_program fp;      /* hold compiled program     */
    bpf_u_int32 maskp;          /* subnet mask               */
    bpf_u_int32 netp;           /* ip                        */
    pcap_if_t * alldevs;

    (void) retval;

    if (readfile)
    {
        /* open a dump file for reading */
        descr = pcap_open_offline(readfile, errbuf);
        if(descr == NULL) errorExit(-1, "pcap_open_offline()", errbuf);
        device = readfile;
    }
    else
    {
        /* grab a device to peek into... */
        if (!device) {
           if (pcap_findalldevs(&alldevs, errbuf)<0)
              errorExit(-1, "pcap_findalldevs()", errbuf);
           if (alldevs == NULL)
              errorExit(-1, "pcap_findalldevs()", "no permitted devices found");

           device = strdup(alldevs->name);
           pcap_freealldevs(alldevs);
        }

        logMsg(LEVEL1, "Network Interface: %s\n", device);

        /* ask pcap for the network address and mask of the device */
        pcap_lookupnet(device, &netp, &maskp, errbuf);

        /* open device for reading in promiscuous mode */
        descr = pcap_open_live(device, BUFSIZ, 1, PCAPWAIT, errbuf);
        if(descr == NULL) errorExit(-1, "pcap_open_live()", errbuf);
    }
    linktype = pcap_datalink(descr);
    logMsg(LEVEL1, "Device '%s' type: %d - %s, description: %s\n",
            device, linktype,
            pcap_datalink_val_to_name(linktype),
            pcap_datalink_val_to_description(linktype));

    /*
     * make sure we're capturing a link type we can handle
     * http://www.tcpdump.org/linktypes.html
     */
    if (linktype != DLT_NULL && linktype != DLT_EN10MB)
        errorExit(-1, device, "is not Ethernet or loopback encapsulation");

    device = NULL;

    if (writefile)
    {
        /* open a dump file for writing */
        dumpfile = pcap_dump_open(descr, writefile);
        if (dumpfile == NULL)
            errorExit(-1, "pcap_dump_open()", pcap_geterr(descr));
    }

    /* do not apply filter if no filter is wanted or reading a dump file */
    if (!nofilter && !readfile)
    {
        /* create filter */
        if (siphost) sprintf(filter_exp, "host %s and ", siphost);
        else filter_exp[0] = '\0';
        sprintf(filter_exp + strlen(filter_exp), "port %d", sipport);

        /* compile the filter expression */
        if (pcap_compile(descr, &fp, filter_exp, 0, netp) == -1)
            errorExit(-1, filter_exp, pcap_geterr(descr));

        /* apply the compiled filter */
        if (pcap_setfilter(descr, &fp) == -1)
            errorExit(-1, filter_exp, pcap_geterr(descr));

        logMsg(LEVEL1, "Filter: %s\n", filter_exp);
    }
    else
    {
        logMsg(LEVEL1, "No filter applied\n");
    }

    while (1)
    {
        if (!readfile) {alarm(PKTWAIT);} /* set timer for SIP packet timeout */
        pcapret = pcap_loop(descr, -1, processPackets, (u_char *) dumpfile);

        /*
        * pcapret values
        *     #:  partial packet count
        *     0:  packet count reached, or end-of-file if file opened with pcap_open_offline
        *     -1: error
        *     -2: pcap_loop() terminated by pcap_breakloop()
        *
        * msgsent values
        *     0x0: no messages logged
        *     0x1: timeout message logged
        *     0x2: error message logged
        */


        if (TERM_flag)
        {
           /* a TERM signal asked us to exit */
           TERMexit();
        }
        if (!readfile)
        {
            logMsg(LEVEL4,
                "Alarm Timeout: pcap_loop(): return = %d, msgsent flag = %d\n",
                pcapret, msgsent);
        }

        if (msgsent ^ 0x3)
        {
            /* need to send either timeout or error message */
            if (pcapret == -2)
            {
                if ((msgsent & 0x1) == 0)
                {
                    /* log only one SIP packet timeout message */
                    logMsg(LEVEL1,
                            "No SIP packets in %d seconds: port %d %s\n",
                            PKTWAIT, sipport, strdate(WITHSEP));
                    msgsent |= 0x1;
                    if (warn && sd)
                    {
                        /*
                        * send clients one warning message if warn option
                        * set and if connected to the NCID server
                        */
                        sprintf(warnmsg, "MSG: No SIP packets in %d seconds: port %d %s\n",
                            PKTWAIT, sipport, strdate(WITHSEP));
                        retval =  write(sd, warnmsg, strlen(warnmsg));
                    }
                }
            }
            else if (pcapret == -1)
            {
                /* It appears that errors from pcap_loop aren't recoverable */
                 errorExit(-1, "pcap_loop", 0);
            }
            else if ((pcapret == 0) && readfile)
            {
                logMsg(LEVEL1, "End of dumpfile reached: %s\n", readfile);
                break;
            }
        }
    }
}


/*
 * Get the next word in a string
 * and return the next piece of string after the word
 *
 * Fetch more words with inptr = getToken(inptr, word);
 * A word is a series of non-space characters, or
 * everything between double quotes, or '='
 */
const char * getToken(const char * inptr, char * wdptr)
{
    char *endptr;

    if (inptr == NULL) return inptr;

    while (*inptr && isspace(*inptr)) inptr++;
    endptr = strchr(inptr, '\n');
    if (*inptr == 0 || *inptr == '\n' || *inptr == '\r') return 0;
    *wdptr = '\0';

    if (*inptr == '"')
    {
        ++inptr;
        if ((endptr = strchr(inptr, '"')) == NULL)
        {
            /* no close quote */
            return 0;
        }
        while (*inptr && inptr != endptr) *wdptr++ = *inptr++;
        if (*inptr) inptr++;
    }
    else while (*inptr && !isspace(*inptr) && *inptr != '"')
            *wdptr++ = *inptr++;

    *wdptr = '\0';

    return inptr;
}

/*
 * Get the next line from a buffer full of \n delimited lines.
 * Advance over space characters,
 * and then copy the rest of the inptr line to lineptr.
 * Returns a pointer to the next line.
 */
const char * getLine(const char * inptr, char * lineptr)
{
    if (inptr == NULL) return inptr;

    while (*inptr && isspace(*inptr)) inptr++;

    do {

       *lineptr = *inptr;

       lineptr++;
       inptr++;

    } while (*inptr != '\n' && *inptr != '\r' && *inptr != '\0');

    /* done */
    *lineptr = '\0';

    return inptr;
}


void cleanup(int error)
{
    /* remove pid file, if it was created */
    if (pid)
    {
        unlink(pidfile);
        logMsg(LEVEL1, "Removed pidfile: %s\n", pidfile);
    }

    /* close dumpfile if open for read or write */
    if (dumpfile) pcap_dump_close(dumpfile);
    if (descr) pcap_close(descr);

    if (test)
    {
        logMsg(LEVEL1, "%s terminated\n",
                readfile ? "Dump read" : "Test mode");
    }
    else if (error != -100 && error != -101 && error != -107)
    {
        /* do not print terminated message, if option error */
        logMsg(LEVEL1, "Terminated:  %s\n", strdate(WITHSEP));
    }
}


/*
 * USRhandler
 */

void    USRhandler(int sig)
{
  switch (sig)
  {
    case SIGHUP:
    case SIGUSR1:
    case SIGUSR2:
      logMsg(LEVEL6, "Received Signal %d: %s\nIgnored: %s\n",
             sig, strsignal(sig), strdate(WITHSEP));
      break;
    case SIGALRM:
      /* alarm timeout breaks out of pcap_loop()
       * to make sure packets are received */
      pcap_breakloop(descr);
      break;
    default:
      /* well that's odd */
      logMsg(LEVEL1, "Received unexpected Signal %d: %s\nIgnored: %s\n",
             sig, strsignal(sig), strdate(WITHSEP));
      break;
  }
}

/*
 * exit on a SIGINT, SIGQUIT, SIGABRT, SIGTERM signal
 */

void TERMhandler(int sig)
{
   /* set a flag so we can close everything neatly before leaving */
   TERM_flag = sig;
   /* break out of pcap_loop() to exit neatly */
   pcap_breakloop(descr);
}

void TERMexit(void)
{
   cleanup(0);

   /* got signal SIGTERM via TERMhandler, clean up and leave */
   logMsg(LEVEL1, "Received Signal %d: %s\nTerminated: %s\n",
           TERM_flag, strsignal(TERM_flag), strdate(WITHSEP));

   /* close open files and sockets */
   exit(0);
}

/*
 * exit on a SIGSEGV signal
 * do minimal clean up after a memory exception
 */

void SEGVhandler(int sig)
{
    /* remove pid file, if it was created */
    if (pid) {
        unlink(pidfile);
    }

    /* and exit */
    _exit(sig);
}

/*
 * set up signal handlers
 */

void setup_signal_handlers(void)
{
   /* SIGHUP, SIGUSR1 and SIGUSR2 not used */
   /* SIGALRM breakes out of the pcap loop */
   static struct sigaction user_sigaction;
   user_sigaction.sa_handler = USRhandler;
   sigfillset( &user_sigaction.sa_mask);
   user_sigaction.sa_flags = 0;
   sigaction(SIGHUP, &user_sigaction, NULL);
   sigaction(SIGUSR1, &user_sigaction, NULL);
   sigaction(SIGUSR2, &user_sigaction, NULL);
   sigaction(SIGALRM, &user_sigaction, NULL);

   /* SIGINT, etc, will stop sip2ncid */
   TERM_flag = 0;
   static struct sigaction leave_sigaction;
   leave_sigaction.sa_handler = TERMhandler;
   sigfillset( &leave_sigaction.sa_mask);
   leave_sigaction.sa_flags = 0;
   sigaction(SIGINT, &leave_sigaction, NULL);  /* Interrupt from keyboard */
   sigaction(SIGQUIT, &leave_sigaction, NULL); /* Quit from keyboard */
   sigaction(SIGABRT, &leave_sigaction, NULL); /* Abort signal from abort(3) */
   sigaction(SIGTERM, &leave_sigaction, NULL); /* Termination signal */

   /* SIGSEGV, will try to remove the lock file */
   static struct sigaction segv_sigaction;
   segv_sigaction.sa_handler = SEGVhandler;
   sigfillset( &segv_sigaction.sa_mask);
   segv_sigaction.sa_flags = 0;
   sigaction(SIGSEGV, &segv_sigaction, NULL);    /* Invalid memory reference */

   /* ignore a bunch of them */
   static struct sigaction ignore_sigaction;
   ignore_sigaction.sa_handler = SIG_IGN;
   sigemptyset( &ignore_sigaction.sa_mask);
   ignore_sigaction.sa_flags = 0;
   sigaction(SIGTSTP, &ignore_sigaction, NULL);
   sigaction(SIGCHLD, &ignore_sigaction, NULL);    /* child process stopped */
   sigaction(SIGPIPE, &ignore_sigaction, NULL);    /* Broken pipe: write to pipe with no readers */
   sigaction(SIGTTIN, &ignore_sigaction, NULL);
   sigaction(SIGTTOU, &ignore_sigaction, NULL);
   sigaction(SIGIO, &ignore_sigaction, NULL);    /* io is waiting */
   sigaction(SIGWINCH, &ignore_sigaction, NULL); /* Window resize signal */

   /* set these back to the default */
   static struct sigaction default_sigaction;
   default_sigaction.sa_handler = SIG_DFL;
   sigfillset( &default_sigaction.sa_mask);
   default_sigaction.sa_flags = 0;
   /* sigaction(SIGSEGV, &default_sigaction, NULL);*/
   sigaction(SIGXCPU, &default_sigaction, NULL);
   sigaction(SIGXFSZ, &default_sigaction, NULL);
}


/* Log error, call cleanup() and exit */
void errorExit(int error, const char * msg, const char * arg)
{
    if (error == -1 && arg == 0)
    {
        /* system error */
        error = errno;
        logMsg(LEVEL1, "%s: %s\n", msg, strerror(errno));
    }
    else if (msg != 0)
    {
        /* internal program error */
        logMsg(LEVEL1, "%s: %s\n", msg, arg);
    }

    cleanup(error);

    exit(error);
}
