/*
 * config.c - This file is part of ncidd.
 *
 * Copyright (c) 2005-2022
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"

int errorStatus;
int inQuotes;
const char * config;
int ignoreConfWord = 0;	/* default, all config lines must be in setword */

int doSet(const char * inptr, int lc);
void doSend(const char * inptr, int lc);

/*
 * Process the config file.
 * Returns:  0 - OK
 *          >0 - config error
 *          <0 - system error
 */

int doConf(void)
{
    char input[BUFSIZ], word[BUFSIZ], *ptr;
    const char * inptr;
    int lc;
    FILE *fp;

    if (config == NULL || (fp = fopen(config, "r")) == NULL)
    {
        logMsg(LEVELE, "No config file: %s\n", config);
        return 0;
    }

    /* read each line of file, one line at a time */
    for (lc = 1; fgets(input, BUFSIZ, fp) != NULL; lc++)
    {
        inptr = getWord(config, input, word, lc);

        /* line containing only <NL> or <CR><NL> or is a comment line*/
        if (inptr == 0 || input[0] == '\r' || word[0] == '#') continue;

        /* remove '\n' and '\r' if there */
        if ((ptr = strchr(input, (int) '\n'))) *ptr = '\0';
        if ((ptr = strchr(input, (int) '\r'))) *ptr = '\0';

        /* first word on line must be either "set", or "send" */
        if (!strcmp(word, "set"))
        {
            if (doSet(inptr, lc))
            {
                errorStatus = -1;
                break;
            }
        }
        else if (!strcmp(word, "send")) doSend(inptr, lc);
        else configError(config, lc, word, ERRCMD);
    }
    (void) fclose(fp);
    logMsg(LEVEL1, "Processed config file: %s\n", config);

    return errorStatus;
}

/*
 * process set lines:
 *        set word = value
 */

int doSet(const char * inptr, int lc)
{
    int num;
    char word[BUFSIZ];

    /* process configuration parameters */
    while ((inptr = getWord(config, inptr, word, lc)))
    {
        if (word[0] == '#')    break; /* rest of line is comment */

        if ((num = findWord(word)) < 0)
        {
            if (!ignoreConfWord)
                configError(config, lc, word, ERRWORD);
            break;
        }
        if (setword[num].type == WORDFLAG)
        {
            ++(*setword[num].value);
            break;
        }

        if (!(inptr = getWord(config, inptr, word, lc)))
        {
            configError(config, lc, word, ERREQA);
            break;
        }

        if (word[0] != '=')
        {
            configError(config, lc, word, ERREQB);
            break;
        }

        if (!(inptr = getWord(config, inptr, word, lc)))
        {
            configError(config, lc, word, ERRARG);
            break;
        }

        /* make sure config file does not override command line */
        if (setword[num].type)
        {
            if (setword[num].type & WORDSTR)
            {
                if (!(*setword[num].buf = strdup(word)))
                    return -1;

                /* set flag, if needed */
                if (setword[num].type & WORDFLAG) ++(*setword[num].value);
            }
            else
            {
                /* strtol indicates an error in errno */
                errno = 0;
                *setword[num].value = strtol(word, NULL, 0);
                if (errno != 0) configError(config, lc, word, ERRNUM);

                /* min is always tested, even if 0, max is not, if 0 */
                if ((*setword[num].value < setword[num].min) ||
                        (setword[num].max &&
                        *setword[num].value > setword[num].max))
                    configError(config, lc, word, ERRNUM);
            }
        }
        else
        {
            logMsg(LEVEL1, "Option: '%s' value in '%s' replaced by command line option\n",
                   setword[num].word, config);
        }
    }
    return 0;
}

int findWord(const char * wdptr)
{
    int i;

    for (i = 0; setword[i].word; i++)
        if (!strcmp(wdptr, setword[i].word)) return i;

    return -1;
}

/*
 * process send lines:
 *        send datatype [datatype] [...]
 */

void doSend(const char * inptr, int lc)
{
    int num;
    char word[BUFSIZ];

    /* process configuration parameters */
    while ((inptr = getWord(config, inptr, word, lc)))
    {
        if (word[0] == '#')    break; /* rest of line is comment */

        if ((num = findSend(word)) < 0)
        {
            if (!ignoreConfWord)
               configError(config, lc, word, ERRWORD);
            break;
        }

        ++(*sendclient[num].value);
    }
}

int findSend(const char * wdptr)
{
    int i;

    for (i = 0; sendclient[i].word; i++)
        if (!strcmp(wdptr, sendclient[i].word)) return i;

    return -1;
}

void configError(const char * file, int lc, const char * word, const char * mesg)
{
    if (*word == 0) return;
    logMsg(LEVELE, "%s: Line %d; %s %s\n", file, lc, mesg, word);
    ++errorStatus;
}

/*
 * a word is either a series of non-space characters,
 * everything between double quotes, or '='
 * returns a new value for inptr (null if nothing is left)
 * and copies the parsed word into wdptr
 * nameptr and lc are the filename and line number for error message
 * e.g.  /etc/ncid/ncidd.conf: Line 486; missing: "
 */

const char * getWord(const char * nameptr, const char * inptr, char * wdptr, int lc)
{
    char *endptr;

    if (inptr == 0) return 0;

    while (*inptr && isspace(*inptr)) inptr++;
    if ((endptr = strchr(inptr, '\n')) != NULL) *endptr = 0;
    if (*inptr == 0) return 0;
    *wdptr = 0;

    inQuotes = 0;
    if (*inptr == '"')
    {
        ++inptr;
        if ((endptr = strchr(inptr, '"')) == NULL)
        {
            configError(nameptr, lc, "\"", ERRMISS);
            return 0;
        }
        while (*inptr && inptr != endptr) *wdptr++ = *inptr++;
        if (*inptr) inptr++;

        /* inQuotes indicates the word is inside "" */
        inQuotes = 1;
    }
    else if (*inptr == '=') *wdptr++ = *inptr++;
    else while (*inptr && !isspace(*inptr) && *inptr != '=' && *inptr != '"')
            *wdptr++ = *inptr++;

    *wdptr = 0;

    return inptr;
}

/*
 * Copy a new set of config words into the global setword struct.
 * The new setword must be smaller than the old setword
 */
int resetConfEntries(struct setword * new_setwords)
{
   int i = 0;

   while(setword[i].word != NULL && new_setwords->word != NULL)
   {
      logMsg(LEVEL9, "resetConfEntries replacing: %d '%s' option with '%s'\n",
                   i, setword[i].word, new_setwords->word);
      memcpy(&setword[i], new_setwords, sizeof(struct setword));
      i++;
      new_setwords++;	/* point to the next row */
   }

   /* set the final entry to NULL */
   setword[i].word = NULL;
   setword[i].type = 0;

   /* no entries for sendclient */
   sendclient[0].word = NULL;

   if (new_setwords->word != NULL)
        /* it didn't all fit */
        return -1;
   else
        return 0;

}
