/*
 * safesend.c - This file is part of ncidd.
 *
 * Copyright (c) 2005-2021
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "safesend.h"

/* per-client table defined by the caller e.g. ncidd.c */
extern struct pollfd polld[];
void errorExit(int error, const char * msg, const char * arg);

/* number of connections that are expected e.g. MAXCONNECT in ncidd.c */
int maxconnections = 0;

/* a fifo of outstanding messages back to the client */
struct QM_FIFO_t
{
        char *		buffer;
        struct QM_FIFO_t *	next;
        char *		start;
        int		remaining;
};


/* QueuedMsg[pos] is for same client/gateway as */
/* in the polld[pos] per-connection stuff */
struct _queuedmsg_ {
        int		connected;	/* a client is connected */
        /* incoming stuff */
        char *		acc;		/* accumulator for incoming bytes */
        int		accptr;
        int		acc_size;	/* size of message accumulator */
        /* outgoing stuff */
        struct QM_FIFO_t *	insert_ptr; /* fifo of outstanding messages back to the client */
        struct QM_FIFO_t *	remove_ptr;
        int		blocked_bytes;	/* bytes on the fifo */
} * QueuedMsg = 0;

void initQueuedMsg(int maxconnect);
struct QM_FIFO_t * newQM_FIFO(const char * buf, int len, struct QM_FIFO_t * nxt);
void  freeQM_FIFO(struct QM_FIFO_t * ms);
void  queue_send(int pos, const char * buf, int len);
void  send_blocked(int pos);
void  wipe_queue(int pos);
void  disconnectClient(int pos);
void  safeSend(int pos, const char * msg );
void  close_socket(int pos);
void  QueuedMsg_begin_connect(int pos);
void  handle_incoming_message(int pos, char * um, int length);
void  handle_accumulated_message(int pos, const char * buf);


/*
=========================================================================
= Method	: initQueuedMsg
= Description	: initialise all of the QueuedMsg struct
=========================================================================
 */
void initQueuedMsg(int maxconnect)
{
   int pos;

   maxconnections = maxconnect;
   QueuedMsg = (struct _queuedmsg_ *)malloc(maxconnections*sizeof(struct _queuedmsg_));
   if (QueuedMsg == NULL ) {
      errorExit(-11, "initQueuedMsg", "safesend out of memory");
   }

   for (pos=0; pos < maxconnections; pos++) {
      /* tell poll to ignore this slot in the table */
      polld[pos].fd = -1;
      polld[pos].events = 0;

      QueuedMsg[pos].connected = 0;

      /* incoming stuff */
      QueuedMsg[pos].acc = NULL;
      QueuedMsg[pos].accptr = 0;
      QueuedMsg[pos].acc_size = 0;
      /* outgoing stuff */
      QueuedMsg[pos].insert_ptr = NULL;
      QueuedMsg[pos].remove_ptr = NULL;
      QueuedMsg[pos].blocked_bytes = 0;
   }
}


/*
=========================================================================
= Method	: close_socket
= Description	: close the client socket
=========================================================================
 */
void	 close_socket(int pos)
{
   if (QueuedMsg[pos].connected) {
      while (polld[pos].fd >= 0) {
         if (close(polld[pos].fd) < 0) {
            if (errno==EINTR) {
            /* interrupted, try again */
            }
            else {
               logMsg(LEVEL1, "Client %d pos %d close error %d: %s\n",
                  polld[pos].fd, pos, errno, strerror(errno));
               /* mark it for reuse */
               polld[pos].fd = -2;
            }
         }
         else {
            /* mark it for reuse */
            polld[pos].fd = -2;
         }
      }
      QueuedMsg[pos].connected = 0;
   }

} /* close_socket */


/*
=========================================================================
= Method	: QueuedMsg_begin_connect
= Description   : a client has connected
=========================================================================
 */
void	 QueuedMsg_begin_connect(int pos)
{
   if (QueuedMsg == 0) {
      errorExit(-11, "QueuedMsg_begin_connect", "safesend not initialized");
   }

   /* make an accumulator for the incoming message fragments */
   if (QueuedMsg[pos].acc == NULL) {
      QueuedMsg[pos].acc_size = 1024;		/* start with 1024 bytes */
      QueuedMsg[pos].accptr = 0;
      QueuedMsg[pos].acc = (char *)malloc(QueuedMsg[pos].acc_size);
   }

   if (QueuedMsg[pos].acc == NULL) {
      errorExit(-11, "QueuedMsg_begin_connect", "safesend out of memory");
   }

   QueuedMsg[pos].connected = 1;
}


/*
=========================================================================
= Method	: handle_incoming_message
= Description   : enters with buffer which may have any fragment of a
=		: line and more than one line of input
=========================================================================
 */
void	 handle_incoming_message(int pos, char * um, int length)
{
   /* logMsg(LEVEL9, "In handle_client_message %d >%s<\n", length,  um); */
   if (QueuedMsg == 0) {
      errorExit(-11, "handle_incoming_message", "safesend not initialized");
   }
   if (QueuedMsg[pos].acc == NULL) {
      errorExit(-11, "handle_incoming_message", "safesend pos not connected");
   }

   /* accumulate characters, looking for end of line */
   int nn = 0;
   /* for all the bytes that just arrived */
   while (nn < length) {
      QueuedMsg[pos].acc[QueuedMsg[pos].accptr] = um[nn];

      if (um[nn] == '\r' || um[nn] == '\n' || um[nn] == '\0') {
         if (um[nn] == '\r') {
            /* just eat 0x0d */
         }

         if (um[nn] == '\n') {
            /* ooo look! - an end of line */
            /* logMsg(LEVEL9, "In handle_client_message got eol %02x\n", um[nn]); */
            QueuedMsg[pos].acc[QueuedMsg[pos].accptr] = '\0';

            /* return to the caller with a complete line */
            handle_accumulated_message(pos, (char*)QueuedMsg[pos].acc);

            QueuedMsg[pos].accptr = 0;

            /* we may have closed the connection */
            if (! QueuedMsg[pos].connected) return;
         }
      }
      else if ((unsigned int)um[nn] <= '\037') {
         /* other control char; just eat it */
         /* unsigned char allows for unicode sequences e.g. C4 8C  */
         /* logMsg(LEVEL9, "In handle_client_message eat control char %02x\n", um[nn]); */
      }
      else {
         /* chars are accepted up to acc_size */
         if (QueuedMsg[pos].accptr < (QueuedMsg[pos].acc_size - 1)) QueuedMsg[pos].accptr++;
      }

      nn++;
   }

}


/*
=========================================================================
= Method	: disconnectClient
= Description	: disconnect and remove client from poll table
=========================================================================
 */
void disconnectClient(int pos)
{
   if (QueuedMsg == 0) {
      /* not intialized */
      errorExit(-11, "disconnectClient", "safesend not initialized");
   }

   close_socket(pos);
   polld[pos].fd = -1;
   polld[pos].events = 0;
   /* callback to allow for more close actions */
   finishDisconnect(pos);

   /* drop incoming stuff */
   QueuedMsg[pos].accptr = 0;
   QueuedMsg[pos].acc_size = 0;
   if ( QueuedMsg[pos].acc != NULL) free(QueuedMsg[pos].acc);
   QueuedMsg[pos].acc = NULL;
   /* drop outgoing stuff */
   wipe_queue(pos);
   QueuedMsg[pos].blocked_bytes = 0;

} /* disconnectClient */


/*
=========================================================================
= Method	: safeSend
= Description	: send the client a pre-formatted reply
=========================================================================
 */
void safeSend (int pos, const char * msg )
{
   int size = strlen(msg);

   if (QueuedMsg == 0) {
      /* not intialized */
      errorExit(-11, "safeSend", "safesend not initialized");
   }

   if (pos >= 0 && pos < maxconnections && polld[pos].fd >= 0) {
      logMsg(LEVEL9, "safeSend %d %d >%s<", polld[pos].fd, size,  msg);
      /* queue it and possibly send it */
      queue_send(pos, msg, size);
   }
   else {
      logMsg(LEVEL6, "safeSend called with stale client pos %d %d >%s<\n", pos, size,  msg);
   }

} /* safeSend */



/*
=========================================================================
= Method	: newQM_FIFO
= Description	: create a linked list item from the message buf
=========================================================================
 */
struct QM_FIFO_t * newQM_FIFO(const char * buf, int len, struct QM_FIFO_t * nxt)
{
   struct QM_FIFO_t * ms;

   ms = (struct QM_FIFO_t *)malloc(sizeof(struct QM_FIFO_t));
   if (ms != NULL ) {
      ms->next = nxt;
      /* the \0 string terminator is not needed */
      ms->buffer = (char *)malloc(len);
      if (ms->buffer != NULL ) {
         ms->start = ms->buffer;
         ms->remaining = len;
         memcpy(ms->buffer, buf, len);
      }
      else {
         /* out of memory */
         free(ms);
         ms = NULL;
      }
   }
   else {
      /* out of memory */
   }
   return ms;
}


/*
=========================================================================
= Method	: freeQM_FIFO
= Description	:
=========================================================================
 */
void freeQM_FIFO(struct QM_FIFO_t * ms)
{
   free(ms->buffer);
   free(ms);
}


/*
=========================================================================
= Method	: queue_send
= Description	: queue a message to be sent and maybe send it
=========================================================================
 */
void	queue_send(int pos, const char * buf, int len)
{
   int previously_blocked;
   struct QM_FIFO_t * ms;

   if (polld[pos].fd < 0) return;	/* socket already closed */
   if (! QueuedMsg[pos].connected) return;	/* socket already closed */

   ms = newQM_FIFO(buf, len, NULL);
   if ( ms == NULL ) {
      errorExit(-11, "safeSend", "safesend out of memory");
   }

   QueuedMsg[pos].blocked_bytes += len;

   if (QueuedMsg[pos].insert_ptr == NULL) {
      previously_blocked = 0;
      QueuedMsg[pos].insert_ptr = ms;
      QueuedMsg[pos].remove_ptr = ms;
   }
   else {
      previously_blocked = 1;
      /* append the newest message */
      QueuedMsg[pos].insert_ptr->next = ms;
      QueuedMsg[pos].insert_ptr = ms;
   }
   if (!previously_blocked) {
      /* try to send it */
      send_blocked(pos);
   }

} /* queue_send */


/*
=========================================================================
= Method	: send_blocked
= Description	: send some more from the blocked output queue
=========================================================================
 */
void	send_blocked(int pos)
{
   if (QueuedMsg == 0) {
      /* not intialized */
      errorExit(-11, "send_blocked", "safesend not initialized");
   }

   struct QM_FIFO_t * ms = QueuedMsg[pos].remove_ptr;
   /* send as much of the queue as possible */
   while (ms != NULL) {
      int sent_bytes = 0;
      do {
         sent_bytes = write(polld[pos].fd, ms->start, ms->remaining);
      } while (sent_bytes<0 && errno == EINTR);

      /* see how it fared */
      if (sent_bytes < 0) {
         if ( errno == EWOULDBLOCK || errno == EAGAIN ) {
            /* no more room in the queue, come back later */
            break;
         }
         else {
            /* other end has closed up */
            /* or some other error */
            logMsg(LEVEL2, "Client %d pos %d Disconnected - send_blocked error %d: %s\n",
                polld[pos].fd, pos, errno, strerror(errno));
            disconnectClient(pos);
            break;
         }

      }
      else {
         logMsg(LEVEL9, "send_blocked bytes %d %d\n", polld[pos].fd, sent_bytes);
         /* hopefully, some was sent */
         QueuedMsg[pos].blocked_bytes -= sent_bytes;
         /* adjust by sent_bytes */
         ms->remaining -= sent_bytes;
         ms->start += sent_bytes;
         if (ms->remaining == 0) {
            /* unlink this one from the chain */
            QueuedMsg[pos].remove_ptr = ms->next;
            if (QueuedMsg[pos].remove_ptr == NULL) QueuedMsg[pos].insert_ptr = NULL;
            freeQM_FIFO(ms);
            /* see if there's more */
            ms = QueuedMsg[pos].remove_ptr;
            /* loop some more */
         }
         else {
            /* come back later */
            break;
         }
      }
   }

   /* ask poll to give Write Event if there's more stuff to be sent */
   if (QueuedMsg[pos].remove_ptr != NULL) {
      polld[pos].events |= POLLOUT;
   }
   else {
      polld[pos].events |= POLLOUT;
      polld[pos].events ^= POLLOUT;
   }

} /* send_blocked */


/*
=========================================================================
= Method	: wipe_queue
= Description	: clear the blocked output queue
=========================================================================
 */
void	wipe_queue(int pos)
{
   while (QueuedMsg[pos].remove_ptr != NULL) {
      struct QM_FIFO_t * ms = QueuedMsg[pos].remove_ptr;
      /* unlink it from the chain */
      QueuedMsg[pos].remove_ptr = ms->next;
      if (QueuedMsg[pos].remove_ptr == NULL) QueuedMsg[pos].insert_ptr = NULL;
      /* delete it */
      freeQM_FIFO(ms);
   }
   QueuedMsg[pos].blocked_bytes = 0;

} /* wipe_queue */
