/*
 * strmatch.c - This file is part of ncidd.
 *
 * Copyright (c) 2005-2023
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "strmatch.h"
#include "logmsg.h"

/* instantiate the pcre error message things */
pcre2_match_data * pcreMatchData;
int pcreErrorNumber;
PCRE2_SIZE pcreErrorOffset;

/*
 * compare an incoming string to a precomputed regular expression
 * the regex can come from any of
 *    regcomp() (regex=1}
 *    pcre2_compile() (regex=2}
 *    store_astrmatch() (regex=0}
 *    store_bwstrmatch() (regex=0}
 *
 * for the case of regex=0 NCID Simple Expressions
 * ^ - at beginning: match from beginning (default for alias)
 * * - at beginning and/or end:  partial match (default for black/white lists)
 * * - by itself, match everything
 * $ - at the end: match to the end of the string
 * ^1? or 1? - at the beginning: match with or without a leading 1
 *
 * If ignore1 set, a leading 1 in "string" and/or "value" is discarded
 * before any of the comparison regex choices
 *
 * a leading ? on the string is always discarded
 *
 * returns: 0 = no match
 *          1 = match
 *          2 = only "*" in value
 *
 * using regex
 * returns: 0 = no match
 *          1 = match
 */

int strmatch(const char * string, void * regexp)
{
  int result = 0;
  char * ptr;

  if (*string && regexp)
  {
    /* skip leading '?', some phone systems do ?<name> */
    if (*string == '?') ++string;

    if (ignore1)
    {
        if (*string == '1') ++string;
    }

    if (regex == 1)
    {
      /* posix regular expressions */
      if (!regexec((regex_t*)regexp, string, 0, NULL, 0)) result = 1;
    }
    else if (regex == 2)
    {
      /* perl regular expressions */
      if (pcre2_match((pcre2_code*)regexp, (PCRE2_SPTR8)string, (int)strlen(string), 0, 0, pcreMatchData, NULL) > 0) result = 1;
    }
    else /* (regex == 0) */
    {
      /* ncidd comparisons */
      struct abwstr * test = (struct abwstr *)regexp;

      if (verbose >= LEVEL9)
      {
       logMsg(LEVEL9, "strmatch %s %c%c%c %d %s %c", string,
          test->skip1?'1':' ',
          test->beginstar?'*':' ',
          test->allstar?'*':' ',
          test->len,
          test->match,
          test->endstar?'*':' '
        );
      }
      /* easy match to * */
      if (test->allstar) return 2;
      /* skip a leading 1 */
      if (test->skip1 && *string == '1') ++string;

      /* look for *something in string */
      if (test->beginstar) {
          if ((ptr = strstr(string, test->match))) {
             /* move past the unwanted part */
             string = ptr;
          }
          else {
             /* nothing matched */
             logMsg(LEVEL9, " -- 0\n");
             return 0;
          }
      }

      if (test->endstar) {
        /* look for something* in string */
        result = !strncmp(string, test->match, test->len);
      }
      else {
        /* look for the exact string */
        result = !strcmp(string, test->match);
      }
      logMsg(LEVEL9, " -- %d\n", result);
    }
  }

  return result;
}


/*
 * store an alias expression for strmatch regex=0
 * supports ^1?*expression*$
 * the resulting abwstr struct stores precomputed flags for the
 * special matching characters ^ 1? * and $ plus a pointer to
 * the remaining substring to be matched
 */
struct abwstr * store_astrmatch(const char * string)
{
   struct abwstr * newncid = (struct abwstr *) malloc(sizeof(struct abwstr));
   /* clear all the flags */
   memset(newncid, 0, sizeof(struct abwstr));
   /* alias lists are implicitly ^expression$ */

   /* ignore leading ? in the alias string */
   if (*string == '?') ++string;

   if (*string == '^') {
      /* alias string is anchored to the beginning */
      string++;
   }
   if (ignore1) {
      if (strncmp(string, "1?", 2)==0) {
         /* not needed */
         string += 2;
      }
      else if (*string == '1') {
         /* not needed */
         string++;
      }
   }
   else {
      if (strncmp(string, "1?", 2)==0) {
         /* skip a leading 1 */
         newncid->skip1 = 1;
         string += 2;
      }
   }

   if (*string == '*') {
      string++;
      if (*string == '\0') {
         /* expression is a single * to match anything */
         newncid->allstar = 1;
      }
      else {
         /* expression begins with * */
         newncid->beginstar = 1;
      }
   }
   int tmplen = strlen(string);
   char * tmpstr = strdup(string);
   if (*(tmpstr + tmplen - 1) == '$') {
      /* remove and ignore the $ */
      *(tmpstr + tmplen - 1) = '\0';
      tmplen--;

   }
   if (*(tmpstr + tmplen - 1) == '*') {
      /* remove and note the * */
      *(tmpstr + tmplen - 1) = '\0';
      tmplen--;
      newncid->endstar = 1;
   }

   newncid->len = tmplen;
   newncid->match = tmpstr;

   return newncid;
}

/*
 * store a black or white list expression for strmatch regex=0
 * supports ^1?*expression*$
 * the resulting abwstr struct stores precomputed flags for the
 * special matching characters ^ 1? * and $ plus a pointer to
 * the remaining substring to be matched
 *
 * black, white lists are implicitly *expression*
 * so that ^419 will match any call from the 419 area code
 */
struct abwstr * store_bwstrmatch(const char * string)
{
   struct abwstr * newncid = (struct abwstr *) malloc(sizeof(struct abwstr));
   /* clear all the flags */
   memset(newncid, 0, sizeof(struct abwstr));

   /* black, white lists are implicitly *expression* */
   newncid->beginstar = 1;
   newncid->endstar = 1;

   if (*string == '^') {
      /* anchor the black/white string to the beginning */
      newncid->beginstar = 0;
      string++;
   }
   if (ignore1) {
      if (strncmp(string, "1?", 2)==0) {
         /* not needed */
         string += 2;
      }
   }
   else {
      if (strncmp(string, "1?", 2)==0) {
         /* skip a leading 1 */
         newncid->skip1 = 1;
         string += 2;
      }
   }

   if (*string == '*') {
      string++;
      if (*string == '\0') {
         /* expression is a single * to match anything */
         newncid->allstar = 1;
      }
      else {
         /* expression begins with * */
         newncid->beginstar = 1;
      }
   }
   int tmplen = strlen(string);
   char * tmpstr = strdup(string);
   if (*(tmpstr + tmplen - 1) == '$') {
      /* remove the $ and turn off * to anchor the string to the end */
      newncid->endstar = 0;
      *(tmpstr + tmplen - 1) = '\0';
      tmplen--;

   }
   if (*(tmpstr + tmplen - 1) == '*') {
      /* remove and note the * */
      *(tmpstr + tmplen - 1) = '\0';
      tmplen--;
      newncid->endstar = 1;
   }

   newncid->len = tmplen;
   newncid->match = tmpstr;

   return newncid;
}

/*
 * free the abwstr struct and the attached match string
 */
void free_abwstr(struct abwstr * ncidreg)
{
   if (ncidreg) {
      free(ncidreg->match);
      free(ncidreg);
   }
}

