/*
 * ncidd.c - This file is part of ncidd.
 *
 * Copyright (c) 2005-2024
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ncidd.h"
#include "nciddmodems.h"
#include "nciddPN.h"

/* globals */
const char * cidlog   = CIDLOG;
const char * datalog  = DATALOG;
const char * country  = DEFAULT_COUNTRY;
const char * language = DEFAULT_LANGUAGE;
const char * logfile  = LOGFILE;
const char * interfaceaddr   = "";
const char * pidfile;

char * progname;
char hostname[HOST_NAME_MAX];
int debug, sendlog, sendinfo;
int mainsock;       /* our listen socket for client connections */
int port = PORT;    /* listen port number */
volatile int HUP_flag, USR1_flag, USR2_flag, TERM_flag; /* interrupt flags */

time_t abw_lastread  = 0;          /* time of last ABW reads */
time_t checkABWtimer = 0;          /* to check for manual ABW changes */
time_t cidcall_lastchange = 0;         /* time of last cidcall change */
time_t checkCidCalltimer = 0;      /* to check for manual cidcall.log changes */
time_t hangupPauseTimer = 0;       /* to turn hangups off for a while */


int cidinput = -1;
int hangup;
int hupmode;
int nomodem, noserial; /* deprecated options, use cidinput */
int verbose = 1;
int ignore1;
int OSXlaunchd, sigpipe, sendlog_override, dialOUT;
long unsigned int cidlogmax = LOGMAXDEFAULT;
pid_t pid;

extern const char * addedmodems;      /* option: default - no added modems */
const char * announce = RECORDING;    /* option: DisconnectedNotInService.rmd" */
char annpath[BUFSIZ];                 /* checked path to announcement */
const char * hupname = HUPEXTENSION;  /* option: default - hangup-nohangup */
char huppath[BUFSIZ];                 /* checked path to hangup extension */
const char * huprmd = 0;              /* option: default - no ext announcent */
char rmdpath[BUFSIZ];                 /* checked path to ext announcement */

struct pollfd polld[MAXCONNECT];
FILE *logptr;
int dialdelay = DIAL_TIMEOUT;         /* how long user gets to pick up before modem hangs up */

extern char LineIDS[]; /* room for 5 lineid's */

/* IPinfo[pos] is for same client/gateway/modem as in polld[pos] */
struct modems_t;
struct ipinfo_t IPinfo[MAXCONNECT];

/* info noted by tcpAccept() that will be stored in IPinfo[] */
struct tcp_accpted_t {
   char IPaddr[20];
   char HostName[MAXIPBUF];
} tcpAccept_ret;

/* modem and per-call info */
struct cid_t gwcid = {0, CID, 0, "", "", "", "", NOMESG, "-"};

/* this is filled by getINFO() to return values from MSG: and NOT: */
struct mesg_t
{
    char date[CIDSIZE];
    char time[CIDSIZE];
    char nmbr[CIDSIZE];
    char name[CIDSIZE];
    char line[CIDSIZE];
    char type[CIDSIZE];
} getINFO_ret;

/* used in handleGWCallinfo() while building an END: line */
struct end_t
{
    char htype[CIDSIZE];
    char ctype[CIDSIZE];
    char  date[CIDSIZE];
    char  time[CIDSIZE];
    char scall[CIDSIZE];
    char pcall[CIDSIZE];
    char ecall[CIDSIZE];
    char  line[CIDSIZE];
    char  nmbr[CIDSIZE];
    char  name[CIDSIZE];
};

struct setword setword[] = {
    /* char *word; int type; char **buf; int *value; int min; int max */
    {"addedmodems", WORDSTR, &addedmodems,  0,          0,       0},
    {"announce",   WORDSTR,  &announce,  0,          0,       0},
    {"audiofmt",   WORDSTR,  &modem_list[0].audiofmt,  0,          0,       0},
    {"blacklist",  WORDSTR,  &blacklist, 0,          0,       0},
    {"cidalias",   WORDSTR,  &cidalias,  0,          0,       0},
    {"cidinput",   WORDNUM,  0,          &cidinput,  0,       3},
    {"cidlog",     WORDSTR,  &cidlog,    0,          0,       0},
    {"cidlogmax",  WORDNUM,  0,          (int*)&cidlogmax, 1, LOGMAXIMUM},
    {"cidnoname",  WORDNUM,  0,          &modem_list[0].cidnoname,    OFF,     ON},
    {"country",    WORDSTR,  &country,   0,          0,       0},
    {"datalog",    WORDSTR,  &datalog,   0,          0,       0},
    {"dialdelay",  WORDNUM,  0,          &dialdelay, 2,       10},
    {"hangup",     WORDNUM,  0,          &hangup,    0,       3},
    {"hupmode",    WORDNUM,  0,          &hupmode,   0,       3},
    {"hupname",    WORDSTR,  &hupname,   0,          0,       0},
    {"huprmd",     WORDSTR,  &huprmd,    0,          0,       0},
    {"gencid",     WORDNUM,  0,          &modem_list[0].gencid,    OFF,     ON},
    {"ignore1",    WORDNUM,  0,          &ignore1,   OFF,     ON},
    {"initcid",    WORDSTR|WORDFLAG, &modem_list[0].initcid, &modem_list[0].setcid, 0,    0},
    {"initstr",    WORDSTR,  &modem_list[0].initstr,   0,          0,       0},
    {"ifaddr",     WORDSTR,  &interfaceaddr,    0,          0,       0},
    {"language",   WORDSTR,  &language,  0,          0,       0},
    {"lineid",     WORDSTR,  &modem_list[0].serial_lineid,    0,          0,       0},
    {"lockfile",   WORDSTR,  &modem_list[0].lockfile,  0,          0,       0},
    {"nomodem",    WORDNUM,  0,          &nomodem,   OFF,     ON},
    {"noserial",   WORDNUM,  0,          &noserial,  OFF,     ON},
    {"pickup",     WORDNUM,  0,          &modem_list[0].pickup,    OFF,     ON},
    {"pidfile",    WORDSTR,  &pidfile,   0,          0,       0},
    {"port",       WORDNUM,  0,          &port,      0,       0},
    {"regex",      WORDNUM,  0,          &regex,     0,       2},
    {"ttyclocal",  WORDNUM,  0,          &modem_list[0].clocal,    OFF,     ON},
    {"ttyport",    WORDSTR,  &modem_list[0].ttyport,   0,          0,       0},
    {"ttyspeed",   WORDSTR,  &modem_list[0].TTYspeed,  0,          0,       0},
    {"nanp_format", WORDNUM, 0,          &nanp_format, 0,       3},
    {"verbose",    WORDNUM,  0,          &verbose,   1,       MAXLEVEL - 2},
    {"whitelist",  WORDSTR,  &whitelist, 0,          0,       0},
    {0,            0,        0,          0,          0,       0}
};

/* associate variables for config file "send cidlog" and "send cidinfo" */
struct sendclient sendclient[] = {
    /* char *word; int *value */
    {  "cidlog",   &sendlog   },
    {  "cidinfo",  &sendinfo  },
    {  0,          0          }
};

/* All line labels of interest to a client, new types must be added */
/* sorted alphabetically here for readability */
const char *lineTags[] =
{
    BLKLINE,
    CIDLINE,
    ENDLINE,
    HUPLINE,
    MSGLINE,
    MWILINE,
    NOTLINE,
    OUTLINE,
    PIDLINE,
    PUTLINE,
    RIDLINE,
    RLYLINE,
    WIDLINE,
    NULL
};

/* server lines sent to clients */
/* sorted alphabetically here for readability */
char *serverTags[] =
{
    "+BLK:",
    "+CID:",
    "+CIDINFO:",
    "+END:",
    "+HUP:",
    "+MSG:",
    "+MWI:",
    "+NOT:",
    "+OUT:",
    "+PID:",
    "+PUT:",
    "+RID:",
    "+RLY:",
    "+WID:",
    NULL
};

void writeClients(const char * inbuf);
void writeLog(const char * logf, const char * logbuf);
void cleanup(void);
void getINFO(const char * bufptr);
void getField(const char * bufptr, const char * field_name, char * mesgptr, const char * noval);
void normalExit(void);
void showConnected(void);
void doIDENT(int pos, const char * buf);
void initPoll(void);

void sendACK(int pos, const char * buf);
void sendCall(const char * linelabel, const struct cid_t * scid, const char * nameptr);
void doPoll(int events);
void acceptNewClient(void);
void readFromClient(int pos);
void sendInfo(int ring, const char * lineid);
void errorExit(int error, const char * msg, const char * arg);
void doCheckABWTimer(void);
void doCidCallTimer(void);
void doHELLO(int pos);
void doOptLine(int pos);
void sendLog(int pos);
void reload(void);
void hex2uchar(const char * buf, unsigned char *ucbuf);
void get_dial_log_nmbr(char * cidnmbr_log, char * cidnmbr_req);
void setup_signal_handlers(void);
void USRhandler(int sig);
void TERMhandler(int sig);
void SEGVhandler(int sig);
void TERMexit(void);
void handleGWCallinfo(int pos, const char * buf);
void handleClientRequest(int pos, const char * buf);
void handleReqReload(int pos, const char * buf);
void handleReqUpdate(int pos, const char * buf);
void handleReqDial(int pos, const char * buf);
void handleReqInfo(int pos, const char * buf);
void handleReqABW(int pos, const char * buf);
void handleReqPause(int pos, const char * buf);
void handleClientWork(const char * buf);
void formatCID(char ** lines, int pos);
void gatewayCID(const char * buf, int pos);
void finishCID(struct cid_t * fcid);
void TCIserial(const char * buf, int pos);
void xdmfCID(const char * buf, int pos);
void modemCID(const char * buf, int pos);
void checkHangupFiles(void);
void checkHangupExtension(void);
void check_cidlog_date(void);
int check_ABW_dates(const char * filename);
int getGWField(const char * bufptr, const char * field_name, char * mesgptr, const char * noval);
int runExternalHangup(const char ** linelabel, const struct cid_t * ecid, const char ** nameptr, const char ** rmdptr);

int getOptions(int argc, char *argv[]);
int tcpOpen(void);
int tcpAccept(void);
int checkVoiceFile(char * path, const char * recording);
int nothex(const char * mesg);
int addPoll(int addfd);
int isSerialDevice(int pos);

char * trimWhitespace(char *sptr, char *eptr);
const char * getLineLabel(int calltype);
unsigned char checksum(unsigned char *ptr, size_t size);

void set_mesg_standard(const char * country);

/*
 *  The main program
 */

int main(int argc, char *argv[])
{
    int events, argind, i, fd, errnum, ret;
    struct stat statbuf;
    struct utsname utsbuf;
    char msgbuf[BUFSIZ];

    /* set up signal handlers for HUP, USR1, USR2, INT, etc */
    setup_signal_handlers();

    /* global containing name of program */
    progname = strrchr(argv[0], (int) '/');
    progname = progname ? progname + 1 : argv[0];

    /* get the nostname */
    if (gethostname(hostname, HOST_NAME_MAX) == -1)
    {
        logMsg(LEVEL1, "gethostname(): %s\n", strerror(errno));
    }
    hostname[HOST_NAME_MAX-1] = '\0';

    /* Get the default config file */
    config = CONFIG;

    /* apply default values for the first modem */
    init_modem(&modem_list[0]);
    modem_list[0].ttyport  = TTYPORT;


    /* process options from the command line */
    argind = getOptions(argc, argv);

    /* should not be any arguments */
    if (argc - argind != 0)
    {
        fprintf(stderr, NOOPT, progname, argv[argind]);
        fprintf(stderr, USAGE, progname);
        exit(0);
    }

    /* open or create logfile */
    logptr = fopen(logfile, "a+");
    errnum = errno;

    logMsg(LEVEL1, "Started: %s\nServer: %s %s\n%s\n",strdate(WITHSEP),
            progname, VERSION, API);

    /* uname system call information */
    if ((ret = uname(&utsbuf) == -1))
    {
        logMsg(LEVELE, "uname system call failed with errno = %d\n", errno);
    }
    else
    {
        logMsg(LEVEL1, "Sysname:  %s\nNodename: %s\nMachine:  %s\nRelease:  %s\nVersion:  %s\n",
             utsbuf.sysname, utsbuf.nodename, utsbuf.machine, utsbuf.release, utsbuf.version);
    }

    /* log command line and any options on separate lines */
    sprintf(msgbuf, "Command line: %s", argv[0]);
    for (i = 1; i < argc; i++)
    {
        if (*argv[i] == '-')
            strcat(strcat(msgbuf, "\n              "), argv[i]);
        else strcat(strcat(msgbuf, " "), argv[i]);
    }
    strcat(msgbuf, NL);
    logMsg(LEVEL1, msgbuf);

    /* check status of logfile */
    if (logptr)
    {
        /* logfile opened */
        logMsg(LEVEL1, "Logfile: %s\n", logfile);
    }
    else
    {
        /* logfile open failed */
        logMsg(LEVELE, "%s: %s\n", logfile, strerror(errnum));
    }

    /*
     * read config file, if present, exit on any errors
     * do not override any options set on the command line
     */
    if ((ret = doConf()))
    {
        if (ret > 0) errorExit(-104, 0, 0);
        else errorExit(-1, progname, 0);
    }

    logMsg(LEVEL1, "Verbose level: %d\n", verbose);

    /* I have CIDSIZE 256 BUFSIZ 8192, safesend clips incoming messages at 1024 */
    logMsg(LEVEL4, "CIDSIZE %d BUFSIZ %d\n", CIDSIZE, BUFSIZ);

    /* for compatibility with ncidd <= 1.6 if cidinput == -1 */
    if (cidinput < 0)
    {
        if (noserial)
        {
            if (hangup || hupmode) cidinput = 0x2;
            else cidinput = 0x3;
        }
        else cidinput = (nomodem & 0x1);
        if (noserial || nomodem)
        {
            logMsg(LEVEL1, "noserial = %d and nomodem = %d were used to create cidinput\n", noserial, nomodem);
        }
    }

    if (noserial || nomodem)
    {
        logMsg(LEVELE, "***WARNING*** noserial and nomodem are obsolete, use cidinput instead\n");
    }

    logMsg(LEVEL1, "cidinput = %d\n", cidinput);

    logMsg(LEVEL1, "sendinfo = %d\n", sendinfo);

    /* check hangup/hupmode against cidinput 1 and 3 */
    if (cidinput & 0x1)
    {
        /* no modem */

        if (hangup)
        {
            sprintf(msgbuf,
            "A modem must be used with the internal hangup option. cidinput=%d hangup=%d", cidinput, hangup);
            errorExit(-110, "Fatal", msgbuf);
        }

        if (hupmode)
        {
            sprintf(msgbuf,
            "A modem must be used with a hangup extension. cidinput=%d hupmode=%d", cidinput, hupmode);
            errorExit(-110, "Fatal", msgbuf);
        }
    }

    /*
     * indicate what is configured to send to the clients
     */
    for (i = 0; sendclient[i].word; i++)
        if (*sendclient[i].value)
        {
            logMsg(LEVEL1, "Configured to send '%s' to clients.\n",
                sendclient[i].word);
        }

    /*
     * indicate location of helper scripts
     */
    logMsg(LEVEL1, "Helper tools:\n    %s\n    %s\n", NCIDUPDATE, NCIDUTIL);

    /*
     *     check if country code is valid
     */
    if (findRegionCode(country) != 0)
    {
      logMsg(LEVEL1, "country code set to : %s \n", country);
      set_mesg_standard(country);
    }
    else
    {
      sprintf(msgbuf, "%s is not a valid country code\n", country);
      errorExit(-110, "Fatal", msgbuf);
    }

    /* NANP country format */
    switch (nanp_format)
    {
      case 0:
        logMsg(LEVEL1, "no format (%d): xxxxxxxxxx\n", nanp_format);
        break;
      case 1:
        logMsg(LEVEL1, "google standard format (%d): (xxx) xxx-xxxx\n", nanp_format);
        break;
      case 2:
        logMsg(LEVEL1, "all-hyphen format (%d): xxx-xxx-xxxx\n", nanp_format);
        break;
      case 3:
        logMsg(LEVEL1, "all-dot format (%d): xxx.xxx.xxxx\n", nanp_format);
        break;
    }


    switch (regex)
    {
        case 0:
            logMsg(LEVEL1, "Using simple expressions for alias/blacklist/whitelist entries\n");
            break;
        case 1:
            logMsg(LEVEL1, "Using posix regular expressions for alias/blacklist/whitelist entries\n");
            break;
        case 2:
            logMsg(LEVEL1, "Using Perl regular expressions for alias/blacklist/whitelist entries\n");
            break;
        default:
            sprintf(msgbuf, "Invalid regex number: %d", regex);
            errorExit(-107, "Fatal", msgbuf);
            break;
    }

    logMsg(LEVEL1,
      "\nBegin: Loading alias, blacklist and whitelist files [%s]\n", strdate(ONLYTIME));

    /* read alias file, if present, exit on any errors */
    if (readAliasFile()) errorExit(-109, 0, 0);

    /* read blacklist and whitelist files, exit on any errors */
    if (readBWlists()) errorExit(-114, 0, 0);

    if (hangup) logMsg(LEVEL1, "%s\n%s\n", BLMSG, WLMSG);
    logMsg(LEVEL1, "%s\n", ignore1 ? IGNORE1 : NOIGNORE1);

    /* time of last ABW reads */
    abw_lastread  = time(0);
    checkABWtimer = time(0) + ABWCHECK;

    logMsg(LEVEL1,
      "End: Loaded alias, blacklist and whitelist files [%s]\n\n", strdate(ONLYTIME));
    if (verbose == LEVEL8) normalExit();

    if (stat(cidlog, &statbuf) == 0)
    {
      logMsg(LEVEL1, "CID logfile: %s\nCID logfile maximum size: %lu bytes\n",
        cidlog, cidlogmax);
    }
    else
    {
        /* Create the call log file if not present
           because writeLog won't create files */
        if ((fd = open(cidlog, O_WRONLY | O_APPEND | O_CREAT,
             S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0)
        {
            logMsg(LEVELE, "%s: %s\n", cidlog, strerror(errno));
        }
        else
        {
          close(fd);
          logMsg(LEVEL1,
            "Created CID logfile: %s\nCID logfile maximum size: %lu bytes\n",
            cidlog, cidlogmax);
        }
    }

    if (stat(datalog, &statbuf) == 0)
    {
        logMsg(LEVEL1, "Data logfile: %s\n", datalog);
    }
    else
    {
        logMsg(LEVEL1, "Data logfile not present: %s\n", datalog);
    }

    logMsg(LEVEL1, "Hostname & Port: %s:%d\n", hostname, port);

    if (cidinput == 0x3)
    {
        /* gateway without modem support */
        logMsg(LEVEL1, "Caller ID from gateways without modem support\n");
        logMsg(LEVEL1, "Maximum number of clients/gateways: %d\n", MAXCLIENTS + 1);
    }
    else
    {
        /*
         * cidinput == 0x0, 0x1, or 0x2
         * There is one or more serial/modem devices
         */
        initSerialDevices();
        checkHangupFiles();
    }

    logMsg(LEVEL1, "ncidd set for country %s\n", country);

    if (isNANPACountry(country))
    {
       logMsg(LEVEL1, "ncidd US phonenumber format %d:", nanp_format);
       if (nanp_format == 0) logMsg(LEVEL1, " not formatted\n");
       else if (nanp_format == 1) logMsg(LEVEL1, " (xxx) xxx-xxxx\n");
       else logMsg(LEVEL1, " xxx-xxx-xxxx\n");
    }

    if (debug || OSXlaunchd)
    {
        if (debug) logMsg(LEVEL1, "Debug Mode\n");
        else logMsg(LEVEL1, "OSX Launchd Mode\n");
    }
    else
    {
        /* fork and exit parent */
        if(fork() != 0) return 0;

        /* close stdin and make fd 0 unavailable */
        close(0);
        if (open("/dev/null",  O_WRONLY | O_SYNC) < 0)
        {
            errorExit(-1, "/dev/null", 0);
        }

        /* become session leader */
        setsid();
    }

    /*
     * Create a pid file
     */
    if (doPID(&pid, pidfile))
    {
        sprintf(msgbuf,"%s already exists", pidfile);
        errorExit(-110, "Fatal", msgbuf);
    }

    /* initialize the polld table */
    initPoll();
    /* initialize the safeSend client table */
    initQueuedMsg(MAXCONNECT);

    /* add the modem or serial device to the polld table, if configured */
   /* 0x3 says there are no serial devices to poll */
    if (cidinput != 0x3) {
        /* using a modem or serial device */
        pollSerialDevices();
    }

    /* initialize server socket */
    if ((mainsock = tcpOpen()) < 0) errorExit(-1, "socket", 0);

    /* add the server listen socket to the polld table, eg port 3333 */
    ret = addPoll(mainsock);
    logMsg(LEVEL3,"NCID connection socket is sd %d pos %d\n", mainsock, ret);

    logMsg(LEVEL1,"End of startup: %s\n",
            strdate(WITHSEP));

    /*
     * loop forever waiting for events from devices and sockets
     *
     * Note: poll() ignores negative fd entried in polld[]
     * and sets revents for every entry to zero or with event flags.
     */
    while (1)
    {
        switch (events = poll(polld, MAXCONNECT, POLL_TIMEOUT))
        {
            case -1:    /* error */
                if (errno != EINTR) /* No error for SIGHUP */
                    errorExit(-1, "poll", 0);
                /* fall through to treat interrupts the same as a time out */

            case 0:    /* time out, without an event */
                /* check to see if DIAL hangup timer is in use */
                doHangupTimers();
                /* check to see if ringing has stopped */
                doRingTimers();
                /* try modem init again */
                doModemLostTimers();
                /* see if ABW files have been edited */
                if (checkABWtimer) doCheckABWTimer();
                /* see if cidcall.log been edited */
                doCidCallTimer();


                if (TERM_flag)
                {
                   /* a TERM signal asked ncidd to exit */
                   TERMexit();
                }
                if (USR1_flag)
                {
                    /* got signal SIGUSR1, replace cidcall.log file with cidcall.log.new */
                    USR1_flag = 0;
                    sprintf (msgbuf, "%s.new", cidlog);
                    if (access (msgbuf, F_OK) == 0)
                    {
                        rename (msgbuf, cidlog);
                        logMsg (LEVEL1,
                        "Replaced %s with %s.new: %s\n", cidlog, cidlog, strdate(ONLYTIME));
                    }
                }
                if (HUP_flag)
                {
                   /* got signal SIGHUP, reload the alias, whitelist, blacklist */
                   HUP_flag = 0;
                   reload();
                }
                if (USR2_flag)
                {
                   /* got signal SIGUSR2, list the connected clients */
                   USR2_flag = 0;
                   showConnected();
                }

                /* if no modem, skip TTY lockfile code */
                if (!(cidinput & 0x1))
                {
                   /* 0x0 and 0x2 */
                   doModemLockTimers();
                }
                break;

            default:    /* 1 or more polld entries have events */
                if (TERM_flag)
                {
                   /* a TERM signal asked ncidd to exit */
                   TERMexit();
                }
                /* check to see if DIAL hangup timer is in use */
                doHangupTimers();
                /* check to see if ringing has stopped */
                doRingTimers();

                doPoll(events);
                break;
        }
    }
}


int getOptions(int argc, char *argv[])
{
    int c, num;
    int option_index = 0;
    static struct option long_options[] = {
        {"alias", 1, 0, 'A'},
        {"announce", 1, 0, 'a'},
        {"audiofmt", 1, 0, 'f'},
        {"blacklist", 1, 0, 'B'},
        {"cidinput", 1, 0, 'j'},
        {"cidlog", 1, 0, 'c'},
        {"cidlogmax", 1, 0, 'M'},
        {"config", 1, 0, 'C'},
        {"datalog", 1, 0, 'd'},
        {"debug", 0, 0, 'D'},
        {"gencid", 1, 0, 'g'},
        {"help", 0, 0, 'h'},
        {"hangup", 1, 0, 'H'},
        {"hupmode", 1, 0, 'm'},
        {"initcid", 1, 0, 'i'},
        {"initstr", 1, 0, 'I'},
        {"lineid", 1, 0, 'e'},
        {"lockfile", 1, 0, 'l'},
        {"logfile", 1, 0, 'L'},
        {"nomodem", 1, 0, 'n'},
        {"noserial", 1, 0, 'N'},
        {"pidfile", 1, 0, 'P'},
        {"port", 1, 0, 'p'},
        {"ifaddr", 1, 0, 'Q'},
        {"regex", 1, 0, 'r'},
        {"send", 1, 0, 's'},
        {"ttyspeed", 1, 0, 'S'},
        {"ttyclocal", 1, 0, 'T'},
        {"ttyport", 1, 0, 't'},
        {"verbose", 1, 0, 'v'},
        {"version", 0, 0, 'V'},
        {"whitelist", 1, 0, 'W'},
        {"osx-launchd", 0, 0, '0'},
        {0, 0, 0, 0}
    };

    while ((c = getopt_long (argc, argv, "a:c:d:e:f:g:hi:j:l:m:n:p:r:s:t:v:A:B:C:DH:I:L:M:N:P:Q:S:T:VW:",
        long_options, &option_index)) != -1)
    {
        switch (c)
        {
            case '0':
                ++OSXlaunchd;
                break;
            case 'A':
                if (!(cidalias = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("cidalias")) >= 0) setword[num].type = 0;
                break;
            case 'B':
                if (!(blacklist = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("blacklist")) >= 0) setword[num].type = 0;
                break;
            case 'C':
                if (!(config = strdup(optarg))) errorExit(-1, progname, 0);
                break;
            case 'D':
                ++debug;
                break;
            case 'H':
                hangup = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(hangup == 0 && *optarg == '0') && hangup > 3))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("hangup")) >= 0) setword[num].type = 0;
                break;
            case 'I':
                if (!(modem_list[0].initstr = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("initstr")) >= 0) setword[num].type = 0;
                break;
            case 'L':
                if (!(logfile = strdup(optarg))) errorExit(-1, progname, 0);
                break;
            case 'M':
                cidlogmax = atoi(optarg);
                if ((num = findWord("cidlogmax")) >= 0)
                {
                    if (cidlogmax < (unsigned) setword[num].min ||
                        cidlogmax > (unsigned) setword[num].max)
                        errorExit(-107, "Invalid number", optarg);
                    setword[num].type = 0;
                }
                break;
            case 'N':
                noserial = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(noserial == 0 && *optarg == '0') && noserial != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("noserial")) >= 0) setword[num].type = 0;
                break;
            case 'P':
                if (!(pidfile = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("pidfile")) >= 0) setword[num].type = 0;
                break;
            case 'S':
                if (!(modem_list[0].TTYspeed = strdup(optarg))) errorExit(-1, progname, 0);
                set_ttyspeed(&modem_list[0]);
                if ((num = findWord("ttyspeed")) >= 0) setword[num].type = 0;
                break;
            case 'T':
                modem_list[0].clocal = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(modem_list[0].clocal == 0 && *optarg == '0') && modem_list[0].clocal != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("ttyclocal")) >= 0) setword[num].type = 0;
                break;
            case 'V': /* version */
                fprintf(stderr, SHOWVER, progname, hostname, port, VERSION, API);
                exit(0);
            case 'W':
                if (!(whitelist = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("whitelist")) >= 0) setword[num].type = 0;
                break;
            case 'a':
                if (!(announce = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("announce")) >= 0) setword[num].type = 0;
                break;
            case 'c':
                if (!(cidlog = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("cidlog")) >= 0) setword[num].type = 0;
                break;
            case 'd':
                if (!(datalog = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("datalog")) >= 0) setword[num].type = 0;
                break;
            case 'e':
                if (!(modem_list[0].serial_lineid = strdup(optarg))) errorExit(-1, progname, 0);
                if (strlen(modem_list[0].serial_lineid) > CIDSIZE -1)
                    errorExit(-113, "string too long", optarg);
                if ((num = findWord("serial_lineid")) >= 0) setword[num].type = 0;
                break;
            case 'f':
                if (!(modem_list[0].audiofmt = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("audiofmt")) >= 0) setword[num].type = 0;
                break;
            case 'g':
                modem_list[0].gencid = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(modem_list[0].gencid == 0 && *optarg == '0') && modem_list[0].gencid != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("gencid")) >= 0) setword[num].type = 0;
                break;
            case 'h': /* help message */
                fprintf(stderr, DESC, progname);
                fprintf(stderr, USAGE, progname);
                exit(0);
            case 'i':
                if (!(modem_list[0].initcid = strdup(optarg))) errorExit(-1, progname, 0);
                ++modem_list[0].setcid;
                if ((num = findWord("initcid")) >= 0) setword[num].type = 0;
                break;
            case 'j':
                cidinput = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(cidinput == 0 && *optarg == '0') && cidinput > 3))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("cidinput")) >= 0) setword[num].type = 0;
                break;
            case 'l':
                if (!(modem_list[0].lockfile = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("lockfile")) >= 0) setword[num].type = 0;
                break;
            case 'm':
                hupmode = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(hupmode == 0 && *optarg == '0') && hupmode > 3))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("hupmode")) >= 0) setword[num].type = 0;
                break;
            case 'n':
                nomodem = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(nomodem == 0 && *optarg == '0') && nomodem != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("nomodem")) >= 0) setword[num].type = 0;
                break;
            case 'p':
                if((port = atoi(optarg)) == 0)
                    errorExit(-101, "Invalid port number", optarg);
                if ((num = findWord("port")) >= 0) setword[num].type = 0;
                break;
            case 'Q':
                if (!(interfaceaddr = strdup(optarg))) errorExit(-1, progname, 0);
                if (strlen(interfaceaddr) > CIDSIZE -1)
                    errorExit(-113, "string too long", optarg);
                if ((num = findWord("interfaceaddr")) >= 0) setword[num].type = 0;
                break;
            case 'r':
                regex = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(regex == 0 && *optarg == '0') && regex != 1 && regex != 2))
                    errorExit(-107, "Invalid regex number", optarg);
                if ((num = findWord("regex")) >= 0) setword[num].type = 0;
                break;
            case 's':
                if ((num = findSend(optarg)) < 0)
                    errorExit(-106, "Invalid send data type", optarg);
                ++(*sendclient[num].value);
                break;
            case 't':
                if (!(modem_list[0].ttyport = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("ttyport")) >= 0) setword[num].type = 0;
                break;
            case 'v':
                verbose = atoi(optarg);
                /* valid range: 1-9 */
                if (strlen(optarg) != 1 || (verbose == 0))
                {
                    verbose = 1;
                    errorExit(-107, "Invalid number", optarg);
                }
                if ((num = findWord("verbose")) >= 0) setword[num].type = 0;
                break;
            case '?': /* bad option */
                fprintf(stderr, USAGE, progname);
                errorExit(-100, 0, 0);
        }
    }
    return optind;
}

/*
 * checkHangupFiles
 * call checkVoiceFile() and checkHangupExtension()
 * to make sure they exist
 */
void checkHangupFiles(void)
{
    int ret;

    if (hangup)
    {
       if (hangup == 3)
       {
          /* check hangup announcement file "announce"
           * sets annpath, can change hangup to 1 */
          if (checkVoiceFile(annpath, announce))
          {
              hangup = 1;
              logMsg(LEVELE, "WARNING: Internal Hangup will use Normal Hangup (hangup=1), no recording file:\n         %s\n", annpath);
          }
          else
          {
              logMsg(LEVEL1, "Internal Hangup recording file: %s\n", annpath);
          }
       }
    }

    if (hupmode)
    {
       /* check to make sure the hangup extension is present and executable */
       checkHangupExtension();

       logMsg(LEVEL1, "Hangup Extension name: %s\nHangup Extension path: %s\n",
                hupname, huppath);

       if (hupmode == 3)
       {
          /* check hangup extensions default announcement file "huprmd"
           * sets rmdpath  */
          ret = 1;
          if (huprmd)
          {
             if ((ret = checkVoiceFile(rmdpath, huprmd)))
                 logMsg(LEVELE, "WARNING: Hangup Extension, recording file not found:\n         %s\n", rmdpath);
          }
          if (ret == 1)
          {
              /* no external hangup voice file, try again with the one for hangup */
             if (checkVoiceFile(rmdpath, announce))
             {
                  hupmode = 1;
                  logMsg(LEVELE, "WARNING: Hangup Extension will use Normal Hangup (hupmode=1), no recording file:\n         %s\n", rmdpath);
             }
          }
          if (hupmode == 3) logMsg(LEVEL1, "Default Hangup Extension Recording: %s\n", rmdpath);
       }
    }
}


/*
 * checkVoiceFile
 * mode is reset to 1 if recording file is not found
 * path is filled in with the absolute path to the file
 * return = 0 for a recording file
 * return = 1 for no recording file
 */
int checkVoiceFile(char * path, const char * recording)
{
    struct stat statbuf;

    if (!recording || *recording == '\0')
    {
        recording = RECORDING;
    }

    if (*recording == '/')
    {
        sprintf(path, "%s", recording);
    }
    else
    {
        sprintf(path, "%s/%s", ANNDIR, recording);
    }

    if (stat(path, &statbuf))
    {
        /* bad */
        return 1;
    }
    else
    {
        return 0;
    }
}


/*
 * return = 1 for a valid hangup extension
 * errrorExit for a problem
 */
void checkHangupExtension(void)
{
    struct stat statbuf;

    if (*hupname == '/')
    {
      /* separate the name and path */
      char * ptr;
      sprintf(huppath, "%s", hupname);
      ptr = strrchr(huppath, '/');
      hupname = strdup(ptr + 1);
    }
    else sprintf(huppath, "%s/%s", EXTDIR, hupname);

    if (strncmp(hupname, HUPNAME "-", strlen(HUPNAME "-")))
    {
      logMsg(LEVELE, "Hangup Extension name must start with \"%s-\": %s\n",
              HUPNAME, hupname);
      errorExit(-114, 0, 0);
    }
    else if (stat(huppath, &statbuf) != 0)
    {
      logMsg(LEVELE, "Hangup Extension not found: %s\n", huppath);
      errorExit(-114, 0, 0);
    }
    else if (!((statbuf.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)) ==
             (S_IXUSR | S_IXGRP | S_IXOTH)))
    {
      logMsg(LEVELE, "Hangup Extension must be executable (try sudo chmod 755 %s)\n", huppath);
      errorExit(-114, 0, 0);
    }
}


/*
 * open a TCP listen socket for clients
 */
int tcpOpen(void)
{
    int     sd, ret, optval;
    static struct  sockaddr_in bind_addr;
    int socksize = sizeof(bind_addr);

    optval = 1;
    if ((sd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        return sd;
    /* allow immediate reuse of the port number on ncidd restarts */
    if((ret = setsockopt(sd, SOL_SOCKET, SO_REUSEADDR,
        &optval, sizeof(optval))) < 0)
        return ret;
    /* send keep-alive probes when connection is idle for 7200s */
    if((ret = setsockopt(sd, SOL_SOCKET, SO_KEEPALIVE,
        &optval, sizeof(optval))) < 0)
        return ret;


    /* bind to a particular interface or INADDR_ANY */
     bind_addr.sin_family = AF_INET;
    if (interfaceaddr[0] == '\0') {
       /*  0.0.0.0, listen to all interfaces on this host */
       bind_addr.sin_addr.s_addr = INADDR_ANY;
    }
    else {
       /* an interface address was specified, so restrict it to one interface */
       bind_addr.sin_addr.s_addr = inet_addr(interfaceaddr);
       if (bind_addr.sin_addr.s_addr == INADDR_NONE) {
          logMsg(LEVELE, "Unable to parse ifaddr: %s, allowing connections from localhost.\n", interfaceaddr);
           bind_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
       }
       else {
          logMsg(LEVEL2, "Allowing connections from ifaddr: %s\n", interfaceaddr);
       }
    }
    memset(bind_addr.sin_zero, 0, 8);
    bind_addr.sin_port = htons(port);   /* port in network byte order */
    if ((ret = bind(sd, (struct sockaddr *)&bind_addr, socksize)) < 0)
    {
        logMsg(LEVEL2, "unable to bind listen socket: %s.\n", strerror(errno));
        close(sd);
        return ret;
    }

    /* listen for new tcp connections from clients */
    if ((ret = listen(sd, 5)) < 0)
    {
        logMsg(LEVEL2, "unable to create listen socket: %s.\n", strerror(errno));
        close(sd);
        return ret;
    }
    return sd;
}


/*
 *  accept a new TCP connection from a client
 */

int  tcpAccept(void)
{
    int sd;
    int ret;
    char tmpbuf[MAXIPBUF-5];    /* -5 to quiet -Wformat-overflow */

    struct  sockaddr_in sa;
    socklen_t sa_len = sizeof(sa);

    if ((sd = accept(mainsock, (struct sockaddr *) &sa, &sa_len)) != -1)
    {
        strcpy(tcpAccept_ret.IPaddr, inet_ntoa(sa.sin_addr));

        /* get the client host name (max domain name is 253) */
        ret = getnameinfo((struct sockaddr *) &sa, sa_len, tmpbuf, sizeof(tmpbuf), NULL, 0, 0);
        if (ret == 0)
        {
           sprintf(tcpAccept_ret.HostName, " [%s]", tmpbuf);
        }
        else if (ret == EAI_AGAIN)
        {
           /* being internet offline gives:  -3, Temporary failure in name resolution */
           strcpy(tcpAccept_ret.HostName, " [Not Available]");
        }
        else
        {
           snprintf(tcpAccept_ret.HostName, sizeof(tcpAccept_ret.HostName)-1, " [hostname lookup error %d, %s]", ret, gai_strerror(ret));
        }
    }

    return sd;
}


/*
 * initialize the poll table with -1 to mark the available slots
 */
void initPoll(void)
{
    int pos;

    for (pos = 0; pos < MAXCONNECT; ++pos)
    {
        polld[pos].fd = -1;    /* tells poll to ignore this slot in the table */
        polld[pos].revents = 0;
        polld[pos].events = 0;
        IPinfo[pos].wants_ack = 0;
    }
}


/*
 * Add client to the poll table
 * slots marked with fd -1 are available ( and are ignored by poll() )
 */
int addPoll(int addfd)
{
    int added = 0, pos;

    for (added = pos = 0; pos < MAXCONNECT; ++pos)
    {
        if ( polld[pos].fd >= 0 ) continue;    /* occupied */
        if ( polld[pos].fd < -2 ) continue;    /* reserved */
        IPinfo[pos].wants_ack = 0;
        polld[pos].revents = 0;
        polld[pos].fd = addfd;
        polld[pos].events = (POLLIN | POLLPRI);            /*  | POLLOUT, will be set later when needed */
        ++added;
        break;
    }
    return added ? pos : -1;
}


/*
 * handle the events from devices and sockets after poll() wakes up
 * the events arg is the number of polld entries with nonzero revents
 */
void doPoll(int events)
{
  int pos;

  /*
   * Poll is configured for POLLIN and POLLPRI events
   * POLLERR, POLLHUP, POLLNVAL events can also happen
   * Poll will also be configured for the POLLOUT event
   * when the outgoing client socket would have blocked.
   * Note: poll() ignores negative fd entried in polld[]
   * and sets revents for every entry to zero or with event flags.
   */

  for (pos = 0; events && pos < MAXCONNECT; ++pos)
  {
    if (!polld[pos].revents) continue; /* no events on this one */
    --events;

    /* log event flags */
    logMsg(LEVEL9, "polld[%d].revents: 0x%X, fd: %d\n",
            pos, polld[pos].revents, polld[pos].fd);

    if (polld[pos].revents & POLLHUP) /* Other end hung up, stream may still have data */
    {
      /* was it a modem or serial device? */
      if (isSerialDevice(pos))
      {
        logMsg(LEVEL2, "tty device fd %d pos %d Disconnected %s\n",
                polld[pos].fd, pos, strdate(WITHSEP));
        startModemLostTimer("tty device disconnected", findModemByPos(pos));
      }
      else
      {
        /* or a client */
        logMsg(LEVEL2, "Client %d pos %d Disconnected %s\n",
                polld[pos].fd, pos, strdate(WITHSEP));
        disconnectClient(pos);
      }
      continue;
    }

    if (polld[pos].revents & POLLERR) /* Poll Error */
    {
      /* was it a modem or serial device? */
      if (isSerialDevice(pos))
      {
        logMsg(LEVEL2, "tty device fd %d pos %d Poll Error %s\n",
                polld[pos].fd, pos, strdate(WITHSEP));
        startModemLostTimer("tty device error", findModemByPos(pos));
      }
      else
      {
        /* or a client */
        logMsg(LEVEL1, "Client %d pos %d Disconnected - Poll Error %s\n",
                polld[pos].fd, pos, strdate(WITHSEP));
        disconnectClient(pos);
      }
      continue;
    }

    if (polld[pos].revents & POLLNVAL) /* Invalid Request, output fd not open */
    {
      /* was it a modem or serial device? */
      if (isSerialDevice(pos))
      {
        logMsg(LEVEL2, "tty device fd %d pos %d Invalid Request %s\n",
                polld[pos].fd, pos, strdate(WITHSEP));
        startModemLostTimer("Invalid Request from Serial device", findModemByPos(pos));
      }
      else
      {
        /* or a client */
        logMsg(LEVEL2, "Client %d pos %d Disconnected - Invalid Request %s\n",
                polld[pos].fd, pos, strdate(WITHSEP));
        disconnectClient(pos);
      }
      continue;
    }

    if (polld[pos].revents & POLLOUT) /* Write Event */
    {
      /* a previously blocked socket has become writable */
      if (isSerialDevice(pos))
         continueModem(pos);
      else
         send_blocked(pos);
    }

    if (polld[pos].revents & (POLLIN | POLLPRI)) /* there is data that can be read */
    {
      /* was it a modem or serial device? */
      if (isSerialDevice(pos))
      {
         /* Modem or device has data to read */
         readFromModem(pos);
      }
      /* or the listen socket eg 3333 */
      else if (polld[pos].fd == mainsock)
      {
         /* ping the existing clients to expose dead ones */
         writeClients("");
         /* accept a new TCP/IP Client Connection */
         acceptNewClient();
      }
      /* or data has arrived from a client */
      else
      {
         /* Client or gateway has data to read */
         readFromClient(pos);
      }
    }
  }
}


/*
 * read from a client or gateway
 */
void readFromClient(int pos)
{
   char buf[BUFSIZ];
   int num;

   if ((num = read(polld[pos].fd, buf, BUFSIZ-1)) < 0)
   {
      logMsg(LEVEL2, "Client %d pos %d read error %d: %s %s\n",
              polld[pos].fd, pos, errno, strerror(errno), strdate(WITHSEP));
      /* drop the client for any read error */
      logMsg(LEVEL2, "Client %d pos %d from %s%s %s {%s} removed %s\n",
              polld[pos].fd, pos, IPinfo[pos].addr, IPinfo[pos].name,
              IPinfo[pos].dialout ? "dialout" : "nodialout",
              IPinfo[pos].ident, strdate(WITHSEP));
      disconnectClient(pos);
   }
   /* read will return 0 for a disconnect */
   else if (num == 0)
   {
      /* TCP/IP Client End Connection */
      logMsg(LEVEL2, "Client %d pos %d from %s%s %s {%s} disconnected %s\n",
              polld[pos].fd, pos, IPinfo[pos].addr, IPinfo[pos].name,
              IPinfo[pos].dialout  ? "dialout" : "nodialout",
              IPinfo[pos].ident, strdate(WITHSEP));
      disconnectClient(pos);
   }
   else
   {
      /*
       * Client sent message to server
       */

      /* Terminate String */
      buf[num] = '\0';

      /*
       * there is no guarantee that the message is complete, it's
       * a stream, so call handle_incoming_message() to
       * accumulate data until a newline character is found
       * which will then call handle_accumulated_message()
       */
      handle_incoming_message(pos, buf, num);
   }
}


/*
 * Process a line from a client.
 * safesend limits the length of the incoming buf to acc_size=1024
 */
void handle_accumulated_message(int pos, const char * buf)
{
   char tmpbuf[BUFSIZ];
   char **svrtag;
   /* additional info obtained from unformatted phone number in MSG and NOT */
   char * infoBuffer;

   /* Make sure there is data in the message line */
   if ((strlen(buf)) != 0)
   {
     /* Look for CALL, CALLINFO, or MSG lines */
     if (strncmp(buf, CALL, strlen(CALL)) == 0)
     {
       /*
        * Found CALL Line
        * See comments for gatewayCID for line format
        */

       logMsg(LEVEL3, "Gateway %d pos %d sent CALL data %s\n",
               polld[pos].fd, pos, strdate(WITHSEP));

       writeLog(datalog, buf);
       if (IPinfo[pos].wants_ack) sendACK(pos, buf);
       gatewayCID(buf + strlen(CALL), pos);
       finishCID(&gwcid);
     }
     else if (strncmp(buf, CALLINFO, strlen(CALLINFO)) == 0)
     {
       /* Found a CALLINFO Line */

       logMsg(LEVEL3, "Gateway %d pos %d sent CALLINFO %s\n",
               polld[pos].fd, pos, strdate(WITHSEP));

        handleGWCallinfo(pos, buf);

     }
     else if (buf[0] == '+' && (buf[4] == ':' || buf[8] == ':'))
     {
       /*
        * Found line from a server gateway
        * for example: "+CID: *", "+HUP: *", "+CIDINFO: *
        * messages require two checks:
        *      "MSG: " and "***",  "NOT: " and "***"
        */

       for (svrtag = serverTags; *svrtag; svrtag++)
       {
           if (!strncmp(buf, *svrtag, strlen(*svrtag)))
           {
               logMsg(LEVEL3, "Server Gateway (sd %d) sent %s %s\n",
                   polld[pos].fd, buf, strdate(WITHSEP));
               if (strncmp(buf, "+CIDINFO", strlen("+CIDINFO")))
               {
                   writeLog(cidlog, buf + 1);
                   /* remember the cidcall.log file date */
                   cidcall_lastchange = time(0);
               }
               writeClients(buf + 1);
               break;
           }
       }
       if (*svrtag == NULL)
       {
           logMsg(LEVEL3, "Unknown Server Gateway (sd %d) line %s %s\n",
                   polld[pos].fd, buf, strdate(WITHSEP));
       }
     }
     else if (!strncmp(buf, MSGLINE, strlen(MSGLINE)))
     {
       /*
        * Found MSG: line
        * MSG: <message>
        * or
        * MSG: <message> ###DATE*mmddyyyy*TIME*hhmm*NAME*<name>*NMBR*<number>*LINE*<id>*MTYPE*IN|OUT|SYS|USER*
        *
        * Format of a MSG: line passed to TCP/IP clients by ncidd:
        *
        * MSG: <message> ***DATE*<mmddyyyy>*TIME*<hhmm>*LINE*<id>*NMBR*<nmbr>*MTYPE*IN|OUT|SYS|USER*FNMBR*<formatted number*NTYPE*FIX/CELL*CTRY*<cc>*LOCA*<location>*CARI*<carrier>*NAME*<name>*
        *
        * Write message to cidlog and all clients
        */

       logMsg(LEVEL3, "Client %d sent text message %s\n", polld[pos].fd, strdate(WITHSEP));
       writeLog(datalog, buf);
       getINFO(buf);      /* fills the global struct getINFO_ret */

       infoBuffer = GetCompleteInfoFromNumber(getINFO_ret.nmbr);

       sprintf(tmpbuf, MESSAGE, buf, getINFO_ret.date, getINFO_ret.time, getINFO_ret.line, getINFO_ret.nmbr, getINFO_ret.type, infoBuffer, getINFO_ret.name);

       writeLog(cidlog, tmpbuf);
       /* remember the cidcall.log file date */
       cidcall_lastchange = time(0);
       writeClients(tmpbuf);
     }
     else if (!strncmp(buf, NOTLINE, strlen(NOTLINE)))
     {
       /*
        * Found NOT: (remote notification) line from a gateway
        * NOT: <message> ###DATE*mmddyyyy*TIME*hhmm*NAME*<name>*NMBR*<number>*LINE*<id>*MTYPE*<IN|OUT|USER>*
        *
        * Format of a NOT: line passed to TCP/IP clients by ncidd:
        *
        * NOT: <message> ***DATE*<mmddyyyy>*TIME*<hhmm>*LINE*<id>*NMBR*<nmbr>*MTYPE*IN|OUT|SYS|USER*FNMBR*<formatted number*NTYPE*FIX/CELL*CTRY*<cc>*LOCA*<location>*CARI*<carrier>*NAME*<name>*
        *
        * Write notice to cidlog and all clients
        *
        * The longest SMS text message is 918 characters.
        * When it is longer than 160 characters then it will be broken
        * into chunks of 153 characters before being sent to the
        * recipient's handset.
        */

       logMsg(LEVEL3, "Gateway %d pos %d sent a notice %s\n",
               polld[pos].fd, pos, strdate(WITHSEP));
       writeLog(datalog, buf);
       if (IPinfo[pos].wants_ack) sendACK(pos, buf);
       getINFO(buf);      /* fills the global struct getINFO_ret */

       /*
        * numbers coming from a mobile phone/gateway are often not dialable
        * (have country code but no international prefix)
        * this will take care of it, before we write them in the cidlog
        */
       normalizeToDialableNumberFromFixedLine(getINFO_ret.nmbr, CIDSIZE);
       infoBuffer = GetCompleteInfoFromNumber(getINFO_ret.nmbr);

       sprintf(tmpbuf, MESSAGE, buf, getINFO_ret.date, getINFO_ret.time, getINFO_ret.line, getINFO_ret.nmbr, getINFO_ret.type, infoBuffer, getINFO_ret.name);

       writeLog(cidlog, tmpbuf);
       /* remember the cidcall.log file date */
       cidcall_lastchange = time(0);
       writeClients(tmpbuf);
     }
     else if (!strncmp(buf, RLYLINE, strlen(RLYLINE)) && strstr(buf, "###"))
     {
       /*
        * Found RLY: (remote relay) line from a gatewaay
        * RLY: <message> ###<info fields>
        * Change ### to *** and write to cidlog and all clients
        */

       logMsg(LEVEL3, "Gateway %d pos %d sent a relay %s\n",
               polld[pos].fd, pos, strdate(WITHSEP));
       writeLog(datalog, buf);
       if (IPinfo[pos].wants_ack) sendACK(pos, buf);
       strcpy(tmpbuf,buf);
       strncpy(strstr(tmpbuf,"###"),"***",3);
       writeLog(cidlog, tmpbuf);
       /* remember the cidcall.log file date */
       cidcall_lastchange = time(0);
       writeClients(tmpbuf);
     }
     else if (strncmp (buf, REQLINE, strlen(REQLINE)) == 0)
     {
       /* Found REQ: line */
        handleClientRequest(pos, buf);
     }
     else if (strncmp (buf, WRKLINE, strlen(WRKLINE)) == 0)
     {
       /* Found WRK: line */
        handleClientWork(buf);
     }
     else if (strncmp (buf, GOODBYE, strlen(GOODBYE)) == 0)
     {
       /*
        * Found GOODBYE line
        * Disconnect the client.
        */
       logMsg(LEVEL2, CLIENT_GOODBYE, polld[pos].fd, pos, strdate(WITHSEP));
       disconnectClient(pos);
     }
     else if (strncmp (buf, IDENT, strlen(IDENT)) == 0)
     {
       /*
        * Found HELLO: IDENT: line, which has arrived too late for
        * the usual HELLO processing
        * get the client IDENT: info from it anyway
        */
       doIDENT(pos, buf);
     }
     else if (strncmp (buf, COMMAND, strlen(COMMAND)) == 0)
     {
       /*
        * Found HELLO: CMD: line, which has arrived too late for
        * the usual HELLO processing
        * quietly discard it
        */
       logMsg(LEVEL3, "Client %d sent late %s\n", polld[pos].fd, buf);
     }
     else
     {
       /*
        * Found unknown data
        */

       logMsg(LEVEL3, "Client %d sent unknown data %s\n",
               polld[pos].fd, strdate(WITHSEP));
       writeLog(datalog, buf);
     }
   }
   else
   {
     /*
      * Found empty line
      * the android client does this on a keepalive timer
      */

       logMsg(LEVEL6, "Client %d sent empty line %s\n",
               polld[pos].fd, strdate(WITHSEP));
   }
}


/*
 * WRK: line
 * called from handle_accumulated_message()
 * Perform the requested work on behalf of the client
 */
void handleClientWork(const char * buf)
{
   char msgbuf[BUFSIZ];
   int ret;

   logMsg(LEVEL2, "%s\n", buf);
   if (strncmp (buf + strlen(WRKLINE), ACPT_LOG, strlen(ACPT_LOG)) == 0)
   {
      /* WRK: ACCEPT LOG or WRK: ACCEPT LOGS
         accept the changes made by REQ: UPDATE or UPDATES */
      if (strstr (buf + strlen(WRKLINE), ACPT_LOGS)) {
          sprintf(msgbuf, "for f in %s.*[0-9]; do mv $f.new $f; done", cidlog);
          ret = system (msgbuf);
          logMsg(LEVEL2, "rc=%d %s [%s]\n", ret, msgbuf, strdate(ONLYTIME));
      }
      sprintf(msgbuf, "%s.new", cidlog);
      if (rename(msgbuf, cidlog)<0)
         logMsg(LEVEL2, "rename failed %s.new %s - %s\n", cidlog, cidlog, strerror(errno));
      else
         logMsg(LEVEL2, "rename %s.new %s [%s]\n", cidlog, cidlog, strdate(ONLYTIME));
   }
   else if (strncmp (buf + strlen(WRKLINE), RJCT_LOG, strlen(RJCT_LOG)) == 0)
   {
      /* WRK: REJECT LOG or WRK: REJECT LOGS
         reject the changes made by REQ: UPDATE or UPDATES */
      if (strstr (buf + strlen(WRKLINE), RJCT_LOGS)) {
        sprintf(msgbuf, "rm %s.*.new",cidlog);
        ret = system (msgbuf);
        logMsg(LEVEL2, "rc=%d %s [%s]\n", ret, msgbuf, strdate(ONLYTIME));
      }
      sprintf(msgbuf, "%s.new", cidlog);
      if(unlink(msgbuf)<0)
         logMsg(LEVEL2, "erase failed %s.new - %s\n", cidlog, strerror(errno));
      else
         logMsg(LEVEL2, "erase %s.new [%s]\n", cidlog, strdate(ONLYTIME));
   }
}


/*
 * REQ: line
 * called from handle_accumulated_message()
 * Perform the requested action and send a response back to the client
 */
void handleClientRequest(int pos, const char * buf)
{
   char msgbuf[BUFSIZ];

   logMsg(LEVEL2, "Received: %s    [%s]%s", buf,
            strdate(WITHSEP), NL);
   if (!strncmp(buf, RELOAD, strlen(RELOAD)))
   {
     /*
      * Found REQ: RELOAD line
      * reload alias, blacklist and whitelist files
      */
      handleReqReload(pos, buf);

   }
   else if (!strncmp(buf, UPDATE, strlen(UPDATE)))
   {
     /*
      * Found REQ: UPDATE or UPDATES
      * invoke cidupdate to update all of the CID call log files
      */
      handleReqUpdate(pos, buf);
   }
   else if (!strncmp(buf, REREAD, strlen(REREAD)))
   {
     /*
      * Found REQ: REREAD
      */
     sendLog(pos);
   }
   else if (!strncmp(buf, DIAL, strlen(DIAL)))
   {
     /*
      * Found REQ: DIAL <number>&&<name>&&<lineid>
      * or    REQ: DIAL_ABORT <number>&&<name>&&<lineid>
      */
     handleReqDial(pos, buf);
   }
   else if (!strcmp(buf, REQ_ACK) || !strcmp(buf, REQ_YO))
   {
     /*
      * Found REQ: ACK or REQ: YO
      * respond with ACK: and echo the request
      */
      if (strstr(buf, ACK)) IPinfo[pos].wants_ack = 1;

      logMsg(LEVEL3, "Sent Client %d pos %d: %s%s %s%s",
              polld[pos].fd, pos, ACKLINE, buf, strdate(WITHSEP), NL);
      sprintf(msgbuf, "%s%s%s", ACKLINE, buf, CRLF);
      safeSend(pos, msgbuf);
   }
   else if (!strncmp(buf, REQ_INFO, strlen(REQ_INFO)))
   {
      /* found a REQ: INFO <nmbr>&&<name>&&<line> line */
      handleReqInfo(pos, buf);
   }
   else if (!strncmp(buf, REQ_PAUSE, strlen(REQ_PAUSE)))
   {
      /* found a REQ: PAUSE <minute(s)> line to set a time to resume hangups */
      handleReqPause(pos, buf);
   }
   else
   {
     /*
      * Found REQ: black/white/alias add/remove
      * which will be passed to ncidutil
      */
      handleReqABW(pos, buf);
   }
}


/*
 * REQ: RELOAD
 * called from handleClientRequest()
 * to reload alias, blacklist and whitelist files
 * Perform the requested action and send a response back to the client
 */
void handleReqReload(int pos, const char * buf)
{
   char tmpbuf[BUFSIZ], msgbuf[BUFSIZ];

  /*
   * Found REQ: RELOAD line
   */
   long position = 0;
   msgbuf[0] = '\0';

   /* remember the current position in the ncidd.log file */
   if (logptr)
   {
       position = ftell (logptr);
   }
   /* reload alias, blacklist and whitelist files */
   reload();

   if (logptr)
   {
      size_t cnt = 0;
      /* read the ncidd.log file from "position" to send to the client */
      fseek (logptr, position, SEEK_SET);
      while (fgets (tmpbuf, sizeof (tmpbuf), logptr) != 0)
      {
          cnt += sizeof (INFOLINE) + strlen(tmpbuf);
          /* stop when msgbuf is full
           * it's ok to break out of here, because the O_APPEND file status flag
           * will make the next log file write jump the the end */
          if (cnt >= sizeof (msgbuf) - 2) break;
          strcat (msgbuf, INFOLINE);
          strcat (msgbuf, tmpbuf);
      }
   }
   else
   {
      strcpy(msgbuf, INFOLINE RELOADED NL);
   }
   logMsg(LEVEL2, BEGIN_DATA NL);
   safeSend(pos, BEGIN_DATA CRLF);

   logMsg(LEVEL2, msgbuf);
   safeSend(pos, msgbuf);

   logMsg(LEVEL2, END_DATA NL);
   safeSend(pos, END_DATA CRLF);
}


/*
 * REQ: UPDATE or UPDATES
 * called from handleClientRequest()
 * to invoke cidupdate and apply the current aliases to cidcall.log(s)
 * Perform the requested action and send a response back to the client
 */
void handleReqUpdate(int pos, const char * buf)
{
   int cnt;
   char msgbuf[BUFSIZ], opt[BUFSIZ];
   char *ptr;

  /*
   * Found REQ: UPDATE or UPDATES
   * invoke cidupdate to update all of the CID call log files
   */
   FILE *respHandle;
   char * updatecmd;
   int status;

   updatecmd = (char *) malloc(strlen(DOUPDATE)+strlen(config)+strlen(cidalias)+strlen(cidlog)+strlen(whitelist)+strlen(blacklist)+sizeof(opt)+20);

   opt[0] = '\0';
   /* explicitly add all the path options for ABW */
   sprintf (updatecmd, DOUPDATE, config, cidalias, cidlog, whitelist, blacklist);
   if (!strncmp(buf, UPDATES, strlen(UPDATES))) strcat(updatecmd, " --multi");
   sprintf(opt, " --regex %d", regex);
   if (ignore1) strcat(opt, " --ignore1");
   if (hangup || hupmode) strcat(opt, " -H 1");
                     else strcat(opt, " -H 0");

   strcat(updatecmd, opt);
   strcat(updatecmd, " < /dev/null 2>&1");

   logMsg(LEVEL4, "Begin: Executing %s [%s]\n", NCIDUPDATE, strdate(ONLYTIME));

   /* invoke /bin/sh -c updatecmd */
   respHandle = popen (updatecmd, "r");

   if (!respHandle) {
      /* when sh -c fails to start */
      logMsg(LEVELE, "popen %s failed - %s\n", NCIDUPDATE, strerror(errno));
      safeSend(pos, "cidupdate failed to run" CRLF);
   }
   else
   {
      strcat(updatecmd, "\n");
      logMsg(LEVEL2, updatecmd);
      strcpy (msgbuf, INFOLINE);
      /* prepend INFO: so all the lines read will start this way */
      ptr = msgbuf + sizeof (INFOLINE) - 1;
      cnt = sizeof (msgbuf) - sizeof (INFOLINE);
      if (fgets (ptr, cnt, respHandle) != NULL)
      {
         /* "denied" will come from the system errno, e.g. Unable to open new call log: Permission denied
          * "no changes" comes from cidupdate's call to nciddalias.c
          */
         if (strstr(msgbuf, NOCHANGES) || strstr(msgbuf, DENIED))
         {
           /* There were no changes to the call log */
           logMsg(LEVEL2, BEGIN_DATA NL);
           safeSend(pos, BEGIN_DATA CRLF);

           logMsg(LEVEL2, msgbuf);
           safeSend(pos, msgbuf);
         }
         else
         {
           /* There were changes to the call log */
           logMsg(LEVEL2, BEGIN_DATA1 NL);
           safeSend(pos, BEGIN_DATA1 CRLF);

           logMsg(LEVEL2, msgbuf);
           safeSend(pos, msgbuf);

           /* return the output from cidupdate */
           while (fgets(ptr, cnt, respHandle))
           {
              logMsg(LEVEL2, msgbuf);
              safeSend(pos, msgbuf);
           }
         }

         status = pclose(respHandle);

         if (status < 0) {
            logMsg(LEVELE, "%s pclose %s\n", NCIDUPDATE, strerror(errno));
         }
         else
         {
            if ( WIFEXITED(status) ) {
               logMsg(LEVEL4, "%s exited with status %d\n", NCIDUPDATE, WEXITSTATUS(status) );
            }
            else {
               logMsg(LEVEL4, "%s abnormal exit\n", NCIDUPDATE);
            }
         }
         logMsg(LEVEL4, "End: Executing %s [%s]\n", NCIDUPDATE, strdate(ONLYTIME));

         logMsg(LEVEL2, END_DATA NL);
         safeSend(pos, END_DATA CRLF);
         free(updatecmd);
      }
   }
}


/*
 * REQ: DIAL <number>&&<name>&&<lineid>
 * REQ: DIAL_ABORT <number>&&<name>&&<lineid>
 * called from handleClientRequest()
 * Perform the requested action and send a response back to the client
 */
void handleReqDial(int pos, const char * buf)
{
   char msgbuf[BUFSIZ];
   char *sptr;

   /*
    * Found REQ: DIAL <number>&&<name>&&<lineid>
    * or    REQ: DIAL_ABORT <number>&&<name>&&<lineid>
    */
   int abort = 0;
   char *eptr1, *eptr2, failreason[100];
   char cidnmbr_req[CIDSIZE];
   char cidname_req[CIDSIZE];
   char cidline_req[CIDSIZE];
   char cidnmbr_log[CIDSIZE];

   logMsg(LEVEL2, BEGIN_DATA2 NL);
   safeSend(pos, BEGIN_DATA2 CRLF);
   sprintf(msgbuf, DIAL_WORKING NL);
   safeSend(pos, msgbuf);

   /* make sure the REQ format is OK */
   eptr1 = eptr2 = 0;
   eptr1 = strstr(buf, "&&");
   if (eptr1) eptr2 = strstr(eptr1 + 2, "&&");
   if (!(eptr1 && eptr2))
   {
      logMsg(LEVEL2, END_RESP NL);
      safeSend(pos, END_RESP CRLF);
      /* dial format error */
      logMsg(LEVEL2, DIAL_LINE_ERR NL, buf);
      safeSend(pos, msgbuf);
      /* give up */
      return;
   }

   /* DIAL or DIAL_ABORT */
   sptr = strchr(buf, 'L'), sptr += 2;
   if (*sptr == 'A') abort = 1, sptr += 6;

   /* get number */
   *eptr1 = '\0';
   strncpy(cidnmbr_req, sptr, CIDSIZE - 1);
   cidnmbr_req[CIDSIZE-1] = '\0';
   *eptr1 = '&';
   sptr = eptr1 + 2;

   /* get name */
   *eptr2 = '\0';
   strncpy(cidname_req, sptr, CIDSIZE - 1);
   cidname_req[CIDSIZE-1] = '\0';
   *eptr2 = '&';
   sptr = eptr2 + 2;

   /* get dial lineid */
   strncpy(cidline_req, sptr, CIDSIZE - 1);
   cidline_req[CIDSIZE-1] = '\0';

   if (abort)
   {
      /* user aborted dial request */
      dialAbort(pos);
      logMsg(LEVEL2, END_RESP NL);
      safeSend(pos, END_RESP CRLF);
   }
   else
   {
     /* remove a dial prefix for the call log */
     get_dial_log_nmbr(cidnmbr_log, cidnmbr_req);

     /* find the corresponding modem and dial the number */
     struct modems_t * a_modem = dialNumber(pos, cidnmbr_req, cidline_req, failreason);
     if (a_modem)
     {
        struct cid_t * modem_cid = a_modem->modem_cid;

        logMsg(LEVEL2, "RESP: Dialed %s on line \"%s\" %s\n", cidnmbr_req, cidline_req, strdate(WITHSEP));
        sprintf(msgbuf, DIALED_NMBR NL, cidnmbr_req, cidline_req);
        safeSend(pos, msgbuf);

         /* this is also used for a log and client RPLY in doHangupTimer() */
         strcpy(modem_cid->cidnmbr, cidnmbr_log);
         strcpy(modem_cid->cidname, cidname_req);
         strcpy(modem_cid->cidline, a_modem->serial_lineid);

         /* server creates OUT line for dialout, otherwise a gateway does it */
         modem_cid->calltype = OUT;
         /* dialOUT means a gateway will log the OUT: line */
         if (!dialOUT)
         {
            /* server creates OUT line for dialout, not a gateway */
            userAlias(modem_cid->cidnmbr, modem_cid->cidname, modem_cid->cidline);

            add_timestamp(modem_cid->ciddate, modem_cid->cidtime);
           /* get additional info from phone number */
            if (modem_cid->infoBuffer != NULL) free(modem_cid->infoBuffer);
            modem_cid->infoBuffer = GetCompleteInfoFromNumber(modem_cid->cidnmbr);
         }

         if (!dialOUT) sendCall(OUTLINE, a_modem->modem_cid, NULL);

        logMsg(LEVEL2, DIALOK NL, dialdelay);
        sprintf(msgbuf, DIALOK CRLF, dialdelay);
        safeSend(pos, msgbuf);
        logMsg(LEVEL2, END_RESP NL);
        safeSend(pos, END_RESP CRLF);
     }
     else /* Dial Number Failed */
     {
        logMsg(LEVEL2, END_RESP NL);
        safeSend(pos, END_RESP CRLF);

        logMsg(LEVEL2, DIAL_FAIL NL, failreason);
        sprintf(msgbuf, DIAL_FAIL CRLF, failreason);
        safeSend(pos, msgbuf);
     }
   }
}

/*
 * the dial request might have a prefix and an added extension
 * e.g. 9-9811234567,271
 * first non-numeric is just one character
 *
 */
void get_dial_log_nmbr(char * cidnmbr_log, char * cidnmbr_req)
{
   size_t prefix, digits;

   /* look for possible prefix */
   prefix = strspn(cidnmbr_req, "0123456789");
   if (prefix == 0) {
      /* not numeric */
      strcpy(cidnmbr_log, cidnmbr_req);
      return;
   }

   if (prefix < 6) {
      prefix++;
      /* count the numeric digits after the prefix */
      digits = strspn(cidnmbr_req+prefix, "0123456789");
      if (digits == 0) {
         digits = prefix-1;
         prefix = 0;
      }
   }
   else {
      /* anything over 6 chars is not a prefix to get an outside line */
      digits = prefix;
      prefix = 0;
   }
   /* copy digits to cidnmbr_log */
   strncpy(cidnmbr_log, cidnmbr_req+prefix, digits);
   cidnmbr_log[digits] = '\0';
}


/*
 * REQ: INFO <nmbr>&&<name>&&<line>
 * called from handleClientRequest()
 * Perform the requested action and send a response back to the client
 */
void handleReqInfo(int pos, const char * buf)
{
   char msgbuf[BUFSIZ];
   char *ptr;

   /* found a REQ: INFO <nmbr>&&<name>&&<line> line */
   char *temp = 0;
   const char * cptr, * whtentry;
   char  name[CIDSIZE], number[CIDSIZE], line[CIDSIZE], dialinfo[2*CIDSIZE+20];
   struct retbw which;

   cptr = buf + strlen(REQ_INFO);

   number[0] = name[0] = line[0] = '\0';
   /* parse out number name line */
   if (strlen(cptr) > 1)
   {
     cptr++;
     /* cheats const char cptr here */
     if ((temp = strstr(cptr, "&&"))) *temp = '\0';
     strncpy (number, cptr, CIDSIZE-1);
     number[CIDSIZE-1] = '\0';
     if (temp)
     {
       cptr += strlen(number) + 2;
       if ((temp = strstr(cptr, "&&"))) *temp = '\0';
       strncpy (name, cptr, CIDSIZE-1);
       name[CIDSIZE-1] = '\0';
     }
     if (temp)
     {
       cptr += strlen(name) + 2;
       strncpy (line, cptr, CIDSIZE-1);
       line[CIDSIZE-1] = '\0';
     }
   }
   else
   {
     /* line was only REQ: INFO\n
      * no reply logs
      * Unable to handle REQ: INFO request - Ignored.
      */
   }

   if (number[0] == '\0' || name[0] == '\0')
   {
      /* number and name must be supplied */
      logMsg(LEVEL1, "Unable to handle request %s - Ignored.\n", buf);
   }
   else
   {
      logMsg(LEVEL4, "Begin: findAlias() [%s]\n", strdate(ONLYTIME));
      temp = findAlias(name, number, line);
      logMsg(LEVEL4, "End: findAlias() [%s]\n", strdate(ONLYTIME));


      logMsg(LEVEL2, BEGIN_DATA3 NL);
      safeSend(pos, BEGIN_DATA3 CRLF);

      /* send back something like INFO: alias NOALIAS */
      logMsg(LEVEL2, INFOLINE "alias %s\n", temp);
      sprintf(msgbuf, INFOLINE "alias %s%s", temp, CRLF);
      safeSend(pos, msgbuf);

      which = onBlackWhite(name, number);
      switch (which.ret)
      {
      case 0:
              temp = "neither";
              break;
      case 1:
              temp = "black name";
              break;
      case 2:
              temp = "white name";
              break;
      case 3:
              temp = "both name";
              break;
      case 4:
              temp = "both number";
              break;
      case 5:
              temp = "black number";
              break;
      case 6:
              temp = "white number";
              break;
      default:
              temp = "";
              break;
      }

      /* which.ret and which.entry are from the whitelist */
      if (which.ret == 3 || which.ret == 4)
      {
        switch (which.ret)
        {
        case 3:
                temp = "both name";
                break;
        case 4:
                temp = "both number";
                break;
        }
        whtentry = which.entry;

        /* which.ret would be the same as for white */
        which = checkBlacklist(name, number);

        /* send back INFO: both name|number "<white entry>" "<black entry>" */
        logMsg(LEVEL2, "%s%s \"%s\" \"%s\"\n", INFOLINE, temp, whtentry, which.entry);
        sprintf (msgbuf, "%s%s \"%s\" \"%s\"\r\n", INFOLINE, temp, whtentry, which.entry);
      }
      else if (which.ret)
      {
        /* send back INFO: <black|white> <name|number> "[<entry>]" */
        logMsg(LEVEL2, "%s%s \"%s\"\n", INFOLINE, temp, which.entry);
        sprintf (msgbuf, "%s%s \"%s\"\r\n", INFOLINE, temp, which.entry);

      }
      else
      {
        /* send back INFO: neither */
        logMsg(LEVEL2, "%s%s\n", INFOLINE, temp);
        sprintf (msgbuf, "%s%s\r\n", INFOLINE, temp);
      }
      safeSend(pos, msgbuf);

      /* tell client if it can use dial */
      ptr = number;
      if (!(cidinput & 0x1))
      {
        /* 0x0 and 0x2 */
        while(*ptr != '\0')
        {
           if (!isdigit((int) *ptr)) break;
           ptr++;
        }
      }
      if ((cidinput & 0x1) || *ptr)
      {
        /* disable dial in the client */
        strcpy(dialinfo, "dial NODIAL");
      }
      else
      {
        /* enable dial in the client */
        sprintf(dialinfo, "dial %s&&%s", number, name);
      }

      /* send back  411 End of response */
      logMsg(LEVEL2, INFOLINE "%s\n" END_RESP NL, dialinfo);
      sprintf (msgbuf, INFOLINE "%s\r\n" END_RESP CRLF, dialinfo);
      safeSend(pos, msgbuf);

   }
}


/*
 * REQ: PAUSE <minute(s)>
 * called from handleClientRequest()
 * to pause hangups and set a time to resume
 * Perform the requested action and send a response back to the client
 */
void handleReqPause(int pos, const char * buf)
{
   char msgbuf[BUFSIZ];
   int hr, min, sec, remainder;

  /*
   * Found REQ: PAUSE line
   */
   const char * ptr = buf + strlen(REQ_PAUSE);

   while (*ptr == ' ') ptr++;

   /* expect a number of minutes and set a time to resume hangups */
   int pause = atoi(ptr);

   if (pause == -1)
   {
      hr = (hangupPauseTimer - time(0))/3600;
      remainder = (hangupPauseTimer - time(0))%3600;
      min = remainder/60;
      sec = remainder%60;
      if (hr < 0 || min < 0 || sec < 0) hr = min = sec = 0;

      if (hr != 0)
        sprintf(msgbuf, "%sRemaining Time: %d %s %d %s %d %s\n", RESPLINE,
                hr, hr == 1 ? "hour" : "hours", min,
                min == 1 ? "minute" : "minutes", sec,
                sec == 1 ? "second" : "seconds");
      else if (min != 0)
        sprintf(msgbuf, "%sRemaining Time: %d %s %d %s\n", RESPLINE, min,
                min == 1 ? "minute" : "minutes", sec,
                sec == 1 ? "second" : "seconds");
      else if (sec != 0)
        sprintf(msgbuf, "%sRemaining Time: %d %s\n", RESPLINE, sec,
                sec == 1 ? "second" : "seconds");
      else
        sprintf(msgbuf, "%sHangup not paused\n", RESPLINE);
   }
   else
   {
      hangupPauseTimer = time(0) + pause*60;

      if (pause > 0)
      {
          hr = pause/60;
          min = pause%60;
          if (hr != 0)
              sprintf(msgbuf, "%sPausing hangups for %d %s %d %s\n",
                      RESPLINE,
                      hr, hr == 1 ? "hour" : "hours",
                      min, min == 1 ? "minute" : "minutes");
          else if (min != 0)
              sprintf(msgbuf, "%sPausing hangups for %d %s\n",
                      RESPLINE,
                      min, min == 1 ? "minute" : "minutes");
      } else sprintf(msgbuf, "%sHangup enabled\n", RESPLINE);
   }

   logMsg(LEVEL2, BEGIN_DATA2 NL);
   safeSend(pos, BEGIN_DATA2 CRLF);

   logMsg(LEVEL2, msgbuf);
   safeSend(pos, msgbuf);

   logMsg(LEVEL2, END_RESP NL);
   safeSend(pos, END_RESP CRLF);
}


/*
 * REQ: black/white/alias add/remove
 * called from handleClientRequest()
 * Perform the requested action and send a response back to the client
 */
void handleReqABW(int pos, const char * buf)
{
   char msgbuf[BUFSIZ], opt[BUFSIZ];

   /*
    * Found REQ: black/white/alias add/remove
    * which will be passed to ncidutil
    */
   const char * filename = "";
   const char * type = "";
   char multi[BUFSIZ];
   const char * cptr;

   multi[0] = opt[0] = '\0';
   sprintf(opt, " --regex %d", regex);
   if (ignore1) strcat(opt, " --ignore1");
   cptr = buf + strlen(REQLINE);
   if (strncmp(cptr, BLK_LST, strlen(BLK_LST)) == 0)
   {
     /*
      * Received: REQ: black add "<string>" "[<comment>]"    [<date> <time>]
      * Received: REQ: black remove "<string>" ""    [<date> <time>]
      */
      filename = blacklist;
      cptr += strlen(BLK_LST);
      type = "Blacklist";
   }
   else if (strncmp(cptr, ALIAS_LST, strlen(ALIAS_LST)) == 0)
   {
     /*
      * Received: REQ: alias add "<number>&&<alias>" "<type>&&<name>"
      * Received: REQ: alias modify "<number>&&<alias>" "<type>&&<name>"
      * Received: REQ: alias remove "<number>&&" "<type>&&<name>"
      */
      filename = cidalias;
      cptr += strlen(ALIAS_LST);
      type = "Alias";
      sprintf (multi, "--multi \"%s %s\"", blacklist, whitelist);
   }
   else if (strncmp(cptr, WHT_LST, strlen(WHT_LST)) == 0)
   {
     /*
      * Received: REQ: white add "<string>" "[<comment>]"    [<date> <time>]
      * Received: REQ: white remove "<string>" ""    [<date> <time>]
      */
      filename = whitelist;
      cptr += strlen(WHT_LST);
      type = "Whitelist";
   }
   else
   {
      /* none of white/black/alias */
      logMsg(LEVEL1, "Unable to handle %s request - Ignored.\n", cptr);
      type = 0;
   }

   /*
    * any of REQ: white, black, alias
    * filename is the path to the ncidd.whitelist, ncidd.blacklist etc
    * cptr points to the rest eg  add "6138285800" "Fri Jul 14 13:22:57 2017"
    */
   if (type != 0 && strlen(cptr) > 4)
   {
       char * ptr;
       char * utilcmd;
       int    cnt;
       int    status;

       /* invoke ncidutil to adjust whitelist, blacklist or alias files */
       FILE        *respHandle;

       cptr++;
       utilcmd = (char *) malloc(strlen(DOUTIL)+strlen(opt)+strlen(multi)+strlen(filename)+strlen(type)+strlen(cptr)+10);

       sprintf (utilcmd, DOUTIL, opt, multi, filename, type, cptr);

       logMsg(LEVEL4, "Begin: Executing %s [%s]\n", NCIDUTIL, strdate(ONLYTIME));

       /* invoke /bin/sh -c utilcmd */
       respHandle = popen (utilcmd, "r");

       if (!respHandle) {
          /* when sh -c fails to start */
          logMsg(LEVELE, "popen %s failed - %s\n", NCIDUTIL, strerror(errno));
          safeSend(pos, "ncidutil failed to run" CRLF);
       }
       else {
          strcat(utilcmd, "\n");
          logMsg(LEVEL2, utilcmd);

          logMsg(LEVEL2, BEGIN_DATA2 NL);
          safeSend(pos, BEGIN_DATA2 CRLF);

          strcpy(msgbuf, RESPLINE);
          /* prepend RESP: so all the lines read will start this way */
          ptr = msgbuf + sizeof (RESPLINE) - 1;
          cnt = sizeof (msgbuf) - sizeof (RESPLINE);
          while (fgets(ptr, cnt, respHandle))
          {
            logMsg(LEVEL2, msgbuf);
            safeSend(pos, msgbuf);
          }

          status = pclose(respHandle);

          if (status < 0) {
             logMsg(LEVELE, "%s pclose %s\n", NCIDUTIL, strerror(errno));
          }
          else
          {
             if ( WIFEXITED(status) ) {
                logMsg(LEVEL4, "%s exited with status %d\n", NCIDUTIL, WEXITSTATUS(status) );
             }
             else {
                logMsg(LEVEL4, "%s abnormal exit\n", NCIDUTIL );
             }
          }
          logMsg(LEVEL4, "End: Executing %s [%s]\n", NCIDUTIL, strdate(ONLYTIME));

          logMsg(LEVEL2, END_RESP NL);
          safeSend(pos, END_RESP CRLF);
          free(utilcmd);
      }
   }
}


/*
 * handle a new connection from a client or gateway
 */
void acceptNewClient(void)
{
   int sd, newp;
   char buf[BUFSIZ], msgbuf[BUFSIZ];

   if ((sd = tcpAccept()) < 0)
   {
      logMsg(LEVEL2, "Connect Error: %s, sd: %d\n", strerror(errno), sd);
   }
   else
   {
      /* Client connected */

      /* set the client socket to non-blocking */
      if (fcntl(sd, F_SETFL, O_NONBLOCK) < 0)
      {
         logMsg(LEVEL2, "NONBLOCK Error: %s, sd: %d\n", strerror(errno), sd);
         close(sd);
      }
      else
      {
         if ((newp = addPoll(sd)) < 0)
         {
            /* sadly there are too many connections for the polld table */
            logMsg(LEVEL1, "Client %d trying to connect.\n", sd);
            logMsg(LEVEL1, TOOMANY, (cidinput & 0x2) ? MAXCLIENTS + 1 : MAXCLIENTS,
                    strdate(WITHSEP), NL);
            sprintf(buf, TOOMANY, (cidinput & 0x2) ? MAXCLIENTS + 1 : MAXCLIENTS,
                    strdate(WITHSEP), CRLF);
            if (write(sd, buf, strlen(buf)) < 0)
            {
               logMsg(LEVEL1, "too many reply %d fails with errno: %d %s [%s]\n", sd, errno, strerror(errno), strdate(ONLYTIME));
            }
            close(sd);
         }
         else
         {
            /* set up the new client connection info */
            if (IPinfo[newp].addr) free(IPinfo[newp].addr);
            if (IPinfo[newp].name) free(IPinfo[newp].name);
            IPinfo[newp].addr = strdup(tcpAccept_ret.IPaddr);
            IPinfo[newp].name = strdup(tcpAccept_ret.HostName);
            logMsg(LEVEL2, "\nClient %d pos %d from %s%s %s connected %s\n",
                    sd, newp, IPinfo[newp].addr, IPinfo[newp].name,
                    IPinfo[newp].dialout ? "dialout" : "nodialout", strdate(WITHSEP));

            QueuedMsg_begin_connect(newp);

            /* send a welcome message */
            logMsg(LEVEL2, "%s%s %s [%s:%d] %s\n", PRETTYINDENT, ANNOUNCE,
                   progname, hostname, port, VERSION);
            sprintf(buf, "%s %s [%s:%d] %s%s", ANNOUNCE, progname, hostname,
                    port, VERSION, CRLF);
            safeSend(newp, buf);

            logMsg(LEVEL2, "%s%s%s\n", PRETTYINDENT, APIANNOUNCE, API);
            sprintf(buf, "%s%s%s", APIANNOUNCE, API, CRLF);
            safeSend(newp, buf);

            /* Look for HELLO to come within 10 milliseconds */
            doHELLO(newp);

            /* client might have sent no_log or send_log to override the config setting */
            if ((sendlog && sendlog_override == -1) || sendlog_override == 1 )
            {
               sendLog(newp);
            }
            else
            {
               /* CID log not sent */
               logMsg(LEVEL2, "%s%s %ld: %s\n", PRETTYINDENT, NOLOGSENT, cidcall_lastchange, cidlog);
               sprintf(msgbuf, "%s %ld%s", NOLOGSENT, cidcall_lastchange, CRLF);
               safeSend(newp, msgbuf);
            }

            /* Send important server options to client */
            doOptLine(newp);

            /* End of startup messages */
            logMsg(LEVEL2, "%s%s\n\n", PRETTYINDENT, ENDSTARTUP);
            sprintf(msgbuf, "%s%s", ENDSTARTUP, CRLF);
            safeSend(newp, msgbuf);
         }
      }
   }
}


/*
 * Process HELLO lines if gateway or client sent HELLO <string> at connect.
 *
 * A message line is permitted without any HELLO lines or after any HELLO lines.
 *
 */
void doHELLO(int pos)
{
   int size, doingHello = 1, got_hello = 0;
   char *sptr, *eptr, buf[BUFSIZ];

   /*
    * use the config sendlog preference, unless this client says otherwise
    * 1- client wants cidcall data, 0- client wants quiet
    */
   sendlog_override = -1;

   if (polld[pos].fd < 0) return;

   while (doingHello)
   {

     usleep(HELLOWAIT); /* Wait time to receive HELLO */

     /* read data from socket without removing the data from buffer */
     size = recv(polld[pos].fd, buf, BUFSIZ, MSG_PEEK);

     /* Check for error */
     if (size == -1)
     {
       if (errno == EAGAIN)
       {
          logMsg(LEVEL2, "%s%s\n", PRETTYINDENT,
               got_hello ? "End of client HELLO data" :
                           "No client HELLO data received");
       }
       else
       {
          logMsg(LEVEL2, "Client %d pos %d Disconnected - recv returned %d: %s\n",
                polld[pos].fd, pos, errno, strerror(errno));
          disconnectClient(pos);
          /* a dead client won't need the logs */
          sendlog_override = 0;
       }
       doingHello = 0;
     }

     /* Check for data */
     if (size > 0)
     {
       buf[size] = '\0';

       /* remove whitespace from start and end of string */
       sptr = trimWhitespace(buf, buf + size - 2);


       /* look for the HELLO label in buf */
       if (!strncmp(sptr, HELLO, strlen(HELLO)))
       {
         if ((eptr = strchr(sptr, (int) '\n')))
         {
           if (*(eptr - 1) == '\r') *(eptr - 1) = '\0';
           *eptr = '\0';
           while (*(eptr + 1) == '\n' || *(eptr + 1) == '\r' || *(eptr + 1) == ' ') eptr++;
           size = (eptr - buf) + 1;
         }
         /* hello line found */
         got_hello = 1;
         logMsg(LEVEL2, "%sClient Sent: %s\n", PRETTYINDENT, buf);

         if (!strncmp(buf, COMMAND, strlen(COMMAND)))
         {
           if ((sptr = strstr(buf, "no_log")))
           {
             /* got "no_log" command */
             sendlog_override = 0;
             logMsg(LEVEL2, "%sClient Command Accepted: %s\n", PRETTYINDENT, sptr);

           }
           else if ((sptr = strstr(buf, "send_log")))
           {
             /* got "send_log" command */
             sendlog_override = 1;
             logMsg(LEVEL2, "%sClient Command Accepted: %s\n", PRETTYINDENT, sptr);
           }
           else
           {
             /*  unknown command */
             logMsg(LEVEL2, "%sUnknown Client Command: %s\n", PRETTYINDENT, buf);
           }
         }
         else if (!strncmp(buf, IDENT, strlen(IDENT)))
         {
           doIDENT(pos, buf);
         }
         else
         {
           logMsg(LEVEL2, "%sUnknown HELLO Line: %s\n", PRETTYINDENT, buf);
         }

         /* read and discard HELLO line from socket buffer */
         size = recv(polld[pos].fd, buf, size, 0);
       }
       else
       {
         /* indicate non-hello line found */
         doingHello = 0;
       }
     }
     else
     {
       doingHello = 0;
     }
   }
   return;
}


/*
 * Process IDENT lines if gateway or client sent IDENT <string>.
*/
void doIDENT(int pos, const char * buf)
{
   if (IPinfo[pos].ident) free(IPinfo[pos].ident);
   IPinfo[pos].ident = strdup(buf + strlen(IDENT));
   logMsg(LEVEL2, "%sClient Ident: %s\n", PRETTYINDENT, buf + strlen(IDENT));

   /* needed for "dial" in case it is enabled */
   if (strstr(buf + strlen(IDENT), GWOUT))
   {
     IPinfo[pos].dialout = 1;
     if (dialOUT >= 0) ++dialOUT;
     logMsg(LEVEL2, "%s       Gateway can handle outgoing calls: dialOUT=%d\n",
            PRETTYINDENT, dialOUT);
   }
   else IPinfo[pos].dialout = 0;
}


/*
 * Send important server options to clients
 * e.g.
 * OPT: country: <2 letter abbreviation>
 * OPT: if modem, send LineIDS: "<line1> <line2> ..."
 * OPT: if hangup, send hangup-<1-3>
 * OPT: if hangup, send hupmode-<1-3>
 * OPT: regex-<0-2>
 * OPT: if ignore1, send ignore1
 */
void doOptLine(int pos)
{
    char msgbuf[BUFSIZ];

    logMsg(LEVEL2, "%sSent 'country: %s' option to client\n", PRETTYINDENT, country);
    sprintf (msgbuf, "%scountry: %s\r\n", OPTLINE, country);
    safeSend(pos, msgbuf);

    if (!(cidinput & 0x1))
    {
        /* 0x0 and 0x2 */
        logMsg(LEVEL2, "%sSent LineIDs to client: %s\n", PRETTYINDENT, LineIDS);
        sprintf (msgbuf, "%sLineIDs: %s\r\n", OPTLINE, LineIDS);
        safeSend(pos, msgbuf);
    }
    if (hangup)
    {
        logMsg(LEVEL2, "%sSent 'hangup-%d' option to client\n", PRETTYINDENT, hangup);
        sprintf (msgbuf, "%shangup-%d\r\n", OPTLINE, hangup);
        safeSend(pos, msgbuf);
    }

    if (hupmode)
    {
        logMsg(LEVEL2, "%sSent 'hupmode-%d' option to client\n", PRETTYINDENT, hupmode);
        sprintf (msgbuf, "%shupmode-%d\r\n", OPTLINE, hupmode);
        safeSend(pos, msgbuf);
    }

    if (ignore1)
    {
        logMsg(LEVEL2, "%sSent 'ignore1' option to client\n", PRETTYINDENT);
        sprintf (msgbuf, "%signore1\r\n", OPTLINE);
        safeSend(pos, msgbuf);
    }

    logMsg(LEVEL2, "%sSent 'regex-%d' option to client\n", PRETTYINDENT, regex);
    sprintf (msgbuf, "%sregex-%d\r\n", OPTLINE, regex);
    safeSend(pos, msgbuf);
}


/*
 * Send a ACK: to gateways that asked for an echo
 */
void sendACK(int pos, const char * buf)
{
    char msgbuf[BUFSIZ];

    logMsg(LEVEL3, "Gateway %d pos %d %s%s %s%s",
            polld[pos].fd, pos, ACKLINE, buf, strdate(WITHSEP), NL);
    sprintf(msgbuf, "%s%s%s", ACKLINE, buf, CRLF);
    safeSend(pos, msgbuf);
}


/*
 * Get or create the INFO from a MSG: or NOT:
 * fills the global struct getINFO_ret
 * adds a \0 to the incoming buffer at the ###
 */
void getINFO(const char * bufptr)
{
    int len;
    char * ptr;

    *getINFO_ret.date = *getINFO_ret.time = '\0';
    len = strlen(bufptr);
    /* this cheats const bufptr here */
    if ((ptr = strstr(bufptr, " ###")) && len > (ptr + 8) - ptr && (*(ptr + 8) == '*'))
    {
        getField(ptr, "DATE", getINFO_ret.date, "");
        getField(ptr, "TIME", getINFO_ret.time, "");
        getField(ptr, "NAME", getINFO_ret.name, NONAME);
        getField(ptr, "NMBR", getINFO_ret.nmbr, NONMBR);
        getField(ptr, "LINE", getINFO_ret.line, NOLINE);
        getField(ptr, "MTYPE", getINFO_ret.type, NOTYPE);
        /* cut the string in bufptr at ### to leave the message part */
        *ptr = '\0';

    }
    else
    {
        /* fill in missing fields */
        strcpy(getINFO_ret.name, NONAME);
        strcpy(getINFO_ret.nmbr, NONMBR);
        strcpy(getINFO_ret.line, NOLINE);
        strcpy(getINFO_ret.type, NOTYPE);
    }
    userAlias(getINFO_ret.nmbr, getINFO_ret.name, getINFO_ret.line);

    if (!*getINFO_ret.date || !*getINFO_ret.time)
    {
        /* no date and time, create both */
        ptr = strdate(NOSEP);
        strncpy(getINFO_ret.date, ptr, 8);
        getINFO_ret.date[8] = '\0';
        strncpy(getINFO_ret.time, ptr + 9, 4);
        getINFO_ret.time[4] = '\0';
    }
}

/*
 * Get a INFO field from a MSG: or NOT:
 * modifies the content of mesgptr
 */
void getField(const char * bufptr, const char * field_name, char * mesgptr, const char * noval)
{
    char tmpbuf[BUFSIZ], *ptr;

    tmpbuf[BUFSIZ -1] = '\0';
    if ((ptr = strstr(bufptr, field_name)))
    {
        if (*(ptr + strlen(field_name) + 1) == '*')
        {
            /* no value for the field, return a default */
            strncpy(mesgptr, noval, CIDSIZE - 1);
        }
        else
        {
            strncpy(tmpbuf, ptr + strlen(field_name) + 1, BUFSIZ -1);
            ptr = strchr(tmpbuf, '*');
            if (ptr) *ptr = '\0';
            strncpy(mesgptr, tmpbuf, CIDSIZE - 1);
        }
    }
    else
    {
        /* field is absent, return a default */
        strncpy(mesgptr, noval, CIDSIZE - 1);
    }
    mesgptr[CIDSIZE-1] = '\0';
}


/*
 * Get a gateway field from a CALL: or CALLINFO: line
 * modifies the content of mesgptr
 * fields are separated by ... and the line ends in +++
 * fields and field names are not separated, e.g. ...NMBR12345...
 */
int getGWField(const char * bufptr, const char * field_name, char * mesgptr, const char * noval)
{
   char *sptr, *eptr;
   int ret = 1;

   if ((sptr = strstr(bufptr, field_name)) != NULL)
   {
      sptr += strlen(field_name);
      if ((eptr = strstr(sptr, "...")) != NULL || (eptr = strstr(sptr, "+++")) != NULL)
      {
         if (eptr != sptr)
         {
            strncpy(mesgptr, sptr, eptr - sptr);
            mesgptr[eptr - sptr] = '\0';
         }
         else {
            /* field is empty, return a default */
            strncpy(mesgptr, noval, CIDSIZE - 1);
         }
      }
      else {
         /* field is absent, return a default */
         strncpy(mesgptr, noval, CIDSIZE - 1);
      }
   }
   else
   {
      /* field is absent, return a default */
      strncpy(mesgptr, noval, CIDSIZE - 1);
      ret = 0;
   }
   mesgptr[CIDSIZE-1] = '\0';

   return ret;
}


/*
 * Format of data from NetCallerID for three types of calls and a message
 *
 *      ###DATE03301423...NMBR4075551212...NAMEWIRELESS CALL+++\r
 *      ###DATE03301423...NMBR...NAME-UNKNOWN CALLER-+++\r
 *      ###DATE03301423...NMBR...NAME+++\r
 *      ###DATE...NMBR...NAME   -MSG OFF-+++\r
 *
 * Gateway CALL Line Format:
 *      where type is one of: BLK CID HUP IN MWI OUT PID PUT RID WID
 *
 *      CALL: ###DATEmmddhhss...CALLtype...LINEidentifier...NMBRnumber...NAMEwords+++\r
 */
void gatewayCID(const char * buf, int pos)
{
   /* Process Caller ID information from a gateway */
   if (strncmp(buf, "###", 3) == 0)
   {
        /*
         * Found NetCallerID box, or a Gateway
         * All information received on one line
         * The Gateway creates a CID, HUP, OUT, PID, PUT Message Line
         * The Gateway contains a LINE and a CALL field
         * The NetCallerID box does not have a LINE or CALL field
         */
        char cidbuf[BUFSIZ];
        char *ptr;

        logMsg(LEVEL4, "Detected NetCallerID or gateway format\n");

        /* Make sure the status field and cidsent are zero */
        gwcid.calltype = CID;
        gwcid.status = 0;
        gwcid.cidsent = 0;
        /* the serial NetCallerID device has the ncidd.conf lineid */
        strcpy(gwcid.cidline, serialLineid(pos));

        if ((ptr = strstr(buf, "DATE")))
        {
            if (*(ptr + 4) == '.')
            {
                /* no date and time, create both */
                ptr = strdate(NOSEP);
                strncpy(gwcid.ciddate, ptr, 8);
                gwcid.ciddate[8] = '\0';
                gwcid.status |= CIDDATE;
                strncpy(gwcid.cidtime, ptr + 9, 4);
                gwcid.cidtime[4] = '\0';
                gwcid.status |= CIDTIME;
            }
            else
            {
                strncpy(gwcid.cidtime, ptr + 8, 4); /* HHMM */
                gwcid.cidtime[4] = '\0';
                gwcid.status |= CIDTIME;

                strncpy(gwcid.ciddate, ptr + 4, 4); /* MMDD */
                gwcid.ciddate[4] = '\0';

                /* need to get year */
                ptr = strdate(ONLYYEAR);        /* returns: YYYY */
                strcat(gwcid.ciddate, ptr);       /* MMDDYYYY */

                gwcid.status |= CIDDATE;
            }
        }

        /*
         * this field is only from a Gateway
         * calltype will be one of:
         * CALLBLK, CALLCID, CALLHUP, CALLIN, CALLMWI,
         * CALLOUT, CALLPID, CALLPUT, CALLRID, CALLWID
         */
        if ((ptr = strstr(buf, CALLIN)) || (ptr = strstr(buf, CALLCID)))
        {
            gwcid.calltype = CID; /* this is an incoming call */
        }
        else if ((ptr = strstr(buf, CALLBLK)))
        {
            gwcid.calltype = BLK; /* this is a blocked call */
        }
        else if ((ptr = strstr(buf, CALLHUP)))
        {
            gwcid.calltype = HUP; /* this is a blacklisted call hangup */
        }
        else if ((ptr = strstr(buf, CALLMWI)))
        {
            gwcid.calltype = MWI; /* this indicates voicemail message waiting */
        }
        else if ((ptr = strstr(buf, CALLOUT))) /* taa: added 'else' to beginning */
        {
            gwcid.calltype = OUT; /* this is an outgoing call */
        }
        else if ((ptr = strstr(buf, CALLPID)))
        {
            gwcid.calltype = PID; /* this is a incoming call from a smart phone */
        }
        else if ((ptr = strstr(buf, CALLPUT)))
        {
            gwcid.calltype = PUT; /* this is a outgoing call from a smart phone */
        }
        else if ((ptr = strstr(buf, CALLRID)))
        {
            gwcid.calltype = RID; /* this is a ringback */
        }
        else if ((ptr = strstr(buf, CALLWID)))
        {
            gwcid.calltype = WID; /* this is call waiting */
        }

        /* get caller telephone line (only from a Gateway) */
        if (getGWField(buf, "LINE", cidbuf, "-"))
           strcpy(gwcid.cidline, cidbuf);

        /* get caller telephone number */
        if (getGWField(buf, "NMBR", cidbuf, NONMBR))
            gwcid.status |= CIDNMBR;	/* if the NMBR field was present */
        builtinAlias(gwcid.cidnmbr, cidbuf);
        /* in case PID or PUT type (smartphone gateway (format from "international-only-digit" --> "(in)national-dialable-only-digit" ) ) */
        normalizeToDialableNumberFromFixedLine(gwcid.cidnmbr, CIDSIZE);

        /* get caller name */
        if (getGWField(buf, "NAME", cidbuf, NONAME))
            gwcid.status |= CIDNAME;	/* if the NAME field was present */
        builtinAlias(gwcid.cidname, cidbuf);

   }
}


/*
 * Found CALLINFO Line
 * parse it to produce an END: line to send to the clients
 *
 * CALLINFO Line Format:
 *  CALLINFO: ###CANCEL...DATE%s...SCALL%S...PCALL%s...ECALL%s...CALLIN...LINE%s...NMBR%s...NAME%s+++
 *  CALLINFO: ###CANCEL...DATE%s...SCALL%S...PCALL%s...ECALL%s...CALLOUT...LINE%s...NMBR%s...NAME%s+++
 *  CALLINFO: ###BYE...DATE%s...SCALL%S...PCALL%s...ECALL%s...CALLIN...LINE%s...NMBR%s...NAME%s+++
 *  CALLINFO: ###BYE...DATE%s...SCALL%S...PCALL%s...ECALL%s...CALLOUT...LINE%s...NMBR%s...NAME%s+++
 */
void handleGWCallinfo(int pos, const char * buf)
{
   char msgbuf[BUFSIZ];
   char *ptr, *sptr, *label;
   struct end_t endcall;
   /* additional info obtained from unformatted phone number in MSG and NOT */
   char * infoBuffer;

   writeLog(datalog, buf);
   if (IPinfo[pos].wants_ack) sendACK(pos, buf);

   /* get end of call date and time */
   label = "DATE";
   if ((sptr = strstr(buf, label)) != NULL)
   {
       /* points to DATEmmddhhmm... */
       sptr += (strlen(label));        /* MMDDHHMM */
       ptr = strdate(ONLYYEAR);        /* returns: YYYY */
       strncpy(endcall.date, sptr, 4); /* MMDD */
       endcall.date[4] = '\0';
       strcat(endcall.date, ptr);      /* MMDDYYYY */

       sptr += (4);                    /* HHMM */
       strncpy(endcall.time, sptr, 4);
       endcall.time[4] = '\0';
   }
   else
   {
       strcpy(endcall.date, "-");
       strcpy(endcall.time, "-");
   }

   /* get end of call start date and extended time */
   getGWField(buf, "SCALL", endcall.scall, "-");

   /* get end of call pickup date and extended time */
   getGWField(buf, "PCALL", endcall.pcall, "-");

   /* get end of call end date and extended time */
   getGWField(buf, "ECALL", endcall.ecall, "-");

   /* get end of call type */
   getGWField(buf, ".CALL", endcall.ctype, "-");

   /* get end of call line label */
   getGWField(buf, "LINE", endcall.line, "-");

   /* get end of call telephone number */
   getGWField(buf, "NMBR", endcall.nmbr, "-");
   normalizeToDialableNumberFromFixedLine(endcall.nmbr, CIDSIZE);

   /* get end of call name */
   getGWField(buf, "NAME", endcall.name, "-");
   userAlias(endcall.nmbr, endcall.name, endcall.line);

   /* get and process end of call termination */
   if (strstr(buf, CANCEL))
   {
       strcpy(endcall.htype, CANCEL);
       sendInfo(-1, endcall.line);
   }
   else if (strstr(buf, BYE))
   {
       strcpy(endcall.htype, BYE);
       sendInfo(-2, endcall.line);
   }
   else if (strstr(buf, BUSY))
   {
       strcpy(endcall.htype, BUSY);
       sendInfo(-3, endcall.line);
   }
   else strcpy(endcall.htype, "-");

   infoBuffer = GetCompleteInfoFromNumber(endcall.nmbr);

   /*
    * This sprintf() probably needs the optional blacklist
    * match name for NAME if ncidd does hangup but it is
    * not simple to add because of other possible calls
    * before this one ends.
    */
   sprintf(msgbuf, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
       ENDLINE,
       HTYPE, endcall.htype,
       DATE,  endcall.date,
       TIME,  endcall.time,
       SCALL, endcall.scall,
       PCALL, endcall.pcall,
       ECALL, endcall.ecall,
       CTYPE, endcall.ctype,
       LINE,  endcall.line,
       NMBR,  endcall.nmbr,
       infoBuffer,
       NAME,  endcall.name,
       STAR);

   /* Log the end of call "END:" line */
   writeLog(cidlog, msgbuf);
   /* remember the cidcall.log file date */
   cidcall_lastchange = time(0);

   /* Send END text line to clients */
   writeClients(msgbuf);
}


/*
 * message from a modem or serial device
 *
 * A USB modem                               (AT+VCID=1)
 *        call modemCID()
 *
 * a USB Modem sending MDMF or SDMF messages (AT+VCID=2)
 *        call xdmfCID()
 *
 * serial TCI multi-pots-line devices
 *        call TCIserial()
 */
void formatCID(char ** lines, int pos)
{
   const char * buf;
   struct modems_t * a_modem = findModemByPos(pos);

   if (a_modem == 0) return;

   if (cidinput == 0)
   {
      /* modem connected */

      while (*lines != NULL) {
         buf = *lines;
         if (strncmp(buf, "RING", 4) == 0 || strncmp(buf, "CID RING", 8) == 0)
         {
             /* RING always goes to modemCID() */
             modemCID(buf, pos);
         }
         else if (a_modem->modemType==2 ||
            (a_modem->modemType==0 &&
            (!strncmp(buf, "80", 2) || /* MDMF CID */
             !strncmp(buf, "04", 2) || /* SDMF CID */
             !strncmp(buf, "82", 2))))   /* MDMF MWI */
         {
             /* USB Modem sending MDMF or SDMF messages (AT+VCID=2) */
             xdmfCID(buf, pos);
             /* remember the modem type to avoid later confusion */
             a_modem->modemType = 2;
         }
         else
         {
             /* serial or USB modem sending multiline CID */
             modemCID(buf, pos);
             /* remember the modem type to avoid later confusion */
             a_modem->modemType = 1;
         }
         /* logMsg(LEVEL5, "free %p\n", *lines); */
         free(*lines);
         lines++;
      }
      finishCID(a_modem->modem_cid);
   }

   else if (cidinput == 0x1)
   {
      /* serial device connected */

      while (*lines != NULL) {
         buf = *lines;
         if (a_modem->modemType==3 ||
            (a_modem->modemType==0 &&
            (strlen(buf) == 70 &&
             isdigit((int) *buf) &&
             *(buf + 9) == '/' &&
             *(buf + 24) == 'M')))
         {
             /* Found TCI serial device */
             TCIserial(buf, pos);
             finishCID(a_modem->modem_cid);
             /* remember the modem type to avoid later confusion */
             a_modem->modemType = 3;
         }

         else if (a_modem->modemType==4 ||
                 (a_modem->modemType==0 &&
                 (strncmp(buf, "###", 3) == 0)))
         {
             /* Found a serial NetCallerID device */
             gatewayCID(buf, pos);
             finishCID(&gwcid);
             /* remember the modem type to avoid later confusion */
             a_modem->modemType = 4;
         }
         free(*lines);
         lines++;
      }
   }

   else  /* cidinput 2 and 3 */
   {
      /* serial device is being ignored */
      while (*lines != NULL) {
         free(*lines);
         lines++;
      }
   }
}


/*
 * see if all of the call information has arrived
 */
void finishCID(struct cid_t * fcid)
{
    const char * ptr;
    const char * linelabel;
    const char * nameptr;
    const char * announceptr;

   if (nothex(fcid->cidmesg) && fcid->status == (CIDALL3 | CIDMESG))
    {
        /*
         * date, time, nmbr and mesg were received
         * name is in the mesg field
         * copy mesg to name
         */
        fcid->status |= CIDNAME;
        strncpy(fcid->cidname, fcid->cidmesg, CIDSIZE - 1);
        fcid->cidname[CIDSIZE-1] = '\0';
        logMsg(LEVEL4, "received date, time, nmbr, mesg\n");
    }

    if (CIDALL3 == (fcid->status & IGNHIGH))
    {
        /* date, time and nmbr were received, set No NAME */
        fcid->status |= CIDNAME;
        strcpy(fcid->cidname,  NONAME);
        logMsg(LEVEL4, "received date, time, nmbr\n");
    }

    if (CIDALT3 == (fcid->status & IGNHIGH))
    {
        /* date, time and name were received, set No Number */
        fcid->status |= CIDNMBR;
        strcpy(fcid->cidnmbr,  NONMBR);
        logMsg(LEVEL4, "received date, time, name\n");
    }

    if (fcid->status == CIDNODT || fcid->status == (CIDNODT | CIDMESG))
    {
        /*
         * number and name received but no date and time
         * mesg may have been received
         */

        logMsg(LEVEL4,"No date and time found -- using current\n");
        ptr = strdate(NOSEP);           /* returns: MMDDYYYY HHMM */
        strncpy(fcid->ciddate, ptr, 8);   /*          0123456789012 */
        fcid->ciddate[8] = '\0';
        strncpy(fcid->cidtime, ptr + 9, 4);
        fcid->cidtime[4] = '\0';
        fcid->status |= (CIDDATE | CIDTIME);
    }

    if (CIDALL4 == (fcid->status & IGNHIGH))
    {
        /*
         * All Caller ID or outgoing call information received.
         *
         * Create the proper line type text line (CID, HUP, OUT, etc.)
         *
         * For OUT text lines (outgoing calls):
         *     the MESG field is not used
         *     the NAME field will be generic if no alias
         *
         * For HUP server generated text lines (hungup call):
         *     the CID label is replaced by a HUP label
         */

       logMsg(LEVEL4, "received %s\n", fcid->status == CIDALL4 ?
           "date, time, nmbr, name" : "date, time, nmbr, name, mesg");

        userAlias(fcid->cidnmbr, fcid->cidname, fcid->cidline);

        /* get additional info from phone number */
        if (fcid->infoBuffer != NULL) free(fcid->infoBuffer);
        fcid->infoBuffer = GetCompleteInfoFromNumber(fcid->cidnmbr);

        linelabel = getLineLabel(fcid->calltype);

        nameptr = fcid->cidname;
        announceptr = annpath;

        /* determine if a hangup check is wanted on proper calltype */
        if ((hangup || hupmode) && fcid->calltype == CID && (hangupPauseTimer < time(0)))
        {
            struct modems_t * a_modem;

            /* can only hangup on CID calls */
            logMsg(LEVEL3,
              "calltype=CID hangup=%d hupmode=%d cidline=%s lineid=%s\n",
                 hangup, hupmode, fcid->cidline, modem_list[0].serial_lineid);

            /* fcid->cidline must match serial_lineid to hangup */
            a_modem = findModemByLineid(fcid->cidline);
            if (a_modem)
            {
                logMsg(LEVEL3, "cidline == lineid: check whitelist\n");
                if (checkWhitelist(fcid->cidname, fcid->cidnmbr).ret)
                {
                   /* the name or number was on the whitelist */
                    logMsg(LEVEL3,
                      "Name or number on the whitelist: skip hangup check\n");
                   /* add the name found after #= in the whitelist */
                   if (listname) nameptr = listname;
                }
                else
                {
                    int dohangup = 0;
                    /* it wasn't on the white list */
                    if (hangup && checkBlacklist(fcid->cidname, fcid->cidnmbr).ret)
                    {
                       /*
                       * hangup phone if on blacklist but not whitelist
                       */
                       dohangup = hangup;
                       /* add the name found after #= in the blacklist */
                       if (listname) nameptr = listname;
                    }
                    else if (hupmode)
                    {
                       /*
                       * hangup phone if external hangup extension says to
                       */
                       dohangup = runExternalHangup(&linelabel, fcid, &nameptr, &announceptr);
                       /* dohangup = hupmode, but may have been downgraded
                        * nameptr may have been extended with the hangup reason
                        * announceptr = rmdpath -or- hermd
                        */
                    }

                    if (dohangup > 0)
                    {
                       /* try to hangup the call */
                       logMsg(LEVEL5, "Processing hangup request\n");

                       if (hangupCall_offhook(dohangup, a_modem) == 0)
                       {
                          /* change CID to HUP if it can go offhook */
                          linelabel = HUPLINE;
                          fcid->calltype = HUP;
                          sendCall(linelabel, fcid, nameptr);
                          fcid->cidsent = 1;
                          if (hangupCall(dohangup, announceptr, a_modem) == 0)
                          {
                             /* make the ring timer go off right away to reset the cid */
                             if (a_modem->ringtimer) a_modem->ringtimer = 1;
                          }
                          else
                          {
                             logMsg(LEVEL3, "hangupCall may have left line off hook\n");
                          }
                          /* indicate hangup completed even if hangup failed for modems providing CID */
                          sendInfo(-4, fcid->cidline);
                       }
                       else
                       {
                          logMsg(LEVEL3, "hangupCall could not go off hook\n");
                       }
                    }
                 }
            }
            else logMsg(LEVEL1, "No modem found for cidline %s - skipped hangup\n", fcid->cidline);
        }

        /* log the call and tell the clients/gateways */
        if (!fcid->cidsent)
           sendCall(linelabel, fcid, nameptr);

        /* Reset caller id info in modem_cid */
        strcpy(fcid->cidmesg, NOMESG); /* default message */
        fcid->status = 0;
        fcid->cidsent = 1;  /* don't do gencid on 2nd RING */
        fcid->calltype = 0;
        if (fcid->infoBuffer!=NULL) free(fcid->infoBuffer);
        fcid->infoBuffer = NULL;
    }
    else if (fcid->status != 0) {
       logMsg(LEVEL2, "finishCID did not receive full call info 0x%04x\n", fcid->status);
    }
}


/*
 * call the external hangup script(s) to see if hangup is wanted
 */
int runExternalHangup(const char ** linelabel, const struct cid_t * ecid, const char ** nameptr, const char ** rmdptr)
{
   char msgbuf[BUFSIZ], exthupcmd[BUFSIZ+(5*CIDSIZE)+60+strlen(ecid->infoBuffer)];
   char *ptr, *sptr;

   FILE *respHandle;
   int len, status, ehupmode, rethupmode;
   char *eptr;
   /* for the announcement file name */
   static char hermd[BUFSIZ];
   /* for adding the hangup reason to the caller name */
   static char hreason[BUFSIZ];

   /*
    * run an external script to decide if hangup is wanted
    */
   sprintf (exthupcmd, DOEXTEND, huppath, ecid->ciddate, ecid->cidtime,
            ecid->cidline, ecid->cidnmbr, hupmode, ecid->infoBuffer?ecid->infoBuffer:" ", ecid->cidname);
   logMsg(LEVEL4, "Begin: Executing %s [%s]\n", hupname, strdate(ONLYTIME));

   /* invoke /bin/sh -c exthupcmd */
   respHandle = popen (exthupcmd, "r");

   if (!respHandle) {
      /* when sh -c fails to start */
      logMsg(LEVELE, "popen %s failed - %s\n", hupname, strerror(errno));
      return 0; /* don't hang up */
   }

   logMsg(LEVEL3, "%s\n", exthupcmd);
   logMsg(LEVEL3, "default hangup mode for %s: %d\n", hupname, hupmode);

   ehupmode = hupmode;  /* effective hangup mode */
   *rmdptr = rmdpath;   /* default announcement for extensions */
   rethupmode = -1;     /* return hangup mode */

   /* read the lines returned from running the extension */
   while (fgets(msgbuf, sizeof(msgbuf), respHandle))
   {
      logMsg(LEVEL3, "%s: %s", hupname, msgbuf);
      if (ehupmode == 3)
      {
         /* check for recording name in hangup extension response */
         if ((ptr = strstr(msgbuf, RECFILE)))
         {
            /* skip over RECFILE */
            sptr = ptr + strlen(RECFILE);
            /* eliminate spaces between RECFILE and recording file name */
            while (sptr && *sptr == ' ') sptr++;
            if ((len = strlen(sptr) - 1))
            {
               /* eliminate '\n' at end of string */
               *(eptr = sptr + len) = '\0';
               /* eliminate spaces at the end of the recording file name */
               for (--eptr; *eptr && *eptr == ' '; --eptr) *eptr = '\0';

               /* if the extension doesn't give a valid recording, just hang up */
               if (checkVoiceFile(hermd, sptr))
               {
                  ehupmode = 1;
                  logMsg(LEVELE, "WARNING: Hangup Extension, recording file not found:\n         %s\n", hermd);
               }
               else
               {
                  /* return a pointer to this static file name */
                  *rmdptr = hermd;
               }
            }
            else
            {
               /* if the extension doesn't give a valid recording, just hang up */
               ehupmode = 1;
               logMsg(LEVEL3, "WARNING: no recording file, using Normal Hangup (mode 1)\n%s\n", huppath);
            }
         }
      }

      /* check for hangup reason in hangup extension response */
      if ((ptr = strstr(msgbuf, HUPREASON)))
      {
         /* skip over "HUPREASON" */
         sptr = ptr + strlen(HUPREASON);
         /* eliminate spaces between HUPREASON and reason text */
         while (sptr && *sptr == ' ') sptr++;
         if ((len = strlen(sptr) - 1))
         {
            /* eliminate '\n' at end of string */
            *(eptr = sptr + len) = '\0';
            /* eliminate spaces at the end of the reason text */
            for (--eptr; *eptr && *eptr == ' '; --eptr) *eptr = '\0';
            /* add the reason to the caller name */
            snprintf(hreason, sizeof(hreason)-1, "%s - %s", *nameptr, sptr);
            /* make sure name and reason does not exceed "CIDSIZE" */
            hreason[CIDSIZE - 1] = '\0';

            /* return a pointer to this static revised caller name */
            *nameptr = hreason;
         }
      }

      /* extension script responds OK or "hangup" */
      if (!strncmp(msgbuf, HUPNAME, strlen(HUPNAME)))
      {
         logMsg(LEVEL3, "using hangup mode: %d\n", ehupmode);
         if (ehupmode == 3)
            logMsg(LEVEL3, "using recording: %s\n", strrchr(*rmdptr, (int) '/'));
         rethupmode = ehupmode;
         break;
      }
      else if (strstr(msgbuf, "OK"))
      {
         rethupmode = 0;  /* don't hang up */
         break;
      }
   }
   if (rethupmode < 0) {
      logMsg(LEVELE, "%s did not return hangup or OK\n", hupname);
      rethupmode = 0;  /* don't hang up */
   }

   status = pclose(respHandle);

   if (status < 0) {
      logMsg(LEVELE, "%s pclose %s\n", hupname, strerror(errno));
   }
   else
   {
      if ( WIFEXITED(status) ) {
         if (WEXITSTATUS(status) != 0)
            logMsg(LEVELE, "%s exited with status %d\n", hupname, (char)(WEXITSTATUS(status)) );
         if (WEXITSTATUS(status) == 127 ||        /* extension not found */
             WEXITSTATUS(status) == 126 ||        /* extension not executable */
             (char)(WEXITSTATUS(status)) < 0) {
            rethupmode = 0; /* don't hang up */
         }
      }
      else {
         logMsg(LEVEL4, "%s abnormal exit\n", hupname );
         rethupmode = 0; /* don't hang up */
      }
   }
   logMsg(LEVEL4, "End: Executing %s [%s]\n", hupname, strdate(ONLYTIME));

   return rethupmode;
}


/*
 * Convert hex string to unsigned characters
 */
void hex2uchar(const char * buf, unsigned char * ucbuf)
{
    int buf_len = strlen(buf)/2, i;
    char tmpchar[5];

    bzero(tmpchar, sizeof(tmpchar));
    for (i = 0; i < buf_len; i++)
    {
        strncpy(tmpchar, buf + 2 * i, 2);
        ucbuf[i] = (unsigned char) strtol(tmpchar, NULL, 16);
    }
}

/*
 * 2's complement 8-bit checksum
 * Modified from: https://stackoverflow.com/questions/3463976/c-file-checksum
 */
unsigned char checksum (unsigned char *ptr, size_t size) {
  unsigned int sum = 0;
  while (size-- != 0)
  {
    sum += *ptr++;
  }
  sum = (sum ^ 0xFF) + 1;
  return (unsigned char)sum;
}


/*
 * Send the call line to clients
 *
 * Format of a CALL line passed to TCP/IP clients by ncidd:
 *      where CALL is one of: BLK CID HUP MWI OUT PID PUT RID WID
 *
 *      CALL: *DATE*mmddyyyy*TIME*hhmm*LINE*<id>*NMBR*number*MESG*NONE*FNMBR*<formatted number>*NTYPE*FIX/CELL*CTRY*<country code>*LOCA*<location>*CARI*<carrier>*NAME*<name>*
 *
 * Format of a MSG or NOT line passed to TCP/IP clients by ncidd:
 *
 *      MSG: <mesg> ***DATE*<date>*TIME*<time>*NAME*<name>*NMBR*<nmbr>*LINE*<line>*MTYPE*<type>*
 */
void sendCall(const char * linelabel, const struct cid_t * scid, const char * nameptr)
{
    char cidbuf[BUFSIZ];

    sprintf(cidbuf, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
        linelabel,
        DATE, scid->ciddate,
        TIME, scid->cidtime,
        LINE, scid->cidline,
        NMBR, scid->cidnmbr,
        MESG, scid->cidmesg,
              scid->infoBuffer?scid->infoBuffer:"",
        NAME, nameptr ? nameptr : scid->cidname,
        STAR);

    /* Log the CID, OUT, HUP, etc. text line */
    writeLog(cidlog, cidbuf);
    /* remember the cidcall.log file date */
    cidcall_lastchange = time(0);

    /*
     * Send the CID, OUT, HUP, etc. text line to clients
     */
    writeClients(cidbuf);

}


/*
 * called from safesend disconnectClient to finish up anything
 * that is specific to ncidd
 */
void finishDisconnect(int pos)
{
   /* remove the client from the IPinfo tables and turn off the ack flag */
   if (pos >= 0 && pos < MAXCONNECT)
   {
      IPinfo[pos].wants_ack = 0;
      if (IPinfo[pos].dialout == 1)
      {
         if (dialOUT > 0) --dialOUT;
         IPinfo[pos].dialout = 0;
      }
      if (IPinfo[pos].addr) free(IPinfo[pos].addr);
      if (IPinfo[pos].name) free(IPinfo[pos].name);
      if (IPinfo[pos].ident) free(IPinfo[pos].ident);
      IPinfo[pos].addr = NULL;
      IPinfo[pos].name = NULL;
      IPinfo[pos].ident = NULL;
   }
}


/*
 * Check if MESG is a hexadecimal number
 * http://ncid.sourceforge.net/doc/NCID-API.html#modem-mesg-hex
 *
 * Returns: 0 - if all characters are hexadecimal
 *          1 - if any character is not hexadecimal
 */

int nothex(const char * mesg)
{
    unsigned int pos;
    int ret = 0;

    for (pos = 0; pos < strlen(mesg); ++pos)
    {
        if (!(ret = isxdigit((int)mesg[pos]))) break;
    }

    return !ret;
}


/*
 * remove whitespace from the start and end of a string
 * sptr, eptr point to start and end
 * when eptr==0, uses strlen() to find the end
 */
char * trimWhitespace(char * sptr, char * eptr)
{
    if (eptr==NULL) eptr = sptr + strlen(sptr);
    while (*sptr && isspace((int) *sptr)) ++sptr;
    while (*eptr && isspace((int) *eptr)) *eptr-- = '\0';

    return sptr;
}


/*
 * fill in the current time and date
 */
void add_timestamp(char * ciddate, char * cidtime)
{
   char * ptr;

   ptr = strdate(NOSEP);

   strncpy(ciddate, ptr, 8);
   ciddate[8] = '\0';
   strncpy(cidtime, ptr + 9, 4);
   cidtime[4] = '\0';
}


/*
 * Send string to all TCP/IP CID clients.
 */

void writeClients(const char * inbuf)
{
    int pos;
    char buf[BUFSIZ];

    logMsg(LEVEL6, "writeClients: %s\n", inbuf);

    strcat(strcpy(buf, inbuf), CRLF);
    for (pos = 0; pos < MAXCONNECT; ++pos)
    {
        if (polld[pos].fd <= 0 || isSerialDevice(pos) ||
            polld[pos].fd == mainsock)
        {
            /* empty polld slots, the modem or serial device and listen port are not clients */
            continue;
        }

        logMsg(LEVEL6, "Sending data to Client %d pos %d at %s%s %s\n",
            polld[pos].fd, pos, IPinfo[pos].addr, IPinfo[pos].name, strdate(WITHSEP));

        safeSend(pos, buf);
    }
}


/*
 * Send log, if log file exists and is under cidlogmax bytes
 * or send that last cidlogmax bytes of it.
 */

void sendLog(int pos)
{
   int linecount;
   const char **ptr;
   char *iptr, *optr, input[BUFSIZ], logbuf[BUFSIZ], msgbuf[BUFSIZ];
   FILE *fp;

   if ((fp = fopen(cidlog, "r")) == NULL)
   {
      logMsg(LEVEL6, "cidlog: %d %s [%s]\n", errno, strerror(errno), strdate(ONLYTIME));
      logMsg(LEVEL1, "%s: %s%s", NOCALL_LOG, cidlog, NL);
      sprintf(msgbuf, "%s%s", NOCALL_LOG, CRLF);
      safeSend(pos, msgbuf);
   }
   else
   {
      /*
       * read each line of file, one line at a time
       * add "LOG" to line tag (CID: becomes CIDLOG:)
       * send line to clients
       */
      iptr = 0;
      linecount = 0;
      logMsg(LEVEL4, "%sBegin: Send call log: %s [%s]\n", PRETTYINDENT, cidlog, strdate(ONLYTIME));

       /* Indicate start of Call Log */
       logMsg(LEVEL2, "%s%s %ld%s", PRETTYINDENT, LOGSTART, cidcall_lastchange, NL);
       sprintf(msgbuf, "%s %ld%s", LOGSTART, cidcall_lastchange, CRLF);
       safeSend(pos, msgbuf);

      /* find a spot cidlogmax bytes from the end of the file */
      if (fseek(fp, -cidlogmax, SEEK_END) == 0)
      {
         /* start part way into the file if it is bigger than cidlogmax
          * skip a first partial line and all of the next
          */
         if (fgets(input, BUFSIZ - sizeof(LINETYPE), fp) != NULL &&
            fgets(input, BUFSIZ - sizeof(LINETYPE), fp) != NULL)
         {
            logMsg(LEVEL1, "%sCall Log too big, sending last %d bytes\n", PRETTYINDENT, cidlogmax);
            safeSend(pos, LOG2BIG CRLF);
         }
         else
         {
            logMsg(LEVELE, "Call Log access error %s\n", strerror(errno));
         }
      }
      else if (errno == EINVAL)
      { /* file is smaller than cidlogmax */
      }
      else
      { /* bad */
         logMsg(LEVELE, "Call Log access error %s\n", strerror(errno));
      }

      while (fgets(input, BUFSIZ - sizeof(LINETYPE), fp) != NULL)
      {
        /* strip <CR> and <LF> */
        if ((iptr = strchr( input, '\r')) != NULL) *iptr = '\0';
        if ((iptr = strchr( input, '\n')) != NULL) *iptr = '\0';

        optr = logbuf;
        iptr = input;
        if (strstr(input, ": ") != NULL)
        {
          /* possible line tag found */
          for (ptr = lineTags; *ptr; ++ptr)
          {
            if (!strncmp(input, *ptr, strlen(*ptr)))
            {
              /* copy line tag, skip ": " */
              for (iptr = input; *iptr != ':';) *optr++ = *iptr++;
              iptr += 2;
              break;
            }
          }
        }
        /*
         * if line "<label>: " found, line begins with "<label>LOG: "
         * if line label not found, line begins with "LOG: "
         */
        strcat(strncat(strcpy(optr, LOGLINE), iptr, BUFSIZ - (iptr - input - 1)), CRLF);
        safeSend(pos, logbuf);
        linecount++;

        optr = logbuf;
      }

      fclose(fp);

        /* Determine if a Call Log was sent */
        if (iptr)
        {
          /* Indicate end of the Call Log */
          logMsg(LEVEL2, "%sSent call log: %s %d lines\n%s%s\n", PRETTYINDENT, cidlog, linecount, PRETTYINDENT, LOGEND);
          sprintf(msgbuf, "%s %ld%s", LOGEND, cidcall_lastchange, CRLF);
          safeSend(pos, msgbuf);
        }
        else
        {
          logMsg(LEVEL2, "%s%s: %s%s", PRETTYINDENT, EMPTYLOG, cidlog, NL);
          sprintf(msgbuf, "%s %ld%s", EMPTYLOG, cidcall_lastchange, CRLF);
          safeSend(pos, msgbuf);
        }

    }

  logMsg(LEVEL4, "%sEnd: Send call log: %s [%s]\n", PRETTYINDENT, cidlog, strdate(ONLYTIME));

}

/*
 * Write log, if logfile exists.
 */

void writeLog(const char * logf, const char * logbuf)
{
    int logfd;
    ssize_t wnum;
    char msgbuf[BUFSIZ];

    /* write to server log */
    logMsg(LEVEL3, "%s\n", logbuf);

    if ((logfd = open(logf, O_WRONLY | O_APPEND)) < 0)
    {
        logMsg(LEVEL6, "%s: %s\n", logf, strerror(errno));
    }
    else
    {
        /* write log entry */
        sprintf(msgbuf, "%s\n", logbuf);
        do {
           wnum = write(logfd, msgbuf, strlen(msgbuf));
        } while (wnum<0 && errno == EINTR);
        close(logfd);
    }
}

/*
 * Send call information
 *
 * Format of CIDINFO line passed to TCP/IP clients by ncidd:
 *
 * CIDINFO: *LINE*<label>*RING*<number>*TIME<hh:mm:ss>
 */

void sendInfo(int ring, const char * lineid)
{
    char buf[BUFSIZ];
    char infoline[CIDSIZE];

    if (!sendinfo) return;

    strcpy(infoline, lineid);

    userAlias(0, 0, infoline);
    sprintf(buf, "%s%s%s%s%d%s%s%s",CIDINFO, LINE, infoline, \
            RING, ring, TIME, strdate(ONLYTIME), STAR);

    logMsg(LEVEL3, "%s\n", buf);
    writeClients(buf);
}


/*
 * does this position in polld go with a serial device ?
 */
int isSerialDevice(int pos)
{
   return (IPinfo[pos].modem != NULL);
}


/*
 * return a string for the call type
 */
const char * getLineLabel(int calltype)
{
    const char * linelabel;
    switch(calltype)
    {
        case BLK:
            linelabel = BLKLINE;
            break;
        case CID:
            linelabel = CIDLINE;
            break;
        case HUP:
            linelabel = HUPLINE;
            break;
        case MWI:
            linelabel = MWILINE;
            break;
        case OUT:
            linelabel = OUTLINE;
            break;
        case PID:
            linelabel = PIDLINE;
            break;
        case PUT:
            linelabel = PUTLINE;
            break;
        case RID:
            linelabel = RIDLINE;
            break;
        case WID:
            linelabel = WIDLINE;
            break;
        default: /* should not happen */
            linelabel = CIDLINE;
            break;
    }
    return linelabel;
}


/*
 * Close all file descriptors and restore tty parameters.
 */
void cleanup(void)
{
   int pos;

   /* restore tty parameters for serial device or modem */
   restoreTTYs();

      /*
       * send hangup to the modem if hangup option given
       * this makes sure modem hangs up if off-hook when ncidd dies
       */
    if (hangup || hupmode) {
       hangupModems();
    }

    /* close open files and sockets */
    for (pos = 0; pos < MAXCONNECT; ++pos)
        if (polld[pos].fd > 0) close(polld[pos].fd);

    /* remove pid file, if it was created */
    if (pid)
    {
        unlink(pidfile);
        logMsg(LEVEL1, "Removed pidfile: %s\n", pidfile);
    }

    /* close log file, if open */
    if (logptr) fclose(logptr);
}

/*
 * exit on a SIGINT, SIGQUIT, SIGABRT, SIGTERM signal
 */

void TERMhandler(int sig)
{
   /* set a flag so ncidd can close everything neatly before leaving */
   TERM_flag = sig;
}

void TERMexit(void)
{
   /* got signal SIGTERM via TERMhandler, clean up and leave */
   logMsg(LEVELE, "Received Signal %d: %s\nTerminated: %s\n",
           TERM_flag, strsignal(TERM_flag), strdate(WITHSEP));
   /* close open files and sockets */
   cleanup();
   exit(0);
}

/*
 * exit on a SIGSEGV signal
 * do minimal clean up after a memory exception
 */

void SEGVhandler(int sig)
{
    /* restore tty parameters */
    restoreTTYs();

    /* remove pid file, if it was created */
    if (pid) {
        unlink(pidfile);
    }

    /* and exit */
    _exit(sig);
}

/*
 * reload alias, blacklist and whitelist files
 */
void reload(void)
{
    logMsg(LEVEL1, "Reloading alias, blacklist and whitelist files [%s]\n",
      strdate(WITHSEP));

    /*
     * Decided not to do a reconfig because it seems like too much work
     * for a small gain:
     *   - open configurable files need to be closed and opened again
     *   - configuration verbose indicators, need to be output again
     * if (doConf()) errorExit(-104, 0, 0);
     */

    /* remove existing aliases to free memory used */
    rmaliases();

    /* reload alias file, but quit on error */
    if (readAliasFile()) {
       sysMesg("The new alias file has a problem", "-");
       errorExit(-109, 0, 0);
    }

    /* remove existing whitelist, blacklist entries to free memory used */
    rmBWlists();

    /* reload whitelist, blacklist entries, but quit on error */
    if (readBWlists()) {
       sysMesg("The new white/blacklist file has a problem", "-");
       errorExit(-114, 0, 0);
    }

    /* time of last ABW reads */
    abw_lastread  = time(0);
    checkABWtimer = time(0) + ABWCHECK;

    logMsg(LEVEL1,
      "End: Reloaded alias, blacklist and whitelist files [%s]\n", strdate(ONLYTIME));

}

/*
 * show connected clients
 */
void showConnected(void)
{
    int pos;
    int cnt = 0;

    logMsg(LEVELR, "Received Signal: show connected at %s\n", strdate(WITHSEP));

    for (pos = 0; pos < MAXCONNECT; ++pos)
    {
        if (polld[pos].fd <= 0 || isSerialDevice(pos) ||
            polld[pos].fd == mainsock)
        {
            /* empty polld slots, the modem and listen port are not clients */
            continue;
        }

        logMsg(LEVELR, "Client %03d pos %03d from %s%s %s {%s} is connected\n",
            polld[pos].fd, pos, IPinfo[pos].addr, IPinfo[pos].name,
            IPinfo[pos].dialout ? "dialout" : "nodialout", IPinfo[pos].ident);
        cnt++;
    }
    if (cnt == 0)
    {
        logMsg(LEVELR, "There are no clients connected at this time\n" );
    }
}

/*
 * check the dates on the ncidd.alias, blacklist and whitelist files
 * to see if something like a manual edit has changed them outside of ncidd
 * if so, reload the ABW caches
 */
void doCheckABWTimer(void)
{
   if (checkABWtimer < time(0)) {
      if (check_ABW_dates(whitelist) ||
          check_ABW_dates(blacklist) ||
          check_ABW_dates(cidalias))
      {
          reload();
          /* reload restarted the timer */
      }
      else {
          /* start the timer again */
          checkABWtimer = time(0) + ABWCHECK;
      }
   }
}


/*
 * return true if an ABW file is newer than the last ABW reload
 * and has been stable for half of the ABWCHECK time
 */
int check_ABW_dates(const char * filename)
{
   struct stat statbuf;

   if (stat(filename, &statbuf) == 0) {
      /* logMsg(LEVEL4, "%s ABW check %d\n", filename, abw_lastread); */
      if ((abw_lastread < statbuf.st_mtime) && (time(0) - statbuf.st_mtime > (ABWCHECK/2))) {
         logMsg(LEVEL4, "%s file changed since %d: size %d timestamp %d\n", filename, abw_lastread, statbuf.st_size, statbuf.st_mtime);
         return 1;
      }
   }
   else {
      /* the file is not readable */
      logMsg(LEVEL1, "%s: %d %s [%s]\n", filename, errno, strerror(errno), strdate(ONLYTIME));
   }
   return 0;
}


/*
 * see if cidcall.log been edited
 */
void doCidCallTimer(void)
{
   /* possibly don't care, when nothing is asking */
   if (checkCidCalltimer < time(0)) {
      check_cidlog_date();

      /* start the timer again */
      checkCidCalltimer = time(0) + 30;
   }
}


/*
 * check the date on the cidcall.log file to see if something like
 * cidupdate changed it outside of ncidd
 */
void check_cidlog_date(void)
{
   struct stat statbuf;

   /* check for outside changes to the cidcall log */
   if (stat(cidlog, &statbuf) == 0) {
/*later- this might need a fuzz if ncidd is slow to update the file */
      /* remember the actual file date */
      if (cidcall_lastchange < statbuf.st_mtime) {
         logMsg(LEVEL4, "CID logfile: %s size: %lu bytes, changed: %d\n", cidlog, statbuf.st_size, statbuf.st_mtime);
         /* remember the file date */
         cidcall_lastchange = statbuf.st_mtime;

      }
   }
   else {
      /* cidcall.log is not readable */
      logMsg(LEVEL1, "cidlog: %d %s [%s]\n", errno, strerror(errno), strdate(ONLYTIME));
   }
}


/*
 * USRhandler
 */

void    USRhandler(int sig)
{
   if ( sig == SIGHUP ) {
      /* SIGHUP reloads the alias, whitelist, blacklist */
      HUP_flag = 1;
   }
   else if ( sig == SIGUSR1 ) {
      /* SIGUSR1 updates the cidcall log by replacing cidcall.log file with cidcall.log.new*/
      USR1_flag = 1;
   }
   else if ( sig == SIGUSR2 ) {
      /* SIGUSR2 lists the connected clients */
      USR2_flag = 1;
   }
   else if ( sig == SIGALRM ) {
      /* SIGALRM when a read or write went too long */
      logMsg(LEVEL6, "Received Signal %d: %s\nIgnored: %s\n", sig, strsignal(sig), strdate(WITHSEP));
   }
   else {
      /* well that's odd */
      logMsg(LEVELE, "Received unexpected Signal %d: %s\nIgnored: %s\n", sig, strsignal(sig), strdate(WITHSEP));
   }
}

/*
 * set up signal handlers for HUP, USR1, USR2, INT, etc
 */

void setup_signal_handlers(void)
{
   /* SIGHUP reloads the alias, whitelist, blacklist */
   /* SIGUSR1 updates the cidcall log */
   /* SIGUSR2 lists the connected clients */
   HUP_flag = 0;
   USR1_flag = 0;
   USR2_flag = 0;
   static struct sigaction user_sigaction;
   user_sigaction.sa_handler = USRhandler;
   sigfillset( &user_sigaction.sa_mask);
   user_sigaction.sa_flags = 0;
   sigaction(SIGHUP, &user_sigaction, NULL);
   sigaction(SIGUSR1, &user_sigaction, NULL);
   sigaction(SIGUSR2, &user_sigaction, NULL);
   sigaction(SIGALRM, &user_sigaction, NULL);

   /* SIGINT, etc, will stop ncidd */
   TERM_flag = 0;
   static struct sigaction leave_sigaction;
   leave_sigaction.sa_handler = TERMhandler;
   sigfillset( &leave_sigaction.sa_mask);
   leave_sigaction.sa_flags = 0;
   sigaction(SIGINT, &leave_sigaction, NULL);    /* Interrupt from keyboard */
   sigaction(SIGQUIT, &leave_sigaction, NULL);    /* Quit from keyboard */
   sigaction(SIGABRT, &leave_sigaction, NULL);    /* Abort signal from abort(3) */
   sigaction(SIGTERM, &leave_sigaction, NULL);    /* Termination signal */

   /* SIGSEGV, will try to remove the lock file */
   static struct sigaction segv_sigaction;
   segv_sigaction.sa_handler = SEGVhandler;
   sigfillset( &segv_sigaction.sa_mask);
   segv_sigaction.sa_flags = 0;
   sigaction(SIGSEGV, &segv_sigaction, NULL);    /* Invalid memory reference */

   /* ignore a bunch of them */
   static struct sigaction ignore_sigaction;
   ignore_sigaction.sa_handler = SIG_IGN;
   sigemptyset( &ignore_sigaction.sa_mask);
   ignore_sigaction.sa_flags = 0;
   sigaction(SIGTSTP, &ignore_sigaction, NULL);
   sigaction(SIGPIPE, &ignore_sigaction, NULL);    /* Broken pipe: write to pipe with no readers */
   sigaction(SIGTTIN, &ignore_sigaction, NULL);
   sigaction(SIGTTOU, &ignore_sigaction, NULL);
   sigaction(SIGIO, &ignore_sigaction, NULL);    /* io is waiting */
   sigaction(SIGWINCH, &ignore_sigaction, NULL); /* Window resize signal */

   /* set these back to the default */
   static struct sigaction default_sigaction;
   default_sigaction.sa_handler = SIG_DFL;
   sigfillset( &default_sigaction.sa_mask);
   default_sigaction.sa_flags = 0;
   /* sigaction(SIGSEGV, &default_sigaction, NULL);*/
   sigaction(SIGXCPU, &default_sigaction, NULL);
   sigaction(SIGXFSZ, &default_sigaction, NULL);
}

/*
 * exit after some fatal error
 */

void errorExit(int error, const char * msg, const char * arg)
{
    if (error == -1)
    {
        /* should not happen */
        if (msg == 0) msg = "oops";

        /*
         * system error
         * print msg, arg should be zero
         */
        error = errno;
        logMsg(LEVELE, "%s: %s\n", msg, strerror(errno));
    }
    else
    {
        /* should not happen */
        if (msg != 0 && arg == 0) arg = "oops";

        /*
         * internal program error
         * print msg and arg if both are not 0
         */
        if (msg != 0 && arg != 0)
        {
            logMsg(LEVELE, "%s: %s\n", msg, arg);
        }
    }

    /* do not print terminated message, or cleanup, if option error */
    if (error != -100 && error != -101 && error != -106 && error != -107 &&
        error != -108 && error != -113)
    {
        logMsg(LEVELE, "Terminated:  %s\n", strdate(WITHSEP));
        cleanup();
    }

    exit(error);
}

/*
 * Normal exit
 */

void normalExit(void)
{
    logMsg(LEVEL1, "Terminated by Verbose Level 8:  %s\n", strdate(WITHSEP));
    cleanup();
    exit(0);
}

/*
 * Tell clients about a server issue
 */
void sysMesg(const char * message, const char * sysline)
{
    char sysmessage[BUFSIZ];
    char sysdate[CIDSIZE];
    char systime[CIDSIZE];
    char * tptr;
    const char * sysname = "ncidd";
    const char * sysnmbr = NONMBR;
    const char * systype = "SYS";

    tptr = strdate(NOSEP);
    strncpy(sysdate, tptr, 8);
    sysdate[8] = '\0';
    strncpy(systime, tptr + 9, 4);
    systime[4] = '\0';

    sprintf(sysmessage, SYSMSG, message, sysdate, systime, sysnmbr,
            sysline, systype, sysname);
    writeLog(cidlog, sysmessage);
    /* remember the cidcall.log file date */
    cidcall_lastchange = time(0);
    writeClients(sysmessage);
    logMsg(LEVEL1, "Server Message: %s\n", message, strdate(WITHSEP));
}


/*
 * turn polling off and reserve the slot in the table
 */
void reservePoll(int pos)
{
   /* addPoll will not reuse this slot,
   while polld[pos].events is preserved */
   polld[pos].fd = -111;
   polld[pos].revents = 0;
}


/*
 * turn polling on for the reserved slot in the table
 */
void restorePoll(int pos, int fd)
{
   polld[pos].fd = fd;
   /* polld[pos].events = pollevents; */
}


/*
 * ask poll to give Write Event when there's more stuff to be sent
 */
void enableWritePoll(int pos, int enable)
{
   if (enable) {
      polld[pos].events |= POLLOUT;
   }
   else {
      polld[pos].events |= POLLOUT;
      polld[pos].events ^= POLLOUT;
   }
}



/*
 * initialize or clear the modem cid
 */
void clear_modem_cid(struct modems_t * a_modem)
{
    struct cid_t * modem_cid = a_modem->modem_cid;

    strcpy(modem_cid->cidmesg, NOMESG); /* default message */

    strcpy(modem_cid->cidline, a_modem->serial_lineid); /* this modem's line indicator */
    modem_cid->status = 0;
    modem_cid->cidsent = 0;
    modem_cid->calltype = CID;

    modem_cid->ciddate[0] = '\0';
    modem_cid->cidtime[0] = '\0';
    modem_cid->cidnmbr[0] = '\0';
    modem_cid->cidname[0] = '\0';

    if (modem_cid->infoBuffer!=NULL) free(modem_cid->infoBuffer);
    modem_cid->infoBuffer = NULL;
}
