/*
 * ncidd.h - This file is part of ncidd.
 *
 * Copyright (c) 2005-2023
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _NCIDD_H
#define _NCIDD_H

#include "nciddalias.h"
#include "nciddhangup.h"
#include "version.h"
#include "libncid.h"

#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include <sys/utsname.h>

#include <sys/socket.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/wait.h>   /* needed for FreeBSD */
#include <netinet/in.h> /* needed for FreeBSD */
#include <arpa/inet.h>
#include <signal.h>

#ifndef CONFIG
#define CONFIG     "/etc/ncid/ncidd.conf"
#endif

#ifndef HOST_NAME_MAX
#define HOST_NAME_MAX 64 /* HOST_NAME_MAX is POSIX.1, Linux defines it as 64 */
#endif

#if (!defined(O_SYNC))
# define O_SYNC 0
#endif

#define UNUSED(var) (void) var

#define SHOWVER     "%s [%s:%d] %s\n%s\n"
#define DESC        "%s - Network CallerID Server\n"
#define NOOPT       "%s: not a option: %s\n"
#define USAGE       "\
Usage:   %s [options]\n\
Options: [-A <file>       | --alias <file>]\n\
         [-a <file>       | --announce <file>]\n\
         [-B <file>       | --blacklist <file>]\n\
         [-C <file>       | --config <file>]\n\
         [-c <file>       | --cidlog <file>]\n\
         [-D              | --debug]\n\
         [-d <file>       | --datalog <file>]\n\
         [-e <identifier> | --lineid <identifier>]\n\
         [-f <command>    | --audiofmt <command>]\n\
         [-g <0|1>        | --gencid <0|1>]\n\
         [-H <0|1|2|3>    | --hangup <0|1|2|3>]\n\
         [-h              | --help]\n\
         [-I <string>     | --initstr <string>]\n\
         [-i <string>     | --initcid <string>]\n\
         [-j <0|1|2|3>    | --cidinput <0|1|2|3>]\n\
         [-L <file>       | --logfile <file>]\n\
         [-l <file>       | --lockfile <file>]\n\
         [-M <MaxBytes>   | --cidlogmax <MaxBytes>]\n\
         [-m <0|1|2|3>    | --hupmode <0|1|2|3>]\n\
         [-P <file>       | --pidfile <file>]\n\
         [-p <portnumber> | --port <portnumber>]\n\
         [-Q <dotted-ip>  | --ifaddr <dotted-ip>]\n\
         [-r <0|1|2>      | --regex <0|1|2>]\n\
         [-S <ttyspeed>   | --ttyspeed <ttyspeed>]\n\
         [-s <datatype>   | --send <datatype>]\n\
         [-T <0|1>        | --ttyclocal <0|1>]\n\
         [-t <ttyport>    | --ttyport <ttyport>]\n\
         [-V              | --version]\n\
         [-v <1-9>        | --verbose <1-9>]\n\
         [-W <file>       | --whitelist <file>]\n\
         [--osx-launchd]\n\
"

/* make sure BUFSIZ on non-Linux is the same as on Linux */
#ifndef _G_BUFSIZ
#undef  BUFSIZ
#define BUFSIZ      8192
#endif

#ifndef TTYPORT
#define TTYPORT     "/dev/ttyACM0"
#endif
#ifndef CIDLOG
#define CIDLOG      "/var/log/cidcall.log"
#endif
#ifndef DATALOG
#define DATALOG     "/var/log/ciddata.log"
#endif
#ifndef LOGFILE
#define LOGFILE     "/var/log/ncidd.log"
#endif
#ifndef PIDFILE
#define PIDFILE     "/var/run/ncidd.pid"
#endif
#ifndef LOCKFILE
#define LOCKFILE    "/var/lock/LCK.."
#endif
#ifndef NCIDUPDATE
#define NCIDUPDATE  "/usr/bin/cidupdate"
#endif
#ifndef NCIDUTIL
#define NCIDUTIL    "/usr/bin/ncidutil"
#endif
#ifndef HUPEXTENSION
#define HUPEXTENSION   "hangup-nohangup"
#endif
#ifndef EXTDIR
#define EXTDIR      "/usr/share/ncid/extensions"
#endif
#ifndef RECORDING
#define RECORDING   "DisconnectedNotInService.rmd"
#endif
#ifndef ANNDIR
#define ANNDIR      "/usr/share/ncid/recordings"
#endif

#define STDOUT      1
#define CHARWAIT    2       /* deciseconds */

/* server messages */
#define ANNOUNCE    "200 Server:"
#define APIANNOUNCE "210 "
#define LOGEND      "250 End of call log"
#define NOLOGSENT   "251 Call log not sent"
#define EMPTYLOG    "252 Call log empty"
#define NOCALL_LOG  "253 No Call log"
#define LOGSTART    "254 Start of call log"
#define ENDSTARTUP  "300 End of connection startup"
#define BEGIN_DATA  "400 Start of data requiring OK"
#define BEGIN_DATA1 "401 Start of data requiring ACCEPT or REJECT"
#define BEGIN_DATA2 "402 Start of data showing status of handled request"
#define BEGIN_DATA3 "403 Start of data defining permitted requests"
#define END_DATA    "410 End of data"
#define END_RESP    "411 End of response"

#define NOCHANGES   "no changes"
#define DENIED      "denied"
#define DOUPDATE    NCIDUPDATE " -C %s -a %s -c %s -W %s -B %s "
#define DOUTIL      NCIDUTIL " %s %s \"%s\" %s %s 2>&1"
#define DOEXTEND    "%s \"*DATE*%s*TIME*%s*LINE*%s*NMBR*%s*MODE*%d%s*NAME*%s\" 2>&1"
#define HUPNAME      "hangup"
#define RECFILE      "Recording:"
#define HUPREASON    "HangupReason:"

/* server warning messages */
#define LOG2BIG      "MSGLOG: Call Log too big for cidlogmax, clipped here"
#define TOOMANY      "MSG: Too many clients connected (%d) %s%s"
#define SYSMSG       "MSG: %s ***DATE*%s*TIME*%s*NMBR*%s*LINE*%s*MTYPE*%s*NAME*%s*"

/* defaults for libphonenumber lookups */
#define DEFAULT_COUNTRY       "US"
#define DEFAULT_LANGUAGE      "en"


#define PORT         3333
#define MAXCLIENTS   50      /* maximun number of clients that can connect */
#define MAXCONNECT   MAXCLIENTS + 2 /* clients + mainsocket + ttyfd */
#define MAXIPBUF     500     /* max number of characters in address or name */
#define POLL_TIMEOUT 200     /* poll() timeout in milliseconds */
#define ABWCHECK     300     /* number of seconds to wait before checking ABW timestamps */

#define CRLF         "\r\n"
#define NL           "\n"
#define CR           "\r"
#define PRETTYINDENT "       "

#define NONAME       "NO NAME"
#define NONMBR       "NO-NUMBER"
#define NOLINE       "NO-LINE"
#define NOTYPE       "USER"
#define NOMESG       "NONE"
#define NOCID        "No Caller ID"

#define LOGMAXDEFAULT 1000000
#define LOGMAXIMUM   100000000

#define BLMSG       "Calls in the blacklist file will be terminated"
#define WLMSG       "Calls in the whitelist file will not be terminated"
#define IGNORE1     "Leading 1 ignored in call & alias/blacklist/whitelist"
#define NOIGNORE1   "Leading 1 in call & alias/blacklist/whitelist not ignored"
#define RELOADED    "Alias, blacklist and whitelist files have been read"

/* Call and message line labels sent to clients */
/* sorted alphabetically here for readability */

#define BLKLINE     "BLK: "
#define CIDLINE     "CID: "
#define ENDLINE     "END: "
#define HUPLINE     "HUP: "
#define MSGLINE     "MSG: "
#define MWILINE     "MWI: "
#define NOTLINE     "NOT: "
#define RLYLINE     "RLY: "
#define OUTLINE     "OUT: "
#define PIDLINE     "PID: "
#define PUTLINE     "PUT: "
#define RIDLINE     "RID: "
#define WIDLINE     "WID: "

/* Other line labels sent to clients */
/* sorted alphabetically here for readability */

#define ACKLINE    "ACK: "
#define INFOLINE   "INFO: "
#define LOGLINE    "LOG: "
#define OPTLINE    "OPT: "
#define REQLINE    "REQ: "
#define RESPLINE   "RESP: "
#define WRKLINE    "WRK: "

#define MESSAGE    "%s ***DATE*%s*TIME*%s*LINE*%s*NMBR*%s*MTYPE*%s%s*NAME*%s*"

#define RELOAD     "REQ: RELOAD"  /* reload alias, blacklist, whitelist files */
#define UPDATE     "REQ: UPDATE"  /* update current CID call log file */
#define UPDATES    "REQ: UPDATES" /* update all of the CID call log files */
#define REREAD     "REQ: REREAD"  /* reread current CID call log file */
#define REQ_INFO   "REQ: INFO"
#define ACPT_LOG   "ACCEPT LOG"
#define ACPT_LOGS  "ACCEPT LOGS"
#define RJCT_LOG   "REJECT LOG"
#define RJCT_LOGS  "REJECT LOGS"
#define BLK_LST    "black"
#define ALIAS_LST  "alias"
#define WHT_LST    "white"
#define ACK        "ACK"
#define YO         "YO"
#define REQ_ACK    "REQ: ACK"
#define REQ_YO     "REQ: YO"
#define REQ_PAUSE  "REQ: PAUSE"

/* optional HELLO lines sent by a client when it connects to the server */
#define HELLO      "HELLO: "
#define IDENT      "HELLO: IDENT: "
#define COMMAND    "HELLO: CMD: "
#define HELLOWAIT  100000 /* Wait 100 ms to receive HELLO */

/* COMMAND line available commands */
#define NOLOG      "no_log"
#define SENDLOG    "send_log"

#define GWOUT      "[OUT]"

#define GOODBYE        "GOODBYE"
#define CLIENT_GOODBYE "Client %d pos %d disconnected - received GOODBYE %s\n"

#define DIAL           "REQ: DIAL"    /* dial caller back */
#define DIAL_WORKING   "RESP: Working ..."
#define DIALOK         "RESP: Pickup phone within %d seconds"
#define DIAL_TIMEOUT   5
#define DIALED_NMBR   "RESP: Dialed %s on line \"%s\""
#define DIAL_LINE_ERR "RPLY: dial - format error: %s"
#define DIAL_FAIL     "RPLY: dial - dial failed, modem returned %s"
#define DIAL_HANGUP   "RPLY: dial - hungup %s on line \"%s\""
#define DIAL_HFAIL    "RPLY: dial - hungup failed, no call in progress"

#define LINETYPE   25
#define ONELINE    "POTS"

/* Call types received only from gateways (list 1 of 2) */
#define IN         30
#define CID        30
#define OUT        31
#define HUP        32
#define BLK        33
#define PID        34
#define PUT        35
#define WID        36
#define RID        37
#define MWI        38

/* Call types received only from gateways (list 2 of 2) */
/* kept in same sequence as (list 1 of 2) for consistency */
#define CALLIN     "CALLIN"
#define CALLCID    "CALLCID"
#define CALLOUT    "CALLOUT"
#define CALLHUP    "CALLHUP"
#define CALLBLK    "CALLBLK"
#define CALLPID    "CALLPID"
#define CALLPUT    "CALLPUT"
#define CALLWID    "CALLWID"
#define CALLRID    "CALLRID"
#define CALLMWI    "CALLMWI"

#define DATE       "*DATE*"
#define TIME       "*TIME*"
#define NMBR       "*NMBR*"
#define FNMBR      "*FNMBR*"
#define NTYPE      "*NTYPE*"
#define CTRY       "*CTRY*"
#define LOCA       "*LOCA*"
#define CARI       "*CARI*"
#define MESG       "*MESG*"
#define NAME       "*NAME*"
#define LINE       "*LINE*"
#define HTYPE      "*HTYPE*"
#define SCALL      "*SCALL*"
#define PCALL      "*PCALL*"
#define ECALL      "*ECALL*"
#define CTYPE      "*CTYPE*"
#define STAR       "*"

#define CIDINFO    "CIDINFO: "
#define RING       "*RING*"

#define CALL       "CALL: "
#define CALLINFO   "CALLINFO: "
#define CANCEL     "CANCEL"
#define BYE        "BYE"
#define BUSY       "BUSY"

/*
 * originally:
 * O - OUT-OF-AREA
 * A - ANONYMOUS
 * P - PRIVATE
 */
#define WORDFOR_O "Unavailable"
#define WORDFOR_A "ANONYMOUS"
#define WORDFOR_P "Withheld"

/*
 * Start of XDMF message definitions
 *
 * An XDMF message starts with one of the following bytes:
 *     SDMF_MESG_TYPE
 *     MDMF_MESG_TYPE
 *     MDMF_MSG_WAITING
 *
 * MDMF_MSG_WAITING is seldom used, it indicates voicemail
 * https://en.wikipedia.org/wiki/Message-waiting_indicator
 */

#define SDMF_MESG_TYPE    0x04
#define MDMF_MESG_TYPE    0x80
#define MDMF_MSG_WAITING  0x82

#define MDMF_VISUAL_IND   0x0B

#define MDMF_DATETIME     0x01
#define MDMF_CALLER_NMBR  0x02
#define MDMF_CALLER_NAME  0x07

#define MDMF_WHY_NO_NMBR  0x04
#define MDMF_WHY_NO_NAME  0x08

/*
 * Two reasons defined for (SDMF, MDMF) WHY_NO_NUMBER
 * and WHY_NO_NAME (MDMF)
 * WITHHELD    = 0x50   'P'RIVATE
 * UNAVAILABLE = 0x4F   'O'UT OF AREA
 */
#define LETTER_P          0x50
#define LETTER_O          0x4F

/*
 * CALL_TYPE_UK is only sent for ETSI FSK and BT FSK
 * In general it can be ignored.
 */
#define CALL_TYPE_UK      0x11
/*
 * There are three call subtypes defined:
 */
#define VOICE_UK          0x01
#define RINGBACK_UK       0x02
#define MSG_WAITING_UK    0x81

#define MDMF_NMSS_UK      0x13  /* Network Message System Status */

/* End of XDMF message definitions */


#define CIDDATE     0x01
#define CIDTIME     0x02
#define CIDNMBR     0x04
#define CIDNAME     0x08
#define CIDMESG     0x10
#define IGNHIGH     0x0F
#define CIDALL3     (CIDDATE | CIDTIME | CIDNMBR)
#define CIDALT3     (CIDDATE | CIDTIME | CIDNAME)
#define CIDALL4     (CIDDATE | CIDTIME | CIDNMBR | CIDNAME)
#define CIDNODT     (CIDNMBR | CIDNAME)

extern const char * hupname;
extern const char * huprmd;
extern const char * cidlog;
extern const char * datalog;
extern const char * logfile;
extern const char * pidfile;
extern const char * announce;
extern int port, hangup, hupmode;
extern int sendlog, sendinfo, ignore1, cidnoname, cidinput;
extern int nomodem, noserial, gencid, verbose, dialdelay;
extern long unsigned int cidlogmax;
extern char LineIDS[];

struct modems_t;

struct cid_t
{
    int status;
    int calltype;   /* CID HUP OUT etc */
    int cidsent;    /* used to give gencid on 2nd ring */
    char ciddate[CIDSIZE];
    char cidtime[CIDSIZE];
    char cidnmbr[CIDSIZE];
    char cidname[CIDSIZE];
    char cidmesg[CIDSIZE];
    char cidline[CIDSIZE];
    char * infoBuffer;
};

/* IPinfo[pos] is for same client/gateway/modem as in polld[pos] */
struct ipinfo_t {
    char * addr;
    char * name;
    char * ident;
    int  dialout;
    int  wants_ack;
    struct modems_t * modem;
};


void errorExit(int error, const char * msg, const char * arg);
void sysMesg(const char * message, const char * sysline);
int addPoll(int addfd);
void reservePoll(int pos);
void restorePoll(int pos, int fd);
void enableWritePoll(int pos, int enable);
void clear_modem_cid(struct modems_t * a_modem);

void writeClients(const char * inbuf);
void writeLog(const char * logf, const char * logbuf);
void sendInfo(int ring, const char * lineid);
void formatCID(char ** lines, int pos);
void finishCID(struct cid_t * fcid);
void checkHangupExtension(void);
int checkVoiceFile(char * path, const char * recording);
char * trimWhitespace(char *sptr, char *eptr);
void add_timestamp(char * ciddate, char * cidtime);
void hex2uchar(const char * buf, unsigned char *ucbuf);
unsigned char checksum(unsigned char *ptr, size_t size);

#endif /* ncidd.h */
