/*
 * nciddPN.cpp - This file is part of ncidd.
 *
 * Copyright (c) 2005-2022
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string>
#include <phonenumbers/phonenumber.pb.h>
#include <phonenumbers/phonenumberutil.h>
#include <phonenumbers/shortnumberinfo.h>
#ifdef HASCARRIERLOOKUP
#include "../libcarrier/phonenumber_offline_carrier.h"
#define HASNTYPELOOKUP
#define HASGEOCODELOOKUP
#endif
#ifdef HASGEOCODELOOKUP
#include <phonenumbers/geocoding/phonenumber_offline_geocoder.h>
#endif
#include "nciddPN.h"

#ifdef __cplusplus
    extern "C" {
#endif
    #include "ncidd.h"
    extern const char * country;
    extern const char * language;
#ifdef __cplusplus
    }
#endif

int nanp_format  = DEFAULT_US_FORMAT;

/*
 * for a two character country code, e.g. US
 * return the numeric 1-64 country code, 0 is "not found"
 */
using i18n::phonenumbers::PhoneNumberUtil;

int findRegionCode(const char * regioncode_c)
{
   std::string region_code(regioncode_c);
   const PhoneNumberUtil& phone_util(*PhoneNumberUtil::GetInstance());

   // return the numeric 1-64 country code, 0 is "not found"
   return phone_util.GetCountryCodeForRegion(region_code);
}

/*
 * for a two character country code, e.g. US
 * return the numeric 1-64 country code, 0 is "not found"
 */
int isNANPACountry(const char * regioncode_c)
{
   std::string region_code(regioncode_c);
   const PhoneNumberUtil& phone_util(*PhoneNumberUtil::GetInstance());

   return (int)phone_util.IsNANPACountry(region_code);
}

/*
 * return an infoBuffer full of fields derived from the cidnmbr
 * by calling libphonenumber, libgeocoding and libcarrier
 * e.g. for  6135991234 country=CA
 * *FNMBR*+1 613-599-1234*NTYPE*FIX/CELL*CTRY*CA*LOCA*Kanata, ON*CARI*Bell Canada
 */

char * GetCompleteInfoFromNumber(const char * cidnmbr_c)
{
   char Formattednmbr_c[CIDSIZE]; /* formatted phone number (international or national if caller's country == ncidd server's country )*/
#ifdef HASNTYPELOOKUP
   const char * Ntype = "-";       /* type of phone number (fixed,mobile, etc) */
#endif
   char callerCountry_c[CIDSIZE];     /* country code of caller  ex : "FR" */
   char infoBuffer[BUFSIZ];

#ifdef HASGEOCODELOOKUP
   char Location[CIDSIZE];    /*location ex: "London" */
#endif
#ifdef HASCARRIERLOOKUP
   char Carrier[CIDSIZE];     /* ex "Vodaphone" */
#endif

   std::string number_raw(cidnmbr_c);
   std::string ourregion(country);
// std::cout << "value of number_raw=" << number_raw  << std::endl;
// std::cout << "value of country=" << country  << std::endl;
   const PhoneNumberUtil& phone_util(*PhoneNumberUtil::GetInstance());
   i18n::phonenumbers::PhoneNumber psnumber;
   PhoneNumberUtil::ErrorType error_type;

   error_type = phone_util.Parse(number_raw, ourregion, &psnumber);
   switch (error_type) {
      case PhoneNumberUtil::NO_PARSING_ERROR:
        logMsg(LEVEL6, "error_type: NO_PARSING_ERROR \n");
        break;

      case PhoneNumberUtil::INVALID_COUNTRY_CODE_ERROR:
        logMsg(LEVEL6, "error_type: INVALID_COUNTRY_CODE_ERROR \n");
        break;

      case PhoneNumberUtil::NOT_A_NUMBER:
        logMsg(LEVEL6, "error_type: NOT_A_NUMBER \n");
        break;

      case PhoneNumberUtil::TOO_SHORT_AFTER_IDD:
        logMsg(LEVEL6, "error_type: TOO_SHORT_AFTER_IDD \n");
        break;

      case PhoneNumberUtil::TOO_SHORT_NSN:
        logMsg(LEVEL6, "error_type: TOO_SHORT_NSN \n");
        break;

      case PhoneNumberUtil::TOO_LONG_NSN:
        logMsg(LEVEL6, "error_type: TOO_LONG_NSN \n");
        break;
   }
   if (error_type == PhoneNumberUtil::NO_PARSING_ERROR) {
//   std::cout << "NO_PARSING_ERROR" << std::endl;
#ifdef HASNTYPELOOKUP
      PhoneNumberUtil::PhoneNumberType phone_number_type;
      phone_number_type = phone_util.GetNumberType(psnumber);
      switch (phone_number_type) {
         case PhoneNumberUtil::FIXED_LINE:
            Ntype = "FIXED";
            break;

         case PhoneNumberUtil::MOBILE:
            Ntype = "CELL";
            break;

         case PhoneNumberUtil::FIXED_LINE_OR_MOBILE:
            Ntype = "FIX/CELL"; // "FIXED_OR_CELL"
            break;

         case PhoneNumberUtil::TOLL_FREE:
            Ntype = "T-FREE";   // "TOLL_FREE"
            break;

         case PhoneNumberUtil::PREMIUM_RATE:
            Ntype = "PR-RATE";  // "PREMIUM_RATE"
            break;

         case PhoneNumberUtil::SHARED_COST:
            Ntype = "SHR-COST"; // "SHARED_COST"
            break;

         case PhoneNumberUtil::VOIP:
            Ntype = "VOIP";
            break;

         case PhoneNumberUtil::PERSONAL_NUMBER:
            Ntype = "PERS-NUM"; // "PERS_NUMBER"
            break;

         case PhoneNumberUtil::PAGER:
            Ntype = "PAGER";
            break;

         case PhoneNumberUtil::UAN:
            Ntype = "UAN"; //
            break;

         case PhoneNumberUtil::VOICEMAIL:
            Ntype = "V-MAIL";   // "VOICEMAIL"
            break;

         case PhoneNumberUtil::UNKNOWN:
         default:
            using i18n::phonenumbers::ShortNumberInfo;
            ShortNumberInfo * short_info = new ShortNumberInfo();
            if (short_info->IsValidShortNumber(psnumber)) {
                logMsg(LEVEL6, "valid short number\n");
            }

            if (short_info->IsCarrierSpecific(psnumber)) {
                logMsg(LEVEL6, "Is Carrier Specific\n");
            }
            if (short_info->IsValidShortNumberForRegion(psnumber, ourregion)) {
               logMsg(LEVEL6, "valid short number for region %s \n", country);

               ShortNumberInfo::ShortNumberCost cost ;
               cost =  short_info->GetExpectedCostForRegion(psnumber, ourregion );
               logMsg(LEVEL6, "cost = %d \n",cost);
               switch (cost ) {
                 case ShortNumberInfo::ShortNumberCost::TOLL_FREE:
                    Ntype = "S_T-FREE";   // "TOLL_FREE"
                    break;

                 case ShortNumberInfo::ShortNumberCost::STANDARD_RATE:
                    Ntype = "S_STD-RATE";   // "STANDARD_RATE"
                    break;

                 case ShortNumberInfo::ShortNumberCost::PREMIUM_RATE:
                    Ntype = "S_PR-RATE";   // "PREMIUM_RATE"
                    break;

                 case ShortNumberInfo::ShortNumberCost::UNKNOWN_COST:
                    Ntype = "S_UNKNOWN";   // "Short Unknown"
                    break;
               }
               if (short_info->IsEmergencyNumber(number_raw, ourregion)) {
                  logMsg(LEVEL6, "Emmergency Number\n");
                  Ntype = "S_EMGRCY";   // "Short Emergency"
               }
            }
            else
            {
               logMsg(LEVEL6, "invalid short number\n");
               Ntype = "-"; /*UNKNOWN*/
            }
            break;
      }
#endif

      Formattednmbr_c[0] = '\0';

      std::string country_code;

      /* look up the caller's country */
      phone_util.GetRegionCodeForNumber(psnumber, &country_code);
      strcpy(callerCountry_c, country_code.c_str());

      logMsg(LEVEL6, "country(receiving):%s country(calling):%s\n", country, callerCountry_c);

      if (0 == strcmp(country, callerCountry_c) ||
          0 == strcmp("ZZ", callerCountry_c)) {
         /* country of calling number is same as country of receiving line (national call)
          * or calling country is unknown (ZZ) */

         if (phone_util.IsNANPACountry(ourregion)) {
            /* formatting options for NANP (North American Numbering Plan) countries */
            /* https://en.wikipedia.org/wiki/North_American_Numbering_Plan */
            if (nanp_format == 0) {
               /* US unformatted */
               logMsg(LEVEL6, "nanp_format == 0 (no format) original_number %s\n", Formattednmbr_c);
               strncpy(Formattednmbr_c, cidnmbr_c, CIDSIZE - 1);
               Formattednmbr_c[CIDSIZE-1] = '\0';
            }
            else if (nanp_format == 2 || nanp_format == 3) {
               /* ncid formatted */
               /* change (416) 555-1234 into 416-555-1234 */
               std::string number_raw(cidnmbr_c);
               std::string ourregion(country);
               std::string formattedNumber;

               phone_util.Format(psnumber, PhoneNumberUtil::NATIONAL, &formattedNumber);
               strncpy(Formattednmbr_c, formattedNumber.c_str(), CIDSIZE);
               Formattednmbr_c[CIDSIZE-1] = '\0';

               /* Formattednmbr_c = "(416) 555-1234" */
               char * p1 = strstr(Formattednmbr_c, "(");
               if (p1 != NULL) {
                  char * p2 = strstr(Formattednmbr_c, ") ");
                  if (p2 != NULL) {
                     char numberbuffer[CIDSIZE];
                     if (nanp_format == 2) {
                         strncpy(numberbuffer, p1 + 1, p2 - 1 - p1);
                         numberbuffer[p2 - 1 - p1] = '-';
                         numberbuffer[p2 - p1] = '\0';
                         /* numberbuffer = "416-" */
                         strncpy(numberbuffer + (p2 - p1), p2 + 2, CIDSIZE - (p2 - p1) - 1);
                         numberbuffer[CIDSIZE-1] = '\0';
                         /* numberbuffer = "416-555-1234" */
                         strcpy(Formattednmbr_c, numberbuffer);
                         /* Formattednmbr_c = "416-555-1234"  */
                         logMsg(LEVEL6, "nanp_format == 2 (all_hyphen) %s\n", Formattednmbr_c);
                     }
                     else {
                         strncpy(numberbuffer, p1 + 1, p2 - 1 - p1);
                         numberbuffer[p2 - 1 - p1] = '.';
                         numberbuffer[p2 - p1] = '\0';
                         /* numberbuffer = "416." */
                         strncpy(numberbuffer + (p2 - p1), p2 + 2, CIDSIZE - (p2 - p1) - 1);
                         numberbuffer[CIDSIZE-1] = '\0';
                         if ((p2 = strchr(numberbuffer, '-'))) *p2 = '.';
                         /* numberbuffer = "416.555.1234" */
                         strcpy(Formattednmbr_c, numberbuffer);
                         /* Formattednmbr_c = "416.555.1234"  */
                         logMsg(LEVEL6, "nanp_format == 3 (all_dot) %s\n", Formattednmbr_c);
                     }
                  }
               }
            }
            else {
               /* nanp_format can be 1 or greater than 2 */
               /* Formattednmbr_c = google formatted = (204) 555-1234 */
               std::string number_raw(cidnmbr_c);
               std::string ourregion(country);
               std::string formattedNumber;

               phone_util.Format(psnumber, PhoneNumberUtil::NATIONAL, &formattedNumber);
               strncpy(Formattednmbr_c, formattedNumber.c_str(), CIDSIZE);
               Formattednmbr_c[CIDSIZE-1] = '\0';
               logMsg(LEVEL6, "nanp_format == %d %s\n", nanp_format, Formattednmbr_c);
            }
         }
         else {
            /* Formattednmbr = google formatted = (204) 555-1234 */
            std::string number_raw(cidnmbr_c);
            std::string ourregion(country);
            std::string formattedNumber;

            phone_util.Format(psnumber, PhoneNumberUtil::NATIONAL, &formattedNumber);
            strncpy(Formattednmbr_c, formattedNumber.c_str(), CIDSIZE);
            Formattednmbr_c[CIDSIZE-1] = '\0';
         }
      }
      else {
         /* international caller */
         std::string number_raw(cidnmbr_c);
         std::string ourregion(country);
         std::string formattedNumber;

         //   std::cout << "NO_PARSING_ERROR" << std::endl;
         phone_util.Format(psnumber, PhoneNumberUtil::INTERNATIONAL, &formattedNumber);
         //std::cout << "value of formattedNumber=" << formattedNumber  << std::endl;
         strncpy(Formattednmbr_c, formattedNumber.c_str(), CIDSIZE);
         Formattednmbr_c[CIDSIZE-1] = '\0';
      }


      if (0 == strlen(Formattednmbr_c)) {
         /*formatted number came back empty, so we put unformatted number (whatever it is ) into it (better than nothing) */
         strncpy(Formattednmbr_c, cidnmbr_c, CIDSIZE - 1);
         Formattednmbr_c[CIDSIZE-1] = '\0';
      }

#ifdef HASGEOCODELOOKUP
      Location[0] = '\0';
      std::string region_description;
      icu::Locale k_Locale = icu::Locale(language);
      //std::cout << "value of number_raw=" << number_raw  << std::endl;
      //std::cout << "value of country=" << country  << std::endl;
      using i18n::phonenumbers::PhoneNumberOfflineGeocoder;
//   std::cout << "NO_PARSING_ERROR" << std::endl;
      PhoneNumberOfflineGeocoder * geocoder = new PhoneNumberOfflineGeocoder();

      region_description = geocoder->GetDescriptionForNumber(psnumber, k_Locale);
      //std::cout << "value of formattedNumber=" << formattedNumber  << std::endl;
      strncpy(Location, region_description.c_str(), CIDSIZE);
      Location[CIDSIZE-1] = '\0';
      if (0 == strlen(Location)) strcpy(Location, "-");
#endif

#ifdef HASCARRIERLOOKUP
      Carrier[0] = '\0';
      std::string carrier_name;
      //std::cout << "value of number_raw=" << number_raw  << std::endl;
      //std::cout << "value of country=" << country  << std::endl;
      using i18n::phonenumbers::PhoneNumberOfflineCarrier;
//   std::cout << "NO_PARSING_ERROR" << std::endl;
      PhoneNumberOfflineCarrier * carrier = new PhoneNumberOfflineCarrier();

      carrier_name = carrier->GetDescriptionForNumber(psnumber, k_Locale);
      //std::cout << "value of formattedNumber=" << formattedNumber  << std::endl;
      strncpy(Carrier, carrier_name.c_str(), CIDSIZE);
      Carrier[CIDSIZE-1] = '\0';
      if (0 == strlen(Carrier)) strcpy(Carrier, "-");
#endif
   }
   else {
      /* cidnmbr_c could not be parsed as a number */
#ifdef HASNTYPELOOKUP
      Ntype = "-"; /*UNKNOWN*/
#endif
      strcpy(callerCountry_c, "ZZ");
      strcpy(Formattednmbr_c, cidnmbr_c);
#ifdef HASGEOCODELOOKUP
      strcpy(Location, "-");
#endif
#ifdef HASCARRIERLOOKUP
      strcpy(Carrier, "-");
#endif
   }

   char FNMBRfragment[CIDSIZE+sizeof(FNMBR)];
   char CTRYfragment[CIDSIZE+sizeof(CTRY)];
   char LOCAfragment[CIDSIZE+sizeof(LOCA)];
   char CARIfragment[CIDSIZE+sizeof(CARI)];
   char NTYPEfragment[CIDSIZE];
   snprintf(FNMBRfragment, sizeof(FNMBRfragment), "%s%s", FNMBR, Formattednmbr_c);
   snprintf(CTRYfragment, sizeof(CTRYfragment), "%s%s", CTRY, callerCountry_c);
#ifdef HASNTYPELOOKUP
   snprintf(NTYPEfragment, sizeof(NTYPEfragment), "%s%s", NTYPE, Ntype);
#else
   NTYPEfragment[0] = '\0';
#endif
#ifdef HASGEOCODELOOKUP
   snprintf(LOCAfragment, sizeof(LOCAfragment), "%s%s", LOCA, Location);
#else
   LOCAfragment[0] = '\0';
#endif
#ifdef HASCARRIERLOOKUP
   snprintf(CARIfragment, sizeof(CARIfragment), "%s%s", CARI, Carrier);
#else
   CARIfragment[0] = '\0';
#endif

   snprintf(infoBuffer, sizeof(infoBuffer), "%s%s%s%s%s",
            FNMBRfragment,
            NTYPEfragment,
            CTRYfragment,
            LOCAfragment,
            CARIfragment);

   logMsg(LEVEL6, "infobuffer: %d %s\n", strlen(infoBuffer), infoBuffer);
   return strdup(infoBuffer);
}

/*
 *
 */
void format_number(char * Formattednmbr_c, const char * cidnmbr_c )
{
   char callerCountry_c[CIDSIZE];     /* country code of caller  ex : "FR" */

   std::string number_raw(cidnmbr_c);
   std::string ourregion(country);
// std::cout << "value of number_raw=" << number_raw  << std::endl;
// std::cout << "value of country=" << country  << std::endl;
   const PhoneNumberUtil& phone_util(*PhoneNumberUtil::GetInstance());
   i18n::phonenumbers::PhoneNumber psnumber;
   PhoneNumberUtil::ErrorType error_type;

   error_type = phone_util.Parse(number_raw, ourregion, &psnumber);
   switch (error_type) {
      case PhoneNumberUtil::NO_PARSING_ERROR:
        logMsg(LEVEL6, "error_type: NO_PARSING_ERROR \n");
        break;

      case PhoneNumberUtil::INVALID_COUNTRY_CODE_ERROR:
        logMsg(LEVEL6, "error_type: INVALID_COUNTRY_CODE_ERROR \n");
        break;

      case PhoneNumberUtil::NOT_A_NUMBER:
        logMsg(LEVEL6, "error_type: NOT_A_NUMBER \n");
        break;

      case PhoneNumberUtil::TOO_SHORT_AFTER_IDD:
        logMsg(LEVEL6, "error_type: TOO_SHORT_AFTER_IDD \n");
        break;

      case PhoneNumberUtil::TOO_SHORT_NSN:
        logMsg(LEVEL6, "error_type: TOO_SHORT_NSN \n");
        break;

      case PhoneNumberUtil::TOO_LONG_NSN:
        logMsg(LEVEL6, "error_type: TOO_LONG_NSN \n");
        break;
  }

   if (error_type == PhoneNumberUtil::NO_PARSING_ERROR) {
//   std::cout << "NO_PARSING_ERROR" << std::endl;

      Formattednmbr_c[0] = '\0';

      std::string country_code;

      /* look up the caller's country */
      phone_util.GetRegionCodeForNumber(psnumber, &country_code);
      strcpy(callerCountry_c, country_code.c_str());
      logMsg(LEVEL6, "country(receiving):%s country(calling):%s\n", country, callerCountry_c);

      if (0 == strcmp(country, callerCountry_c) ||
          0 == strcmp("ZZ", callerCountry_c)) {
         /* country of calling number is same as country of receiving line (national call)
          * or calling country is unknown (ZZ) */

         if (phone_util.IsNANPACountry(ourregion)) {
            /* formatting options for NANP (North American Numbering Plan) countries */
            /* https://en.wikipedia.org/wiki/North_American_Numbering_Plan */
            if (nanp_format == 0) {
               /* US unformatted */
               logMsg(LEVEL6, "nanp_format == 0 (no format) original_number %s\n", Formattednmbr_c);
               strncpy(Formattednmbr_c, cidnmbr_c, CIDSIZE - 1);
               Formattednmbr_c[CIDSIZE-1] = '\0';
            }
            else if (nanp_format == 2 || nanp_format == 3) {
               /* ncid formatted */
               /* change (416) 555-1234 into 416-555-1234 */
               std::string number_raw(cidnmbr_c);
               std::string ourregion(country);
               std::string formattedNumber;

               phone_util.Format(psnumber, PhoneNumberUtil::NATIONAL, &formattedNumber);
               strncpy(Formattednmbr_c, formattedNumber.c_str(), CIDSIZE);
               Formattednmbr_c[CIDSIZE-1] = '\0';

               /* Formattednmbr_c = "(416) 555-1234" */
               char * p1 = strstr(Formattednmbr_c, "(");
               if (p1 != NULL) {
                  char * p2 = strstr(Formattednmbr_c, ") ");
                  if (p2 != NULL) {
                     char numberbuffer[CIDSIZE];
                     if (nanp_format == 2) {
                         strncpy(numberbuffer, p1 + 1, p2 - 1 - p1);
                         numberbuffer[p2 - 1 - p1] = '-';
                         numberbuffer[p2 - p1] = '\0';
                         /* numberbuffer = "416-" */
                         strncpy(numberbuffer + (p2 - p1), p2 + 2, CIDSIZE - (p2 - p1) - 1);
                         numberbuffer[CIDSIZE-1] = '\0';
                         /* numberbuffer = "416-555-1234" */
                         strcpy(Formattednmbr_c, numberbuffer);
                         /* Formattednmbr_c = "416-555-1234"  */
                         logMsg(LEVEL6, "nanp_format == 2 (all_hyphen) %s\n", Formattednmbr_c);
                     }
                     else {
                         strncpy(numberbuffer, p1 + 1, p2 - 1 - p1);
                         numberbuffer[p2 - 1 - p1] = '.';
                         numberbuffer[p2 - p1] = '\0';
                         /* numberbuffer = "416." */
                         strncpy(numberbuffer + (p2 - p1), p2 + 2, CIDSIZE - (p2 - p1) - 1);
                         numberbuffer[CIDSIZE-1] = '\0';
                         if ((p2 = strchr(numberbuffer, '-'))) *p2 = '.';
                         /* numberbuffer = "416.555.1234" */
                         strcpy(Formattednmbr_c, numberbuffer);
                         /* Formattednmbr_c = "416.555.1234"  */
                         logMsg(LEVEL6, "nanp_format == 3 (all_dot) %s\n", Formattednmbr_c);
                     }
                  }
               }
            }
            else {
               /* nanp_format can be 1 or greater than 2 */
               /* Formattednmbr_c = google formatted = (204) 555-1234 */
               std::string number_raw(cidnmbr_c);
               std::string ourregion(country);
               std::string formattedNumber;

               phone_util.Format(psnumber, PhoneNumberUtil::NATIONAL, &formattedNumber);
               strncpy(Formattednmbr_c, formattedNumber.c_str(), CIDSIZE);
               Formattednmbr_c[CIDSIZE-1] = '\0';
               logMsg(LEVEL6, "nanp_format == %d %s\n", nanp_format, Formattednmbr_c);
            }
         }
         else {
            /* Formattednmbr = google formatted = (204) 555-1234 */
            std::string number_raw(cidnmbr_c);
            std::string ourregion(country);
            std::string formattedNumber;

            phone_util.Format(psnumber, PhoneNumberUtil::NATIONAL, &formattedNumber);
            strncpy(Formattednmbr_c, formattedNumber.c_str(), CIDSIZE);
            Formattednmbr_c[CIDSIZE-1] = '\0';
         }
      }
      else {
         /* international caller */
         std::string number_raw(cidnmbr_c);
         std::string ourregion(country);
         std::string formattedNumber;

         //   std::cout << "NO_PARSING_ERROR" << std::endl;
         phone_util.Format(psnumber, PhoneNumberUtil::INTERNATIONAL, &formattedNumber);
         //std::cout << "value of formattedNumber=" << formattedNumber  << std::endl;
         strncpy(Formattednmbr_c, formattedNumber.c_str(), CIDSIZE);
         Formattednmbr_c[CIDSIZE-1] = '\0';
      }


      if (0 == strlen(Formattednmbr_c)) {
         /*formatted number came back empty, so we put unformatted number (whatever it is ) into it (better than nothing) */
         strncpy(Formattednmbr_c, cidnmbr_c, CIDSIZE - 1);
         Formattednmbr_c[CIDSIZE-1] = '\0';
      }

   }
   else {
      /* cidnmbr_c could not be parsed as a number */
      strcpy(Formattednmbr_c, cidnmbr_c);
   }
}


/*
 * changes the numbers from MSG, NOT and gateway CALL lines
 * to a number that can be dialled from our country
 * this is not called for modem NMBRs
 */
void normalizeToDialableNumberFromFixedLine(char * cidnmbr_c, int cidnmbr_size)
{
   char callerCountry_c[CIDSIZE];      /* country code of caller  ex : "FR" */

   std::string number_raw(cidnmbr_c);
   std::string ourregion(country);

   logMsg(LEVEL6, "normalizeToDialableNumberFromFixedLine %s\n", cidnmbr_c);

   using i18n::phonenumbers::PhoneNumberUtil;
   const PhoneNumberUtil& phone_util(*PhoneNumberUtil::GetInstance());

   if (phone_util.IsPossibleNumberForString(number_raw, ourregion)) {
      /* it might be a phone number */
      std::string country_code;
      i18n::phonenumbers::PhoneNumber psnumber;

      PhoneNumberUtil::ErrorType error_type;
      error_type = phone_util.Parse(number_raw, ourregion, &psnumber);
      if (error_type == PhoneNumberUtil::NO_PARSING_ERROR) {
         phone_util.GetRegionCodeForNumber(psnumber, &country_code);
         strcpy(callerCountry_c, country_code.c_str());

         if (0 == strcmp(country, callerCountry_c)) {
            /* caller country is ours */
            std::string formattedNumber;

            phone_util.Format(psnumber, PhoneNumberUtil::NATIONAL, &formattedNumber);
            phone_util.NormalizeDigitsOnly(&formattedNumber);

            strncpy(cidnmbr_c, formattedNumber.c_str(), cidnmbr_size);
            cidnmbr_c[cidnmbr_size-1] = '\0';
            logMsg(LEVEL6, "normalized national dialable nmbr=:%s\n", cidnmbr_c);
         }
         else {
            /* caller country is international */
            // Formats a phone number for out-of-country dialing purposes.
            //
            // Note this function takes care of the case for calling inside of NANPA
            // and between Russia and Kazakhstan (who share the same country calling
            // code). In those cases, no international prefix is used. For regions which
            // have multiple international prefixes, the number in its INTERNATIONAL
            // format will be returned instead.
            std::string formattedNumber;

            phone_util.FormatOutOfCountryCallingNumber(psnumber, ourregion, &formattedNumber);
            phone_util.NormalizeDigitsOnly(&formattedNumber);
            // Normalizes a string of characters representing a phone number. This
            // converts wide-ascii and arabic-indic numerals to European numerals, and
            // strips punctuation and alpha characters.

            strncpy(cidnmbr_c, formattedNumber.c_str(), cidnmbr_size);
            cidnmbr_c[cidnmbr_size-1] = '\0';
         }
      }
      else {
         /* can't parse as a number
          * don't do anything to cidnmbr_c  */
      }
   }
   else {
      /* doesn't look like a phone number
       * don't do anything to cidnmbr_c  */
   }

   logMsg(LEVEL6, "exit normalizeToDialableNumberFromFixedLine number=%s\n", cidnmbr_c);
}
