/*
 * nciddalias.c - This file is part of ncidd.
 *
 * Copyright (c) 2005-2023
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ncidd.h"

const char * cidalias = CIDALIAS;

struct alias *aliasHead = NULL;
struct alias *aliasCurrent = NULL;

int readAliasFile(void);
char * findAlias(const char * name, const char * nmbr, const char * line);
void builtinAlias(char * to, const char * from);
void nextAlias(void);
void userAlias(char *nmbr, char *name, char *line);
void getAlias(const char * fnptr, const char * inptr, int lc);
void setAlias(const char * fnptr, const char * inptr, int lc, char *wdptr, int type);
void rmaliases(void);

/*
 * Built-in Aliases for O, P and A
 * builtinAlias is a conditional strncpy of CIDSIZE
 */

void builtinAlias(char * to, const char * from)
{
   char * eptr;
    if (!strcmp(from, "O")) strcpy(to, WORDFOR_O);
    else if (!strcmp(from, "P")) strcpy(to, WORDFOR_P);
    else if (!strcmp(from, "A")) strcpy(to, WORDFOR_A);
    else {
       /* trim leading space */
       while (*from && isspace((int) *from)) ++from;
       strncpy(to, from, CIDSIZE);
       to[CIDSIZE-1] = '\0';
       /* trim trailing space */
       eptr = to + strlen(to);
       while (*eptr && isspace((int) *eptr)) *eptr-- = '\0';
    }
}

/*
 * User defined aliases.
 * Replace name, number, line with its alias
 * these are all CIDSIZE buffers or null
 */

void userAlias(char *nmbr, char *name, char *line)
{
    struct alias *node = aliasHead;
    struct alias *nextnode;
    void * regexp = NULL;

    int matched = 0;
    int namechanged = 0;
    /*debug int cnt = 0; */

    logMsg(LEVEL4, "Begin: userAlias() nmbr=%s name=%s line=%s\n", nmbr==0?"0":nmbr, name==0?"0":name, line==0?"0":line);

    while (node != NULL)
    {
        nextnode = node->next;
        matched = 0;
        if (regex == 2)
        {
            /* perl regular expression */
            regexp = (void*)node->info.pcreg;
        }
        else if (regex == 1)
        {
            /* expressions are regex */
            regexp = (void*)node->info.preg;
        }
        else /* (regex == 0) */
        {
            /* NCID Simple expressions */
            regexp = (void*)node->info.ncidreg;
        }
        switch (node->info.type)
        {
            case NMBRNAME:
                /* change both name and/or number if caller matches node->info.from */
                if (name!=0 &&
                    strmatch(name, regexp) &&
                    strcmp(name, node->info.to)!=0)
                {
                    /* copy if it has changed */
                    strcpy(name, node->info.to);
                    matched = 1;
                    namechanged = 1;
                }
                if (nmbr!=0 &&
                    strmatch(nmbr, regexp) &&
                    strcmp(nmbr, node->info.to)!=0)
                {
                    strcpy(nmbr, node->info.to);
                    matched = 1;
                }
                break;
            case NMBRONLY:
                /* change number if caller number matches node->info.from */
                if (nmbr!=0 &&
                    strmatch(nmbr, regexp) &&
                    strcmp(nmbr, node->info.to)!=0)
                {
                    strcpy(nmbr, node->info.to);
                    matched = 1;
                }
                break;
            case NMBRDEP:
                /* change any number if caller name matches node->info.depend */
                if (name!=0 && nmbr!=0 &&
                    strmatch(name, regexp) &&
                    strcmp(nmbr, node->info.to)!=0)
                {
                    strcpy(nmbr, node->info.to);
                    matched = 1;
                }
                break;
            case NAMEONLY:
                /* change name if caller name matches node->info.from */
                if (name!=0 &&
                    strmatch(name, regexp) &&
                    strcmp(name, node->info.to)!=0)
                {
                    strcpy(name, node->info.to);
                    matched = 1;
                    namechanged = 1;
                }
                break;
            case NAMEDEP:
                /* change any name if caller number matches node->info.depend */
                if (name!=0 && nmbr!=0 &&
                    strmatch(nmbr, regexp) &&
                    strcmp(name, node->info.to)!=0)
                {
                    strcpy(name, node->info.to);
                    matched = 1;
                    namechanged = 1;
                }
                break;
            case LINEONLY:
                /* change line if called line matches node->info.from */
                if (line!=0 &&
                    strmatch(line, regexp) &&
                    strcmp(line, node->info.to)!=0)
                {
                    strcpy(line, node->info.to);
                    matched = 1;
                }
                break;
        }
        /*debug logMsg(LEVEL4, "rule %-4.3d %s %s %s %s\n", cnt, node->info.type_txt, node->info.from, nmbr, name);
           cnt++; */
        if (matched) node->info.matched++;
        node = nextnode;
    }
    if (!namechanged && name)
    {
        /* remove * delimiters from the name */
        char * sf = starfree(name, CIDSIZE);
        strcpy(name, sf);
        free(sf);
    }

    logMsg(LEVEL4, "End: userAlias() nmbr=%s name=%s line=%s [%s]\n", nmbr==0?"0":nmbr, name==0?"0":name, line==0?"0":line, strdate(ONLYTIME));
}

/*
 * Process the alias file.
 * Returns: 0 - no errors
 *          # of errors
 */

int readAliasFile(void)
{
    char input[BUFSIZ], word[BUFSIZ], msgbuf[BUFSIZ], *ptr;
    const char * inptr;
    int lc, i;
    int max_temp = 0, max_type_txt = 0, max_from = 0, max_to = 0;
    FILE *fp;
    struct alias *node;
    struct alias *nextnode;

    if ((fp = fopen(cidalias, "r")) == NULL)
    {
        logMsg(LEVELE, "No alias file: %s\n", cidalias);
        return 0;
    }

    /* read each line of file, one line at a time */
    for (lc = 1; fgets(input, BUFSIZ, fp) != NULL; lc++)
    {
        /* line containing only <NL> or <CR><NL> or is a comment line*/
        if (input[0] == '\n' || input[0] == '\r' || input[0] == '#') continue;

        /* remove '\n' and '\r' if there */
        if ((ptr = strchr(input, (int) '\n'))) *ptr = '\0';
        if ((ptr = strchr(input, (int) '\r'))) *ptr = '\0';

        inptr = getWord(cidalias, input, word, lc);
        if (inptr == 0) continue;

        /* first word on line must be "alias" */
        if (!strcmp(word, "alias")) getAlias(cidalias, inptr, lc);
        else configError(cidalias, lc, word, ERRCMD);
    }
    (void) fclose(fp);
    logMsg(LEVEL1, "Processed alias file: %s\n", cidalias);

    logMsg(LEVEL1, "Alias Table:\n");

    /* log the number of entries at LEVEL1 and list them at LEVEL7 */
    node = aliasHead;
    if (!errorStatus && node)
    {
        for (i = 0; node != NULL; i++)
        {
            nextnode = node->next;
            node = nextnode;
        }
        logMsg(LEVEL1, "    Number of Entries: %d\n", i);

        /*
         * Make the alias display pretty by determining maximum column
         * widths and create human readable 'type'.
         */

        node = aliasHead;
        while(node != NULL)
        {
            nextnode = node->next;
            switch (node->info.type)
            {
                case NMBRNAME:
                    node->info.type_txt = NMBRNAME_TXT ;
                    break;

                case NMBRONLY:
                    node->info.type_txt = NMBRONLY_TXT ;
                    break;

                case NMBRDEP:
                    node->info.type_txt = NMBRDEP_TXT ;
                    break;

                case NAMEONLY:
                    node->info.type_txt = NAMEONLY_TXT ;
                    break;

                case NAMEDEP:
                    node->info.type_txt = NAMEDEP_TXT ;
                    break;

                case LINEONLY:
                    node->info.type_txt = LINEONLY_TXT ;
                    break;

                default:
                    node->info.type_txt = "UNKNOWN" ;
            }

            max_temp = strlen(node->info.type_txt);
            if (max_temp > max_type_txt) max_type_txt=max_temp;

            max_temp = strlen(node->info.from);
            if (max_temp > max_from) max_from=max_temp;

            max_temp = strlen(node->info.to);
            if (max_temp > max_to) max_to=max_temp;

            node = nextnode;
        }

        max_temp = strlen("FROM");
        if (max_temp > max_from) max_from=max_temp;

        max_temp = strlen("TO");
        if (max_temp > max_to) max_to=max_temp;

        logMsg(LEVEL7, "    %-4s %s%-*s %s%-*s %s%-*s %s\n",
               "SLOT",
               "TYPE", max_type_txt - (int) strlen("TYPE"), "",
               "FROM", max_from - (int) strlen("FROM"), "",
               "TO", max_to - (int) strlen("TO"), "",
               "DEPEND");
        logMsg(LEVEL7, "    %-4s %s%-*s %s%-*s %s%-*s %s\n",
               "----",
               "----", max_type_txt - (int) strlen("TYPE"), "",
               "----", max_from - (int) strlen("FROM"), "",
               "--", max_to - (int) strlen("TO"), "",
               "------");

        node = aliasHead;
        for (i = 0; node != NULL; i++)
        {
            nextnode = node->next;
            sprintf(msgbuf, "    %-4.3d %s%-*s %s%-*s %s%-*s ",
                    i,
                    node->info.type_txt,
                    max_type_txt - (int) strlen(node->info.type_txt), "",
                    node->info.from,
                    max_from - (int) strlen(node->info.from), "",
                    node->info.to,
                    max_to - (int) strlen(node->info.to), "");

            if (node->info.depend)
                strcat(strcat(strcat(msgbuf, "\""), node->info.depend), "\"");
            strcat(msgbuf, NL);

            logMsg(LEVEL7, msgbuf);
            node = nextnode;
        }
    }
    else logMsg(LEVEL1, "    Number of Entries: 0\n");

    return errorStatus;
}

/*
 * process alias lines:
 *        alias from = to
 *        alias NMBR from = to [if name_value]
 *        alias NAME from = to [if number_value]
 *        alias LINE from = to
 */

void getAlias(const char * fnptr, const char * inptr, int lc)
{
    char word[BUFSIZ];

    inptr = getWord(fnptr, inptr, word, lc);

    if (word[0] == '#') return; /* rest of line is comment */

    if (!strcmp(word, "NMBR")) setAlias(fnptr, inptr, lc, word, NMBRONLY);
    else if (!strcmp(word, "NAME")) setAlias(fnptr, inptr, lc, word, NAMEONLY);
    else if (!strcmp(word, "LINE")) setAlias(fnptr, inptr, lc, word, LINEONLY);
    else setAlias(fnptr, inptr, lc, word, NMBRNAME);
}

void setAlias(const char * fnptr, const char * inptr, int lc, char *wdptr, int type)
{
    int err = 0;
    char errbuf[512];

    nextAlias();
    aliasCurrent->info.matched = 0;

    if (type == NMBRNAME || (inptr = getWord(fnptr, inptr, wdptr, lc)))
    {
        if (strlen(wdptr) > CIDSIZE - 1) {
            configError(cidalias, lc, wdptr, ERRLONG);
            return;
        }

        /* store the FROM string and the compiled expression */
        aliasCurrent->info.from = strdup(wdptr);
        if (regex == 1)
        {
            aliasCurrent->info.preg = (regex_t*)malloc(sizeof(regex_t));
            err = regcomp(aliasCurrent->info.preg, wdptr, REG_EXTENDED|REG_NOSUB);
            if (err)
            {
                regerror(err, aliasCurrent->info.preg, errbuf, sizeof(errbuf));
                configError(cidalias, lc, wdptr, errbuf);
                return;
            }
        }
        else if (regex == 2)
        {
            aliasCurrent->info.pcreg = pcre2_compile((PCRE2_SPTR8)wdptr, PCRE2_ZERO_TERMINATED, 0, &pcreErrorNumber, &pcreErrorOffset, NULL);

            if (aliasCurrent->info.pcreg == NULL)
            {
                PCRE2_UCHAR pcreErrorMessage[256];
                pcre2_get_error_message(pcreErrorNumber, pcreErrorMessage, sizeof(pcreErrorMessage));
                sprintf(errbuf, "pcre2_compile failed (offset: %d), %s\n", (int)pcreErrorOffset, pcreErrorMessage);
                configError(cidalias, lc, wdptr, errbuf);
                return;
            }
            /* alloc space for the unused result */
            if (pcreMatchData == NULL) pcreMatchData = pcre2_match_data_create(4, NULL);
        }
        else /* (regex == 0) */
        {
            aliasCurrent->info.ncidreg = store_astrmatch(wdptr);
        }

        /* look for the TO string  = ["]TO["] */
        if ((inptr = getWord(fnptr, inptr, wdptr, lc)))
        {
            if (*wdptr == '=')
            {
                if ((inptr = getWord(fnptr, inptr, wdptr, lc)))
                {
                    if (strlen(wdptr) > CIDSIZE - 1) {
                        configError(cidalias, lc, wdptr, ERRLONG);
                        return;
                    }
                    /* store the TO string with * delimiters removed */
                    aliasCurrent->info.to = starfree(wdptr, CIDSIZE);
                    if (type == NMBRNAME)
                    {
                        aliasCurrent->info.depend = 0;
                        aliasCurrent->info.type = type;
                    }
                    else
                    {
                        if ((inptr = getWord(fnptr, inptr, wdptr, lc)))
                        {
                            if (wdptr[0] == '#')
                            {
                                /* found a trailing comment */
                                aliasCurrent->info.depend = 0;
                                aliasCurrent->info.type = type;
                            }
                            else if (strcmp(wdptr, "if"))
                            {
                                /* should be "if" */
                                configError(cidalias, lc, wdptr, ERRIF);
                            }
                            else if ((inptr = getWord(fnptr, inptr, wdptr, lc)))
                            {
                                if (strlen(wdptr) > CIDSIZE - 1) {
                                    configError(cidalias, lc, wdptr, ERRLONG);
                                    return;
                                }
                                aliasCurrent->info.depend = strdup(wdptr);
                                if (regex == 1)
                                {
                                    if (aliasCurrent->info.preg) { regfree(aliasCurrent->info.preg); free(aliasCurrent->info.preg); }
                                    aliasCurrent->info.preg = (regex_t*)malloc(sizeof(regex_t));
                                    err = regcomp(aliasCurrent->info.preg,
                                                  wdptr, REG_EXTENDED|REG_NOSUB);
                                    if (err)
                                    {
                                        regerror(err, aliasCurrent->info.preg, errbuf, sizeof(errbuf));
                                        configError(cidalias, lc, wdptr, errbuf);
                                    }
                                }
                                else if (regex == 2)
                                {
                                    if (aliasCurrent->info.pcreg) pcre2_code_free(aliasCurrent->info.pcreg);
                                    aliasCurrent->info.pcreg = pcre2_compile((PCRE2_SPTR8)wdptr, PCRE2_ZERO_TERMINATED, 0, &pcreErrorNumber, &pcreErrorOffset, NULL);

                                    if (aliasCurrent->info.pcreg == NULL)
                                    {
                                        PCRE2_UCHAR pcreErrorMessage[256];
                                        pcre2_get_error_message(pcreErrorNumber, pcreErrorMessage, sizeof(pcreErrorMessage));
                                        sprintf(errbuf, "pcre2_compile failed (offset: %d), %s\n", (int)pcreErrorOffset, pcreErrorMessage);
                                        configError(cidalias, lc, wdptr, errbuf);
                                    }
                                    /* alloc space for the unused result */
                                    if (pcreMatchData == NULL) pcreMatchData = pcre2_match_data_create(4, NULL);
                                }
                                else /* (regex == 0) */
                                {
                                    if (aliasCurrent->info.ncidreg) free_abwstr(aliasCurrent->info.ncidreg);
                                    aliasCurrent->info.ncidreg = store_astrmatch(wdptr);
                                }
                                aliasCurrent->info.type = type + 1;
                            }
                            else configError(cidalias, lc, wdptr, ERRARG);
                        }
                        else
                        {
                            /* alias has no if condition */
                            aliasCurrent->info.depend = 0;
                            aliasCurrent->info.type = type;
                        }
                    }
                }
                else configError(cidalias, lc, wdptr, ERRARG);
            }
            else configError(cidalias, lc, wdptr, ERREQB);
        }
        else configError(cidalias, lc, wdptr, ERREQA);
    }
    else configError(cidalias, lc, wdptr, ERRARG);
}

void nextAlias(void)
{
    struct alias *node;
    if (!(node = (struct alias *) malloc(sizeof(struct alias))))
        errorExit(-1, progname, 0);
    memset(node, 0x00, sizeof(struct alias));
    node->next = NULL;
    if (aliasHead == NULL) aliasHead = node;
    else aliasCurrent->next = node;
    aliasCurrent = node;
}

/*
 * Frees all alias nodes
 */
void rmaliases(void)
{
    struct alias *node = aliasHead;
    struct alias *nextnode;

    while (node != NULL)
    {
        nextnode = node->next;
        if (regex == 1) { regfree(node->info.preg); free(node->info.preg); }
        else if (regex == 2) pcre2_code_free(node->info.pcreg);
        else free_abwstr(node->info.ncidreg);
        if (node->info.from) free(node->info.from);
        if (node->info.to) free(node->info.to);
        if (node->info.depend) free(node->info.depend);
        free(node);
        node = nextnode;
    }

    aliasHead = NULL;
    aliasCurrent = NULL;
}

/*
 * Find an alias and return the type of alias.
 * If multiple aliases, return the last one.
 */
char * findAlias(const char * name, const char * nmbr, const char * line)
{
    char calltype[CIDSIZE +10], linetype[CIDSIZE + 10];
    static char findalias_ret[(2*sizeof(NMBRNAME_TXT)+8) + (2 * CIDSIZE)];

    struct alias *node = aliasHead;
    struct alias *nextnode;
    void * regexp = NULL;

    /* set defaults */
    strcat(strcpy(calltype, NOALIAS_TXT), " \"\"");
    strcat(strcpy(linetype, NOALIAS_TXT), " \"\"");

    while (node != NULL)
    {
        nextnode = node->next;

        if (regex == 2)
        {
            /* perl regular expression */
            regexp = (void*)node->info.pcreg;
        }
        else if (regex == 1)
        {
            /* regex regular expression */
            regexp = (void*)node->info.preg;
        }
        else
        {
            /* ncid simple expression */
            regexp = (void*)node->info.ncidreg;
        }

        if (node->info.depend)
        {
            /* NAMEDEP or NMBRDEP for node->info.depend */
            if (strmatch(nmbr, regexp) ||
                strmatch(name, regexp))
                (void) snprintf(calltype, CIDSIZE + 10, "%s \"%s\"",
                         node->info.type_txt, node->info.depend);
        }
        else
        {
            /*
             * When modifying these aliases, only the alias is known.
             * Searches are for both name/nmbr/line and alias
             */

            /* NAMEONLY, NMBRONLY, or NMBRNAME for node->info.from */
            if (strmatch(nmbr, regexp) ||
                strmatch(name, regexp) ||
                strcmp(nmbr, node->info.to)==0 ||
                strcmp(name, node->info.to)==0 )
                (void) snprintf(calltype, CIDSIZE + 10, "%s \"%s\"",
                         node->info.type_txt, node->info.from);

            /* LINEONLY for node->info.from */
            else if (strmatch(line, regexp) ||
                     strcmp(line, node->info.to)==0 )
                (void) snprintf(linetype, CIDSIZE + 10, "%s \"%s\"",
                         node->info.type_txt, node->info.from);
        }
        node = nextnode;
    }

    if (*line)
    {
        /* there is a line name so new format */
        sprintf(findalias_ret, "%s %s ", calltype, linetype);
    }
    else
    {
        /* there is no line name so old format */
        sprintf(findalias_ret, "%s", calltype);
    }

    return findalias_ret;
}

/*
 * copy a string and remove unwanted '*' characters
 * returns a pointer to a strdup'd string, clipped to clipsiz
 * that will need to be free()'d later
 *
 * if the expanded last star won't fit into clipsiz, it will be skipped
 */
char * starfree(const char * src, size_t clipsiz)
{
    const char * utf8star = "\xe2\x98\x85"; // small five pointed star \u2605
    char * starpos;
    char starsout[BUFSIZ];

    /* clipsiz is usually CIDSIZE, but */
    if (clipsiz > BUFSIZ) clipsiz = BUFSIZ;

    /* is there a * in the cid entry? */
    starpos = strchr(src, '*');
    if (starpos == NULL || (size_t)(starpos-src)>clipsiz ) {
        /* no, just copy the string */
        strncpy(starsout, src, clipsiz-1);
        starsout[clipsiz-1] = '\0';
    }
    else {
        /* copy the string while changing the stars */
        const char * from = src;
        char * pdest = starsout;
        size_t size = 0;
        while (*from != '\0' && size < clipsiz-1) {
            /* is it a star? */
            if (*from == '*') {
                if (size+3 < clipsiz-1) {
                    /* substitute a utf8 character */
                    *pdest = utf8star[0];
                    *(pdest+1) = utf8star[1];
                    *(pdest+2) = utf8star[2];
                    size += 3;
                    pdest += 3;
                }
            }
            else {
                /* normal text */
                *pdest = *from;
                if (size < clipsiz-1) {
                    size++;
                    pdest++;
                }
            }
            from++;
        }
        *pdest = '\0';


    }
    return strdup(starsout);
}


/*
 * show the number of matches for each alias
 */
void showAliasScores(void)
{

    char msgbuf[BUFSIZ];
    struct alias *node;
    int nc;
    int matched = 0;

    logMsg(LEVEL1, "Alias Scores:\n");

    node = aliasHead;
    if (!errorStatus && node)
    {
        matched = 0;
        while(node != NULL)
        {
            matched += node->info.matched;
            node = node->next;
        }
        if (matched == 0)
            logMsg(LEVELR, "There were no changes from the alias file\n\n");
        else
            logMsg(LEVELR, "There were %d changes from the alias file\n\n", matched);

        /*
         * Make the alias display pretty by determining maximum column
         * widths and create human readable 'type'.
         */

        if (matched > 0) {
            node = aliasHead;
            while(node != NULL)
            {
                switch (node->info.type)
                {
                    case NMBRNAME:
                        node->info.type_txt = NMBRNAME_TXT ;
                        break;

                    case NMBRONLY:
                        node->info.type_txt = NMBRONLY_TXT ;
                        break;

                    case NMBRDEP:
                        node->info.type_txt = NMBRDEP_TXT ;
                        break;

                    case NAMEONLY:
                        node->info.type_txt = NAMEONLY_TXT ;
                        break;

                    case NAMEDEP:
                        node->info.type_txt = NAMEDEP_TXT ;
                        break;

                    case LINEONLY:
                        node->info.type_txt = LINEONLY_TXT ;
                        break;

                    default:
                        node->info.type_txt = "UNKNOWN" ;
                }

                if (node->info.matched > 0) {
                    nc = sprintf(msgbuf, "    (%s) Changed \"%s\" to \"%s\"",
                                 node->info.type_txt,
                                 node->info.from,
                                 node->info.to);
                    if (node->info.depend)
                        nc += sprintf(msgbuf+nc, " for \"%s\"", node->info.depend);
                    sprintf(msgbuf+nc, " %d time%c\n", node->info.matched, node->info.matched==1?' ':'s' );

                    logMsg(LEVELR, msgbuf);
                }
                node = node->next;
            }
        }
    }
}
