/*
 * nciddexpandMESG.c - This file is part of ncidd.
 *
 * Copyright (c) 2005-2022
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "ncidd.h"
#include "nciddPN.h"


#define STD_UNKNOWN   0
#define STD_BELLCORE  1
#define STD_ETSI      2
#define STD_NTT       3

unsigned int standard = STD_UNKNOWN;

/* MESG fields BELLCORE */
#define BELL_CALLING_NUMBER               0x02
#define BELL_CALLED_NUMBER_DDN            0x03

#define BELL_CALL_QUALIFIER_LONG_DISTANCE 0x06014C

#define BELL_REDIR_UNCONDIDITIONAL        0x050100
#define BELL_REDIR_BUSY                   0x050101
#define BELL_REDIR_UNANSWERED             0x050102

#define BELL_VISUAL_MWI_INDICATOR_ON      0x0B01FF
#define BELL_VISUAL_MWI_INDICATOR_OFF     0x0B0100
/* END of MESG fields BELLCORE */

/* MESG fields ETSI */
#define ETSI_CALLING_NUMBER               0x02
#define ETSI_CALLED_NUMBER_DDN            0x03
#define ETSI_FIRST_REDIRECTING_NUMBER          0x12
#define ETSI_LAST_REDIRECTING_NUMBER           0x1A

#define ETSI_FORWARD_UNKNOWN               0x150100
#define ETSI_FORWARD_BUSY                  0x150101
#define ETSI_FORWARD_NO_REPLY              0x150102
#define ETSI_FORWARD_UNCONDITIONAL         0x150103
#define ETSI_FORWARD_DEFLECTED_AFTER_ALERT 0x150104
#define ETSI_FORWARD_DEFLECTED_IMEDIATE    0x150105
#define ETSI_FORWARD_UNREACHABLE           0x150106

#define ETSI_USER_UNKNOWN           0x160100
#define ETSI_USER_VOICE             0x160101
#define ETSI_USER_TEXT              0x160102
#define ETSI_USER_VPN               0x160103
#define ETSI_USER_MOBILE            0x160104
#define ETSI_USER_MOBILE_VPN        0x160105
#define ETSI_USER_FAX               0x160106
#define ETSI_USER_VIDEO             0x160107
#define ETSI_USER_EMAIL             0x160108
#define ETSI_USER_OPERATOR          0x160109
#define ETSI_USER_ORDINARY          0x16010A
#define ETSI_USER_PRIORITY          0x16010B
#define ETSI_USER_DATA              0x16010C
#define ETSI_USER_TEST              0x16010D
#define ETSI_USER_TELEMETRIC        0x16010E
#define ETSI_USER_PAY_PHONE         0x16010F

#define ETSI_TYPE_VOICE               0x110101
#define ETSI_TYPE_RINGBACK_WHEN_FREE  0x110102
#define ETSI_TYPE_CALL_NAME_DELIVERY  0x110103
#define ETSI_TYPE_CALL_RETURN         0x110104
#define ETSI_TYPE_ALARM               0x110105
#define ETSI_TYPE_DOWNLOAD_FUNCTION   0x110106
#define ETSI_TYPE_REVERSE_CHARGE      0x110107
#define ETSI_TYPE_EXTERNAL            0x110110
#define ETSI_TYPE_INTERNAL            0x110111
#define ETSI_TYPE_MONITORING          0x110150
#define ETSI_TYPE_MESSAGE_WAITING     0x110181


#define ETSI_VISUAL_MWI_INDICATOR_ON     0x0B01FF
#define ETSI_VISUAL_MWI_INDICATOR_OFF    0x0B0100

/* voice mail queue management */
#define ETSI_NB_MSG_WAITING              0x13

#define ETSI_MSG_IDENTIFIER              0x0D
#define ETSI_MSG_IDENTIFIER_REMOVED      0x0D00
#define ETSI_MSG_IDENTIFIER_ ADDED       0x0DFF

#define ETSI_NETWORK_INFO                0xE0
/* this contains the number of the last caller that left a voicemail  message . */
/* END of MESG fields ETSI */


/* MESG fields NTT */
#define NTT_CALLING_NUMBER                       0x02
#define NTT_CALLED_NUMBER_DDN                    0x09

#define NTT_CALLING_NUMBER_UNKNOWN_UNKNOWN       0x210100
#define NTT_CALLING_NUMBER_UNKNOWN_ISDN          0x210101
#define NTT_CALLING_NUMBER_UNKNOWN_DATA          0x210103
#define NTT_CALLING_NUMBER_UNKNOWN_TELEX         0x210104
#define NTT_CALLING_NUMBER_UNKNOWN_DOMESTIC      0x210108
#define NTT_CALLING_NUMBER_UNKNOWN_PRIVATE       0x210109

#define NTT_CALLING_NUMBER_INTERNATIONAL_UNKNOWN       0x210110
#define NTT_CALLING_NUMBER_INTERNATIONAL_ISDN          0x210111
#define NTT_CALLING_NUMBER_INTERNATIONAL_DATA          0x210113
#define NTT_CALLING_NUMBER_INTERNATIONAL_TELEX         0x210114
#define NTT_CALLING_NUMBER_INTERNATIONAL_DOMESTIC      0x210118
#define NTT_CALLING_NUMBER_INTERNATIONAL_PRIVATE       0x210119

#define NTT_CALLING_NUMBER_DOMESTIC_UNKNOWN       0x210120
#define NTT_CALLING_NUMBER_DOMESTIC_ISDN          0x210121
#define NTT_CALLING_NUMBER_DOMESTIC_DATA          0x210123
#define NTT_CALLING_NUMBER_DOMESTIC_TELEX         0x210124
#define NTT_CALLING_NUMBER_DOMESTIC_DOMESTIC      0x210128
#define NTT_CALLING_NUMBER_DOMESTIC_PRIVATE       0x210129

#define NTT_CALLING_NUMBER_NETWORK_UNKNOWN       0x210130
#define NTT_CALLING_NUMBER_NETWORK_ISDN          0x210131
#define NTT_CALLING_NUMBER_NETWORK_DATA          0x210133
#define NTT_CALLING_NUMBER_NETWORK_TELEX         0x210134
#define NTT_CALLING_NUMBER_NETWORK_DOMESTIC      0x210138
#define NTT_CALLING_NUMBER_NETWORK_PRIVATE       0x210139

#define NTT_CALLING_NUMBER_LOCAL_UNKNOWN       0x210140
#define NTT_CALLING_NUMBER_LOCAL_ISDN          0x210141
#define NTT_CALLING_NUMBER_LOCAL_DATA          0x210143
#define NTT_CALLING_NUMBER_LOCAL_TELEX         0x210144
#define NTT_CALLING_NUMBER_LOCAL_DOMESTIC      0x210148
#define NTT_CALLING_NUMBER_LOCAL_PRIVATE       0x210149

#define NTT_CALLING_NUMBER_ABBREVIATED_UNKNOWN       0x210160
#define NTT_CALLING_NUMBER_ABBREVIATED_ISDN          0x210161
#define NTT_CALLING_NUMBER_ABBREVIATED_DATA          0x210163
#define NTT_CALLING_NUMBER_ABBREVIATED_TELEX         0x210164
#define NTT_CALLING_NUMBER_ABBREVIATED_DOMESTIC      0x210168
#define NTT_CALLING_NUMBER_ABBREVIATED_PRIVATE       0x210169


#define NTT_REDIR_UNDEF                  0x0B0100
#define NTT_REDIR_BUSY                   0x0B0101
#define NTT_REDIR_UNANSWERED             0x0B0102
#define NTT_REDIR_UNCONDITIONAL          0x0B0115

#define NTT_REDIRECTING_NUMBER           0x0B

#define NTT_PUBLIC_MESSAGE               0x24
/* END of MESG fields NTT */


void set_mesg_standard(const char * country);
void expand_mesg(const char * hexcode_mesg, struct cid_t * modem_cid);

/*
 * set standard depending on country
 * called from ncidd.c after the country has bee validated
 */
void set_mesg_standard(const char * country)
{


   if (0 == strcmp(country, "US") || 0 == strcmp(country, "CA") || \
       0 == strcmp(country, "AU") || 0 == strcmp(country, "BR") || \
       0 == strcmp(country, "NZ") || 0 == strcmp(country, "CN") || \
       0 == strcmp(country, "HK")) {
      standard = STD_BELLCORE;
      logMsg(LEVEL3, "Using BELLCORE STD to decode hexcode MESG  (standard=%d) \n", standard);
   }


   if (0 == strcmp(country, "UK") || 0 == strcmp(country, "FR") || \
       0 == strcmp(country, "SP") || 0 == strcmp(country, "NO") || \
       0 == strcmp(country, "IR") || 0 == strcmp(country, "TW") || \
       0 == strcmp(country, "DE") || 0 == strcmp(country, "AR") || \
       0 == strcmp(country, "CL") || 0 == strcmp(country, "EC") || \
       0 == strcmp(country, "IT") || 0 == strcmp(country, "BE") || \
       0 == strcmp(country, "NL")) {
      standard = STD_ETSI;
      logMsg(LEVEL3, "using ETSI STD to decode hexcode MESG  (standard=%d) \n", standard);
   }

   if (0 == strcmp(country, "JP")) {
      logMsg(LEVEL3, "Using NTT STD to decode hexcode MESG  (standard=%d) \n", standard);
      standard = STD_NTT;
   }
}
/*
 *
 */
void expand_mesg(const char * hexcode_mesg, struct cid_t * modem_cid)
{
   unsigned int param_of_length_one           = 0;
   unsigned int param_and_value_of_length_one = 0;
   unsigned int param_of_variable_length      = 0;
   int mesg_count                             = 0;
   unsigned int msg_id                        = 0;
   char string_param_value[CIDSIZE];
   char temp_buffer[CIDSIZE];
   char formatted_number[CIDSIZE];
   unsigned char unformatted_number[CIDSIZE];
   const char * trans_buffer                  = 0;
   char formatted_result[CIDSIZE*2];
   const char * translate_buffer              = 0;
   char string_param[4];
   char * endptr;

   /*in case we do not find anything, we return the hex code as received  */
   translate_buffer = hexcode_mesg;

   /* hex2uchar does not set end of string */
   bzero(unformatted_number, sizeof(unformatted_number));

   /* there are a lot of parameters of length 1  and not so many of length > 1 (mainly
    * contain numbers or names ), So we separate the two cases .
    */
   logMsg(LEVEL6, "we got this (unprocessed  by modem) code MESG=%s\n", translate_buffer);

   if (strlen(hexcode_mesg) == 6) {
      strncpy(string_param, hexcode_mesg, 2);
      string_param[2] = '\0';
      param_of_length_one = strtol(string_param, &endptr, 16);
      if (endptr==string_param || *endptr!='\0') {
         logMsg(LEVELE, "bad MESG parameter %s\n", hexcode_mesg);
         param_of_length_one = param_and_value_of_length_one = 999;
      }
      else {
         param_and_value_of_length_one = strtol(hexcode_mesg, &endptr, 16);
         if (endptr==hexcode_mesg || *endptr!='\0') {
            logMsg(LEVELE, "bad MESG value %s\n", hexcode_mesg);
            param_of_length_one = param_and_value_of_length_one = 999;
         }
      }
      strncpy(string_param_value, hexcode_mesg + 4, CIDSIZE-1);
   }
   else {
      strncpy(string_param, hexcode_mesg, 2);
      string_param[2] = '\0';
      param_of_variable_length = strtol(string_param, &endptr, 16);
      if (endptr==string_param || *endptr!='\0') {
         logMsg(LEVELE, "bad MESG parameter %s\n", hexcode_mesg);
         param_of_variable_length = 999;
      }
      strncpy(string_param_value, hexcode_mesg + 4, CIDSIZE-1);
   }


   switch (standard) {
      case STD_BELLCORE:
         logMsg(LEVEL6, "in BELLCORE STD case \n");
         switch (param_and_value_of_length_one) {
            case BELL_CALL_QUALIFIER_LONG_DISTANCE:
               translate_buffer = "Long distance";
               break;

            case BELL_REDIR_UNCONDIDITIONAL:
               translate_buffer = " Unconditional forward ";
               break;

            case BELL_REDIR_BUSY:
               translate_buffer = " Forward on busy";
               break;

            case BELL_REDIR_UNANSWERED:
               translate_buffer = "Forward on no reply";
               break;

            case BELL_VISUAL_MWI_INDICATOR_ON:
               logMsg(LEVEL6, "MSG_WAITING = YES (MWI)\n");
               strcpy(modem_cid->cidname, " Message(s) Waiting");
               strcpy(modem_cid->cidnmbr, "Voicemail");
               modem_cid->status         |= CIDNAME | CIDNMBR;
               modem_cid->calltype = MWI;
               modem_cid->cidsent  = 0;
               break;

            case BELL_VISUAL_MWI_INDICATOR_OFF:
               logMsg(LEVEL6, "MSG_WAITING = NO (MWI)\n");
               strcpy(modem_cid->cidname, "No Messages Waiting");
               strcpy(modem_cid->cidnmbr, "Voicemail");
               modem_cid->status         |= CIDNAME | CIDNMBR;
               modem_cid->calltype = MWI;
               modem_cid->cidsent  = 0;
               break;
         }


         switch (param_of_variable_length) {
            case BELL_CALLING_NUMBER:
               hex2uchar(string_param_value, unformatted_number);
               format_number(formatted_number, (char*)unformatted_number);
               snprintf(formatted_result, sizeof(formatted_result), "From:%s", formatted_number);
               translate_buffer = formatted_result;
               break;

            case BELL_CALLED_NUMBER_DDN:
               hex2uchar(string_param_value, unformatted_number);
               format_number(formatted_number, (char*)unformatted_number);
               snprintf(formatted_result, sizeof(formatted_result), "From:%s", formatted_number);
               translate_buffer = formatted_result;
               break;
         }
         break;

      case STD_ETSI:
         logMsg(LEVEL6, "in ETSI STD case \n");
         switch (param_and_value_of_length_one) {
            case ETSI_FORWARD_UNKNOWN:
               translate_buffer = "Forward ";
               break;

            case ETSI_FORWARD_BUSY:
               translate_buffer = "Forward on busy";
               break;

            case ETSI_FORWARD_NO_REPLY:
               translate_buffer = "Forward on no reply";
               break;

            case ETSI_FORWARD_UNCONDITIONAL:
               translate_buffer = "Unconditional forward ";
               break;

            case ETSI_FORWARD_DEFLECTED_AFTER_ALERT:
               translate_buffer = "Deflected after alert";
               break;

            case ETSI_FORWARD_DEFLECTED_IMEDIATE:
               translate_buffer = "Deflected immediate";
               break;

            case ETSI_FORWARD_UNREACHABLE:
               translate_buffer = "Forward on unreachable";
               break;

            case ETSI_USER_UNKNOWN:
               translate_buffer = "";
               break;

            case ETSI_USER_VOICE:
               translate_buffer = "Voice user";
               break;

            case ETSI_USER_TEXT:
               translate_buffer = "Text";
               break;

            case ETSI_USER_VPN:
               translate_buffer = "VPN";
               break;

            case ETSI_USER_MOBILE:
               translate_buffer = "Mobile";
               break;

            case ETSI_USER_MOBILE_VPN:
               translate_buffer = "Mobile VPN";
               break;

            case ETSI_USER_FAX:
               translate_buffer = "Fax";
               break;

            case ETSI_USER_VIDEO:
               translate_buffer = "Video";
               break;

            case ETSI_USER_EMAIL:
               translate_buffer = "Email";
               break;

            case ETSI_USER_OPERATOR:
               translate_buffer = "Operator";
               break;

            case ETSI_USER_ORDINARY:
               translate_buffer = "";
               break;

            case ETSI_USER_PRIORITY:
               translate_buffer = "Priority user";
               break;

            case ETSI_USER_DATA:
               translate_buffer = "Data";
               break;

            case ETSI_USER_TEST:
               translate_buffer = "Test";
               break;

            case ETSI_USER_TELEMETRIC:
               translate_buffer = "Telemetric";
               break;

            case ETSI_USER_PAY_PHONE:
               translate_buffer = "PayPhone";
               break;

            case ETSI_TYPE_VOICE:
               translate_buffer = "Voice";
               break;

            case ETSI_TYPE_RINGBACK_WHEN_FREE:
               translate_buffer = "Ring back when free";
               modem_cid->calltype = RID;
               break;

            case ETSI_TYPE_CALL_NAME_DELIVERY:
               translate_buffer = "Call name delivery";
               break;

            case ETSI_TYPE_CALL_RETURN:
               translate_buffer = "Call return";
               break;

            case ETSI_TYPE_ALARM:
               translate_buffer = "Alarm";
               break;

            case ETSI_TYPE_DOWNLOAD_FUNCTION:
               translate_buffer = "Download function";
               break;

            case ETSI_TYPE_REVERSE_CHARGE:
               translate_buffer = "Charge back";
               break;

            case ETSI_TYPE_EXTERNAL:
               translate_buffer = "External";
               break;

            case ETSI_TYPE_INTERNAL:
               translate_buffer = "Internal";
               break;

            case ETSI_TYPE_MONITORING:
               translate_buffer = "Monitoring";
               break;

            case ETSI_TYPE_MESSAGE_WAITING:
               translate_buffer = "Mesg waiting";
               break;

            case ETSI_VISUAL_MWI_INDICATOR_ON:
               logMsg(LEVEL6, "MSG_WAITING = YES (MWI)\n");
               strcpy(modem_cid->cidname, " Message(s) Waiting");
               strcpy(modem_cid->cidnmbr, "Voicemail");
               modem_cid->status         |= CIDNAME | CIDNMBR;
               modem_cid->calltype = MWI;
               modem_cid->cidsent  = 0;
               return;

            case ETSI_VISUAL_MWI_INDICATOR_OFF:
               logMsg(LEVEL6, "MSG_WAITING = NO (MWI)\n");
               strcpy(modem_cid->cidname, "No Messages Waiting");
               strcpy(modem_cid->cidnmbr, "Voicemail");
               modem_cid->status         |= CIDNAME | CIDNMBR;
               modem_cid->calltype = MWI;
               modem_cid->cidsent  = 0;
               return;

            default:
               logMsg(LEVEL6, "nothing found for parameter of length one: %02x \n", param_of_length_one);
               break;
         }

         /* special case for number of messages waiting */
         if (param_of_length_one == ETSI_NB_MSG_WAITING) {
            logMsg(LEVEL6, "(matches with ETSI_NB_MSG_WAITING) parameter of length one: %02x \n", param_of_length_one);
            mesg_count = strtol(string_param_value, &endptr, 16);
            if (endptr==string_param_value || *endptr!='\0')
               logMsg(LEVELE, "bad MESG value %s\n", hexcode_mesg);
            snprintf(temp_buffer, sizeof(temp_buffer), " (%d)", mesg_count);
            strncat(modem_cid->cidname, temp_buffer, CIDSIZE-1 - strlen(modem_cid->cidname));
         }

         logMsg(LEVEL6, "Just before switch number param_of_variable_length=%02x\n", param_of_variable_length);

         switch (param_of_variable_length) {
            case ETSI_CALLING_NUMBER:
               hex2uchar(string_param_value, unformatted_number);
               format_number(formatted_number, (char*)unformatted_number);
               snprintf(formatted_result, sizeof(formatted_result), "From:%s", formatted_number);
               translate_buffer = formatted_result;
               break;

            case ETSI_FIRST_REDIRECTING_NUMBER:
               hex2uchar(string_param_value, unformatted_number);
               logMsg(LEVEL6, "Unformatted redirected number\n", unformatted_number);
               format_number(formatted_number, (char*)unformatted_number);
               logMsg(LEVEL6, "formatted redirected number\n", formatted_number);
               snprintf(formatted_result, sizeof(formatted_result), "From (first):%s", formatted_number);
               translate_buffer = formatted_result;
               break;

            case ETSI_LAST_REDIRECTING_NUMBER:
               logMsg(LEVEL6, "redirected number (hexa) \n", string_param_value);
               hex2uchar(string_param_value, unformatted_number);
               logMsg(LEVEL6, "Unformatted redirected number\n", unformatted_number);
               format_number(formatted_number, (char*)unformatted_number);
               logMsg(LEVEL6, "formatted redirected number\n", formatted_number);
               snprintf(formatted_result, sizeof(formatted_result), "From (last):%s", formatted_number);
               translate_buffer = formatted_result;
               break;

            case ETSI_CALLED_NUMBER_DDN:
               hex2uchar(string_param_value, unformatted_number);
               format_number(formatted_number, (char*)unformatted_number);
               snprintf(formatted_result, sizeof(formatted_result), "DDN:%s", formatted_number);
               translate_buffer = formatted_result;
               break;

            case ETSI_NETWORK_INFO:
               hex2uchar(string_param_value, unformatted_number);
               /*Number of last user that left a voicemail message*/
               format_number(formatted_number, (char*)unformatted_number);
               snprintf(formatted_result, sizeof(formatted_result), "Last voicemail NB:%s", (char*)formatted_number);
               translate_buffer = formatted_result;
               break;

            case ETSI_MSG_IDENTIFIER:
               /*ETSI_MSG_IDENTIFIER=0D    0D0300xxxx (added)   or  0D03FFxxxx (removed)  */
               trans_buffer = "MSg identifier:";
               logMsg(LEVEL6, "MSg identifier: parameter value  (hexa) \n", string_param_value);
               logMsg(LEVEL6, " string_param_value[3] (hexa) %c \n", string_param_value[3]);
               logMsg(LEVEL6, " string_param_value[4] (hexa) %c \n", string_param_value[4]);
               if ((string_param_value[3] == '0') && (string_param_value[3] == '0')) {
                  trans_buffer = "added VoiceMail Msg : ";
               }
               if ((string_param_value[3] == 'F') && (string_param_value[3] == 'F')) {
                  trans_buffer = "removed VoiceMail Msg : ";
               }
               msg_id = strtol(string_param_value + 2, &endptr, 16);
               if (endptr==(string_param_value + 2) || *endptr!='\0') {
                  logMsg(LEVELE, "bad MESG value %s\n", hexcode_mesg);
                  break;
               }
               snprintf(temp_buffer, sizeof(temp_buffer), " (%d)", msg_id);
               snprintf(formatted_result, sizeof(formatted_result), "%s%s", trans_buffer, (char*)temp_buffer);
               translate_buffer = formatted_result;
               break;
         }
         break;





      case STD_NTT: {
         logMsg(LEVEL6, "in NTT STD case \n");
         switch (param_and_value_of_length_one) {
            case NTT_CALLING_NUMBER_UNKNOWN_UNKNOWN:
               translate_buffer = "";
               break;

            case NTT_CALLING_NUMBER_UNKNOWN_ISDN:
               translate_buffer = "ISDN";
               break;

            case NTT_CALLING_NUMBER_UNKNOWN_DATA:
               translate_buffer = "Data";
               break;

            case NTT_CALLING_NUMBER_UNKNOWN_TELEX:
               translate_buffer = "Telex";
               break;

            case NTT_CALLING_NUMBER_UNKNOWN_DOMESTIC:
               translate_buffer = "Domestic";
               break;

            case NTT_CALLING_NUMBER_UNKNOWN_PRIVATE:
               translate_buffer = "Private";
               break;

            case NTT_CALLING_NUMBER_INTERNATIONAL_UNKNOWN:
               translate_buffer = "International";
               break;

            case NTT_CALLING_NUMBER_INTERNATIONAL_ISDN:
               translate_buffer = "International ISDN";
               break;

            case NTT_CALLING_NUMBER_INTERNATIONAL_DATA:
               translate_buffer = "International Data";
               break;

            case NTT_CALLING_NUMBER_INTERNATIONAL_TELEX:
               translate_buffer = "International Telex";
               break;

            case NTT_CALLING_NUMBER_INTERNATIONAL_DOMESTIC:
               translate_buffer = "International Domestic";
               break;

            case NTT_CALLING_NUMBER_INTERNATIONAL_PRIVATE:
               translate_buffer = "International Private";
               break;

            case NTT_CALLING_NUMBER_DOMESTIC_UNKNOWN:
               translate_buffer = "Domestic";
               break;

            case NTT_CALLING_NUMBER_DOMESTIC_ISDN:
               translate_buffer = "Domestic ISDN";
               break;

            case NTT_CALLING_NUMBER_DOMESTIC_DATA:
               translate_buffer = "Domestic Data";
               break;

            case NTT_CALLING_NUMBER_DOMESTIC_TELEX:
               translate_buffer = "Domestic Telex";
               break;

            case NTT_CALLING_NUMBER_DOMESTIC_DOMESTIC:
               translate_buffer = "Domestic";
               break;

            case NTT_CALLING_NUMBER_DOMESTIC_PRIVATE:
               translate_buffer = "Domestic Private";
               break;

            case NTT_CALLING_NUMBER_NETWORK_UNKNOWN:
               translate_buffer = "Network";
               break;

            case NTT_CALLING_NUMBER_NETWORK_ISDN:
               translate_buffer = "Network ISDN";
               break;

            case NTT_CALLING_NUMBER_NETWORK_DATA:
               translate_buffer = "Network DATA ";
               break;

            case NTT_CALLING_NUMBER_NETWORK_TELEX:
               translate_buffer = "Network Telex";
               break;

            case NTT_CALLING_NUMBER_NETWORK_DOMESTIC:
               translate_buffer = "Network domestic";
               break;

            case NTT_CALLING_NUMBER_NETWORK_PRIVATE:
               translate_buffer = "Network Private";
               break;

            case NTT_CALLING_NUMBER_ABBREVIATED_UNKNOWN:
               translate_buffer = "Abbreviated";
               break;

            case NTT_CALLING_NUMBER_ABBREVIATED_ISDN:
               translate_buffer = "Abbreviated ISDN";
               break;

            case NTT_CALLING_NUMBER_ABBREVIATED_DATA:
               translate_buffer = "Abbreviated Data";
               break;

            case NTT_CALLING_NUMBER_ABBREVIATED_TELEX:
               translate_buffer = "Abbreviated Telex";
               break;

            case NTT_CALLING_NUMBER_ABBREVIATED_DOMESTIC:
               translate_buffer = "Abbreviated Domestic";
               break;

            case NTT_CALLING_NUMBER_ABBREVIATED_PRIVATE:
               translate_buffer = "Abbreviated Private";
               break;

            case NTT_REDIR_UNDEF:
               translate_buffer = "Forward";
               break;

            case NTT_REDIR_BUSY:
               translate_buffer = "Forward on busy:";
               break;

            case NTT_REDIR_UNANSWERED:
               translate_buffer = "Forward on no answer";
               break;

            case NTT_REDIR_UNCONDITIONAL:
               translate_buffer = "Forward Unconditionnal";
               break;
         }


         switch (param_of_variable_length) {
            case NTT_CALLING_NUMBER:
               hex2uchar(string_param_value, unformatted_number);
               format_number(formatted_number, (char*)unformatted_number);
               snprintf(formatted_result, sizeof(formatted_result), "From%s", formatted_number);
               translate_buffer = formatted_result;
               break;

            case NTT_REDIRECTING_NUMBER:
               hex2uchar(string_param_value, unformatted_number);
               format_number(formatted_number, (char*)unformatted_number);
               snprintf(formatted_result, sizeof(formatted_result), "From%s", formatted_number);
               translate_buffer = formatted_result;
               break;

            case NTT_CALLED_NUMBER_DDN:
               hex2uchar(string_param_value, unformatted_number);
               format_number(formatted_number, (char*)unformatted_number);
               snprintf(formatted_result, sizeof(formatted_result), "DDN:%s", formatted_number);
               translate_buffer = formatted_result;
               break;

            case NTT_PUBLIC_MESSAGE:
               hex2uchar(string_param_value, unformatted_number);
               /* no need to format, this is not a number, but a public message */
               snprintf(formatted_result, sizeof(formatted_result), "Public Message:%s", (char* )unformatted_number);
               translate_buffer = formatted_result;
               break;
         }
      }
      break;
   }

   size_t addlen = strlen(translate_buffer);
   if (addlen > 0) {
      if (!(modem_cid->status & CIDMESG))
      {
         if (CIDSIZE - addlen > 0) {
            strcpy(modem_cid->cidmesg, translate_buffer);
         }
         else {
            logMsg(LEVEL6, "MESG discarded %s\n", translate_buffer);
         }
      }
      else
      {
         /* there is already a previous code MESG */
         size_t curlen = strlen(modem_cid->cidmesg);
         if (CIDSIZE - (curlen + addlen + 2) > 0) {
            modem_cid->cidmesg[curlen] = ',';
            strcpy(modem_cid->cidmesg+curlen+1, translate_buffer);
         }
         else {
            logMsg(LEVEL6, "MESG discarded %s\n", translate_buffer);
         }
      }
      modem_cid->cidmesg[CIDSIZE-1] = '\0';
      modem_cid->status |= CIDMESG;
      modem_cid->cidsent = 0;
   }
}
