/*
 * nciddhangup.c - This file is part of ncidd.
 *
 * Copyright (c) 2005-2022
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ncidd.h"

/* globals */
int regex;
const char * blacklist = BLACKLIST;
const char * listname;
const char * whitelist = WHITELIST;

struct list *blkHead = NULL, *blkCurrent = NULL,
            *whtHead = NULL, *whtCurrent = NULL;

int doList(const char * filename, list_t **listHead, list_t **listCurrent);
struct retbw onList(const char * namep, const char * nmbrp, int flag, list_t **listHead);

void addEntry(const char * fnptr, const char * inptr, char * wdptr, int lc, list_t **listHead, list_t **listCurrent);
void nextEntry(list_t **listHead, list_t **listCurrent);
void rmEntries(list_t **listHead, list_t **listCurrent);

/*
 * Process the blacklist or whitelist file
 * Returns: 0 - no errors
 *          # of errors
 */
int doList(const char * filename, list_t **listHead, list_t **listCurrent)
{
    char input[BUFSIZ], word[BUFSIZ], msgbuf[BUFSIZ], *ptr;
    const char * inptr;
    struct list *node, *nextnode;
    int lc, i, bl;
    int max_temp = 0, max_entry = 0;
    FILE *fp;

    bl = !strcmp(blacklist, filename) ? 1 : 0;

    if ((fp = fopen(filename, "r")) == NULL)
    {
        logMsg(LEVELE, "%s file missing: %s\n",
               bl ? "Blacklist" : "Whitelist", filename);
        return 1;
    }

    /* read each line of file, one line at a time */
    for (lc = 1; fgets(input, BUFSIZ, fp) != NULL; lc++)
    {
        /* line containing only <NL> or <CR><NL> or is a comment line */
        if (input[0] == '\n' || input[0] == '\r' || input[0] == '#') continue;

        /* remove '\n' and '\r' if there */
        if ((ptr = strchr(input, (int) '\n'))) *ptr = '\0';
        if ((ptr = strchr(input, (int) '\r'))) *ptr = '\0';

        inptr = getWord(filename, input, word, lc);
        if (inptr == 0) continue;

        /* get search strings on line */
        addEntry(filename, inptr, word, lc, listHead, listCurrent);
    }
    (void) fclose(fp);
    logMsg(LEVEL1, "Processed %s file: %s\n",
           bl ? "blacklist" : "whitelist", filename);

    logMsg(LEVEL1, "%s Table:\n", bl ? "Blacklist" : "Whitelist");

    /* log the number of entries at LEVEL1 and list them at LEVEL7 */
    if (!errorStatus && *listHead)
    {
        node = *listHead;
        for (i = 0; node != NULL; i++)
        {
            nextnode = node->next;
            node = nextnode;
        }

        logMsg(LEVEL1, "    Number of Entries: %d\n", i);

        /*
         * Make the List display pretty by determining maximum column width.
         */

        node = *listHead;
        while(node != NULL)
        {
            nextnode = node->next;
            max_temp = strlen(node->entry);
            if (max_temp > max_entry) max_entry = max_temp;
            node = nextnode;
        }

        logMsg(LEVEL7, "    %-4s %s%-*s %s\n",
               "SLOT",
               "ENTRY", max_entry - (int) strlen("ENTRY") + 2, "",
               "MATCH NAME");

        logMsg(LEVEL7, "    %-4s %s%-*s %s\n",
               "----",
               "-----", max_entry - (int) strlen("ENTRY") + 2, "",
               "----------");

        node = *listHead;
        for (i = 0; node != NULL; i++)
        {
            nextnode = node->next;
            sprintf(msgbuf, "    %-4.3d \"%s\"%-*s ",
                    i,
                    node->entry,
                    max_entry - (int) strlen(node->entry), "");

            if (node->newname)
                strcat(strcat(strcat(msgbuf, "\""), node->newname), "\"");
            strcat(msgbuf, NL);

            logMsg(LEVEL7, msgbuf);
            node = nextnode;
        }
    }
    else
    {
        logMsg(LEVEL1, "    Number of Entries: 0\n");
    }

    return errorStatus;
}

/*
 * Process blacklist or whitelist file lines
 */
void addEntry(const char * fnptr, const char * inptr, char * wdptr, int lc, list_t **listHead, list_t **listCurrent)
{
    int err;
    char errbuf[512], *cptr, *eptr;

    /* process the blacklist or whitelist words */
    do
    {
        /*
         * If inQuotes = 1, the word is inside "" and is not a comment.
         * Allows an entry in the blacklist or whitelist to begin with a "#".
         */
        if (!inQuotes && *wdptr == '#') break;    /* rest of line is comment */
        nextEntry(listHead, listCurrent);
        if (strlen(wdptr) > ENTRYSIZE - 1) {
            configError(fnptr, lc, wdptr, ERRLONG);
            return;
        }
        /* store the entry to be matched */
        (*listCurrent)->entry = strdup(wdptr);
        /* cheats const char inptr here */
        if ((cptr = strstr(inptr, "#=")))
        {
            /* store the name that comes after #= for later use
             * if the caller is found in white or blacklist
             */
            cptr += 2;

            /* remove any leading spaces */
            while (*cptr && (*cptr == ' ' || *cptr == '\t')) ++cptr;

            /* remove a trailing comment from the #= new name */
            if ((eptr = strstr(cptr, "#"))) *eptr = '\0';
            /* remove any trailing spaces */
            eptr = cptr;
            while (*eptr) eptr++;
            for (--eptr; *eptr && (*eptr == ' ' || *eptr == '\t'); --eptr) *eptr = '\0';

            /* store the #= newname with * delimiters removed */
            (*listCurrent)->newname = starfree(cptr, CIDSIZE);
        }
        else (*listCurrent)->newname = 0;

        if (regex == 1)
        {
            /* expressions are regex */
            (*listCurrent)->preg = (regex_t*)malloc(sizeof(regex_t));
            err = regcomp((*listCurrent)->preg, wdptr, REG_EXTENDED|REG_NOSUB);
            if (err)
            {
                regerror(err, (*listCurrent)->preg, errbuf, sizeof(errbuf));
                configError(fnptr, lc, wdptr, errbuf);
            }
        }
        else if (regex == 2)
        {
            /* expressions are pcre */
            (*listCurrent)->pcreg = pcre2_compile((PCRE2_SPTR8)wdptr, PCRE2_ZERO_TERMINATED, 0, &pcreErrorNumber, &pcreErrorOffset, NULL);

            if ((*listCurrent)->pcreg == NULL)
            {
                PCRE2_UCHAR pcreErrorMessage[256];
                pcre2_get_error_message(pcreErrorNumber, pcreErrorMessage, sizeof(pcreErrorMessage));
                sprintf(errbuf, "pcre2_compile failed (offset: %d), %s\n", (int)pcreErrorOffset, pcreErrorMessage);
                configError(fnptr, lc, wdptr, errbuf);
            }
            /* alloc space for the unused result */
            if (pcreMatchData == NULL) pcreMatchData = pcre2_match_data_create(4, NULL);
        }
        else /* (regex == 0) */
        {
            /* NCID Simple expressions */
            (*listCurrent)->ncidreg = store_bwstrmatch(wdptr);
        }
    }
    while ((inptr = getWord(fnptr, inptr, wdptr, lc)) != 0);
}

/*
 * Adds a node to the list for a new entry
 */
void nextEntry(list_t **listHead, list_t **listCurrent)
{
    list_t *node;

    if (!(node = (list_t *) malloc(sizeof(list_t)))) errorExit(-1, progname, 0);
    memset(node, 0x00, sizeof(list_t));
    node->next = NULL;
    if (*listHead == NULL) *listHead = node;
    else (*listCurrent)->next = node;
    *listCurrent = node;
}

/*
 * Frees all list nodes
 */
void rmEntries(list_t **listHead, list_t **listCurrent)
{
    struct list *node = *listHead;
    struct list *nextnode;

    while (node != NULL)
    {
        nextnode = node->next;
        if (regex == 1) { regfree(node->preg); free(node->preg); }
        else if (regex == 2) pcre2_code_free(node->pcreg);
        else free_abwstr(node->ncidreg);
        if (node->entry) free(node->entry);
        if (node->newname) free(node->newname);
        free(node);
        node = nextnode;
    }

    *listHead = NULL;
    *listCurrent = NULL;
}


/*
 * Check if call is in whitelist file
 *
 * retbw.ret is 0 if call not in whitelist
 * retbw.ret is 1 if call is in whitelist
 */

struct retbw checkWhitelist(const char * namep, const char * nmbrp)
{
    return onList(namep, nmbrp, 0, &whtHead);
}

/*
 * Check if call is in blacklist file
 *
 * A whitelist match prevents a blacklist match.
 *
 * retbw.ret is 0 if call not in blacklist
 * retbw.ret is 1 if call is in blacklist
 */

struct retbw checkBlacklist(const char * namep, const char * nmbrp)
{
    return onList(namep, nmbrp, 0, &blkHead);
}


/*
 * Compare blacklist or whitelist strings to name and number
 *
 * If "flag" is zero, log a match message
 *   Return = 1  if the number or name matches
 *   Return = 0  if no match
 *
 * If "flag" is one, log a match message
 *   Return = 3  if the number matches
 *   Return = 1  if the name matches
 *   Return = 0  if no match
 */
struct retbw onList(const char * namep, const char * nmbrp, int flag, list_t **listHead)
{
    int i, pos, nbrMatch = 0;
    struct retbw retbw = { 0, "" };
    const char *ptr, *listptr;
    list_t *node, *nextnode;
    void * regexp = NULL;

    listptr = (listHead == &blkHead ? "Blacklist" : "Whitelist");
    node = *listHead;
    if (!node)
    {
        logMsg(LEVEL3, "%s empty\n", listptr);
        return retbw;
    }

    logMsg(LEVEL4, "Begin: Search %s file nmbr=%s name=%s [%s]\n", listptr, nmbrp==0?"0":nmbrp, namep==0?"0":namep, strdate(ONLYTIME));

    for (i = 0; node != 0; i++)
    {
        nextnode = node->next;
        ptr = node->entry;
        pos = i;

        if (regex == 2)
        {
            /* perl regular expression */
            regexp = (void*)node->pcreg;
        }
        else if (regex == 1)
        {
            /* regex regular expression */
            regexp = (void*)node->preg;
        }
        else /* (regex == 0) */
        {
            /* ncid simple expression */
            regexp = (void*)node->ncidreg;
        }

        if (strmatch(namep, regexp))
        {
            retbw.ret = 1;
            retbw.entry = ptr;
            break;
        }
        if (strmatch(nmbrp, regexp))
        {
            retbw.ret = 1,
            nbrMatch = 2;
            retbw.entry = ptr;
            break;
        }

        node = nextnode;
    }

    logMsg(LEVEL3, "Checked %s for match flag=%d ret=%d nmbrmatch=%d\n",
           listptr, flag, retbw.ret, nbrMatch);
    if (flag) retbw.ret += nbrMatch;
    if (retbw.ret)
    {
        listname = node->newname;
        logMsg(LEVEL3, "%s Match #%.2d: %s    number: %s    name: %s\n",
               listptr, pos, ptr, nmbrp, namep);
    }

    return retbw;
}

/*
 * Check for a name or a number being in the black or white list
 *
 * Return = 0  not on either list
 * Return = 1  name is on the blacklist and not on the whitelist
 * Return = 5  number is on the blacklist and not on the whitelist
 * Return = 3  name is on the whitelist and on the blacklist
 * Return = 4  number is on the whitelist and on the blacklist
 * Return = 2  name is on the whitelist and not on the blacklist
 * Return = 6  number is on the whitelist and not on the blacklist
 *
 * Bits 0 and 1:
 *      0 - On neither list
 *      1 - Name or number on blacklist
 *      2 - Name or number on whitelist
 *
 * Bit 2:
 *      0 - Name is on the list
 *      1 - Number is on the list
 */
struct retbw onBlackWhite (const char * namep, const char * nmbrp)
{
    struct retbw retbw, blkret;

    /* retbw.ret will be either 0, 1, or 3 */
    retbw = onList(namep, nmbrp, 1, &blkHead);
    if (retbw.ret) retbw.ret =  2 * retbw.ret - 1;
    blkret = retbw;

    retbw = onList(namep, nmbrp, 1, &whtHead);
    if (retbw.ret)
    {
        retbw.ret =  2 * retbw.ret;
        if (blkret.ret) retbw.ret = (retbw.ret == 1 ? 3 : 4);
    }
    else
    {
        retbw = blkret;
    }

    return retbw;
}


/*
 * remove existing whitelist, blacklist entries to free memory used
 */
void rmBWlists(void)
{
    rmEntries(&blkHead, &blkCurrent);

    rmEntries(&whtHead, &whtCurrent);
}

/*
 * read the whitelist, blacklist file entries into memory
 */
int readBWlists(void)
{
    /* reload blacklist file but quit on error */
    if (doList(blacklist, &blkHead, &blkCurrent)) return 114;

    /* reload whitelist file but quit on error */
    if (doList(whitelist, &whtHead, &whtCurrent)) return 114;

    return 0;
}
