/*
 * nciddmodems.c - This file is part of ncidd.
 *
 * Copyright (c) 2005-2024
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "ncidd.h"
#include "nciddmodems.h"

/* globals */
const char * addedmodems = 0;        /* option: default - no added modems */
int modemModes;                      /* used temporarily in discovery */

/* modem and per-call info */
time_t lastmodemlocktimer  = 0;      /* to check for modem locks */

char modembuf[BUFSIZ];               /* used temporarily, has modem response from initModem() */

char LineIDS[CIDSIZE * 5];           /* room for 5 lineid's */

/* IPinfo[pos] is for same client/gateway/modem as in polld[pos] */
struct modems_t;
extern struct ipinfo_t  IPinfo[];

/* List of popular country codes as returned by QUERYATGCI ("AT+GCI?")
 * http://doc.slitaz.org/en:handbook:pstn:countries
 * http://www.cisco.com/c/en/us/td/docs/routers/access/modem/AT/wic/command/reference/atwic/atwic3.html
 *
 * Country codes have their origin based on fascimile standards.
 * The definitive guide for all country codes is maintained by the
 * International Telegraph Union (ITU), a specialized agency of the
 * United Nations. Not all modem manufacturers use a consistent list
 * of country codes and not all modem manufactures even use the
 * complete list. The ITU's complete list is available here:
 *     http://www.itu.int/dms_pub/itu-t/opb/sp/T-SP-T.35-2012-OAS-PDF-E.pdf
 */
char *modemGCI[] =
{
     "00 Japan",
     "04 Germany",
     "09 Australia",
     "0A Austria",
     "0F Belgium",
     "16 Brazil",
     "1B Bulgaria",
     "20 Canada",
     "26 China",
     "2D Cyprus",
     "2E Czech Republic",
     "31 Denmark",
     "3C Finland",
     "3D France",
     "42 Germany",
     "46 Greece",
     "50 Hong Kong",
     "51 Hungary",
     "52 Iceland",
     "53 India",
     "57 Ireland",
     "58 Israel",
     "59 Italy",
     "61 Korea",
     "68 Liechtenstein",
     "69 Luxembourg",
     "6C Malaysia",
     "73 Mexico",
     "7B Netherlands",
     "7E New Zealand",
     "82 Norway",
     "89 Philippines",
     "8A Poland",
     "8B Portugal",
     "8C Singapore",
     "9F South Africa",
     "A0 Spain",
     "A5 Sweden",
     "A6 Switzerland",
     "A9 Thailand",
     "AE Turkey",
     "B4 United Kingdom",
     "B5 United States",
     "B8 Russia",
     "F6 TBR-21(Default)",
     "F7 Lithuania",
     "F8 Latvia",
     "F9 Estonia",
     "FB Slovakia",
     "FC Slovenia",
     "FE Taiwan",
     NULL
};


/* externs - options shared by all modems */
extern int hangup;
extern int hupmode;
extern int dialdelay;
extern const char * announce;
extern char annpath[];
extern const char * hupname;
extern char huppath[];
extern const char * huprmd;
extern char rmdpath[];

/* a list of modem objects */
int next_modem = 0;
struct modems_t modem_list[MAXMODEM];

void initSerialDevices(void);

void configModem(struct modems_t * a_modem);
void configFirstModem(struct modems_t * a_modem);
void addMoreDevices(void);

void init_modem(struct modems_t * a_modem);

void checkHangupMode(int *mode, struct modems_t * a_modem);
void changeTTY(int raw, struct modems_t * a_modem);

void openandCheckSerial(struct modems_t * a_modem);
void checkSerialOptions(struct modems_t * a_modem);
void set_ttyspeed(struct modems_t * a_modem);
void make_lockfile_name(struct modems_t * a_modem);
void readFromModem(int pos);
void continueModem(int pos);
void doHangupTimers(void);
void doHangupTimer(struct modems_t * a_modem);
void doRingTimers(void);
void doRingTimer(struct modems_t * a_modem);
int restartRingTimer(int pos);
void doModemLockTimers(void);
void doModemLockTimer(struct modems_t * a_modem);
void doModemLostTimers(void);
void doModemLostTimer(struct modems_t * a_modem);
void startModemLostTimer(const char * msg, struct modems_t * a_modem);
void TCIserial(const char * buf, int pos);
void xdmfCID(const char * buf, int pos);
void modemCID(const char * buf, int pos);
void restoreTTY(struct modems_t * a_modem);

int initModem(const char * ptr, int maxtry, struct modems_t * a_modem);

int CheckForLockfile(struct modems_t * a_modem);
int makeLockfile(struct modems_t * a_modem);
int removeLockfile(struct modems_t * a_modem);
int CheckLockfilePath(struct modems_t * a_modem);
int openSerial(struct modems_t * a_modem);
int checkModem(struct modems_t * a_modem);
int checkModemCapabilities(struct modems_t * a_modem);
int setModem(struct modems_t * a_modem);
int initModem(const char * ptr, int maxtry, struct modems_t * a_modem);
int reAttachModem(struct modems_t * a_modem);
int openTTY(struct modems_t * a_modem);
int nothex(const char * mesg);
int setTTY(struct modems_t * a_modem);
int checkSCID(struct modems_t * a_modem);

const char * getSpeedLabel(int ttyspeed);
void expand_mesg(const char * hexcode_mesg, struct cid_t * modem_cid);

/*
 * Configure serial devices or modems
 * for cidinput == 0x0, 0x1, or 0x2
 */
void initSerialDevices(void)
{
    struct modems_t * a_modem = NULL;

    /*
     * cidinput == 0x0, 0x1, or 0x2
     * TTY port needs to be set
     */

    /*
     * the first modem/serial device is described by the ncidd.conf file
     * and its options, so transfer these and init this device
     */
    a_modem = &modem_list[next_modem];
    a_modem->my_mit = next_modem;
    next_modem++;

    /* fill in the empty parts
     * init_modem() was done for this modem in main() before
     * reading command line options and the main conf file
     * fill in specific parts from the config
     */
    configFirstModem(a_modem);

    /* open the device and ask it questions */
    openandCheckSerial(a_modem);


    /* add more serial devices from the addedmodems list */
    addMoreDevices();

    logMsg(LEVEL1, "%s declares %d modem%c and/or serial device%c\n", config, next_modem, (next_modem==1?' ':'s'), (next_modem==1?' ':'s'));
    logMsg(LEVEL1, "Maximum number of clients/gateways: %d\n", MAXCLIENTS+1-next_modem);
}


/*
 * init the struct of per-modem information
 */
void init_modem(struct modems_t * a_modem)
{
    a_modem->is_modem = 0;
/*       a_modem->ttyfd;                the modem file descriptor */
    a_modem->ttyport = NULL;             /* the modem device name */
    a_modem->serial_lineid = ONELINE;    /* the lineid */
    a_modem->initstr  = INITSTR;
    a_modem->initcid  = INITCID1;
    a_modem->setcid = 0;
    a_modem->gencid = 0;
    a_modem->TTYspeed = NULL;
    a_modem->ttyspeed = TTYSPEED;
    a_modem->cidnoname = 0;
    a_modem->pickup = 1;
    a_modem->audiofmt = XPRESS;
    a_modem->modem_ext_locked = 0;
    a_modem->modem_locked = 0;
    a_modem->modemType = 0;
    memset(a_modem->doesmode, 0, 4*sizeof(unsigned char));
    a_modem->lockfile = NULL;            /* the modem lock file name */
    a_modem->pollpos = -1;               /* remember modem pollfd during lock */
/*       a_modem->pollevents;        remember modem pollfd during lock */
    a_modem->checked_hangup_mode = 0;    /* checked modem capabilities */
    a_modem->nodatacnt = 0;
    a_modem->nodatalowest = 0;
    a_modem->modemlostcount = 0;
    a_modem->clocal = 0;
    a_modem->ringcnt = 0;                /* one ringie-dingie ... */
    a_modem->ringtimer = 0;              /* to detect end of ringing */
    a_modem->modemlosttimer = 0;         /* to try modem init again */
    a_modem->hanguptimer = 0;
    a_modem->dialclientpos = -1;
    a_modem->msgfd = -1;
    a_modem->modem_cid = (struct cid_t *) calloc(1, sizeof(struct cid_t));
    a_modem->lidx = 0;
    a_modem->state = MODEMINITIALIZING;

}


/*
 * apply the config file for a single modem or serial device
 */

void configFirstModem(struct modems_t * a_modem)
{
    char aliasbuf[CIDSIZE];

    /*
     * lineid could have been changed in either ncidd.conf or ncidd.alias
     * this sets cid.cidline to lineid after any changes to it
     */
    strncpy(aliasbuf, a_modem->serial_lineid, CIDSIZE -1);
    aliasbuf[CIDSIZE-1] = '\0';
    userAlias(0, 0, aliasbuf);
    a_modem->serial_lineid = strdup(aliasbuf);

    logMsg(LEVEL1, "Telephone Line Identifier: %s\n", a_modem->serial_lineid);

    /*initialize the modem cid */
    clear_modem_cid(a_modem);

    /* Add lineid to LineIDS */
    sprintf(LineIDS, "\"%s\"", a_modem->serial_lineid);

}


/*
 * hang up outgoing DIAL calls after a few seconds (dialdelay)
 * - or -
 * hang up after the outgoing FAX tones, announcement or off-hook pause
 */
void doHangupTimers(void)
{
    int mit = 0;
    while (modem_list[mit].is_modem && mit < MAXMODEM) {
       if (modem_list[mit].hanguptimer) doHangupTimer(&modem_list[mit]);
       mit++;
    }
}


/*
 * hang up an outgoing DIAL call after a few seconds (dialdelay)
 * - or -
 * hang up after the outgoing FAX tones, announcement or off-hook pause
 */
void doHangupTimer(struct modems_t * a_modem)
{
   int ret = 0;
   char msgbuf[BUFSIZ];
   struct cid_t * modem_cid = a_modem->modem_cid;

   if (a_modem->hanguptimer < time(0))
   {
      /* turn the timer off */
      a_modem->hanguptimer = 0;
      if (a_modem->state == MODEMHANGUP ||
          a_modem->state == MODEMANNOUNCING)
      {
         /* waiting after hangup or outgoing FAX/voice message  */

         /* Hangup the call */
         ret = initModem(HANGUP, HANGUPTRY, a_modem);
         logMsg(LEVEL5, "hangup: HANGUP sent, return code is %d\n", ret);

         /* if Hangup does not indicate OK try 1 more time  */
         if (ret != 0)
         {
             /* Send AT to get modem OK */
             ret = initModem(GETOK, HANGUPTRY, a_modem);
             logMsg(LEVEL5, "hangup try 2: GETOK sent, return code is %d\n", ret);

             /* Send HANGUP to get modem OK */
             ret = initModem(HANGUP, HANGUPTRY, a_modem);
             logMsg(LEVEL5, "hangup try 2: HANGUP sent, return code is %d\n", ret);
         }

         if (a_modem->state == MODEMANNOUNCING)    /* FAX or VOICE hangup mode */
         {
             /* go back to data mode */
             ret = initModem(DATAMODE, HANGUPTRY, a_modem);
             logMsg(LEVEL5, "hangup: DATAMODE sent, return code is %d\n", ret);
         }

         /* take tty port out of raw mode */
         (void) tcsetattr(a_modem->ttyfd, TCSAFLUSH, &a_modem->ntty);


         /* remove TTY port lockfile */
         removeLockfile(a_modem);

         /* make the ring timer go off right away to do sendInfo 0 */
         if (a_modem->ringtimer) a_modem->ringtimer = 1;
      }
      else
      {
         /* dialled an outgoing call, a_modem->state == MODEMDIALLED */

         /* hang up the call, log it and send notice to client */
         initModem(HANGUP, READTRY, a_modem);
         logMsg(LEVEL2, "dial - hungup %s on line \"%s\" at %s\n",
                 modem_cid->cidnmbr, modem_cid->cidline, strdate(ONLYTIME));
         if (a_modem->dialclientpos > 0) {
            sprintf(msgbuf, DIAL_HANGUP CRLF, modem_cid->cidnmbr, modem_cid->cidline);
            safeSend(a_modem->dialclientpos, msgbuf);
         }
      }

      /* Reset caller id info in modem_cid */
      clear_modem_cid(a_modem);
      a_modem->state = MODEMWAITING;

   }
}


/*
 * the modem hasn't said RING in a while (RINGWAIT seconds)
 */
void doRingTimers(void)
{
    int mit = 0;
    while (modem_list[mit].is_modem && mit < MAXMODEM) {
       if (modem_list[mit].ringtimer) doRingTimer(&modem_list[mit]);
       mit++;
    }
}


/*
 * the modem hasn't said RING in a while (RINGWAIT seconds)
 */
void doRingTimer(struct modems_t * a_modem)
{
   if (a_modem->ringtimer < time(0))
   {
       /* turn the timer off */
       a_modem->ringtimer = 0;
       logMsg(LEVEL5, "ring: %d timeout at: %s\n", a_modem->ringcnt, strdate(ONLYTIME));
       a_modem->ringcnt = 0;

       /* tell clients that ringing stopped
        * (cidline might have added distinctive ring) */
       sendInfo(0, a_modem->modem_cid->cidline);

       clear_modem_cid(a_modem);

   }
}


/*
 * increment the RING count and reset the timer
 */
int restartRingTimer(int pos)
{
   struct modems_t * a_modem = findModemByPos(pos);

   if (a_modem != NULL)
   {
      a_modem->ringcnt++;
      /* timer to announce that ringing has stopped */
      a_modem->ringtimer = time(0) + RINGWAIT;
      return a_modem->ringcnt;
   }
   else
   {
      return 0;
   }
}


/*
 * try the modem init again in a while
 */
void startModemLostTimer(const char * msg, struct modems_t * a_modem)
{
   char msgbuf[BUFSIZ];
   const char * devicetype = "Modem";

   if (a_modem == 0) return;
   if (a_modem->modemlosttimer > 0) return;
   if (cidinput==0x1) devicetype = "Caller ID Device";
   logMsg(LEVEL1, "%s %s Not Connected - %s at: %s\n", a_modem->ttyport, devicetype, msg, strdate(ONLYTIME));

   /* stop polling, but reserve the slot in polld[] */
   if (a_modem->pollpos >= 0) reservePoll(a_modem->pollpos);
   if (a_modem->ttyfd >= 0) close(a_modem->ttyfd);
   a_modem->ttyfd = -1;

   if (a_modem->modemlostcount == 0) {
      sprintf(msgbuf, "%s %s lost", a_modem->ttyport, devicetype);
      sysMesg(msgbuf, a_modem->serial_lineid);
   }
   /* give up at 5 minutes */
   if (a_modem->modemlostcount > 5) {
      if (cidinput == 0x2) {
         /* support modem open fail not fatal, just remove it
          * this is not recoverable without a restart */
         sprintf(msgbuf, "%s %s still lost, will be ignored", a_modem->ttyport, devicetype);
         sysMesg(msgbuf, a_modem->serial_lineid);
         a_modem->modemlosttimer = 0;
         return;
      }
      else {
         sprintf(msgbuf, "%s %s still lost, server shutting down now", a_modem->ttyport, devicetype);
         sysMesg(msgbuf, a_modem->serial_lineid);
         errorExit(-104, "Fatal", msg);
      }
   }
   a_modem->modemlostcount++;
   a_modem->modemlosttimer = time(0) + MODEMWAIT;
}



/*
 * try the modem init again for each lost modem
 */
void doModemLostTimers(void)
{
    int mit = 0;
    while (modem_list[mit].is_modem && mit < MAXMODEM) {
       if (modem_list[mit].modemlosttimer) doModemLostTimer(&modem_list[mit]);
       mit++;
    }
}

/*
 * try the modem init again for a lost modem
 */
void doModemLostTimer(struct modems_t * a_modem)
{
   char msgbuf[BUFSIZ];
   const char * devicetype = "Modem";
   if (cidinput==0x1) devicetype = "Caller ID Device";

   if (a_modem->modemlosttimer < time(0))
   {
       /* turn the timer off */
       a_modem->modemlosttimer = 0;
       logMsg(LEVEL5, "Modem Lost timer expires, trying again: %s\n", strdate(ONLYTIME));

       if (reAttachModem(a_modem) == 0 && a_modem->modemlosttimer == 0) {
           a_modem->modemlostcount = 0;

           /* tell clients that the modem is back */
           sprintf(msgbuf, "%s %s connected", a_modem->ttyport, devicetype);
           sysMesg(msgbuf, a_modem->serial_lineid);
       }
   }
}


/*
 * try the modem init again,
 * restarting the timer if it fails
 * The modem is expected to be the same as the lost one
 */
int reAttachModem(struct modems_t * a_modem)
{
   const char * devicetype = "Modem";

   if (openSerial(a_modem) != 0) return -1;

   if (cidinput == 0)
   {
       /* set modem to do caller ID */
       if (setModem(a_modem) != 0) return -1;
   }
   else if (cidinput==0x1)
   {
       devicetype = "Caller ID Device";
   }
   else if (cidinput == 0x2)
   {
       /* modem used to support gateway */
       if (initModem(a_modem->initstr, READTRY, a_modem) != 0) {
          startModemLostTimer("Failed to initialize modem", a_modem);
          return -1;
       }
   }

   /* has the modem been here before? */
   if (a_modem->pollpos >= 0) {
      /* put the modem back into the polld table */
      restorePoll(a_modem->pollpos, a_modem->ttyfd);
      logMsg(LEVEL3,"%s %d is fd %d pos %d\n", devicetype, a_modem->my_mit, a_modem->ttyfd, a_modem->pollpos);
   }
   else {
      /* first time finding the modem */
      if (checkModemCapabilities(a_modem) < 0) return -1;

      /* add the modem to the polld table */
      a_modem->pollpos = addPoll(a_modem->ttyfd);
      logMsg(LEVEL3,"%s %d is fd %d pos %d\n", devicetype, a_modem->my_mit, a_modem->ttyfd, a_modem->pollpos);

        /* link the modem_list to the serial_line list */
        a_modem->state = MODEMWAITING;
        if (a_modem->pollpos>=0) {
           IPinfo[a_modem->pollpos].modem = a_modem;
        }
   }
   return 0;
}


/*
 * Configure serial device or modem
 * for cidinput == 0x0, 0x1, or 0x2
 */
void openandCheckSerial(struct modems_t * a_modem)
{

    checkSerialOptions(a_modem);
    if (openSerial(a_modem) != 0) return;
    checkModemCapabilities(a_modem);
}


/*
 * Check the cidinput mode against the hangup/hupmode setting
 * and make sure the supporting voice/extension files are present
 * Describe the settings for the log file.
 */
void checkSerialOptions(struct modems_t * a_modem)
{
   /*
    * noserial and nomodem are obsolete (now used to set cidinput)
    *      noserial = 1: serial port not used
    *      noserial = 0: serial port used for Caller ID
    *      nomodem  = 1: serial device
    *      nomodem  = 0: modem
    *
    * cidinput = 0: Caller ID from a modem and optional gateways
    * cidinput = 1: Caller ID from a serial or USB device and optional gateways
    * cidinput = 2: Caller ID from a gateway with modem support
    * cidinput = 3: Caller ID from gateways without modem support
    */

   /* If the tty port speed was set, map it to the correct integer. */
   set_ttyspeed(a_modem);

   /* Create lock file name from TTY port device name */
   make_lockfile_name(a_modem);

   /* set or reset the TTYspeed string from the current ttyspeed value */
   if (a_modem->TTYspeed != NULL) free((void*)a_modem->TTYspeed);
   a_modem->TTYspeed = getSpeedLabel(a_modem->ttyspeed);

   logMsg(LEVEL1, "TTY port speed: %s\n", a_modem->TTYspeed);
   logMsg(LEVEL1, "TTY lock file: %s\n", a_modem->lockfile);
   logMsg(LEVEL1, "TTY port control signals %s\n", a_modem->clocal ? "disabled" : "enabled");

   if (cidinput == 0x1)
   {
     /* serial device is only wanted for caller ID */
     logMsg(LEVEL1, "Caller ID from a serial or USB device and optional gateways\n");
   }
   else
   {
      /*
       * noserial == 0 && nomodem == 0
       * noserial == 1 && nomodem == 0
       * cidinput == 0x0 or 0x2
       */

      if (cidinput == 0x2)
      {
        /* modem used for gateway support; hangup and dial */

        if (!hangup && !hupmode)
            logMsg(LEVEL1, "Modem will only be used to dial number\n");
        else
            logMsg(LEVEL1, "Modem will only be used to dial number and hangup calls\n");
      }
      else /* cidinput == 0 */
      {
        /* modem used for callerid, hangup and dial */

        /* get modem capabilities */
        logMsg(LEVEL1, "Modem will be used for Caller ID\n");

        if (a_modem->gencid)
           logMsg(LEVEL1, "Handles modem calls without Caller ID (gencid=1)\n");
        else
           logMsg(LEVEL1, "Does not handle modem calls without Caller ID (gencid=0)\n");

        logMsg(LEVEL1, "Caller ID from a modem and optional gateways\n");
        logMsg(LEVEL1, "Modem also used to %s\n", (hangup || hupmode)
                ? "dial number and terminate calls" : "dial number");
      }

      /* cidinput == 0x0 or 0x2 do dial and hangup */
      logMsg(LEVEL1, "Dial Modem: %s\nDial Timeout: %d seconds\n",
              a_modem->serial_lineid, dialdelay);
      logMsg(LEVEL1, "LineIDS: %s\n", LineIDS);

   }
}


/*
 * Configure serial device or modem
 * for cidinput == 0x0, 0x1, or 0x2
 */
int openSerial(struct modems_t * a_modem)
{
   /* say it's a modem */
   a_modem->is_modem = 1;

   /* Open tty port; start a timer if it fails */
   if (openTTY(a_modem) < 0)
   {
       /* perhaps ModemManager was interfering at boot time */
       startModemLostTimer("serial device open TTY failed", a_modem);
       return -1;
   }

   logMsg(LEVEL1, "TTY port opened: %s\n", a_modem->ttyport);

   if (setTTY(a_modem) < 0)
   {
       logMsg(LEVELE, "Set TTY error:  %s\n", strerror(errno));
       startModemLostTimer("serial device setTTY", a_modem);
       return -1;
   }

   return 0;
}


/*
 * Check modem capabilities
 * for cidinput== 0x0, tells modem to do caller ID
 */
int checkModemCapabilities(struct modems_t * a_modem)
{
   int ret;

   /*
    * noserial == 0 && nomodem == 0
    * noserial == 1 && nomodem == 0
    * cidinput == 0x0 or 0x2
   */
   if (cidinput == 0x1)
   {
      /* serial callerid device, nothing more to do  */
   }
   else if (cidinput == 0x2)
   {
      /* modem used for gateway support; hangup and dial */

      ret = initModem(a_modem->initstr, READTRY, a_modem);
      if (ret)
      {
         startModemLostTimer("Gateway support modem not found", a_modem);
         return -1;
      }
      if (!hangup && !hupmode)
      {
         logMsg(LEVEL1, "Modem detected\nModem only used to dial number\n");
      }
      else
      {
         logMsg(LEVEL1, "Modem detected\nModem only used to dial number and hangup calls\n");
         /* get modem capabilities */
         if (checkModem(a_modem) < 0) return -1;
      }
   }
   else /* cidinput == 0 */
   {
      /* modem used for callerid, hangup and dial */

      /* get modem capabilities, as modemModes */
      if (checkModem(a_modem) != 0) return -1;
      /* set modem to do caller ID */
      if (setModem(a_modem) != 0) return -1;

      logMsg(LEVEL1, "Modem detected\nModem used for Caller ID\n");
   }

   /* do this section just once, without repeating for a lost modem */
   if (!a_modem->checked_hangup_mode)
   {
      if (hangup)
      {
         /* check hangup mode against modem abilities from checkModem() */
         checkHangupMode(&hangup, a_modem);
      }

      if (hupmode)
      {
         /* check hangup mode against modem abilities from checkModem() */
         checkHangupMode(&hupmode, a_modem);
      }
      a_modem->checked_hangup_mode = 1;
   }
   return 0;
}


/*
 * If the tty port speed was set, map it to the correct integer.
 * set the modem ttyspeed from the string TTYspeed
 */
void set_ttyspeed(struct modems_t * a_modem)
{
      if (a_modem->TTYspeed)
      {
        if (!strcmp(a_modem->TTYspeed, "115200")) a_modem->ttyspeed = B115200;
        else if (!strcmp(a_modem->TTYspeed, "38400")) a_modem->ttyspeed = B38400;
        else if (!strcmp(a_modem->TTYspeed, "19200")) a_modem->ttyspeed = B19200;
        else if (!strcmp(a_modem->TTYspeed, "9600")) a_modem->ttyspeed = B9600;
        else if (!strcmp(a_modem->TTYspeed, "4800")) a_modem->ttyspeed = B4800;
        else if (!strcmp(a_modem->TTYspeed, "2400")) a_modem->ttyspeed = B2400;
        else if (!strcmp(a_modem->TTYspeed, "1200")) a_modem->ttyspeed = B1200;
        else errorExit(-108, "Invalid TTY port speed set in config file", a_modem->TTYspeed);
      }
}


const char * getSpeedLabel(int ttyspeed)
{

   switch(ttyspeed)
   {
     case B1200:
       return strdup("1200");
     case B2400:
       return strdup("2400");
     case B4800:
       return strdup("4800");
     case B9600:
       return strdup("9600");
     case B19200:
       return strdup("19200");
     case B38400:
       return strdup("38400");
     case B115200:
       return strdup("115200");
     default:
       return strdup("TTYspeed?");
   }
}


/*
 * Create lock file name from TTY port device name
 * and check for an existing lock file
 */

void make_lockfile_name(struct modems_t * a_modem)
{
   if (!a_modem->lockfile)
   {
        const char * ptr;
        char * newlockfile;
        if ((ptr = strrchr(a_modem->ttyport, '/'))) ptr++;
        else ptr = a_modem->ttyport;

        newlockfile = (char *) malloc(strlen(LOCKFILE) + strlen(ptr) + 1);
        if (newlockfile !=  NULL) {
           strcat(strcpy(newlockfile, LOCKFILE), ptr);
           a_modem->lockfile = newlockfile;
        }
        else errorExit(-1, progname, 0);
   }

   /* check TTY port lock file */
   if (CheckForLockfile(a_modem)) errorExit(-102, "Exiting - TTY lockfile exists", a_modem->lockfile);
   if ((hangup || hupmode) && CheckLockfilePath(a_modem)) errorExit(-102, "Exiting - TTY lockfile not writable", a_modem->lockfile);
}


/*
 * Open tty port; tries to make sure the open does
 * not hang if port in use, or not restored after use
 */

int openTTY(struct modems_t * a_modem)
{
    /*
     * open the modem tty device
     * O_RDWR    read/write
     * O_NOCTTY  don't become a controlling tty
     * O_NDELAY  don't hang here at the open
     */
    if ((a_modem->ttyfd = open(a_modem->ttyport, O_RDWR | O_NOCTTY | O_NDELAY)) < 0)
    {
        logMsg(LEVELE, "TTY port %s open error:  %s\n", a_modem->ttyport, strerror(errno));
        a_modem->ttyfd = -1;
        return -1;
    }

    /* change back to blocking I/O */
    if (fcntl(a_modem->ttyfd, F_SETFL, fcntl(a_modem->ttyfd, F_GETFL, 0) & ~O_NDELAY) < 0)
    {
        logMsg(LEVELE, "TTY port fcntl error:  %s\n", strerror(errno));
        return -1;
    }

    return 0;
}


/*
 * save tty settings and switch to cooked/ICANON mode
 *
 * Returns:  0 - OK
 *          -1 - error
 */
int setTTY(struct modems_t * a_modem)
{
    /* Flush port */
    if (tcflush(a_modem->ttyfd, TCIOFLUSH) < 0) return -1;

    /* Save tty port settings */
    if (tcgetattr(a_modem->ttyfd, &a_modem->otty) < 0) return -1;

    /* Configure the rtty structure for raw mode */
    a_modem->rtty.c_iflag = (IGNBRK | IGNPAR);
    a_modem->rtty.c_cflag = (CS8 | CREAD | HUPCL | CRTSCTS);
    a_modem->rtty.c_cc[VMIN] = 0;
    a_modem->rtty.c_cc[VTIME] = 1;
    if (a_modem->clocal) a_modem->rtty.c_cflag |= CLOCAL;
    if (cfsetispeed(&a_modem->rtty, a_modem->ttyspeed) < 0) return -1;
    if (cfsetospeed(&a_modem->rtty, a_modem->ttyspeed) < 0) return -1;

    /* Set the tty port for raw mode from rtty */
    if (tcsetattr(a_modem->ttyfd, TCSANOW, &a_modem->rtty) < 0) return -1;

    /* delay needed after changes for some devices */
    usleep(1000);

    /* Save the raw tty port to the ntty structure */
    if (tcgetattr(a_modem->ttyfd, &a_modem->ntty) < 0) return -1;

    /* Add the ICANON flag to the ntty structure for ICANON mode */
    a_modem->ntty.c_lflag = ICANON;

    /* Set the tty port to ICANON mode from ntty structure */
    if (tcsetattr(a_modem->ttyfd, TCSANOW, &a_modem->ntty) < 0) return -1;

    /* delay needed after changes for some devices */
    usleep(1000);

    logMsg(LEVEL1, "TTY port initialized in ICANON mode.\n");

    return 0;
}


/*
 * restore the old tty settings
 */
void restoreTTYs(void)
{
   int mit;
   for (mit = 0; mit < MAXMODEM; ++mit)
   {
      if (modem_list[mit].is_modem) {
         restoreTTY(&modem_list[mit]);
      }
   }
}

void restoreTTY(struct modems_t * a_modem)
{
    if (a_modem->ttyfd > 0)
    {
        /* restore tty parameters for serial device or modem */
        tcflush(a_modem->ttyfd, TCIOFLUSH);
        tcsetattr(a_modem->ttyfd, TCSANOW, &a_modem->otty);
    }
}


/*
 * hang up modems
 */
void hangupModems(void)
{
   int mit;
   for (mit = 0; mit < MAXMODEM; ++mit)
   {
      if (modem_list[mit].is_modem) {
         (void) initModem(HANGUP, 0, &modem_list[mit]);
      }
   }
}



/*
 * change to raw mode if raw = 1
 * change to icanon mode if raw = 0
 */

void changeTTY(int raw, struct modems_t * a_modem)
{
    if (a_modem->ttyfd < 0) return;

    if (tcflush(a_modem->ttyfd, TCIOFLUSH) < 0)
    {
        logMsg(LEVELE, "TTY port flush error:  %s\n", strerror(errno));
    }
/*canon if (!raw) {
   logMsg(LEVEL1, "TTY port left in RAW mode.\n");
   return;
}*/

    if (raw)
       tcsetattr(a_modem->ttyfd, TCSANOW, &a_modem->rtty);
    else
       tcsetattr(a_modem->ttyfd, TCSANOW, &a_modem->ntty);

    logMsg(LEVEL1, "TTY port changed to %s mode.\n", raw ? "RAW" : "ICANON");

    /* delay needed after changes for some devices */
    usleep(1000);

}


/*
 * Check modem for Call Waiting (SCID?)
 */
int checkSCID(struct modems_t * a_modem)
{
    int ret;
    const char *ptr;
    char *eptr;

    a_modem->widflag = 0;

    /* Query Snooping CID (Type II) */
    logMsg(LEVEL3, "\nQuery Snooping CID (Type II)\n");
    if (!(ret = initModem(QUERYATSCID, READTRY, a_modem)))
    {
        ptr = modembuf + 10;
        while (*ptr == '\r' || *ptr == '\n') ++ptr;
        if ((eptr = strchr(ptr, (int) '\r')))
        {
            *eptr = '\0';
            a_modem->widflag = atoi(ptr);
            logMsg(LEVEL1, "Modem Snooping CID (off = 0 on = 1 or 2): %s\n\n", ptr);
            *eptr = '\r';
        }
        else logMsg(LEVEL1, "Cannot determine if modem Snooping CID\n\n");
    }
    else
    {
        logMsg(LEVEL1, "Modem Snooping CID query failed: ret =%d\n\n", ret);
    }
         if (a_modem->widflag)
            logMsg(LEVEL1, "Modem set for CallerID and Call Waiting: AT-SCID=%d\n", a_modem->widflag);
         else
            logMsg(LEVEL1, "Modem set for CallerID.\n");
    return a_modem->widflag;
}


/*
 * Check some modem capabilities and hangup options
 * returns modemModes
 */
int checkModem(struct modems_t * a_modem)
{
   int ret;
   char countryGCI[GCISIZE], *ptr, *eptr, **gci;

   logMsg(LEVEL1, "Checking Modem Capabilities...\n");

   /* Set the tty port to raw mode */
   changeTTY(RAW, a_modem);

   /* initialize modem */
   if ((ret = initModem(a_modem->initstr, READTRY, a_modem)) == 4)
   {
      startModemLostTimer("No modem found", a_modem);
      return -1;
   }
   else
   {
      logMsg(LEVEL1, "Modem Initialized\n");
   }

   /* Query modem software */
   logMsg(LEVEL1, "\nQuery modem software\n");
   if (!(ret = initModem(QUERYATI3, READTRY, a_modem)))
   {
      ptr = modembuf + 7;
      while (*ptr == '\r' || *ptr == '\n') ++ptr;
      if ((eptr = strchr(ptr, (int) '\r')))
      {
         *eptr = '\0';
         logMsg(LEVEL1, "Modem Identifier: %s\n", ptr);
         *eptr = '\r';
      }
      else logMsg(LEVEL1, "Cannot determine Modem Identifier\n");
   }
   else
   {
      logMsg(LEVEL1, "Modem Identifier query failed: ret =%d\n", ret);
   }

   /* Query chipset version */
   logMsg(LEVEL1, "\nQuery chipset version\n");
   if (!(ret = initModem(QUERYATPV, READTRY, a_modem)))
   {
      ptr = modembuf + 7;
      while (*ptr == '\r' || *ptr == '\n') ++ptr;
      if ((eptr = strchr(ptr, (int) '\r')))
      {
         *eptr = '\0';
         logMsg(LEVEL1, "Modem Chipset Version: %s\n", ptr);
         *eptr = '\r';
      }
      else logMsg(LEVEL1, "Cannot determine Modem Chipset Version\n");
   }
   else
   {
       logMsg(LEVEL1, "Modem Chipset Version query failed: ret =%d\n", ret);
   }

   /* Query modem country setting (United States = B5) */
   logMsg(LEVEL1, "\nQuery modem country setting (United States = B5)\n");
   if (!(ret = initModem(QUERYATGCI, READTRY, a_modem)))
   {
       if ((ptr = strstr(modembuf, "GCI:")))
       {
           if (strchr(ptr, (int) ' ')) ptr += 5;
           else ptr += 4;
           memset(countryGCI, 0, GCISIZE);
           strncpy(countryGCI, ptr, 2);
           for (gci = modemGCI; *gci != NULL; ++gci)
           {
               if (!strncmp(countryGCI, *gci, 2))
               {
                   strncpy(countryGCI, *gci, GCISIZE - 2);
                   break;
                }
           }
           logMsg(LEVEL1, "Modem country code: %s\n", countryGCI);
       }
       else logMsg(LEVEL1, "Modem country code cannot be determined\n");
   }
   else
   {
       logMsg(LEVEL1, "Modem country code query failed: ret = %d\n", ret);
   }

   /* Query active profile */
   logMsg(LEVEL1, "\nQuery active profile\n");
   if (!(ret = initModem(QUERYATV, READTRY, a_modem)) || ret == 3)
   {
       if (!(ptr = strchr(modembuf, (int) '\n') + 1))
           logMsg(LEVEL1, "Cannot retrieve modem Active Profile\n");
       if ((eptr = strstr(modembuf, "OK")))
       {
           *(eptr - 2) = '\0';
           if (strstr(modembuf, "ACTIVE PROFILE"))
              logMsg(LEVEL1, "Modem %s", ptr);
           else
              logMsg(LEVEL1, "Modem Active Profile settings: %s", ptr);
       }
       else logMsg(LEVEL1, "Failed to retrieve the modem Active Profile\n");
   }
   else
   {
       logMsg(LEVEL1, "Modem active profile query failed: ret = %d\n", ret);
   }

   /* Query modem modes supported */
   logMsg(LEVEL1, "\nQuery modem modes supported \n");
   if (!(ret = initModem(QUERYATFCLASS, READTRY, a_modem)))
   {
       /* Modes start with 0 then optional mode numbers for example: 0,1,1.0,2,8 */
       if (strncmp(modembuf, "0", 1))
       {
           modemModes |= 0x1;
           logMsg(LEVEL1, "Modem supports Data Mode\n");
       }
       if (strstr(modembuf, ",1"))
       {
           modemModes |= 0x2;
           logMsg(LEVEL1, "Modem supports FAX Mode 1\n");
       }
       if (strstr(modembuf, ",2"))
       {
           modemModes |= 0x4;
           logMsg(LEVEL1, "Modem supports FAX Mode 2\n");
       }
       if (strstr(modembuf, ",8"))
       {
           modemModes |= 0x8;
           logMsg(LEVEL1, "Modem supports VOICE Mode\n");
       }
           logMsg(LEVEL1, "\n");
   }
   else
   {
       logMsg(LEVEL1, "Modem modes query failed: ret = %d\n", ret);
   }

   /* Set the tty port to ICANON mode */
   changeTTY(NOTRAW, a_modem);

   return 0;
}


/*
 * Check if hangup mode is supported by modem
 * (as returned by checkModem as modemModes)
 * if not, set mode for a normal hangup
 * Log hangup mode
 * If FAX mode, log if pickup is enabled or not
 * If voice mode, log voice parameters
 */
void checkHangupMode(int *mode, struct modems_t * a_modem)
{
    int ret;
    char msgbuf[BUFSIZ], *ptr, *eptr;

    /* if hangup option set, do checks on hangup modes  */
    if (*mode > 0)
    {
        /* determine if FAX or VOICE mode can be used, if not change it */
        switch (*mode)
        {
            case 1: /* normal hangup */
                a_modem->doesmode[1] = 1;
                break;

            case 2: /* FAX mode */
                if (!(modemModes & 0x2) && !(modemModes & 0x4))
                {
                    /* fax not supported */
                    a_modem->doesmode[2] = 0;
                    logMsg(LEVELE, "WARNING: %s using Normal Hangup (mode 1), modem does not support FAX Hangup\n",
                        mode == &hangup ? "Internal Hangup" : "Hangup Extension");
                }
                else
                {
                    a_modem->doesmode[2] = 2;
                    logMsg(LEVEL1, "Pickup %s for FAX hangup\n",
                    a_modem->pickup == 1 ? "enabled" : "not enabled");
                }
                break;

            case 3: /* VOICE mode */
                if (!(modemModes & 0x8))
                {
                    /* voice not supported */
                    a_modem->doesmode[3] = 0;
                    logMsg(LEVELE, "WARNING: %s using Normal Hangup (mode 1), modem does not support Announce Hangup\n",
                        mode == &hangup ? "Internal Hangup" : "Hangup Extension");
                }
                else
                {
                    a_modem->doesmode[3] = 3;
                    /* set raw mode */
                    changeTTY(RAW, a_modem);

                    /* Query Voice Sampling Methods (audio data formats) */
                    ret = initModem(VOICEMODE, READTRY, a_modem);
                    if (!(ret = initModem(QUERYATFMI, READTRY, a_modem)) || ret == 3)
                    {
                        if ((ptr = strstr(modembuf, "FMI")))
                        {
                            ptr += (strlen("FMI") + 3);
                            eptr = strchr(ptr, (int) '\r');
                            *(eptr + 2) = '\0'; /* skip over "\r\n" */
                            logMsg(LEVEL1, "Manufacturer: %s", ptr);
                        }
                        else logMsg(LEVEL1, "Manufacturer not found\n");
                    }
                    else logMsg(LEVEL1, "Modem AT+FMI query failed\n");
                    if (!(ret = initModem(QUERYATVSM, READTRY, a_modem)) || ret == 3)
                    {
                      if ((ptr = strstr(modembuf, "VSM")))
                      {
                        ptr = strchr(modembuf, (int) '\n') + 1;
                        eptr = strrchr(modembuf, (int) ',');
                        eptr = strchr(eptr, (int) '\r');
                        *(eptr + 2) = '\0'; /* skip over "\r\n" */
                        logMsg(LEVEL1, "Modem Voice Sampling Methods:\n%s", ptr);
                      }
                      else logMsg(LEVEL1, "Modem Voice Sampling methods not found\n");
                    }
                    else logMsg(LEVEL1, "Modem AT+VSM=? query failed\n");
                    logMsg(LEVEL1, "Modem Voice Sampling Method selected: %s\n", a_modem->audiofmt);

                    ret = initModem(DATAMODE, READTRY, a_modem);

                    /* restore ICANON mode */
                    changeTTY(NOTRAW, a_modem);
                }
                break;

            default: /* should never happen */
                logMsg(LEVELE, "Unknown %s mode (%d)\n",
                    mode == &hangup ? "Internal Hangup" : "Hangup Extension", *mode);
                errorExit(-104, 0, 0);
                break;
        }

        /* log hangup option used */
        sprintf(msgbuf, "%s Mode = ",
            mode == &hangup ? "Internal Hangup" : "Hangup Extension");
        switch (a_modem->doesmode[*mode])
        {
            case 0:
            case 1:
                strcat(msgbuf, "1: hangup");
                break;
            case 2:
                strcat(msgbuf, "2: generate FAX tones then hangup");
                break;
            case 3:
                strcat(msgbuf, "3: play an announcement then hangup");
                break;
        }
        strcat(msgbuf, " on a blacklisted call\n");
        logMsg(LEVEL1, msgbuf);
    }
    else
    {
        logMsg(LEVEL1, "%s mode = %d: disabled\n",
            mode == &hangup ? "Internal Hangup" : "Hangup Extension", *mode);
    }
}


/*
 * Configure the modem
 * returns:  0 if successful
 *          -1 if cannot read from or write to modem
 * exits program if major problem
 */
int setModem(struct modems_t * a_modem)
{
   int ret;

   if (*a_modem->initstr)
   {
      changeTTY(RAW, a_modem);
      /* Try to initialize modem */
      if ((ret = initModem(a_modem->initstr, READTRY, a_modem)))
      {
         switch (ret)
         {
            case 1: /* CONNECT */
                logMsg(LEVEL1, "Modem returned \"CONNECT\".\n");
                break;
            case 2: /* ERROR */
            case 3: /* incomplete or unexpected response from modem */
                startModemLostTimer("Unable to initialize modem", a_modem);
                return -1;
            case 4: /* no response */
                startModemLostTimer("No modem found", a_modem);
                return -1;
            case -1: /* cannot read from or write to modem */
                startModemLostTimer("Unable to initialize modem", a_modem);
                return -1;

         }
      }
      else
      {
         /* OK */
         logMsg(LEVEL1, "Modem initialized.\n");
      }
   }
   else
   {
      /* initstr is null */
      logMsg(LEVEL1, "Initialization string for modem is null.\n");
   }

   if (*a_modem->initcid)
   {
      /* try to initialize modem for CID */
      if ((ret = initModem(a_modem->initcid, READTRY, a_modem)) < 0) return -1;

      if (ret && !a_modem->setcid)
      {
         /*default CID string 1 failed, try default CID string 2 */
         a_modem->initcid = INITCID2;
         if ((ret = initModem(a_modem->initcid, READTRY, a_modem)) < 0) return -1;
      }

      if (ret)
      {
         /* CID initialization failed */
         tcsetattr(a_modem->ttyfd, TCSANOW, &a_modem->otty);
         startModemLostTimer("Unable to set modem CallerID", a_modem);
         return -1;
      }
      else
      {
         /* CID initialization succeeded */
         checkSCID(a_modem);
      }
   }
   else
   {
      /* initcid is null */
      logMsg(LEVEL1, "CID initialization string for modem is null.\n");
   }
   changeTTY(NOTRAW, a_modem);

   return 0;
}


/*
 * Initialize modem
 * expects:  ptr     - initialization string to send
 *           maxwait - 0 or more times to wait READWAIT milliseconds
 * returns:  0 modem returns OK
 *           1 modem returns "CONNECT"
 *           2 modem returns "ERROR"
 *           3 incomplete or unexpected response from modem
 *           4 no response from modem
 *           5 modem returns "NO DIALTONE"
 *           6 modem returns "NO CARRIER"
 *          -1 modem read or write error
 */
int initModem(const char * ptr, int maxwait, struct modems_t * a_modem)
{
    int num, wnum, size, midx, ret = 4;
    int readcnt, waitcnt;
    int was_idle;

    if (a_modem->ttyfd < 0) {
       logMsg(LEVEL5, "Modem fd is not open\n");
       return -1;
    }

    /* Flush port */
    if (tcflush(a_modem->ttyfd, TCIOFLUSH) < 0) {
       logMsg(LEVEL5, "Modem flush error: %s.\n", strerror(errno));
       return -1;
    }

    /* send string to modem */
    strcat(strncpy(modembuf, ptr, BUFSIZ - 2), CRLF);
    size = strlen(modembuf);
    if ((num = write(a_modem->ttyfd, modembuf, size)) < 0)
    {
      if (errno != EINTR && errno != EAGAIN)
          logMsg(LEVEL2, "Modem write error: %s.\n", strerror(errno));
      else
          logMsg(LEVEL5, "Modem write error: %s.\n", strerror(errno));
      return -1;
    }
    logMsg(LEVEL3, "Sent Modem %d of %d characters: \n%s", num, size, modembuf);
    if (verbose >= 7) hexdump(modembuf, size);

    /*
     * read until OK, CONNECT, ERROR, or "Press any key"  response detected
     * or number of tries exceeded
     */
    size = 0;
    readcnt = 0;
    waitcnt = 0;
    was_idle = 0;
    while (waitcnt < maxwait)
    {
        /* use poll to wait and see if there is something to read */
        int pret;
        struct pollfd serialfd;
        serialfd.fd = a_modem->ttyfd;
        serialfd.events = POLLIN | POLLERR | POLLHUP | POLLNVAL;
        serialfd.revents = 0;
        pret = poll(&serialfd, 1, READWAIT);
        if (pret < 0) {
           if (errno != EINTR) {
              logMsg(LEVEL2, "Modem polling error: %s.\n", strerror(errno));
              return -1;
           }
        }
        else if (pret == 0) {
           /* timed out, will allow maxwait*READWAIT milliseconds */
           waitcnt++;
           was_idle = 1;
           logMsg(LEVEL9, "Modem poll %d millisecond timeout.\n", READWAIT);
        }
        else  if (serialfd.revents & (POLLERR | POLLNVAL)) {
             logMsg(LEVEL2, "Modem polling error: ttyfd %d not open.\n", a_modem->ttyfd);
             return -1;
        }
        else if (serialfd.revents & POLLIN) {
           /* there is something to read */
           readcnt++;
           was_idle = 0;
           if ((num = read(a_modem->ttyfd, modembuf + size, BUFSIZ - size - 1)) < 0)
           {
              if (errno != EINTR && errno != EAGAIN) {
                 logMsg(LEVEL2, "Modem read error: %s.\n", strerror(errno));
                 return -1;
              }
           }
           if (num > 0) {
              /* remove \0 bytes by changing them to \r
               * in case the tcflush was not recent enough
               */
              for (midx = size; midx < size+num; midx++) {
                 if (modembuf[midx] == '\0') modembuf[midx] = '\r';
              }
           }
           size += num;
       }

       if (size > 0 && was_idle)
       {
           /* check response */
           modembuf[size] = '\0';
           if (strstr(modembuf, "OK"))
           {
               ret = 0;
               break;
           }
           else if (strstr(modembuf, "CONNECT"))
           {
               ret = 1;
               break;
           }
           else if (strstr(modembuf, "ERROR"))
           {
               ret = 2;
               break;
           }
           else if (strstr(modembuf, "NO DIALTONE"))
           {
               ret = 5;
               break;
           }
           else if (strstr(modembuf, "NO CARRIER"))
           {
               ret = 6;
               break;
           }
           else if (strstr(modembuf, "Press any key"))
           {
               do {
                  wnum = write(a_modem->ttyfd, CRLF, 2);
               } while (wnum<0 && errno == EINTR);
           }
           else
           {
               ret = 3;
           }
       }
    }
    modembuf[size] = '\0';

    if (size > 0)
    {
        logMsg(LEVEL3, "Modem response: %d characters in %d %s %d %s:\n%s",
            size,
            readcnt, readcnt == 1 ? "read" : "reads",
            waitcnt, waitcnt == 1 ? "wait" : "waits",
            modembuf);
        if (verbose >= HEXLEVEL) hexdump(modembuf, size);
    }
    else
    {
        /* maxwait can be zero to not read a modem response */
        if (maxwait)
           logMsg(LEVEL3, "No Modem Response\n");
        else
            logMsg(LEVEL3, "Skipped read for a modem response\n");
    }

    return ret;
}


/*
 * add all the modems/serial devices to polld
 */
void pollSerialDevices(void)
{
   int mit;
   int pollpos;
   for (mit = 0; mit < MAXMODEM; ++mit)
   {
      if (modem_list[mit].is_modem && modem_list[mit].ttyfd>0) {
         pollpos = addPoll(modem_list[mit].ttyfd);
         logMsg(LEVEL3,"%s %d is fd %d pos %d\n", (cidinput & 0x1) ? "Caller ID Device" : "Modem", mit, modem_list[mit].ttyfd, pollpos);

         /* link the modem_list to the IPinfo list */
         modem_list[mit].pollpos = pollpos;
         modem_list[mit].state = MODEMWAITING;
         if (pollpos>=0) {
            IPinfo[pollpos].modem = &modem_list[mit];
         }
     }
   }
}

/*
 * read from a modem
 */
void readFromModem(int pos)
{
   char buf[BUFSIZ];
   char msgbuf[BUFSIZ];
   int num;
   char * ptr;
   const char * bptr;
   struct modems_t * a_modem = findModemByPos(pos);

   if (a_modem == 0) return;
   if (!a_modem->modem_ext_locked)
   {
     /* Modem or device has data to read */
     a_modem->lidx = 0;
     a_modem->ilines[a_modem->lidx] = NULL;
     do {
       if ((num = read(a_modem->ttyfd, buf, BUFSIZ-1)) < 0)
       {
         sprintf(msgbuf, "Serial device %d pos %d read error: %s\n",
                 a_modem->ttyfd, pos, strerror(errno));
         errorExit(-112, "Fatal", msgbuf);
       }

       /* Modem or device returned no data */
       else if (!num)
       {
         a_modem->nodatacnt++;
         /* if no data 10 times in a row, something wrong */
         if (a_modem->nodatacnt >= 10)
         {
           sprintf(buf,
             "%sSerial device %d pos %d returned no data in %d tries. Terminated  %s",
             MSGLINE, a_modem->ttyfd, pos, a_modem->nodatacnt, strdate(WITHSEP));
           writeClients(buf);
           errorExit(-112, "Fatal", "Serial device returned no data");
         }
         else
         {
           logMsg(LEVEL2, "Serial device %d pos %d returned no data in try #%d %s\n", a_modem->ttyfd, pos, a_modem->nodatacnt, strdate(WITHSEP));
         }
       }
       else
       {
         /* Modem or device returned data */
         a_modem->nodatacnt = 0;

         /* Terminate String */
         buf[num] = '\0';

         /* strip <CR> and <LF> */
         if ((ptr = strchr(buf, '\r'))) *ptr = '\0';
         if ((ptr = strchr(buf, '\n'))) *ptr = '\0';

         logMsg(LEVEL5, "Serial fd %d pos %d sent %d bytes  %s\n", a_modem->ttyfd, pos, num, strdate(WITHSEP));
         writeLog(datalog, buf);

         /* remove blanks */
         bptr = buf;
         while (*bptr && isspace(*bptr)) bptr++;

         if (*bptr != '\0') {
            a_modem->ilines[a_modem->lidx] = strdup(bptr);
            /* logMsg(LEVEL5, "line %d %p\n", a_modem->lidx, a_modem->ilines[a_modem->lidx]); */
            a_modem->lidx++;
            a_modem->ilines[a_modem->lidx] = NULL;
            if (a_modem->lidx == MAXMODEMLINES-1) {
               /* pass a null terminated array of lines */
               formatCID(a_modem->ilines, pos);
               a_modem->lidx = 0;
            }
         }
       }
       if (num > 0) {
          num = 0;
          /* use poll to see if there is more */
          struct pollfd serialfd;
          serialfd.fd = a_modem->ttyfd;
          serialfd.events = POLLIN;
          serialfd.revents = 0;
          poll(&serialfd, 1, 200);
          num = (serialfd.revents & POLLIN);
       }
     } while (num > 0);
     /* pass a null terminated array of lines */
     if (a_modem->lidx > 0)
        formatCID(a_modem->ilines, pos);
   }
   else {
      logMsg(LEVEL5, "Serial fd %d pos %d is locked\n", a_modem->ttyfd, pos);
   }
}


/*
 * continue a paused serial output now that poll says the device is writable again
 */
void continueModem(int pos)
{
   /* char buf[BUFSIZ];
   char msgbuf[BUFSIZ];
   int num; */
   struct modems_t * a_modem = findModemByPos(pos);
   if (a_modem == 0) return;
   /* this isn't finished but is called at POLLOUT */
}


/*
 * a message from a modem
 *
 * Format of data from modem:
 *
 *      DATE = 0330            -or-  DATE=0330
 *      TIME = 1423            -or-  TIME=1423
 *      NMBR = 4075551212      -or-  NMBR=4075551212
 *      NAME = WIRELESS CALL   -or-  NAME=WIRELESS CALL
 *
 * kortex Modems  France Telecom:
 *      CID RING
 *
 *      CALLING MSG
 *
 *      DATE TIME=11/24 22:10
 *
 *      NBR=0632727751
 *
 *      END MSG
 *
 */
void modemCID(const char * buf, int pos)
{
    const char * ptr;
    struct cid_t * modem_cid = findModemCidByPos(pos);
    struct modems_t * a_modem = findModemByPos(pos);

    if (a_modem == 0) return;

    /*
     * At a RING
     * Could be "RING", "RING X" or "CID RING"
     *
     * US systems send Caller ID between the 1st and 2nd ring
     * Some non-US systems send Caller ID before 1st ring.
     * e.g. BT in UK, KPN in Netherlands
     *
     * If generate Caller ID set and Caller ID not received, generate
     * a generic Caller ID at RING 2.
     *
     * Clear Caller ID info between rings.
     */
    if (strncmp(buf, "RING", 4) == 0 || strncmp(buf, "CID RING", 8) == 0)
    {
        /*
         * If distinctive ring, save line indicator, it will be
         * "RING A", "RING B", etc.
         * append the letter to the lineid
         * RING A/B is discarded if the system sends CID ahead of the 1st RING
         */
         int ringcnt = restartRingTimer(pos);
         if (ringcnt == 1)
         {
             if (strlen(buf) == 6)
             {
                 strcpy(modem_cid->cidline, serialLineid(pos));
                 strcat(modem_cid->cidline, buf + 5);
             }
             else
             {
                 strcpy(modem_cid->cidline, serialLineid(pos));
             }
         }
         modem_cid->calltype = CID;

        /*
         * If ring information is wanted, send it to the clients.
         */
        sendInfo(ringcnt, modem_cid->cidline);

        if (modem_cid->status != 0)
        {
            /* caller ID has arrived before this RING */
            if (!(modem_cid->status & CIDDATE) ||
                !(modem_cid->status & CIDTIME)) {
               /* generate date and time, if missing */
               char * sptr;
               ptr = strdate(NOSEP);     /* returns: MMDDYYYY HHMM */
               for (sptr = modem_cid->ciddate; *ptr && *ptr != ' ';) *sptr++ = *ptr++;
               *sptr = '\0';
               for (sptr = modem_cid->cidtime, ptr++; *ptr;) *sptr++ = *ptr++;
               *sptr = '\0';
               modem_cid->status |= (CIDDATE | CIDTIME);
            }
            finishCID(modem_cid);
        }
        else if (a_modem->gencid && modem_cid->cidsent == 0 && ringcnt == 2 )
        {
            /*
             * gencid = 1: generate a Caller ID if it is not received
             * gencid = 0: do not generate a Caller ID
             *
             * CID information always received between before RING 2
             * no CID information received, so create one.
             */
            char * sptr;
            ptr = strdate(NOSEP);     /* returns: MMDDYYYY HHMM */
            for (sptr = modem_cid->ciddate; *ptr && *ptr != ' ';) *sptr++ = *ptr++;
            *sptr = '\0';
            for (sptr = modem_cid->cidtime, ptr++; *ptr;) *sptr++ = *ptr++;
            *sptr = '\0';
            strcpy(modem_cid->cidnmbr, "RING");
            strcpy(modem_cid->cidname, NOCID);
            modem_cid->status = (CIDDATE | CIDTIME | CIDNMBR | CIDNAME);
            /* process the generated caller ID */
            finishCID(modem_cid);
        }
        else
        {
            /*
             * CID not here yet or already processed
             * Make sure status and cidsent are zero
             */
            modem_cid->status = 0;
            modem_cid->cidsent = 0;
            return;
        }
    } /* RING */

    /*
     * A Mac Mini Motorola Jump  modem sends "^PR^PX\n\n" before the CID
     * information.  It also sends "^P.^XNAME"
     */
    else if ((ptr = strstr(buf, "\020R\020X")))
    {
        modem_cid->status = 0;
        modem_cid->cidsent = 0;
        strcpy(modem_cid->cidline, serialLineid(pos));
    }

    else if (strncmp(buf, "DATE", 4) == 0)
    {
        if ((ptr = strchr(buf, '='))) ++ptr;
        else ptr = buf + 4; /* there was no equals */
        if (*ptr == ' ') ++ptr;

        /* NetTalk sends no date but modem has line: "DATE=" */
        if (*ptr)
        {
            if (strstr(buf, "TIME"))
            {
                /* kortex Modems send
                 * DATE TIME=11/24 22:10 */
                strncpy(modem_cid->ciddate, ptr, 2);
                strncpy(modem_cid->ciddate + 2, ptr + 3, 2);
                modem_cid->ciddate[4] = '\0';
                strncpy(modem_cid->cidtime, ptr + 6, 2);
                strncpy(modem_cid->cidtime + 2, ptr + 9, 3);
                modem_cid->status |= CIDTIME;
            }
            else
            {
                /* Conexant and other modems send
                 * DATE = 0330 */
                strncpy(modem_cid->ciddate, ptr, CIDSIZE - 1);
                modem_cid->ciddate[CIDSIZE-1] = '\0';
            }

            /* need to get year */
            ptr = strdate(ONLYYEAR);        /* returns: YYYY */
            strcat(modem_cid->ciddate, ptr);       /* MMDDYYYY */

            modem_cid->status |= CIDDATE;
        }
        modem_cid->cidsent = 0;
    }
    else if (strncmp(buf, "TIME", 4) == 0)
    {
        if ((ptr = strchr(buf, '='))) ++ptr;
        else ptr = buf + 4; /* there was no equals */
        if (*ptr == ' ') ++ptr;
        if (*ptr)
        {
            /* NetTalk sends no time but modem has line: "TIME=" */
            strncpy(modem_cid->cidtime, ptr, CIDSIZE - 1);
            modem_cid->cidtime[CIDSIZE-1] = '\0';
            modem_cid->status |= CIDTIME;
        }
        modem_cid->cidsent = 0;
    }
    /*
     * Some modems send DDN_NMBR, DDN, or NBR instead of NMBR.
     * This will catch the 4 cases.
     */
    else if (!strncmp(buf, "NMBR", 4) ||
             !strncmp(buf, "DDN", 3) ||
             !strncmp(buf, "NBR", 3))
    {
        /* some telcos send NMBR = ##########, then NMBR = O to mask it */
        if (!(modem_cid->status & CIDNMBR))
        {
            if ((ptr = strchr(buf, '='))) ++ptr;
            else while (*ptr && !isblank((int) *ptr)) ++ptr; /* this should never happen */
            if (*ptr == ' ') ++ptr;
            builtinAlias(modem_cid->cidnmbr, ptr);
            modem_cid->status |= CIDNMBR;
            modem_cid->cidsent = 0;
        }

        /*
         * Check if this is a Call Waiting CID (WID).
         * If CID arrives more than 200ms before RING, or more than 6 seconds after RING
         * and modem snooping is on, it will marked as a WID.
         */
        if (a_modem->widflag == 1 && a_modem->ringcnt == 0)
        {
           logMsg(LEVEL4, "Number without RING is Call Waiting (WID)\n");
           modem_cid->calltype = WID;
        }

        /* cidnoname option for when Telco won't send a useful NAME */
        if (a_modem->cidnoname)
        {
            modem_cid->status |= CIDNAME;
            strncpy(modem_cid->cidname, NONAME, CIDSIZE - 1);
        }
    }
    /*
     * Using strstr() instead of strncmp() because a Mac
     * Mini Jump modem sent '^P.^PXNAME' instead of 'NAME'.
     * At this point the string was converted to '?.?XNAME'
     */
    else if ((ptr = strstr(buf, "NAME")))
    {
        /*
         * if cidnoname set, NAME is not wanted or
         * if NAME already received, discard the second one
         */
        if (!(a_modem->cidnoname) && !(modem_cid->status & CIDNAME))
        {
            char * sptr;
            char tempname[100];

            if (*(ptr + 4) == '=') ptr += 5;
            else ptr += 7;
            if (*ptr == ' ') ++ptr;
            /* remove any trailing spaces */
            strncpy(tempname, ptr, sizeof(tempname));
            tempname[sizeof(tempname)-1]=  '\0';
            for (sptr = tempname; *sptr; ++sptr);
            for (--sptr; *sptr && *sptr == ' '; --sptr) *sptr = '\0';

            builtinAlias(modem_cid->cidname, tempname);
            modem_cid->status |= CIDNAME;
            modem_cid->cidsent = 0;
        }
    }
    else if (strncmp(buf, "MSG_WAITING = YES", 18) == 0)
    {
           logMsg(LEVEL4, "MSG_WAITING = YES (MWI)\n");
           strcpy(modem_cid->cidname, " Message(s) Waiting");
           strcpy(modem_cid->cidnmbr, "Voicemail");
           modem_cid->status |= CIDNAME | CIDNMBR;
           modem_cid->calltype = MWI;
           modem_cid->cidsent = 0;
    }
    else if (strncmp(buf, "MSG_WAITING = NO", 17) == 0)
    {
           logMsg(LEVEL4, "MSG_WAITING = NO (MWI)\n");
           strcpy(modem_cid->cidname, "No Messages Waiting");
           strcpy(modem_cid->cidnmbr, "Voicemail");
           modem_cid->status |= CIDNAME | CIDNMBR;
           modem_cid->calltype = MWI;
           modem_cid->cidsent = 0;
    }

    else if (strncmp(buf, "MESG", 4) == 0)
    {
#ifdef HIDEMESGS
        /* minimal interpretation of the code */
        int curlen, addlen;

        ptr = buf;
        if (*(ptr + 4) == '=') ptr += 5;
        else ptr += 7;
        if (*ptr == ' ') ++ptr;
        /* Comb out some of the noise where the modem has not handled valid info.
         * The modems appear to process these
         *     040150 - P private
         *     04014F - O unavailable/out-of-area BT/Bell
         * but don't process
         *     110101 - BT voice call
         *     110102 - BT ring-back-when-free call
         *     110181 - BT message waiting call
         *     06014C - Bell Canada Long Distance
         */
        while (*ptr != '\0') {
           if (!strncmp(ptr, "110101", 6)) {
              ptr += 6;
              logMsg(LEVEL3, "110101 is a BT voice call\n");
           }
           else if (!strncmp(ptr, "110102", 6)) {
              ptr += 6;
              logMsg(LEVEL3, "110102 is a BT ring-back-when-free call\n");
           }
           else if (!strncmp(ptr, "110181", 6)) {
              ptr += 6;
              logMsg(LEVEL3, "110181 is a BT message waiting call\n");
           }
           else if (!strncmp(ptr, "06014C", 6)) {
              ptr += 6;
              logMsg(LEVEL3, "06014C is Bell Canada Long Distance\n");
           }
           /* stop if the code is unknown, though it should be code-length-content */
           else break;
        }
        addlen = strlen(ptr);
        if (addlen > 0) {
           if (!(modem_cid->status & CIDMESG))
           {
              if (CIDSIZE - addlen > 0) {
                 strcpy(modem_cid->cidmesg, ptr);
              }
              else {
                 logMsg(LEVEL3, "MESG discarded %s\n", ptr);
              }
           }
           else
           {
              /* there is already a previous code MESG */
              curlen = strlen(modem_cid->cidmesg);
              if (CIDSIZE - (curlen + addlen + 2) > 0) {
                 modem_cid->cidmesg[curlen] = ',';
                 strcpy(modem_cid->cidmesg+curlen+1, ptr);
              }
              else {
                 logMsg(LEVEL3, "MESG discarded %s\n", ptr);
              }
           }
           modem_cid->cidmesg[CIDSIZE-1] = '\0';
           modem_cid->status |= CIDMESG;
           modem_cid->cidsent = 0;
        }
#else
        /* use nciddexpandMESG to interpret the code */
        ptr = buf;
        if (*(ptr + 4) == '=') ptr += 5;
        else ptr += 7;
        if (*ptr == ' ') ++ptr;

        if (strlen(ptr) > 0)
        {
            expand_mesg(ptr, modem_cid);
        }
#endif
    }
}


/*
 * TCI serial device
 * All information received on one line
 * Only inbound calls
 * No MESG field
 *
 * Format of data from serial TCI multi-pots-line devices. Should also
 * work for serial Whozz Calling devices that have the output format
 * (usually switch 7) set to ON for TCI.
 *
 * Fixed field format, should always 70 characters per line.
 *
 *                1         2         3         4         5         6
 *      0123456789012345678901234567890123456789012345678901234567890123456789
 *      01      9/03      2:00 PM            PRIVATE
 *      02      9/03      2:15 PM            PRIVATE           PRIVATE
 *      03      9/03      2:23 PM       407-555-1212           WIRELESS CALLER
 */
void TCIserial(const char * buf, int pos)
{
    char cidbuf[BUFSIZ], tmpbuf[BUFSIZ];
    char *ptr, *sptr, *tptr;
    int i;
    struct cid_t * modem_cid = findModemCidByPos(pos);

    logMsg(LEVEL4, "Detected TCI serial device format\n");

    /* Make sure the status field and cidsent are zero */

    modem_cid->status = 0;
    modem_cid->cidsent = 0;
    /* copy data to working buffer and init pointer */
    strncpy(cidbuf, buf, BUFSIZ - 1);
    cidbuf[BUFSIZ-1] = '\0';
    ptr = cidbuf;
   /*
    * TELEPHONE LINE
    * positions: 0-1
    */
    *(ptr + 2) = '\0';
    /* if the lineid is still the default, use the line number */
    if (strcmp(serialLineid(pos), ONELINE) == 0)
    {
        strcpy(modem_cid->cidline, ptr);
    }
    else {
        /* otherwise append the line number */
        strcpy(modem_cid->cidline, serialLineid(pos));
        strcat(modem_cid->cidline, ptr);
    }

    /*
     * MONTH
     * positions 7-8
     * force leading zero
     */
    *(ptr + 9) = '\0';
    sprintf(modem_cid->ciddate, "%02d", atoi(ptr + 7));

    /*
     * DAY
     * positions 10-11
     */
    *(ptr + 12) = '\0';
    strcat(modem_cid->ciddate, ptr + 10);

    /* need to get year */
    tptr = strdate(ONLYYEAR);        /* returns: YYYY */
    strcat(modem_cid->ciddate, tptr);       /* MMDDYYYY */

    modem_cid->status |= CIDDATE;

    /*
     * HOUR
     * positions 17-18
     * AM/PM positions 23-24
     */
    *(ptr + 19) = '\0';
    i = atoi(ptr + 17);
    if (strncmp(ptr + 23, "PM", 2) == 0) { if (i < 12) i = i + 12; }
    else if (i == 12) i = 0;
    sprintf(modem_cid->cidtime, "%02d", i);

    /*
     * MINUTE
     * positions 20-21
     */
    *(ptr + 22) = '\0';
    strcat(modem_cid->cidtime, ptr + 20);
    modem_cid->status |= CIDTIME;

    /*
     * NMBR
     * positions 29-43
     * right justified (15 characters)
     */
    *(ptr + 44) = '\0';
    sptr = trimWhitespace(ptr + 29, ptr + 43);
    if (isdigit((int) *sptr))
    {
        /* remove hyphens from a number, builtinAlias() not needed */
        tptr = tmpbuf;
        while(*sptr)
        {
            if (*sptr != '-') *tptr++ = *sptr++;
            else ++sptr;
        }
        *tptr = '\0';
        strncpy(modem_cid->cidnmbr, tmpbuf, CIDSIZE - 1);
        modem_cid->cidnmbr[CIDSIZE-1] = '\0';
    }
    else builtinAlias(modem_cid->cidnmbr, sptr);
    modem_cid->status |= CIDNMBR;

    /*
     * NAME
     * positions 55-69
     * left justified (15 characters)
     */
    *(ptr + 54) = '\0';
    sptr = trimWhitespace(ptr + 55, ptr + 69);
    if (*sptr) builtinAlias(modem_cid->cidname, sptr);
    else strcpy(modem_cid->cidname, NONAME);
    modem_cid->status |= CIDNAME;
}


/*
 * USB Modem sending MDMF or SDMF messages (AT+VCID=2)
 * All information received as one HEX string
 * Inbound calls (MDMF and SDMF) and Message Waiting (MDMF)
 * No MESG field
 */
void xdmfCID(const char * buf, int pos)
{
   unsigned char ucbuf[BUFSIZ], *ptr, chksum, reclen, nmss;
   char *tptr;
   size_t msglen;
   struct cid_t * modem_cid = findModemCidByPos(pos);

   logMsg(LEVEL4, "Detected MDMF or SDMF message\n");

   bzero(ucbuf, BUFSIZ);
   hex2uchar(buf, ucbuf);
   if (verbose >= HEXLEVEL) hexdump(ucbuf, strlen(buf) / 2);

   modem_cid->status = 0;

   ptr = ucbuf;
   modem_cid->calltype = CID;
   strcpy(modem_cid->cidline, serialLineid(pos));
   modem_cid->cidsent = 0;

   chksum = checksum(ptr, (msglen = strlen(buf) / 2));
   logMsg(LEVEL4, "Checksum %s: 0x%.2x\n", chksum ? "bad" : "good", chksum);

   modem_cid->ciddate[0] = modem_cid->cidtime[0] = modem_cid->cidname[0] = modem_cid->cidnmbr[0] = '\0';

   if (*ptr == SDMF_MESG_TYPE)
   {
     logMsg(LEVEL4, "Detected SDMF message\n");

     /* SDMF date and time are at fixed byte positions */
     reclen = 8;

     /* start at second byte */
     ptr += 2;

     strncat(modem_cid->ciddate, (char *) ptr, 4);
     modem_cid->status |= CIDDATE;

     strncat(modem_cid->cidtime, (char *) ptr + 4, 4);
     modem_cid->status |= CIDTIME;

     logMsg(LEVEL4, "Date & Time: %s %s\n", modem_cid->ciddate, modem_cid->cidtime);

     /* need to get year */
     tptr = strdate(ONLYYEAR);        /* returns: YYYY */
     strcat(modem_cid->ciddate, tptr);       /* MMDDYYYY */

     ptr += reclen;

     /*
      * SDMF has NMBR only, no NAME
      * starts at a fixed position
      * length varies
      */

     strcpy(modem_cid->cidname, NONAME);
     modem_cid->status |= CIDNAME;

     reclen = msglen - reclen - 3;

     strncat(modem_cid->cidnmbr, (char *) ptr, reclen);
     modem_cid->status |= CIDNMBR;

     if (reclen == 1)
     {
       logMsg(LEVEL4, "Why No Number: %c\n", *ptr);
       switch (*ptr)
       {
          case LETTER_P:
            strcpy(modem_cid->cidnmbr, WORDFOR_P);
            break;
          case LETTER_O:
            strcpy(modem_cid->cidnmbr, WORDFOR_O);
            break;
       }
     }

     logMsg(LEVEL4, "Calling Number: %s\n", modem_cid->cidnmbr);

     logMsg(LEVEL4, "Message Checksum: 0x%.2x\n", *(ucbuf + msglen - 1));
   }
   else
   {
     logMsg(LEVEL4, "Detected MDMF message\n");

     /* Check packet ID  for call type */
     if (*ptr == MDMF_MSG_WAITING) modem_cid->calltype = MWI;

     /* start at second parameter */
     ptr += 2;

     while (*ptr != 0)
     {
       switch (*ptr)
       {
         case MDMF_VISUAL_IND:
           logMsg(LEVEL4, "Parameter 0x%.2x: US Visual Indicator\n", *ptr);
           ++ptr;
           reclen = *ptr++;
           if (*ptr) strcpy(modem_cid->cidname, "Message(s) Waiting");
           else strcpy(modem_cid->cidname, "No Messages Waiting");
           strcpy(modem_cid->cidnmbr, "Message");
           modem_cid->status = CIDNAME | CIDNMBR;
           ptr += reclen;
           break;
         case CALL_TYPE_UK:
           ++ptr;
           reclen = *ptr++;
           logMsg(LEVEL4, "Parameter 0x%.2x: UK Call Type 0x%.2x; ",
                  *(ptr - 2), *ptr);
           switch (*ptr)
           {
             case VOICE_UK:
               modem_cid->calltype = CID;
               logMsg(LEVEL4, "CallType 0x%.2x\n", modem_cid->calltype);
               break;
             case RINGBACK_UK:
               modem_cid->calltype = RID;
               logMsg(LEVEL4, "CallType 0x%.2x\n", modem_cid->calltype);
               break;
             case MSG_WAITING_UK:
               modem_cid->calltype = MWI;
               ++ptr;
               nmss = *ptr++;
               reclen = *ptr++;
               switch (*ptr)
               {
                 case 0:
                   strcpy(modem_cid->cidname, "No Messages Waiting");
                   break;
                 case 1:
                   strcpy(modem_cid->cidname, "1 Message Waiting");
                   break;
                 default:
                   sprintf(modem_cid->cidname, "%u Messages Waiting", *ptr);
                   break;
               }
               logMsg(LEVEL4, "CallType 0x%.2x; NMSS 0x%.2x\n",
                      modem_cid->calltype, nmss);
               strcpy(modem_cid->cidnmbr, "Voicemail");
               modem_cid->status = CIDNAME | CIDNMBR;
               break;
           }
           ptr += reclen;
           break;
         case MDMF_DATETIME:
           logMsg(LEVEL4, "Parameter 0x%.2x: Date & Time\n", *ptr);
           ++ptr;
           reclen = *ptr++;
           strncat(modem_cid->ciddate, (char *) ptr, 4);
           modem_cid->status |= CIDDATE;

           strncat(modem_cid->cidtime, (char *) ptr + 4, 4);
           modem_cid->status |= CIDTIME;

           /* need to get year */
           tptr = strdate(ONLYYEAR);        /* returns: YYYY */
           strcat(modem_cid->ciddate, tptr);       /* MMDDYYYY */

           ptr += reclen;
           break;
         case MDMF_CALLER_NMBR:
           logMsg(LEVEL4, "Parameter 0x%.2x: Calling Number\n", *ptr);
           ++ptr;
           reclen = *ptr++;
           if (reclen)
           {
             strncat(modem_cid->cidnmbr, (char *) ptr, reclen);
             modem_cid->status |= CIDNMBR;
           }
           ptr += reclen;
           break;
         case MDMF_WHY_NO_NMBR:
           logMsg(LEVEL4, "Parameter 0x%.2x: Why No Number\n", *ptr);
           ++ptr;
           reclen = *ptr++;
           if (*ptr == LETTER_P) strcpy(modem_cid->cidnmbr, WORDFOR_P);
           else if (*ptr == LETTER_O) strcpy(modem_cid->cidnmbr, WORDFOR_O);
           else strcpy(modem_cid->cidnmbr, NONMBR);
           modem_cid->status |= CIDNMBR;
           ++ptr;
           break;
         case MDMF_CALLER_NAME:
           logMsg(LEVEL4, "Parameter 0x%.2x: Calling Name\n", *ptr);
           ++ptr;
           reclen = *ptr++;
           if (reclen)
           {
             strncat(modem_cid->cidname, (char *) ptr, reclen);
             modem_cid->status |= CIDNAME;
           }
           ptr += reclen;
           break;
         case MDMF_WHY_NO_NAME:
           logMsg(LEVEL4, "Parameter 0x%.2x: Why No Name\n", *ptr);
           ++ptr;
           reclen = *ptr++;
           if (*ptr == LETTER_P) strcpy(modem_cid->cidname, WORDFOR_P);
           else if (*ptr == LETTER_O) strcpy(modem_cid->cidname, WORDFOR_O);
           else strcpy(modem_cid->cidname, NONAME);
           modem_cid->status |= CIDNAME;
           ++ptr;
           break;
         default:
           ++ptr;
           if (ptr == ucbuf + msglen)
           {
             logMsg(LEVEL4, "Parameter 0x%.2x: Message Checksum\n",
                    *(ptr - 1));
           }
           else
           {
             logMsg(LEVEL4, "Parameter 0x%.2x: Unknown\n", *(ptr - 1));
             /* skip parameter data */
             reclen = *ptr++;
             ptr += reclen;
           }
           break;
       }
     }

     /* make sure there is a date, time, number and name */
     if (modem_cid->cidnmbr[0] == '\0')
     {
       strcpy(modem_cid->cidnmbr, NONMBR);
       modem_cid->status |= CIDNMBR;
     }
     if (modem_cid->cidname[0] == '\0')
     {
       strcpy(modem_cid->cidname, NONAME);
       modem_cid->status |= CIDNAME;
     }
     if (modem_cid->ciddate[0] == '\0' || modem_cid->cidtime[0] == '\0')
     {
       tptr = strdate(NOSEP);
       strncat(modem_cid->ciddate, tptr, 8);
       strncat(modem_cid->cidtime, tptr + 9, 4);
       modem_cid->status |= (CIDDATE | CIDTIME);
     }
   }
}


/*
 * Check if lockfile present and has an active process number in it
 * ret == 0 if lockfile is not present or lockfile had stale process number
 * ret == 1 if lockfile present and process number active, or error
 */

int CheckForLockfile(struct modems_t * a_modem)
{
    int kret, found = 0, lockp;
    static unsigned int sentmsg = 0;
    FILE *fp;
    char lockbuf[BUFSIZ];
    struct stat statbuf;

    if (a_modem->lockfile != 0)
    {
        if (stat(a_modem->lockfile, &statbuf) == 0)
        {
            found = 1;
            if ((fp = fopen(a_modem->lockfile, "r")) == NULL)
            {
                if (!(sentmsg & 1))
                {
                    logMsg(LEVELE, "%s: %s\n", a_modem->lockfile, strerror(errno));
                    sentmsg |= 1;
                }
            }
            else
            {
                if (fgets(lockbuf, BUFSIZ - 1, fp) != NULL)
                {
                    lockbuf[BUFSIZ - 1] = '\0';
                    lockp = atoi(lockbuf);
                    if (lockp > 0)
                    {
                        /* lockfile contains a process number */
                        found = lockp;
                        if (lockp == getpid())
                           kret = 1;
                        else
                           kret = kill(lockp, 0);
                        if (kret && errno != EPERM)
                        {
                            /* the error is not permission denied */
                            if (unlink(a_modem->lockfile))
                            {
                                if (!(sentmsg & 2))
                                {
                                    logMsg(LEVELE,
                                        "Failed to remove stale lockfile: %s - %s\n",
                                        a_modem->lockfile, strerror(errno));
                                    sentmsg |= 2;
                                }
                            }
                            else
                            {
                                logMsg(LEVEL3, "Removed stale lockfile: %s\n",
                                        a_modem->lockfile);
                                found = 0;
                            }
                        }
                    }
                }
                fclose(fp);
            }
        }
        else if (errno != ENOENT)
        {
           logMsg(LEVELE, "Failed to access lockfile path: %s - %s\n",
                                        a_modem->lockfile, strerror(errno));
        }
        /* else
           logMsg(LEVELE, "Normal case, lockfile absent: %s - %s\n",
                                        a_modem->lockfile, strerror(errno)); */
    }

    /* clear all locally latched failures if lockfile no longer present */
    if (found == 0) sentmsg = 0;

    return found;
}

/*
 * touch the TTY lockfile
 * called from hangupCall_offhook()
 * returns 0 for success, or the locking pid
 */
int makeLockfile(struct modems_t * a_modem)
{
    FILE *lockptr;
    int lockpid;

    a_modem->modem_locked = 0;
    if ((lockpid = CheckForLockfile(a_modem))!=0) return lockpid;

    /* Create TTY port lockfile */
    if ((lockptr = fopen(a_modem->lockfile, "w")) == NULL)
    {
        logMsg(LEVELE, "%s: %s\n", a_modem->lockfile, strerror(errno));
        a_modem->modem_locked = 0;
        return 1;
    }
    else
    {
        /* write process number to lockfile */
        fprintf(lockptr, "%d\n", getpid());
        a_modem->modem_locked = 1;

        fclose(lockptr);
    }

    return 0;
}


/*
 * remove the TTY lockfile
 * called from hangupCall_offhook() or doHangupTimer()
 */
int removeLockfile(struct modems_t * a_modem)
{
    a_modem->modem_locked = 0;
    if (unlink(a_modem->lockfile))
    {
        logMsg(LEVELE, "Failed to remove my lockfile: %s - %s\n", a_modem->lockfile, strerror(errno));
        return 1;
    }

    return 0;
}


/*
 * test the lockfile path by trying to make a lockfile
 * called from hangupCall
 */
int CheckLockfilePath(struct modems_t * a_modem)
{
    FILE *lockptr;

    /* Create TTY port lockfile */
    if ((lockptr = fopen(a_modem->lockfile, "w")) == NULL)
    {
        logMsg(LEVELE, "%s: %s\n", a_modem->lockfile, strerror(errno));
        return 1;
    }
    else
    {
        /* write process number to lockfile */
        fprintf(lockptr, "%d\n", getpid());

        fclose(lockptr);
    }

    /* remove TTY port lockfile */
    if (unlink(a_modem->lockfile))
    {
        logMsg(LEVELE, "Failed to remove my lockfile: %s - $s\n", a_modem->lockfile, strerror(errno));
        return 1;
    }

    return 0;
}


/*
 * The modem lockfile timer checks frequently to see if a lockfile
 * has been created, probably by minicom.
 * When the lockfile exists, the modem file descriptor is closed,
 * and when the lockfile has been removed, the modem is again added to the
 * poll table.
 */
void doModemLockTimers(void)
{
    int mit = 0;
    if (lastmodemlocktimer == time(0)) return;
    lastmodemlocktimer = time(0);

    while (modem_list[mit].is_modem && mit < MAXMODEM) {
       doModemLockTimer(&modem_list[mit]);        /* or just point to the modem */
       mit++;
    }
}

void doModemLockTimer(struct modems_t * a_modem)
{
   char msgbuf[BUFSIZ];

   if (a_modem->modem_locked) return;  /* ncidd made the lockfile */

   /* TTY port lockfile */
   if (CheckForLockfile(a_modem))
   {
      if (!a_modem->modem_ext_locked)
      {
         /* lockfile just found */
         /* stop polling, but reserve the slot in polld[] */
         if (a_modem->pollpos >= 0) reservePoll(a_modem->pollpos);
         if (a_modem->ttyfd >= 0) close(a_modem->ttyfd);
         a_modem->ttyfd = -1;
         a_modem->modem_ext_locked = 1;
         a_modem->ringtimer = 0;

         /* tell clients that the modem is in use */
         sprintf(msgbuf, "TTY %s in use: releasing modem", a_modem->ttyport);
         logMsg(LEVEL1, "%s %s\n", msgbuf, strdate(WITHSEP));
         sysMesg(msgbuf, a_modem->serial_lineid);
      }
   }
   else if (a_modem->modem_ext_locked)
   {
      /* lockfile just went away */
      a_modem->modem_ext_locked = 0;

      /* tell clients that the modem is free */
      sprintf(msgbuf, "TTY %s free: using modem again", a_modem->ttyport);
      logMsg(LEVEL1, "%s %s\n", msgbuf, strdate(WITHSEP));
      sysMesg(msgbuf, a_modem->serial_lineid);

      /* restore tty poll events */
      reAttachModem(a_modem);
   }
}


/*
 * return the index of the serial device for this polld position
int findModemIdxByPos(int pos);
int findModemIdxByPos(int pos)
{
//Ed this should be sensible when not a modem
   return IPinfo[pos].modem_idx;
}
 */


/*
 * return the index of the serial device for this polld position
 */
struct cid_t * findModemCidByPos(int pos)
{
   if (IPinfo[pos].modem)
      return IPinfo[pos].modem->modem_cid;
   else
      return 0;
}


/*
 * return the index of the serial device for this polld position
 */
struct modems_t * findModemByPos(int pos)
{
   return IPinfo[pos].modem;
}


/*
 * return the serial device line ID for this polld position
 */
const char * serialLineid(int pos)
{
   if (IPinfo[pos].modem)
      return IPinfo[pos].modem->serial_lineid;
   else
      return ONELINE;
}



/*
 * return a modem matching this lineid
 */
struct modems_t * findModemByLineid(const char * lineid)
{
    int mit = 0;
    /* find a modem matching the lineid */
    while (modem_list[mit].is_modem && mit < MAXMODEM) {
        if (strcmp(modem_list[mit].serial_lineid, lineid)==0) {
            if (modem_list[mit].ttyfd >= 0)
               return &modem_list[mit];
            else
               logMsg(LEVEL3, "Modem %s for lineid \"%s\" is offline\n", modem_list[mit].ttyport, lineid);
        }
        mit++;
    }

    logMsg(LEVEL3, "No modem found for lineid \"%s\"\n", lineid);
    return NULL;
}

/* find the corresponding modem and dial the number
 * returns the index of the modem used
 * or a NULL with text in failreason
 */
struct modems_t *dialNumber(int dialclientpos, const char * number, const char * lineid, char * failreason)
{
    int mret;
    char msgbuf[BUFSIZ];
    int mit;
    struct modems_t * a_modem = NULL;

    /* find a modem matching the lineid */
    for (mit = 0; mit < MAXMODEM; ++mit)
    {
        if (modem_list[mit].is_modem && strcmp(modem_list[mit].serial_lineid, lineid)==0)
        {
            a_modem = &modem_list[mit];
            logMsg(LEVEL2, "DIAL lineid \"%s\" found\n", lineid);
            break;
        }
    }
    if (a_modem == NULL)
    {
        logMsg(LEVEL2, "DIAL lineid \"%s\" not found\n", lineid);
        /* or just take the first one */
        for (mit = 0; mit < MAXMODEM; ++mit)
        {
            if (modem_list[mit].is_modem)
            {
                a_modem = &modem_list[mit];
                break;
            }
        }
    }

    if (a_modem != NULL && a_modem->hanguptimer == 0)
    {
        /* found a modem and not already in use */
        initModem("AT", READTRY, a_modem);

        sprintf(msgbuf, "ATDT%s;%s", number, CRLF);
        mret = initModem(msgbuf, READTRY, a_modem);

        if (mret == 0)
        {
            /* start timer to hang up after dialdelay seconds */
            a_modem->hanguptimer = time(0) + dialdelay;
            a_modem->dialclientpos = dialclientpos;
            a_modem->state = MODEMDIALLED;
            return a_modem;
        }
        else
        {
            initModem(HANGUP, READTRY, a_modem);
            switch (mret)
            {
                case 4:
                    strcpy(failreason, "NO RESPONSE");
                    break;
                case 5:
                    strcpy(failreason, "NO DIALTONE");
                    break;
                case -1:
                    strcpy(failreason, "In USE");
                    break;
                default:
                    sprintf(failreason, "error code %d", mret);
                    break;
            }
            return NULL;
        }
    }
    else {
        logMsg(LEVEL2, "DIAL lineid \"%s\" already in use\n", lineid);
        sprintf(failreason, "DIAL lineid \"%s\" already in use\n", lineid);
        return NULL;
    }
}


/*
 * abort a DIAL in progress
 */
void dialAbort(int dialclientpos)
{
    /* find the modem where dialclientpos matches the client pos
     *  and a hangup timer is running
     */
    struct modems_t * a_modem = NULL;
    int mit;
    for (mit = 0; mit < MAXMODEM; ++mit)
    {
        if (modem_list[mit].is_modem &&
            (dialclientpos == modem_list[mit].dialclientpos) &&
            (modem_list[mit].hanguptimer != 0)) {
            a_modem = &modem_list[mit];
            break;
        }
    }

    if (a_modem != NULL && a_modem->hanguptimer)
    {
        a_modem->hanguptimer = 1;      /* make it go off right away */
        doHangupTimer(a_modem);
        a_modem->dialclientpos = -1;
    }
    else {
        char msgbuf[BUFSIZ];
        /* no call in progress */
        sprintf(msgbuf, DIAL_HFAIL CRLF);
        safeSend(dialclientpos, msgbuf);
    }
}


/*
 * Hangup Call -- go off hook
 * return = 0 modem is off hook
 * return != 0 error
 */
int hangupCall_offhook(int mode, struct modems_t * a_modem)
{
    int ret = 0, pret, lockpid;

    if (a_modem == NULL) return 1;
    if (! a_modem->is_modem) return 1;
    if (a_modem->ttyfd < 0) {
       logMsg(LEVELE, "Unable to go off hook, modem %s not connected\n", a_modem->ttyport);
       return 1;
    }

    /* try to create the tty lock file.
       if it's already present, don't try to use the modem */
    if ((lockpid = makeLockfile(a_modem))!=0) {
       logMsg(LEVELE, "Unable to lock modem for hangup mode %d, locked by %d\n",mode, lockpid);
       return 1;
    }


    /* downgrade the hangup mode if the modem doesn't support it */
    if (!a_modem->doesmode[mode]) mode = 1;

    /* put tty port in raw mode */
    if (!(ret = tcsetattr(a_modem->ttyfd, TCSANOW, &a_modem->rtty) < 0))
    {
        /* Send AT to get modem OK after switch to raw mode */
        ret = initModem(GETOK, HANGUPTRY, a_modem);
        if (ret != 0) {
           startModemLostTimer("Modem is not responding", a_modem);
           return 1;
        }

        if (mode == 1)    /* hangup mode */
        {
            /* Pick up the call */
            ret = initModem(PICKUP, HANGUPTRY, a_modem);
            logMsg(LEVEL5, "hangup mode %d: PICKUP sent, return code is %d\n",mode,ret);
            a_modem->state = MODEMHANGUP;
        }
        else if (mode == 2)   /* FAX hangup mode */
        {
            /* Set FAX mode */
            ret = initModem(FAXMODE, HANGUPTRY, a_modem);    /* AT+FCLASS=1 */
            logMsg(LEVEL5, "hangup mode %d: FAXMODE sent, return code is %d\n",mode,ret);
            a_modem->state = MODEMANNOUNCING;

            if (a_modem->pickup)
            {
                /* PICKUP required for some (mostly USB) modems */
                ret = initModem(PICKUP, HANGUPTRY, a_modem);
                logMsg(LEVEL5, "hangup mode %d: PICKUP sent, return code is %d\n",mode,ret);
                if (ret != 0) startModemLostTimer("Modem is not responding", a_modem);
            }
        }
        else /* Voice Hangup Mode (mode == 3) */
        {
            /*
             *  Set Voice Mode:             AT+FCLASS=8  (return == 0) OK
             *  speaker volume normal:      AT+VGT=128   (return == 0) OK
             *  Select Compress Method:     AT+VSM=128   (return == 0) OK
             *  Answer Call:                AT+VLS=1     (return == 0) OK
             *  Select Voice Transfer Mode: AT+VTX       (return == 1) CONNECT
             *  send voice data to modem
             *  end of voice data:          <DLE><ETX>   (return == 0) OK
             *  Go on hook in data mode:   ATH          (return == 0) OK
             *
             *  NOTE: AT+VSM=128 is the default used for the CONEXANT
             *        CX93001 chipset used in a lot of modems.
             */

            ret = initModem(VOICEMODE, HANGUPTRY, a_modem);  /* AT+FCLASS=8 */
            logMsg(LEVEL5, "hangup mode %d: VOICEMODE sent, return code is %d\n",mode,ret);
            a_modem->state = MODEMANNOUNCING;

            ret = initModem(SPKRVOL, HANGUPTRY, a_modem);    /* AT+VGT=128 */
            ret = initModem(a_modem->audiofmt, HANGUPTRY, a_modem);   /* AT+VSM=130 */
            ret = initModem(ANSCALL, HANGUPTRY, a_modem);    /* AT+VLS=1   */
            ret = initModem(NOECHO, HANGUPTRY, a_modem);     /* Echo Off   */
            ret = initModem(VOICEXFR, HANGUPTRY, a_modem);   /* AT+VTX     */
            if (ret == 1 ) ret = 0; /* returned CONNECT */
        }
    }
    else {
       logMsg(LEVELE, "Unable to set raw mode on modem - %s\n", strerror(errno));
    }

    if (ret)
    {
        /* unable to go off-hook, try to hang up and leave */
        pret = initModem(HANGUP, HANGUPTRY, a_modem);
        logMsg(LEVEL5, "hangup mode %d: HANGUP sent, return code is %d\n", mode, pret);

        /* if Hangup does not indicate OK try 1 more time  */
        if (pret != 0)
        {
            /* Send AT to get modem OK */
            pret = initModem(GETOK, HANGUPTRY, a_modem);
            logMsg(LEVEL5, "hangup mode %d try 2: GETOK sent, return code is %d\n", mode, pret);

            /* Send HANGUP to get modem OK */
            pret = initModem(HANGUP, HANGUPTRY, a_modem);
            logMsg(LEVEL5, "hangup mode %d try 2: HANGUP sent, return code is %d\n", mode, pret);
        }

        if (mode != 1)    /* FAX or VOICE hangup mode */
        {
            /* set data mode */
            pret = initModem(DATAMODE, HANGUPTRY, a_modem);
            logMsg(LEVEL5, "hangup mode %d: DATAMODE sent, return code is %d\n", mode, pret);
        }

        /* take tty port out of raw mode */
        (void) tcsetattr(a_modem->ttyfd, TCSAFLUSH, &a_modem->ntty);


        /* remove TTY port lockfile */
        removeLockfile(a_modem);
        a_modem->state = MODEMWAITING;
    }

    return ret;
}

/*
 * 1- go off-hook for a while
 * 2- send some fax tones
 * 3- do a voice announcement
 * then sleep for a hangupdelay before sending "ATH0"
 */
int hangupCall(int mode, const char * recording, struct modems_t * a_modem)
{
    int ret = 0, rcnt, wcnt, tcnt = 0;
    unsigned int hangupdelay;
    char voicebuf[BUFSIZ];

    (void) wcnt;

    if (a_modem == NULL) return 1;
    if (! a_modem->is_modem) return 1;

    /* defer the ring timer until after hangup finishes */
    if (a_modem->ringtimer) a_modem->ringtimer = time(0) + 180;

    if (mode == 1)    /* hangup mode */
    {
        /* set off-hook delay to make sure to hangup call */
        hangupdelay = HANGUPDELAY;
    }
    else if (mode == 2)   /* FAX hangup mode */
    {
        /* generate FAX tones */
        ret = initModem(FAXANS, HANGUPTRY, a_modem);
        logMsg(LEVEL5, "hangup mode %d: FAXANS sent, return code is %d\n", mode, ret);

        /* set off-hook delay so caller hears annoying fax tones */
        hangupdelay = FAXDELAY;
    }
    else /* Voice Hangup Mode (mode == 3) */
    {
        /*  send voice data to modem
         *  end of voice data:          <DLE><ETX>   (return == 0) OK
         *  Go on hook in data mode:   ATH          (return == 0) OK
         */

        /* send voice file */
        if ((a_modem->msgfd = open(recording, O_RDONLY)) == -1)
        {
            logMsg(LEVELE, "%d: %s\n", a_modem->msgfd, strerror(errno));
            ret = initModem(DLEETX, VOICETRY, a_modem); /* <DLE><ETX> */
            hangupdelay = HANGUPDELAY;
        }
        else
        {
            /*
             * rcnt == BUFSIZ or remaining bytes of file at end
             */
            logMsg(LEVEL3, "Sending file %s, to modem in %d byte chunks\n", recording, BUFSIZ);
            do
            {
                rcnt = read(a_modem->msgfd, voicebuf, BUFSIZ);
                if (rcnt > 0 ) wcnt = write(a_modem->ttyfd, voicebuf, rcnt);
                tcnt = tcnt + rcnt;
            }
            while (rcnt > 0);
            close(a_modem->msgfd);
            logMsg(LEVEL3, "Sent %d bytes to modem for announcement\n", tcnt);
            ret = initModem(DLEETX, VOICETRY, a_modem); /* <DLE><ETX> */

            /* set delay after announcement */
            hangupdelay = VOICEDELAY;
        }
    }

    /* off-hook delay for all hangup modes */
    /* start timer to hang up after hangupdelay seconds */
    a_modem->hanguptimer = time(0) + hangupdelay;
    logMsg(LEVEL3, "staying off-hook for %d %s\n", hangupdelay, hangupdelay == 1 ? "second" : "seconds");

    return ret;
}


/*
 * read the config file for a single modem or serial device
 */

void configModem(struct modems_t * a_modem)
{
    struct setword serial_setword[] = {
        /* char *word; int type; char **buf; int *value; int min; int max */
        {"audiofmt",   WORDSTR,  &a_modem->audiofmt,  0,          0,       0},
        {"cidnoname",  WORDNUM,  0,          &a_modem->cidnoname,   OFF,     ON},
        {"gencid",     WORDNUM,  0,          &a_modem->gencid,    OFF,     ON},
        {"initcid",    WORDSTR|WORDFLAG, &a_modem->initcid, &a_modem->setcid, 0,    0},
        {"initstr",    WORDSTR,  &a_modem->initstr,   0,          0,       0},
        {"lineid",     WORDSTR,  &a_modem->serial_lineid,    0,          0,       0},
        {"lockfile",   WORDSTR,  &a_modem->lockfile,  0,          0,       0},
        {"pickup",     WORDNUM,  0,          &a_modem->pickup,    OFF,     ON},
        {"ttyclocal",  WORDNUM,  0,          &a_modem->clocal,    OFF,     ON},
        {"ttyport",    WORDSTR,  &a_modem->ttyport,   0,          0,       0},
        {"ttyspeed",   WORDSTR,  &a_modem->TTYspeed,  0,          0,       0},
        {0,            0,        0,          0,          0,       0}
    };

    int ret;
    char aliasbuf[CIDSIZE];

    ret = resetConfEntries((struct setword *)&serial_setword);
    if (ret != 0) errorExit(-104, 0, 0);

    /* the new conf file name was set before doConfig was called
       extern const char * config; */
    logMsg(LEVEL1, "reading serial line config file %s\n", config);

    /* read the conf file */
    if ((ret = doConf()))
    {
        if (ret > 0) errorExit(-104, 0, 0);
        else errorExit(-1, progname, 0);
    }

    /* fill in parts that were not in the config */

    /* these were filled in by the config */
    /* a_modem->ttyfd;        the modem file descriptor */
    /* a_modem->ttyport        the modem device name */
    /* a_modem->serial_lineid        the lineid */
    /* a_modem->TTYspeed         */
    /* a_modem->ttyspeed         */
    /* a_modem->lockfile         the modem lock file name */
    /* if (a_modem->audiofmt == NULL) a_modem->audiofmt = XPRESS; */
    if (a_modem->ttyport == NULL) {
        errorExit(-108, "A ttyport device name must be configured", config);
    }
    if (a_modem->serial_lineid == NULL) {
        /* use the device name when lineid was not in the config file */
        const char * ptr = strrchr(a_modem->ttyport, '/');
        if (ptr) ptr++;
        else ptr = a_modem->ttyport;

        a_modem->serial_lineid = strdup(ptr);
    }

    /* alias the lineid */
    strncpy(aliasbuf, a_modem->serial_lineid, CIDSIZE -1);
    aliasbuf[CIDSIZE-1] = '\0';
    userAlias(0, 0, aliasbuf);
    a_modem->serial_lineid = strdup(aliasbuf);

    logMsg(LEVEL1, "Telephone Line Identifier: %s\n", a_modem->serial_lineid);

    clear_modem_cid(a_modem);

    /* append the lineid to LineIDS */
    sprintf(aliasbuf, " \"%s\"", a_modem->serial_lineid);
    strcat(LineIDS, aliasbuf);

}


/*
 * add more serial devices from the addedmodems list
 */
void addMoreDevices(void)
{
    struct stat buf;
    u_short entryType;
    char entryname[BUFSIZ];
    struct modems_t * a_modem = NULL;

    /* config is the complete pathname to ncidd.conf */
    char * configpath;
    char * nextconfig;
    char * saveptr;
    const char * mainconfig = config;

    char whitespace[] = " \n\t\r";
    if (addedmodems)
    {
        char more[strlen(addedmodems)+1];

        /* get the directory for ncidd.conf */
        configpath = strdup(mainconfig);
        char * confdir = strrchr(configpath, '/');
        confdir = confdir ? confdir + 1 : configpath;
        *confdir = '\0';
        /* make the addedmodems list writeable */
        strcpy(more, addedmodems);

        nextconfig = strtok_r(more, whitespace, &saveptr);
        while (nextconfig && next_modem < MAXMODEM)
        {
            logMsg(LEVEL1, "Adding Modem: %s\n", nextconfig);
            /* if the entry is not xx.conf, skip it */
            const char * tail = strrchr(nextconfig, '.');
            if ((tail==NULL) || (strcmp(tail, ".conf")!=0)) {
                logMsg(LEVEL1, "Tossed: %s\n", nextconfig);
                nextconfig = strtok_r(NULL, whitespace, &saveptr);
                continue;
            }
            sprintf(entryname, "%s%s", configpath, nextconfig);
            logMsg(LEVEL1, "entryname: %s\n", entryname);

            /* make sure it's a file or link */
            if(stat(entryname, &buf)==0) {
                entryType = buf.st_mode;
                entryType = entryType & 0170000;
                /* printf("%d %s\n",buf.st_blocks,entryname); */
                switch (entryType) {
                    case S_IFREG: /* regular file */
                        a_modem = &modem_list[next_modem];
                        a_modem->my_mit = next_modem;
                        /* set the config file name for configModem */
                        config = entryname;

                        /* fill in the empty parts */
                        init_modem(a_modem);
                        /* remove the default lineid and ttyport */
                        a_modem->ttyport = NULL;
                        a_modem->serial_lineid = NULL;
                        /* fill in specific parts from the config */
                        configModem(a_modem);

                        /* open the device and ask it questions */
                        openandCheckSerial(a_modem);

                        next_modem++;
                        break;

                    case S_IFLNK: /* symbolic link */
                    case S_IFDIR: /* directory */
                    case S_IFCHR: /* character special */
                    case S_IFBLK: /* block special */
                    case S_IFSOCK: /* socket */
                    default:
                        break;
                }
            }
            else {
                /* exit on config errors */
                logMsg(LEVELE, "Unable to find added modem %s: %s\n", entryname, strerror(errno));
                errorExit(-104, 0, 0);
            }

            nextconfig = strtok_r(NULL, whitespace, &saveptr);
        }
        config = mainconfig;
    }

    if (addedmodems)
        logMsg(LEVEL1, "Added Modems: %s\n", addedmodems);
}
