/*
 * nciddmodems.h - This file is part of ncidd.
 *
 * Copyright (c) 2005-2024
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _NCIDDMODEMS_H
#define _NCIDDMODEMS_H
#include <termios.h>

#define MAXMODEMLINES  20 /* largest burst of lines from a modem */
#define MAXMODEM       5 /* max modems in modem_cid and modem_list */

enum modemstate_t { MODEMWAITING, MODEMINITIALIZING, MODEMHANGUP, MODEMANNOUNCING, MODEMDIALLED, MODEMSTATEMAX };

struct modems_t
{
    int        is_modem;
    int        my_mit;
    int        ttyfd; /* the modem file descriptor */
    const char *      ttyport;       /* the modem device name */
    const char *      serial_lineid; /* the lineid */
    const char *      initstr;
    const char *      initcid;
    const char *      TTYspeed;
    const char *      audiofmt;
    int               cidnoname;
    int               ttyspeed;
    int               pickup;
    int               setcid;
    int               gencid;
    int               modem_locked;   /* modem locked by ncidd for hangup */
    int               modem_ext_locked; /* modem locked by another process */
    int               modemType;
    const char *      lockfile;       /* the modem lock file name */
    int               pollpos;        /* remember modem pollfd during lock */
    int               checked_hangup_mode;
    unsigned char     doesmode[4];    /* is a hangup/hupmode supported */
    int               nodatacnt;
    int               nodatalowest;
    int               modemlostcount;
    int               clocal;
    struct termios    otty, rtty, ntty;
    int               ringcnt;        /* one ringie-dingie ... */
    int               widflag;        /* modem supports call waiting (WID) */
    time_t            ringtimer;      /* to detect end of ringing */
    time_t            modemlosttimer; /* to try modem init again */
    time_t            hanguptimer;
    int               dialclientpos;  /* the client that asked for DIAL */
    int               msgfd;
    struct cid_t *    modem_cid;
    int               lidx;
    char *            ilines[MAXMODEMLINES]; /* a burst of lines sent by the modem */
    enum modemstate_t state;
};
extern struct modems_t    modem_list[MAXMODEM];

#define READWAIT    100     /* milliseconds */
#define READTRY     60      /* number of times to EEADWAIT for a character 6 seconds */
#define TTYSPEED    B115200
#define RAW         1
#define NOTRAW      0

#define INITSTR       "AT Z S0=0 E1 V1 Q0"
#define INITCID1      "AT+VCID=1"
#define INITCID2      "AT#CID=1"
#define QUERYATI3     "ATI3"
#define QUERYATGCI    "AT+GCI?"
#define QUERYATFCLASS "AT+FCLASS=?"
#define QUERYATVSM    "AT+VSM=?"
#define QUERYATFMI    "AT+FMI"
#define QUERYATV      "AT&V"
#define QUERYATPV     "AT-PV"
#define QUERYATSCID   "AT-SCID?"

#define RINGWAIT     8       /* number of seconds to wait for another RING */
#define GCISIZE      50      /* array size used for countryGCI[] */
#define MODEMWAIT    45      /* number of seconds to wait before trying init again */

#define NOECHO      "ATE0"
#define PICKUP      "ATH1"
#define HANGUP      "ATH0"
#define DATAMODE    "AT+FCLASS=0"

#define FAXANS      "ATA"
#define FAXMODE     "AT+FCLASS=1"
#define FAXDELAY    10   /* seconds */

#define VOICEMODE   "AT+FCLASS=8"
#define SPKRVOL     "AT+VGT=128"

/*
 * Some chipsets take 1 argument, some 2 arguments and some 4 arguments
 * for the voice data formats shown with AT+VSM=?
 * each () indicates an additional argument
 * 4 arguments can always be given, modems seem to ignore extra arguments
 */
#define XPRESS      "AT+VSM=130" // 130,"UNSIGNED PCM",8,0,8000,0,0

#define ANSCALL     "AT+VLS=1"
#define VOICEXFR    "AT+VTX"
#define DLEETX      "\x10\x03"
#define VOICETRY    300 /* number of times to try DLEETX (30 seconds max) */
#define VOICEDELAY  2   /* delay before hangup in seconds */

#define GETOK       "AT"
#define HANGUPTRY   8   /* number of times to try to hangup (0.8 seconds max) */
#define HANGUPDELAY 2   /* delay before hangup in seconds */




void initSerialDevices(void);
void pollSerialDevices(void);

void doModemLockTimers(void);
void doHangupTimers(void);
void doRingTimers(void);
void doModemLostTimers(void);
void startModemLostTimer(const char * msg, struct modems_t * a_modem);

void restoreTTYs(void);
void hangupModems(void);

const char * serialLineid(int pos);

struct cid_t *        findModemCidByPos(int pos);

struct modems_t *    findModemByPos(int pos);
struct modems_t *    findModemByLineid(const char * lineid);

void init_modem(struct modems_t * a_modem);
void set_ttyspeed(struct modems_t * a_modem);


void readFromModem(int pos);
void continueModem(int pos);

int hangupCall_offhook(int mode, struct modems_t * a_modem);
int hangupCall(int mode, const char * recording, struct modems_t * a_modem);
struct modems_t * dialNumber(int dialclientpos, const char * number, const char * lineid, char * failreason);
void dialAbort(int pos);

#endif /* nciddmodems.h */
