/*
 * cidupdate.c - This file is part of ncidd.
 *
 * developed by hacking ncidd.c
 *
 * Copyright (c) 2021
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef __cplusplus
    extern "C" {
#endif
#include "ncidd.h"
#ifdef __cplusplus
       }
#endif
#include "../server/nciddPN.h"

/* globals */
const char * cidlog   = CIDLOG;
char newcidlog[BUFSIZ];

char * progname;
char yesno[BUFSIZ];
char * yesnop;
int debug;
FILE * logptr;
FILE * cidfile;
int hangup;
int hupmode;
int verbose = 0;        /* below the usual default to get only errors and responses */
int ignore1;
int do_multi = 0;
int do_rename = 0;        /* the interactive query can be skipped with -y */

const char * country  = DEFAULT_COUNTRY;
const char * language = DEFAULT_LANGUAGE;
char new_fnmbr[BUFSIZ];


/* this is filled by breakLINE3() to return values from CID: PID: HUP: etc */
static struct breakline_t
{
    const char * part[4];
    char field[4][BUFSIZ];
    char * name;
    char * nmbr;
    char * line;
    int fnmbr;
} breakline;


/* this is a reduced set of the options in ncidd.conf, allowed by ignoreConfWord=1 */
struct setword setword[] = {
    /* char *word; int type; char **buf; int *value; int min; int max */
    {"blacklist",  WORDSTR,  &blacklist, 0,          0,       0},
    {"cidalias",   WORDSTR,  &cidalias,  0,          0,       0},
    {"cidlog",     WORDSTR,  &cidlog,    0,          0,       0},
    {"hangup",     WORDNUM,  0,          &hangup,    0,       3},
    {"hupmode",    WORDNUM,  0,          &hupmode,   0,       3},
    {"ignore1",    WORDNUM,  0,          &ignore1,   OFF,     ON},
    {"regex",      WORDNUM,  0,          &regex,     0,       2},
    {"whitelist",  WORDSTR,  &whitelist, 0,          0,       0},
    {"country",    WORDSTR,  &country,   0,          0,       0},
    {"language",   WORDSTR,  &language,  0,          0,       0},
    {"nanp_format",  WORDNUM, 0,         &nanp_format, 0,     3},
    {0,            0,        0,          0,          0,       0}
};


/* associate variables for config file "send cidlog" and "send cidinfo"
 * also allowed to be empty by ignoreConfWord=1 */
struct sendclient sendclient[] = {
    /* char *word; int *value */
    {  0,          0          }
};

/* All line labels of interest to a client, new types must be added */
/* sorted alphabetically here for readability */
const char *lineTags[] =
{
    BLKLINE,
    CIDLINE,
    ENDLINE,
    HUPLINE,
    MSGLINE,
    MWILINE,
    NOTLINE,
    OUTLINE,
    PIDLINE,
    PUTLINE,
    RIDLINE,
    RLYLINE,
    WIDLINE,
    NULL
};

void processLog(const char * cidlog);
void applyWhiteBlack(const char * cidnmbr, char * cidname);
void openLog(const char * logf);
void writeLog(const char * logbuf);
void cleanup(void);
void breakLINE3(const char * bufptr);
void repair_fnmbr(void);
void normalExit(void);


void errorExit(int error, const char * msg, const char * arg);
int getOptions(int argc, char *argv[]);


#define CIDUPUSAGE       "\
Usage:\n\
   cidupdate [--help|-h] [--version|-V]\n\
   cidupdate [options]\n\
\n\
Note: The defaults for these options will be obtained\n\
from the ncidd.conf file (except for -C -M -v and -y)\n\
\n\
Options: [-a <file>       | --aliasfile <file>]\n\
         [-B <file>       | --blacklist <file>]\n\
         [-C <file>       | --config <file>]\n\
         [-c <file>       | --cidlog <file>]\n\
         [-H <0|1|2|3>    | --hangup <0|1|2|3>]\n\
         [-h              | --help]\n\
         [-i              | --ignore1]\n\
         [-M              | --multi]\n\
         [-r <0|1|2>      | --regex <0|1|2>]\n\
         [-V              | --version]\n\
         [-v <1-9>        | --verbose <1-9>]\n\
         [-W <file>       | --whitelist <file>]\n\
         [-y              | --yes]\n\
"
#define CIDUPOPTIONS "\
\n\
  Where:\n\
    -h, --help\n\
           Displays the help message and exits.\n\
\n\
    -v, --verbose <1-9>\n\
           Output information while processing is occurring. Set the level\n\
           to a higher number for more information. Levels range from 1 to\n\
           9, but not all levels are used.\n\
           Default: verbose = 0\n\
\n\
    -V, --version\n\
           Displays the version and exits.\n\
\n\
    -a <file>, --alias <file>\n\
    -A <file>, --aliasfile <file>\n\
           Set the alias file to <file>\n\
           Default: from ncidd.conf (usually /etc/ncid/ncidd.alias)\n\
\n\
    -B <file>, --blacklist <file>\n\
           Set the blacklist file to <file>\n\
           Default: from ncidd.conf (usually /etc/ncid/ncidd.blacklist)\n\
\n\
    -W <file>, --whitelist <file>\n\
           Set the whitelist file to <file>\n\
           Default: from ncidd.conf (usually /etc/ncid/ncidd.whitelist)\n\
\n\
    -H <0|1|2|3>, --hangup <0|1|2|3>\n\
           When non-zero, use the whitelist and blacklist files to obtain\n\
           the #= match names\n\
           Default: from ncidd.conf\n\
\n\
    -c <file>, --cidlog <file>\n\
           Set the call log file to <file>\n\
           Default: /var/log/cidcall.log\n\
\n\
    -C <file>, --config <file>\n\
           Get the default options from <file>\n\
           Default: /etc/ncid/ncidd.conf\n\
\n\
    -i, --ignore1\n\
           This is a US/Canada option only.\n\
\n\
           A leading one in an alias definition and in the calling number is\n\
           ignored.\n\
\n\
           Normally an alias requires the calling number as it appears in\n\
           the cidcall.log. In the US a leading 1 may or may not be provided\n\
           in incoming or outgoing calls.\n\
\n\
           Default: The leading 1 must be in the alias if it is in the\n\
           calling number.\n\
\n\
    --multi\n\
           Updates the current cidcall.log file and previous call files\n\
           numbered cidcall.log.1 to cidcall.log.9 \n\
           Default: Updates only one cidcall.log file\n\
\n\
    -r, --regex <0-2>\n\
           If regex = 0, uses Simple Expressions.\n\
           If regex = 1, uses Posix Regular Expressions.\n\
           If regex = 2, uses Perl-compatible Expressions.\n\
           Default: from ncidd.conf (where the default is 0)\n\
\n\
    -y, --yes\n\
           Skips the interactive prompt for overwriting the old call log\n\
           with the generated new one\n\
"

/*
 *  The main program
 */

int main(int argc, char *argv[])
{
    int argind, i, ret;
    struct stat statbuf;
    char msgbuf[BUFSIZ];


    /* global containing name of program */
    progname = strrchr(argv[0], (int) '/');
    progname = progname ? progname + 1 : argv[0];

    /* Get the default ncidd config file,
     * allow file to have options which are not in setword
     */
    config = CONFIG;
    ignoreConfWord = 1;

    /* process options from the command line */
    argind = getOptions(argc, argv);

    /* should not be any arguments */
    if (argc - argind != 0)
    {
        fprintf(stderr, NOOPT, progname, argv[argind]);
        fprintf(stderr, CIDUPUSAGE);
        exit(0);
    }

    /* make logMsg write to stdout */
    logptr = fdopen(STDOUT, "a");;

    /* note that verbose is set to 0 so these need -v 1 option to be seen */
    logMsg(LEVEL1, "Started: %s\nCompiled %s %s\n",strdate(WITHSEP),
            progname, VERSION);

    /* log command line and any options on separate lines */
    sprintf(msgbuf, "Command line: %s", argv[0]);
    for (i = 1; i < argc; i++)
    {
        if (*argv[i] == '-')
            strcat(strcat(msgbuf, "\n              "), argv[i]);
        else strcat(strcat(msgbuf, " "), argv[i]);
    }
    strcat(msgbuf, NL);
    logMsg(LEVEL1, msgbuf);

    /*
     * read config file, if present, exit on any errors
     * do not override any options set on the command line
     */
    if ((ret = doConf()))
    {
        if (ret > 0) errorExit(-104, 0, 0);
        else errorExit(-1, progname, 0);
    }

    logMsg(LEVEL1, "Verbose level: %d\n", verbose);
    logMsg(LEVEL1, "Config File: %s\n", config);
    logMsg(LEVEL1, "Alias File: %s\n", cidalias);
    logMsg(LEVEL1, "Blacklist File: %s\n", blacklist);
    logMsg(LEVEL1, "Whitelist File: %s\n", whitelist);

    switch (regex)
    {
        case 0:
            logMsg(LEVEL1, "Using simple expressions for alias/blacklist/whitelist entries\n");
            break;
        case 1:
            logMsg(LEVEL1, "Using posix regular expressions for alias/blacklist/whitelist entries\n");
            break;
        case 2:
            logMsg(LEVEL1, "Using Perl regular expressions for alias/blacklist/whitelist entries\n");
            break;
        default:
            sprintf(msgbuf, "Invalid regex number: %d", regex);
            errorExit(-107, "Fatal", msgbuf);
            break;
    }
    if (hangup) {
        logMsg(LEVEL1, "Using additional names from the blacklist & whitelist #= descriptions\n");
    }
    if (ignore1) {
        logMsg(LEVEL1, "Ignoring 1 in alias definition and calling number\n");
    }
    if (do_multi) {
        logMsg(LEVEL1, "Update current and previous call logs\n");
    } else {
        logMsg(LEVEL1, "Update current call log\n");
    }
    if (do_rename) {
        logMsg(LEVEL1, "Skip interactive prompt\n");
    }

    logMsg(LEVEL1,
      "\nBegin: Loading alias, blacklist and whitelist files [%s]\n", strdate(ONLYTIME));

    /*
     * read alias file, if present, exit on any errors
     */
    if (readAliasFile()) errorExit(-109, 0, 0);

    if (hangup || hupmode) {
       /* read blacklist and whitelist files, exit on any errors */
       if (readBWlists()) errorExit(-114, 0, 0);

       if (hangup) logMsg(LEVEL1, "%s\n%s\n", BLMSG,  WLMSG);
       logMsg(LEVEL1, "%s\n", ignore1 ? IGNORE1 : NOIGNORE1);

       logMsg(LEVEL1, "End: Loaded alias, blacklist, and whitelist files [%s]\n\n", strdate(ONLYTIME));
    }
    else {
       logMsg(LEVEL1, "End: Loaded alias file [%s]\n\n", strdate(ONLYTIME));
    }
    if (verbose == LEVEL8) normalExit();

    /* Read cidlog and write a new one */
    processLog(cidlog);

    /* repeat for --multi  */
    if (do_multi) {
       /* repeat for .1 .2 .3 to .9 */
       int try_post;
       char cidlogx[BUFSIZ];
       for (try_post=1; try_post<=9; try_post++) {
          sprintf(cidlogx, "%s.%d", cidlog, try_post);
          if (stat(cidlogx, &statbuf) == 0) {
             /* Read cidlog and write a new one */
             processLog(cidlogx);
          }
       }
    }

    /* show which alias entries were used for changes, if any */
    /* (NMBRONLY) Changed "4075551212" to "----------" 1 time */
    showAliasScores();

    /*  ask for permission to do the rename, unless -y/--yes */
    sprintf(newcidlog, "%s.new", cidlog);
    if (!do_rename) {
        if (isatty(0)) logMsg(LEVELR, "rename %s to %s? Y/n\n", newcidlog, cidlog );
        yesnop = fgets(yesno, sizeof(yesno), stdin);
        /* accept y or yes or the old a */
        if (yesnop != 0 && ((strcasecmp(yesno, "y") == 0) || (strcasecmp(yesno, "yes") == 0) || (strcmp(yesno, "a") == 0)))
           do_rename = 1;
    }
    if (do_rename) {
        /* rename the new files */
        if (rename(newcidlog, cidlog)!=0) {
           logMsg(LEVELR, "failed to rename %s to %s - %s\n", newcidlog, cidlog, strerror(errno) );
        }
        else {
           logMsg(LEVELR, "rename %s to %s\n", newcidlog, cidlog );
        }
        if (do_multi) {
           /* repeat for .1 .2 .3 to .9 */
           int try_post;
           char cidlogx[BUFSIZ];
           char newcidlogx[BUFSIZ];
           for (try_post=1; try_post<=9; try_post++) {
              sprintf(cidlogx, "%s.%d", cidlog, try_post);
              sprintf(newcidlogx, "%s.%d.new", cidlog, try_post);
              if (stat(newcidlogx, &statbuf) == 0) {
                 if (rename(newcidlogx, cidlogx)!=0) {
                    logMsg(LEVELR, "failed to rename %s to %s - %s\n", newcidlogx, cidlogx, strerror(errno) );
                 }
                 else {
                    logMsg(LEVELR, "rename %s to %s\n", newcidlogx, cidlogx );
                 }
              }
           }
        }
    }

    cleanup();
    exit(0);
}


int getOptions(int argc, char *argv[])
{
    int c, num;
    int option_index = 0;
    static struct option long_options[] = {
        {"alias", 1, 0, 'A'},
        {"aliasfile", 1, 0, 'A'},
        {"blacklist", 1, 0, 'B'},
        {"cidlog", 1, 0, 'c'},
        {"config", 1, 0, 'C'},
/*         {"debug", 0, 0, 'D'}, */
        {"help", 0, 0, 'h'},
        {"hangup", 1, 0, 'H'},
        {"ignore1", 0, 0, 'i'},
        {"multi", 0, 0, 'M'},
        {"regex", 1, 0, 'r'},
        {"verbose", 1, 0, 'v'},
        {"version", 0, 0, 'V'},
        {"whitelist", 1, 0, 'W'},
        {"yes", 0, 0, 'y'},
        {0, 0, 0, 0}
    };

    while ((c = getopt_long (argc, argv, "a:A:B:c:C:DhH:Mir:R:v:VW:y",
        long_options, &option_index)) != -1)
    {
        switch (c)
        {
            case 'a':
            case 'A':
                if (!(cidalias = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("cidalias")) >= 0) setword[num].type = 0;
                break;
            case 'B':
                if (!(blacklist = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("blacklist")) >= 0) setword[num].type = 0;
                break;
            case 'c':
                if (!(cidlog = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("cidlog")) >= 0) setword[num].type = 0;
                break;
            case 'C':
                if (!(config = strdup(optarg))) errorExit(-1, progname, 0);
                break;
            case 'D':
                ++debug;
                break;
            case 'h': /* help message */
                fprintf(stderr, "%s - Network CallerID Log Update\n", progname);
                fprintf(stderr, "%s%s", CIDUPUSAGE, CIDUPOPTIONS);
                exit(0);
            case 'H':
                hangup = atoi(optarg);
                hupmode = hangup;
                if (strlen(optarg) != 1 ||
                    (!(hangup == 0 && *optarg == '0') && hangup > 3))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("hangup")) >= 0) setword[num].type = 0;
                hupmode = hangup;
                if ((num = findWord("hupmode")) >= 0) setword[num].type = 0;
                break;
            case 'i':
                ignore1 = 1;
                break;
           case 'M':
                do_multi = 1;
                break;
            case 'r':
            case 'R':
                regex = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(regex == 0 && *optarg == '0') && regex != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("regex")) >= 0) setword[num].type = 0;
                break;
            case 'v':
                verbose = atoi(optarg);
                /* valid range: 1-9 */
                if (strlen(optarg) != 1 || (verbose == 0))
                {
                    verbose = 1;
                    errorExit(-107, "Invalid number", optarg);
                }
                if ((num = findWord("verbose")) >= 0) setword[num].type = 0;
                break;
            case 'V': /* version */
                fprintf(stderr, "%s %s\n", progname, VERSION);
                exit(0);
            case 'W':
                if (!(whitelist = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("whitelist")) >= 0) setword[num].type = 0;
                break;
            case 'y':
                do_rename = 1;
                break;
            case '?': /* bad option */
                fprintf(stderr, CIDUPUSAGE);
                errorExit(-100, 0, 0);
        }
    }
    return optind;
}


/*
 * Break the line at LINE NMBR and NAME
 * CID: *DATE*07242017*TIME*1435*LINE*POTS*NMBR*6135550555*MESG*NONE*NAME*FLINTSTONE F*
 * NOT: and MSG: skip to *** first
 */
void breakLINE3(const char * bufptr)
{
    static char parts[BUFSIZ];
    char * ptr;
    char * ptr2;
    int tagfound = 0;
    int partctr = 0;
    /* init empty bits and fields */
    breakline.part[0] = "\0";
    breakline.field[0][0] = '\0';
    breakline.part[1] = "\0";
    breakline.field[1][0] = '\0';
    breakline.part[2] = "\0";
    breakline.field[2][0] = '\0';
    breakline.part[3] = "\0";
    breakline.name = breakline.nmbr = breakline.line = NULL;
    breakline.fnmbr = -1;

    /* copy everything to the static array */
    strcpy(parts, bufptr);

    /* break it into parts and fields */
    breakline.part[0] = parts;

    /* for NOT: and MSG: skip to the last *** */
    if (strncmp(parts, "NOT: ", 5) == 0 || strncmp(parts, "MSG: ", 5) == 0 ) {
        ptr = parts;
        while ((ptr2 = strstr(ptr, "***")) != 0) ptr = ptr2 + 2;
        if (ptr == parts) return;        /* without changing anything */
    }
    else {
        /* skip CID: to first * */
        ptr = strchr(parts, '*');
    }

/* CID: *DATE*07242017*TIME*1435*LINE*POTS*NMBR*6135550555*MESG*NONE*NAME*FLINTSTONE F*
 * becomes
 * part[0] -> CID: *DATE*07242017*TIME*1435*LINE*
 * field[0] = POTS             (aka line)
 * part[1] -> *NMBR*
 * field[1] = 6135550555       (aka nmbr)
 * part[2] ->*MESG*NONE*NAME*
 * field[2] = FLINTSTONE F     (aka name)
 * part[3] -> *
 * (there is no field 3)
 * the part that holds FNMBR is noted as fnmbr (probably 2)
 */
    while (partctr < 3) {
        if (ptr == 0) break;
        if (strncmp(ptr, "*NAME*", 6) == 0) {
            breakline.name = breakline.field[partctr];
            tagfound = 1;
        }
        else if (strncmp(ptr, "*NMBR*", 6) == 0) {
            breakline.nmbr = breakline.field[partctr];
            tagfound = 1;
        }
        else if (strncmp(ptr, "*LINE*", 6) == 0) {
            breakline.line = breakline.field[partctr];
            tagfound = 1;
        }
        else if (strncmp(ptr, "*FNMBR*", 7) == 0) {
            /* in case of a change to nmbr, note the part that holds fnmbr */
            breakline.fnmbr = partctr;
        }

        if (tagfound) {
            /* copy out the field and find its end */
            strcpy(breakline.field[partctr], ptr+6);
            ptr2 = strchr(breakline.field[partctr], '*');
            if (ptr2 != 0) *ptr2 = '\0';
            /* cut the part at the trailing star */
            *(ptr+6) = '\0';
            ptr += strlen(breakline.field[partctr]) + 6;
            /* clip longer fields to CIDSIZE */
            if (strlen(breakline.field[partctr]) > CIDSIZE-1)
                breakline.field[partctr][CIDSIZE-1] = '\0';

            /* look again */
            tagfound = 0;
            partctr++;
            breakline.part[partctr] = ptr;
            ptr = strchr(ptr, '*');
        }
        else {
            /* keep looking */
            ptr = strchr(ptr+1, '*');
        }
    }

    /*  bit0 CID: *DATE*07242017*TIME*1435*LINE f0 POTS bit1 *NMBR f1 6135550555 bit2 *MESG*NONE*NAME f2 FLINTSTONE F bit3 * */
    logMsg(LEVEL6, "bit0 %s f0 %s bit1 %s f1 %s bit2 %s f2 %s bit3 %s\n",
         breakline.part[0], breakline.field[0],
         breakline.part[1], breakline.field[1],
         breakline.part[2], breakline.field[2],
         breakline.part[3]);
}
/*
 * Break the previously untouched FNMBR part up when an alias changes the number
 * *blah*blah*FNMBR*844-867-5309*NTYPE*-*CTRY*ZZ*LOCA*-*CARI*-*blah*blah*
 * repair the libphonenumber content
 */
void repair_fnmbr(void)
{
    char acc[BUFSIZ];
    const char * ptr;
    const char * ptr2;
    int found = 0;

    /* copy the part where FNMBR was seen to remove FNMBR NTYPE CTRY LOCA CARI */
    ptr = breakline.part[breakline.fnmbr];
    acc[0] = '\0';

    while (!found) {
        if (ptr == 0) break;

        if (strncmp(ptr, "*FNMBR*", 7) == 0) {
            found = 7;
        }
        else if (strncmp(ptr, "*NTYPE*", 7) == 0) {
            found = 7;
        }
        else if (strncmp(ptr, "*CTRY*", 6) == 0) {
            found = 6;
        }
        else if (strncmp(ptr, "*LOCA*", 6) == 0) {
            found = 6;
        }
        else if (strncmp(ptr, "*CARI*", 6) == 0) {
            found = 6;
        }

        if (found) {
            /* skip the label */
            ptr += found;
            /* skip the data field and look again */
            ptr = strchr(ptr, '*');
            found = 0;
        }
        else {
            /* copy this label/field */
            ptr2 = strchr(ptr+1, '*');
            if (ptr2 == 0) {
               strcat(acc, ptr);
               break;
            }
            else {
               strncat(acc, ptr, (ptr2 - ptr));
               /* keep looking */
               ptr = ptr2;
            }
        }
    }

    /* put in a new FNMBR part with FNMBR NTYPE CTRY LOCA CARI */
    char argnmbr[CIDSIZE];
    char * infoBuffer;
    strcpy(argnmbr, breakline.nmbr);
    normalizeToDialableNumberFromFixedLine(argnmbr, CIDSIZE);
    infoBuffer = GetCompleteInfoFromNumber(argnmbr);
    /* now infoBuffer = "*FNMBR*613-373-8863*NTYPE*FIX/CELL*CTRY*CA*LOCA*Ontario*CARI*Bell Canada*" */
    if (infoBuffer==NULL) infoBuffer = strerror(errno);
    strcpy(new_fnmbr, infoBuffer);
    strcat(new_fnmbr, acc);
    if (infoBuffer!=NULL) free(infoBuffer);

    breakline.part[breakline.fnmbr] = new_fnmbr;
}


/*
 * process the CID log
 */
void processLog(const char * cidlog)
{
    int isCall, len;
    const char **ltptr;
    char *iptr, oldcidline[BUFSIZ], logbuf[BUFSIZ];
    FILE *fp;
    char * assyptr;
    char oldnumber[BUFSIZ];
    int partctr = 0;
    int partlen, fieldlen;

    if ((fp = fopen(cidlog, "r")) == NULL)
    {
        logMsg(LEVELR, "Unable to open call log: %s - %s\n", cidlog, strerror(errno));
        errorExit(-1, "Unable to open call log", 0);
    }
    else
    {
        sprintf(newcidlog, "%s.new", cidlog);
        logMsg(LEVEL1, "Output CID logfile: %s\n", newcidlog);

        openLog(newcidlog);

        /*
         * read each line of the cidcall.log file
         * parse out the fields
         * pass unknown lines out to the new file
         * do alias processing on the lines
         */
        while (fgets(oldcidline, BUFSIZ - sizeof(LINETYPE), fp) != NULL)
        {
            /* strip <CR> and <LF> */
            if ((iptr = strchr(oldcidline, '\r')) != NULL) *iptr = '\0';
            if ((iptr = strchr(oldcidline, '\n')) != NULL) *iptr = '\0';
            logMsg(LEVEL4, "\n\t %s\n", oldcidline);

            /* test for known line tags, eg CID: */
            isCall = 0;
            for (ltptr = lineTags; *ltptr; ++ltptr)
            {
                if (strncmp(oldcidline, *ltptr, strlen(*ltptr)) == 0)
                {
                    /* found a call line */
                    isCall = 1;
                    /* parse the fields */
                    /* getINFO(oldcidline); */
                    breakLINE3(oldcidline);
                    /* returns breakline.name, breakline.nmbr, breakline.line */

                    /* if breakline failed, leave the line as-is */
                    if (breakline.name == NULL ||
                    breakline.nmbr == NULL ||
                    breakline.line == NULL) {
                       isCall = 0;
                       break;
                    }

                    /* store the nmbr */
                    strcpy(oldnumber, breakline.nmbr);
                    /* run it through the alias */
                    userAlias(breakline.nmbr, breakline.name, breakline.line);
                    /* pick up =# names from white and black lists */
                    applyWhiteBlack(breakline.nmbr, breakline.name);
                    /* iff the nmbr changed, clobber the FNMBR and libphonenumber bits
                     */
                    if (breakline.fnmbr && strcmp(oldnumber, breakline.nmbr) != 0) {
                       repair_fnmbr();
                    }

                    /* rebuild the line from the parts and fields */
                    assyptr = logbuf;
                    len = 0;

                    for (partctr=0; partctr<=3; partctr++) {
                       partlen = strlen(breakline.part[partctr]);
                       if (partlen > 0) {
                          len += partlen;
                          if (len > BUFSIZ) break;
                          memcpy(assyptr, breakline.part[partctr], partlen);
                          assyptr += partlen;
                       }
                       fieldlen = strlen(breakline.field[partctr]);
                       if (fieldlen > 0) {
                          len += fieldlen;
                          if (len > BUFSIZ) break;
                          memcpy(assyptr, breakline.field[partctr], fieldlen);
                          assyptr += fieldlen;
                       }
                       *assyptr = '\0';
                    }
                    break;
                }
            }

            if (isCall) {
                /* the reassembled call line */
                writeLog(logbuf);
            }
            else {
                /* the line wasn't a call */
                writeLog(oldcidline);
            }
        }

        (void) fclose(fp);
        (void) fclose(cidfile);

    }
}


void applyWhiteBlack(const char * cidnmbr, char * cidname)
{
    if (hangup || hupmode)
    {
        if (checkWhitelist(cidname, cidnmbr).ret)
        {
           /* the name or number was on the whitelist */
           /* add the name found after #= in the whitelist */
           if (listname) strcpy(cidname, listname);
           logMsg(LEVEL4, "whitelist #= changed the name %s\n", cidname);
        }
        else
        {
            /* it wasn't on the white list */
            if (hangup && checkBlacklist(cidname, cidnmbr).ret)
            {
               /* add the name found after #= in the blacklist */
               if (listname) strcpy(cidname, listname);
               logMsg(LEVEL4, "blacklist #= changed the name %s\n", cidname);
            }
         }
    }
}

/*
 * open new cidlog for write
 */
void openLog(const char * logf)
{
    cidfile = fopen(logf, "w+");
    if (cidfile == NULL)
    {
        errorExit(-1, "Unable to open new call log", logf);
    }
}


/*
 * Write to new cidlog file
 */
void writeLog(const char * logbuf)
{
    logMsg(LEVEL3, "%s\n", logbuf);

    /* write log entry */
    fprintf(cidfile, "%s\n", logbuf);
}


/*
 * Close all file descriptors
 */
void cleanup(void)
{
    /* close log file, if open */
    if (logptr) fclose(logptr);
}


/*
 * exit after some fatal error
 */
void errorExit(int error, const char * msg, const char * arg)
{
    if (error == -1)
    {
        /* should not happen */
        if (msg == 0) msg = "oops";

        /*
         * system error
         * print msg, arg should be zero
         */
        error = errno;
        logMsg(LEVELE, "%s: %s\n", msg, strerror(errno));
    }
    else
    {
        /* should not happen */
        if (msg != 0 && arg == 0) arg = "oops";

        /*
         * internal program error
         * print msg and arg if both are not 0
         */
        if (msg != 0 && arg != 0)
        {
            logMsg(LEVELE, "%s: %s\n", msg, arg);
        }
    }

    /* do not print terminated message, or cleanup, if option error */
    if (error != -100 && error != -101 && error != -106 && error != -107 &&
        error != -108 && error != -113)
    {
        logMsg(LEVELE, "Terminated:  %s\n", strdate(WITHSEP));
        cleanup();
    }

    exit(error);
}

/*
 * Normal exit
 */

void normalExit(void)
{
    logMsg(LEVEL1, "Terminated by Verbose Level 8:  %s\n", strdate(WITHSEP));
    cleanup();
    exit(0);
}
