/*
 * ncidnumberinfo.cpp - This file is part of ncidd.
 *
 * developed by hacking ncidd.c
 * needs  sudo apt-get install libphonenumber-dev libicu-dev

 *
 * Copyright (c) 2020-2021
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

  #ifdef __cplusplus
    extern "C" {
  #endif
#include "ncidd.h"
  #ifdef __cplusplus
       }
  #endif
#include "../server/nciddPN.h"


/* globals */
char * progname;
FILE * logptr;
int debug;
int verbose = 0;        /* below the usual default to get only errors and responses */
const char * country  = DEFAULT_COUNTRY;
const char * language = DEFAULT_LANGUAGE;
const char * refnmbr = NULL;


void cleanup(void);
void normalExit(void);
void errorExit(int error, const char * msg, const char * arg);
int getOptions(int argc, char *argv[]);


#define PNHUSAGE "\
Usage:\n\
   %s [--help|-h] [--version|-V]\n\
   %s [options] <phonenumber> [country] [language]\n\
\n\
Options: [-h              | --help]\n\
         [-n <0-2>        | --nanp-format <0-2>]\n\
         [-V              | --version]\n\
         [-v <1-9>        | --verbose <1-9>]\n\
"
#define PNHOPTIONS "\
\n\
  Where:\n\
    -h, --help\n\
           Displays the help message and exits.\n\
\n\
    -v, --verbose <1-9>\n\
           Output information while processing is occurring. Set the level\n\
           to a higher number for more information. Levels range from 1 to\n\
           9, but not all levels are used.\n\
           Default: verbose = 0\n\
\n\
    -n, --nanp-format <0-3>\n\
        For NANP countries there is a choice of phone number formats:\n\
        (https://en.wikipedia.org/wiki/North_American_Numbering_Plan)\n\
\n\
        0 - phone number not formatted: 2045551234\n\
        1 - google standard format:     (204) 555-1234\n\
        2 - all-hyphen format:          204-555-1234 (default)\n\
        3 - all-dot format:             204.555.1234\n\
\n\
    -V, --version\n\
           Displays the version and exits.\n\
\n\
"

/*
 *  The main program
 */

int main(int argc, char *argv[])
{
    int argind, i;
    char msgbuf[BUFSIZ];
    /* additional info obtained from unformatted phone number in MSG and NOT */
    char * infoBuffer;


    /* global containing name of program */
    progname = strrchr(argv[0], (int) '/');
    progname = progname ? progname + 1 : argv[0];

    /* process options from the command line */
    argind = getOptions(argc, argv);

    /* make logMsg write to stdout */
    logptr = fdopen(STDOUT, "a");

    // 1st arg is a string of phone number digits
    if (argind >= argc) {
       logMsg(LEVELE, "A phone number must be given: ");
       logMsg(LEVELE, "%s <phonenumber> [country] [language]\n", progname);
       exit(-3);
    }
    else {
       refnmbr = argv[argind];
       argind++;
       if (argind < argc) country = argv[argind];
       argind++;
       if (argind < argc) language = argv[argind];
       argind++;
       while (argind < argc) {
          logMsg(LEVEL4, "arg %d %s\n", argind, argv[argind]);
          argind++;
       }
    }

    /* note that verbose is set to 0 so these need -v 1 option to be seen */
    logMsg(LEVEL1, "Started: %s\n%s %s\n",strdate(WITHSEP),
            progname, VERSION);

    /* log command line and any options on separate lines */
    sprintf(msgbuf, "Command line: %s", argv[0]);
    for (i = 1; i < argc; i++)
    {
        if (*argv[i] == '-')
            strcat(strcat(msgbuf, "\n              "), argv[i]);
        else strcat(strcat(msgbuf, " "), argv[i]);
    }
    strcat(msgbuf, NL);
    logMsg(LEVEL1, msgbuf);

    logMsg(LEVEL1, "Verbose level: %d\n", verbose);

    /* NANP country format */
    switch (nanp_format)
    {
      case 0:
        logMsg(LEVEL1, "no format (%d): xxxxxxxxxx\n", nanp_format);
        break;
      case 1:
        logMsg(LEVEL1, "google standard format (%d): (xxx) xxx-xxxx\n", nanp_format);
        break;
      case 2:
        logMsg(LEVEL1, "all-hyphen format (%d): xxx-xxx-xxxx\n", nanp_format);
        break;
    }

    if (verbose == LEVEL8) normalExit();

      /* const char * refnmbr = "6131234567"; */
      logMsg(LEVEL3, "cidnmbr = %s\n", refnmbr);

      char argnmbr[CIDSIZE];
      strncpy(argnmbr, refnmbr, CIDSIZE-1);
      argnmbr[CIDSIZE-1] = '\0';
      normalizeToDialableNumberFromFixedLine(argnmbr, CIDSIZE);
      infoBuffer = GetCompleteInfoFromNumber(argnmbr);
      if (infoBuffer==NULL)
      {
         infoBuffer = strerror(errno);
      }
      printf("%s\n", infoBuffer);

    cleanup();
    exit(0);
}


int getOptions(int argc, char *argv[])
{
    int c;
    int option_index = 0;
    static struct option long_options[] = {
        {"help", 0, 0, 'h'},
        {"nanp-format", 1, 0, 'n'},
        {"verbose", 1, 0, 'v'},
        {"version", 0, 0, 'V'},
        {0, 0, 0, 0}
    };

    while ((c = getopt_long (argc, argv, "hn:v:V",
        long_options, &option_index)) != -1)
    {
        switch (c)
        {
            case 'h': /* help message */
                fprintf(stderr, "%s - Network CallerID Phone Number Helper\n", progname);
                fprintf(stderr, PNHUSAGE, progname, progname);
                fprintf(stderr, PNHOPTIONS);
                exit(0);
            case 'n':
                nanp_format = atoi(optarg);
                /* valid range: 0-2 */
                if (nanp_format < 0 || nanp_format > 3)
                {
                    fprintf(stderr,
                      "number format %d should be within 0-3, using %d\n",
                        nanp_format, DEFAULT_US_FORMAT);
                    nanp_format = 1;
                }
                break;
            case 'v':
                verbose = atoi(optarg);
                /* valid range: 1-9 */
                if (strlen(optarg) != 1 || (verbose == 0))
                {
                    verbose = 1;
                    errorExit(-107, "Invalid number", optarg);
                }
                break;
            case 'V': /* version */
                fprintf(stderr, "%s %s\n", progname, VERSION);
                exit(0);
            case '?': /* bad option */
                fprintf(stderr, PNHUSAGE, progname, progname);
                errorExit(-100, 0, 0);
        }
    }
    return optind;
}

/*
 * Close all file descriptors
 */
void cleanup(void)
{
}


/*
 * exit after some fatal error
 */
void errorExit(int error, const char * msg, const char * arg)
{
    if (error == -1)
    {
        /* should not happen */
        if (msg == 0) msg = "oops";

        /*
         * system error
         * print msg, arg should be zero
         */
        error = errno;
        logMsg(LEVELE, "%s: %s\n", msg, strerror(errno));
    }
    else
    {
        /* should not happen */
        if (msg != 0 && arg == 0) arg = "oops";

        /*
         * internal program error
         * print msg and arg if both are not 0
         */
        if (msg != 0 && arg != 0)
        {
            logMsg(LEVELE, "%s: %s\n", msg, arg);
        }
    }

    /* do not print terminated message, or cleanup, if option error */
    if (error != -100 && error != -101 && error != -106 && error != -107 &&
        error != -108 && error != -113)
    {
        logMsg(LEVELE, "Terminated:  %s\n", strdate(WITHSEP));
        cleanup();
    }

    exit(error);
}

/*
 * Normal exit
 */

void normalExit(void)
{
    logMsg(LEVEL1, "Terminated by Verbose Level 8:  %s\n", strdate(WITHSEP));
    cleanup();
    exit(0);
}
